/***********************************************************************************************************************************
Error Type Definition

Automatically generated by 'build-code error' -- do not modify directly.
***********************************************************************************************************************************/
#ifndef COMMON_ERROR_ERROR_AUTO_H
#define COMMON_ERROR_ERROR_AUTO_H

/***********************************************************************************************************************************
Error type declarations
***********************************************************************************************************************************/
ERROR_DECLARE(AssertError);
ERROR_DECLARE(ChecksumError);
ERROR_DECLARE(ConfigError);
ERROR_DECLARE(FileInvalidError);
ERROR_DECLARE(FormatError);
ERROR_DECLARE(CommandRequiredError);
ERROR_DECLARE(OptionInvalidError);
ERROR_DECLARE(OptionInvalidValueError);
ERROR_DECLARE(OptionRequiredError);
ERROR_DECLARE(PgRunningError);
ERROR_DECLARE(ProtocolError);
ERROR_DECLARE(PathNotEmptyError);
ERROR_DECLARE(FileOpenError);
ERROR_DECLARE(FileReadError);
ERROR_DECLARE(ParamRequiredError);
ERROR_DECLARE(ArchiveMismatchError);
ERROR_DECLARE(ArchiveDuplicateError);
ERROR_DECLARE(VersionNotSupportedError);
ERROR_DECLARE(PathCreateError);
ERROR_DECLARE(CommandInvalidError);
ERROR_DECLARE(HostConnectError);
ERROR_DECLARE(LockAcquireError);
ERROR_DECLARE(BackupMismatchError);
ERROR_DECLARE(FileSyncError);
ERROR_DECLARE(PathOpenError);
ERROR_DECLARE(PathSyncError);
ERROR_DECLARE(FileMissingError);
ERROR_DECLARE(DbConnectError);
ERROR_DECLARE(DbQueryError);
ERROR_DECLARE(DbMismatchError);
ERROR_DECLARE(FileRemoveError);
ERROR_DECLARE(PathRemoveError);
ERROR_DECLARE(StopError);
ERROR_DECLARE(TermError);
ERROR_DECLARE(FileWriteError);
ERROR_DECLARE(FeatureNotSupportedError);
ERROR_DECLARE(ArchiveCommandInvalidError);
ERROR_DECLARE(LinkExpectedError);
ERROR_DECLARE(LinkDestinationError);
ERROR_DECLARE(HostInvalidError);
ERROR_DECLARE(PathMissingError);
ERROR_DECLARE(FileMoveError);
ERROR_DECLARE(BackupSetInvalidError);
ERROR_DECLARE(TablespaceMapError);
ERROR_DECLARE(LinkMapError);
ERROR_DECLARE(FileCloseError);
ERROR_DECLARE(DbMissingError);
ERROR_DECLARE(DbInvalidError);
ERROR_DECLARE(ArchiveTimeoutError);
ERROR_DECLARE(FileModeError);
ERROR_DECLARE(ArchiveDisabledError);
ERROR_DECLARE(FileOwnerError);
ERROR_DECLARE(UserMissingError);
ERROR_DECLARE(FileExistsError);
ERROR_DECLARE(MemoryError);
ERROR_DECLARE(CryptoError);
ERROR_DECLARE(ParamInvalidError);
ERROR_DECLARE(PathCloseError);
ERROR_DECLARE(FileInfoError);
ERROR_DECLARE(JsonFormatError);
ERROR_DECLARE(KernelError);
ERROR_DECLARE(ServiceError);
ERROR_DECLARE(ExecuteError);
ERROR_DECLARE(RepoInvalidError);
ERROR_DECLARE(CommandError);
ERROR_DECLARE(AccessError);
ERROR_DECLARE(ClockError);
ERROR_DECLARE(RuntimeError);
ERROR_DECLARE(UnhandledError);
ERROR_DECLARE(UnknownError);

#endif
