/***********************************************************************************************************************************
Config Parse Rules

Automatically generated by 'build-code config' -- do not modify directly.
***********************************************************************************************************************************/

/***********************************************************************************************************************************
Rule Strings
***********************************************************************************************************************************/
#define PARSE_RULE_VAL_STR(value)                                   PARSE_RULE_U32_2(parseRuleValStr##value)
#define PARSE_RULE_VAL_STRID(value)                                 PARSE_RULE_VAL_STR(QT_##value##_QT)

static const StringPubConst parseRuleValueStr[] =
{
    PARSE_RULE_STRPUB("-1"),                                                                                              // val/str
    PARSE_RULE_STRPUB("-20"),                                                                                             // val/str
    PARSE_RULE_STRPUB("-5"),                                                                                              // val/str
    PARSE_RULE_STRPUB("-7"),                                                                                              // val/str
    PARSE_RULE_STRPUB("/tmp/pgbackrest"),                                                                                 // val/str
    PARSE_RULE_STRPUB("/var/lib/pgbackrest"),                                                                             // val/str
    PARSE_RULE_STRPUB("/var/log/pgbackrest"),                                                                             // val/str
    PARSE_RULE_STRPUB("/var/spool/pgbackrest"),                                                                           // val/str
    PARSE_RULE_STRPUB("0"),                                                                                               // val/str
    PARSE_RULE_STRPUB("0B"),                                                                                              // val/str
    PARSE_RULE_STRPUB("0s"),                                                                                              // val/str
    PARSE_RULE_STRPUB("1"),                                                                                               // val/str
    PARSE_RULE_STRPUB("100ms"),                                                                                           // val/str
    PARSE_RULE_STRPUB("1024"),                                                                                            // val/str
    PARSE_RULE_STRPUB("12"),                                                                                              // val/str
    PARSE_RULE_STRPUB("128KiB"),                                                                                          // val/str
    PARSE_RULE_STRPUB("128MiB"),                                                                                          // val/str
    PARSE_RULE_STRPUB("15m"),                                                                                             // val/str
    PARSE_RULE_STRPUB("15s"),                                                                                             // val/str
    PARSE_RULE_STRPUB("16KiB"),                                                                                           // val/str
    PARSE_RULE_STRPUB("16MiB"),                                                                                           // val/str
    PARSE_RULE_STRPUB("19"),                                                                                              // val/str
    PARSE_RULE_STRPUB("1B"),                                                                                              // val/str
    PARSE_RULE_STRPUB("1GiB"),                                                                                            // val/str
    PARSE_RULE_STRPUB("1MiB"),                                                                                            // val/str
    PARSE_RULE_STRPUB("1PiB"),                                                                                            // val/str
    PARSE_RULE_STRPUB("1TiB"),                                                                                            // val/str
    PARSE_RULE_STRPUB("1d"),                                                                                              // val/str
    PARSE_RULE_STRPUB("1h"),                                                                                              // val/str
    PARSE_RULE_STRPUB("1m"),                                                                                              // val/str
    PARSE_RULE_STRPUB("2"),                                                                                               // val/str
    PARSE_RULE_STRPUB("20MiB"),                                                                                           // val/str
    PARSE_RULE_STRPUB("22"),                                                                                              // val/str
    PARSE_RULE_STRPUB("256"),                                                                                             // val/str
    PARSE_RULE_STRPUB("256KiB"),                                                                                          // val/str
    PARSE_RULE_STRPUB("2MiB"),                                                                                            // val/str
    PARSE_RULE_STRPUB("3"),                                                                                               // val/str
    PARSE_RULE_STRPUB("30m"),                                                                                             // val/str
    PARSE_RULE_STRPUB("31m"),                                                                                             // val/str
    PARSE_RULE_STRPUB("32"),                                                                                              // val/str
    PARSE_RULE_STRPUB("32KiB"),                                                                                           // val/str
    PARSE_RULE_STRPUB("360"),                                                                                             // val/str
    PARSE_RULE_STRPUB("3600"),                                                                                            // val/str
    PARSE_RULE_STRPUB("443"),                                                                                             // val/str
    PARSE_RULE_STRPUB("4MiB"),                                                                                            // val/str
    PARSE_RULE_STRPUB("4PiB"),                                                                                            // val/str
    PARSE_RULE_STRPUB("512KiB"),                                                                                          // val/str
    PARSE_RULE_STRPUB("5432"),                                                                                            // val/str
    PARSE_RULE_STRPUB("5MiB"),                                                                                            // val/str
    PARSE_RULE_STRPUB("6"),                                                                                               // val/str
    PARSE_RULE_STRPUB("64KiB"),                                                                                           // val/str
    PARSE_RULE_STRPUB("65535"),                                                                                           // val/str
    PARSE_RULE_STRPUB("7d"),                                                                                              // val/str
    PARSE_RULE_STRPUB("8432"),                                                                                            // val/str
    PARSE_RULE_STRPUB("8KiB"),                                                                                            // val/str
    PARSE_RULE_STRPUB("8MiB"),                                                                                            // val/str
    PARSE_RULE_STRPUB("9"),                                                                                               // val/str
    PARSE_RULE_STRPUB("900"),                                                                                             // val/str
    PARSE_RULE_STRPUB("999"),                                                                                             // val/str
    PARSE_RULE_STRPUB("9999999"),                                                                                         // val/str
    PARSE_RULE_STRPUB("accept-new"),                                                                                      // val/str
    PARSE_RULE_STRPUB("aes-256-cbc"),                                                                                     // val/str
    PARSE_RULE_STRPUB("asc"),                                                                                             // val/str
    PARSE_RULE_STRPUB("auto"),                                                                                            // val/str
    PARSE_RULE_STRPUB("azure"),                                                                                           // val/str
    PARSE_RULE_STRPUB("blob.core.windows.net"),                                                                           // val/str
    PARSE_RULE_STRPUB("bz2"),                                                                                             // val/str
    PARSE_RULE_STRPUB("cifs"),                                                                                            // val/str
    PARSE_RULE_STRPUB("count"),                                                                                           // val/str
    PARSE_RULE_STRPUB("debug"),                                                                                           // val/str
    PARSE_RULE_STRPUB("default"),                                                                                         // val/str
    PARSE_RULE_STRPUB("desc"),                                                                                            // val/str
    PARSE_RULE_STRPUB("detail"),                                                                                          // val/str
    PARSE_RULE_STRPUB("diff"),                                                                                            // val/str
    PARSE_RULE_STRPUB("error"),                                                                                           // val/str
    PARSE_RULE_STRPUB("fingerprint"),                                                                                     // val/str
    PARSE_RULE_STRPUB("full"),                                                                                            // val/str
    PARSE_RULE_STRPUB("gcs"),                                                                                             // val/str
    PARSE_RULE_STRPUB("gz"),                                                                                              // val/str
    PARSE_RULE_STRPUB("host"),                                                                                            // val/str
    PARSE_RULE_STRPUB("immediate"),                                                                                       // val/str
    PARSE_RULE_STRPUB("incr"),                                                                                            // val/str
    PARSE_RULE_STRPUB("info"),                                                                                            // val/str
    PARSE_RULE_STRPUB("json"),                                                                                            // val/str
    PARSE_RULE_STRPUB("latest"),                                                                                          // val/str
    PARSE_RULE_STRPUB("localhost"),                                                                                       // val/str
    PARSE_RULE_STRPUB("lsn"),                                                                                             // val/str
    PARSE_RULE_STRPUB("lz4"),                                                                                             // val/str
    PARSE_RULE_STRPUB("md5"),                                                                                             // val/str
    PARSE_RULE_STRPUB("n"),                                                                                               // val/str
    PARSE_RULE_STRPUB("name"),                                                                                            // val/str
    PARSE_RULE_STRPUB("none"),                                                                                            // val/str
    PARSE_RULE_STRPUB("num"),                                                                                             // val/str
    PARSE_RULE_STRPUB("off"),                                                                                             // val/str
    PARSE_RULE_STRPUB("path"),                                                                                            // val/str
    PARSE_RULE_STRPUB("pause"),                                                                                           // val/str
    PARSE_RULE_STRPUB("pg"),                                                                                              // val/str
    PARSE_RULE_STRPUB("pgbackrest"),                                                                                      // val/str
    PARSE_RULE_STRPUB("pod-id"),                                                                                          // val/str
    PARSE_RULE_STRPUB("posix"),                                                                                           // val/str
    PARSE_RULE_STRPUB("postgres"),                                                                                        // val/str
    PARSE_RULE_STRPUB("prefer"),                                                                                          // val/str
    PARSE_RULE_STRPUB("preserve"),                                                                                        // val/str
    PARSE_RULE_STRPUB("progress"),                                                                                        // val/str
    PARSE_RULE_STRPUB("promote"),                                                                                         // val/str
    PARSE_RULE_STRPUB("repo"),                                                                                            // val/str
    PARSE_RULE_STRPUB("s3"),                                                                                              // val/str
    PARSE_RULE_STRPUB("sas"),                                                                                             // val/str
    PARSE_RULE_STRPUB("service"),                                                                                         // val/str
    PARSE_RULE_STRPUB("sftp"),                                                                                            // val/str
    PARSE_RULE_STRPUB("sha1"),                                                                                            // val/str
    PARSE_RULE_STRPUB("sha256"),                                                                                          // val/str
    PARSE_RULE_STRPUB("shared"),                                                                                          // val/str
    PARSE_RULE_STRPUB("shutdown"),                                                                                        // val/str
    PARSE_RULE_STRPUB("ssh"),                                                                                             // val/str
    PARSE_RULE_STRPUB("standby"),                                                                                         // val/str
    PARSE_RULE_STRPUB("storage.googleapis.com"),                                                                          // val/str
    PARSE_RULE_STRPUB("strict"),                                                                                          // val/str
    PARSE_RULE_STRPUB("text"),                                                                                            // val/str
    PARSE_RULE_STRPUB("time"),                                                                                            // val/str
    PARSE_RULE_STRPUB("tls"),                                                                                             // val/str
    PARSE_RULE_STRPUB("token"),                                                                                           // val/str
    PARSE_RULE_STRPUB("trace"),                                                                                           // val/str
    PARSE_RULE_STRPUB("warn"),                                                                                            // val/str
    PARSE_RULE_STRPUB("web-id"),                                                                                          // val/str
    PARSE_RULE_STRPUB("xid"),                                                                                             // val/str
    PARSE_RULE_STRPUB("y"),                                                                                               // val/str
    PARSE_RULE_STRPUB("zst"),                                                                                             // val/str
    PARSE_RULE_STRPUB(CFGOPTDEF_CONFIG_PATH),                                                                             // val/str
    PARSE_RULE_STRPUB(CFGOPTDEF_CONFIG_PATH "/" PROJECT_CONFIG_FILE),                                                     // val/str
    PARSE_RULE_STRPUB(CFGOPTDEF_CONFIG_PATH "/" PROJECT_CONFIG_INCLUDE_PATH),                                             // val/str
};

typedef enum
{
    parseRuleValStrQT_DS_1_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_DS_20_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_DS_5_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_DS_7_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_FS_tmp_FS_pgbackrest_QT,                                                                       // val/str/enum
    parseRuleValStrQT_FS_var_FS_lib_FS_pgbackrest_QT,                                                                // val/str/enum
    parseRuleValStrQT_FS_var_FS_log_FS_pgbackrest_QT,                                                                // val/str/enum
    parseRuleValStrQT_FS_var_FS_spool_FS_pgbackrest_QT,                                                              // val/str/enum
    parseRuleValStrQT_0_QT,                                                                                          // val/str/enum
    parseRuleValStrQT_0B_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_0s_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_1_QT,                                                                                          // val/str/enum
    parseRuleValStrQT_100ms_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_1024_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_12_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_128KiB_QT,                                                                                     // val/str/enum
    parseRuleValStrQT_128MiB_QT,                                                                                     // val/str/enum
    parseRuleValStrQT_15m_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_15s_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_16KiB_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_16MiB_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_19_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_1B_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_1GiB_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_1MiB_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_1PiB_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_1TiB_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_1d_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_1h_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_1m_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_2_QT,                                                                                          // val/str/enum
    parseRuleValStrQT_20MiB_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_22_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_256_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_256KiB_QT,                                                                                     // val/str/enum
    parseRuleValStrQT_2MiB_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_3_QT,                                                                                          // val/str/enum
    parseRuleValStrQT_30m_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_31m_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_32_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_32KiB_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_360_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_3600_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_443_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_4MiB_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_4PiB_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_512KiB_QT,                                                                                     // val/str/enum
    parseRuleValStrQT_5432_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_5MiB_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_6_QT,                                                                                          // val/str/enum
    parseRuleValStrQT_64KiB_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_65535_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_7d_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_8432_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_8KiB_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_8MiB_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_9_QT,                                                                                          // val/str/enum
    parseRuleValStrQT_900_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_999_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_9999999_QT,                                                                                    // val/str/enum
    parseRuleValStrQT_accept_DS_new_QT,                                                                              // val/str/enum
    parseRuleValStrQT_aes_DS_256_DS_cbc_QT,                                                                          // val/str/enum
    parseRuleValStrQT_asc_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_auto_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_azure_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_blob_DT_core_DT_windows_DT_net_QT,                                                             // val/str/enum
    parseRuleValStrQT_bz2_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_cifs_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_count_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_debug_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_default_QT,                                                                                    // val/str/enum
    parseRuleValStrQT_desc_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_detail_QT,                                                                                     // val/str/enum
    parseRuleValStrQT_diff_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_error_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_fingerprint_QT,                                                                                // val/str/enum
    parseRuleValStrQT_full_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_gcs_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_gz_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_host_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_immediate_QT,                                                                                  // val/str/enum
    parseRuleValStrQT_incr_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_info_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_json_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_latest_QT,                                                                                     // val/str/enum
    parseRuleValStrQT_localhost_QT,                                                                                  // val/str/enum
    parseRuleValStrQT_lsn_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_lz4_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_md5_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_n_QT,                                                                                          // val/str/enum
    parseRuleValStrQT_name_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_none_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_num_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_off_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_path_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_pause_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_pg_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_pgbackrest_QT,                                                                                 // val/str/enum
    parseRuleValStrQT_pod_DS_id_QT,                                                                                  // val/str/enum
    parseRuleValStrQT_posix_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_postgres_QT,                                                                                   // val/str/enum
    parseRuleValStrQT_prefer_QT,                                                                                     // val/str/enum
    parseRuleValStrQT_preserve_QT,                                                                                   // val/str/enum
    parseRuleValStrQT_progress_QT,                                                                                   // val/str/enum
    parseRuleValStrQT_promote_QT,                                                                                    // val/str/enum
    parseRuleValStrQT_repo_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_s3_QT,                                                                                         // val/str/enum
    parseRuleValStrQT_sas_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_service_QT,                                                                                    // val/str/enum
    parseRuleValStrQT_sftp_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_sha1_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_sha256_QT,                                                                                     // val/str/enum
    parseRuleValStrQT_shared_QT,                                                                                     // val/str/enum
    parseRuleValStrQT_shutdown_QT,                                                                                   // val/str/enum
    parseRuleValStrQT_ssh_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_standby_QT,                                                                                    // val/str/enum
    parseRuleValStrQT_storage_DT_googleapis_DT_com_QT,                                                               // val/str/enum
    parseRuleValStrQT_strict_QT,                                                                                     // val/str/enum
    parseRuleValStrQT_text_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_time_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_tls_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_token_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_trace_QT,                                                                                      // val/str/enum
    parseRuleValStrQT_warn_QT,                                                                                       // val/str/enum
    parseRuleValStrQT_web_DS_id_QT,                                                                                  // val/str/enum
    parseRuleValStrQT_xid_QT,                                                                                        // val/str/enum
    parseRuleValStrQT_y_QT,                                                                                          // val/str/enum
    parseRuleValStrQT_zst_QT,                                                                                        // val/str/enum
    parseRuleValStrCFGOPTDEF_CONFIG_PATH,                                                                            // val/str/enum
    parseRuleValStrCFGOPTDEF_CONFIG_PATH_SP_QT_FS_QT_SP_PROJECT_CONFIG_FILE,                                         // val/str/enum
    parseRuleValStrCFGOPTDEF_CONFIG_PATH_SP_QT_FS_QT_SP_PROJECT_CONFIG_INCLUDE_PATH,                                 // val/str/enum
} ParseRuleValueStr;

/***********************************************************************************************************************************
Rule Ints
***********************************************************************************************************************************/
#define PARSE_RULE_VAL_INT(value)                                   PARSE_RULE_U32_1(parseRuleValInt##value)

static const int parseRuleValueInt[] =
{
    -20,                                                                                                                  // val/int
    -7,                                                                                                                   // val/int
    -5,                                                                                                                   // val/int
    -1,                                                                                                                   // val/int
    0,                                                                                                                    // val/int
    1,                                                                                                                    // val/int
    2,                                                                                                                    // val/int
    3,                                                                                                                    // val/int
    6,                                                                                                                    // val/int
    9,                                                                                                                    // val/int
    12,                                                                                                                   // val/int
    19,                                                                                                                   // val/int
    22,                                                                                                                   // val/int
    32,                                                                                                                   // val/int
    256,                                                                                                                  // val/int
    360,                                                                                                                  // val/int
    443,                                                                                                                  // val/int
    900,                                                                                                                  // val/int
    999,                                                                                                                  // val/int
    1024,                                                                                                                 // val/int
    3600,                                                                                                                 // val/int
    5432,                                                                                                                 // val/int
    8432,                                                                                                                 // val/int
    65535,                                                                                                                // val/int
    9999999,                                                                                                              // val/int
};

static const uint8_t parseRuleValueIntStrMap[] =
{
    parseRuleValStrQT_DS_20_QT,                                                                                    // val/int/strmap
    parseRuleValStrQT_DS_7_QT,                                                                                     // val/int/strmap
    parseRuleValStrQT_DS_5_QT,                                                                                     // val/int/strmap
    parseRuleValStrQT_DS_1_QT,                                                                                     // val/int/strmap
    parseRuleValStrQT_0_QT,                                                                                        // val/int/strmap
    parseRuleValStrQT_1_QT,                                                                                        // val/int/strmap
    parseRuleValStrQT_2_QT,                                                                                        // val/int/strmap
    parseRuleValStrQT_3_QT,                                                                                        // val/int/strmap
    parseRuleValStrQT_6_QT,                                                                                        // val/int/strmap
    parseRuleValStrQT_9_QT,                                                                                        // val/int/strmap
    parseRuleValStrQT_12_QT,                                                                                       // val/int/strmap
    parseRuleValStrQT_19_QT,                                                                                       // val/int/strmap
    parseRuleValStrQT_22_QT,                                                                                       // val/int/strmap
    parseRuleValStrQT_32_QT,                                                                                       // val/int/strmap
    parseRuleValStrQT_256_QT,                                                                                      // val/int/strmap
    parseRuleValStrQT_360_QT,                                                                                      // val/int/strmap
    parseRuleValStrQT_443_QT,                                                                                      // val/int/strmap
    parseRuleValStrQT_900_QT,                                                                                      // val/int/strmap
    parseRuleValStrQT_999_QT,                                                                                      // val/int/strmap
    parseRuleValStrQT_1024_QT,                                                                                     // val/int/strmap
    parseRuleValStrQT_3600_QT,                                                                                     // val/int/strmap
    parseRuleValStrQT_5432_QT,                                                                                     // val/int/strmap
    parseRuleValStrQT_8432_QT,                                                                                     // val/int/strmap
    parseRuleValStrQT_65535_QT,                                                                                    // val/int/strmap
    parseRuleValStrQT_9999999_QT,                                                                                  // val/int/strmap
};

typedef enum
{
    parseRuleValIntDS_20,                                                                                            // val/int/enum
    parseRuleValIntDS_7,                                                                                             // val/int/enum
    parseRuleValIntDS_5,                                                                                             // val/int/enum
    parseRuleValIntDS_1,                                                                                             // val/int/enum
    parseRuleValInt0,                                                                                                // val/int/enum
    parseRuleValInt1,                                                                                                // val/int/enum
    parseRuleValInt2,                                                                                                // val/int/enum
    parseRuleValInt3,                                                                                                // val/int/enum
    parseRuleValInt6,                                                                                                // val/int/enum
    parseRuleValInt9,                                                                                                // val/int/enum
    parseRuleValInt12,                                                                                               // val/int/enum
    parseRuleValInt19,                                                                                               // val/int/enum
    parseRuleValInt22,                                                                                               // val/int/enum
    parseRuleValInt32,                                                                                               // val/int/enum
    parseRuleValInt256,                                                                                              // val/int/enum
    parseRuleValInt360,                                                                                              // val/int/enum
    parseRuleValInt443,                                                                                              // val/int/enum
    parseRuleValInt900,                                                                                              // val/int/enum
    parseRuleValInt999,                                                                                              // val/int/enum
    parseRuleValInt1024,                                                                                             // val/int/enum
    parseRuleValInt3600,                                                                                             // val/int/enum
    parseRuleValInt5432,                                                                                             // val/int/enum
    parseRuleValInt8432,                                                                                             // val/int/enum
    parseRuleValInt65535,                                                                                            // val/int/enum
    parseRuleValInt9999999,                                                                                          // val/int/enum
} ParseRuleValueInt;

/***********************************************************************************************************************************
Rule Sizes
***********************************************************************************************************************************/
#define PARSE_RULE_VAL_SIZE(value)                                  PARSE_RULE_U32_1(parseRuleValSize##value)

static const int64_t parseRuleValueSize[] =
{
    0,                                                                                                                   // val/size
    1,                                                                                                                   // val/size
    8192,                                                                                                                // val/size
    16384,                                                                                                               // val/size
    32768,                                                                                                               // val/size
    65536,                                                                                                               // val/size
    131072,                                                                                                              // val/size
    262144,                                                                                                              // val/size
    524288,                                                                                                              // val/size
    1048576,                                                                                                             // val/size
    2097152,                                                                                                             // val/size
    4194304,                                                                                                             // val/size
    5242880,                                                                                                             // val/size
    8388608,                                                                                                             // val/size
    16777216,                                                                                                            // val/size
    20971520,                                                                                                            // val/size
    134217728,                                                                                                           // val/size
    1073741824,                                                                                                          // val/size
    1099511627776,                                                                                                       // val/size
    1125899906842624,                                                                                                    // val/size
    4503599627370496,                                                                                                    // val/size
};

static const uint8_t parseRuleValueSizeStrMap[] =
{
    parseRuleValStrQT_0B_QT,                                                                                      // val/size/strmap
    parseRuleValStrQT_1B_QT,                                                                                      // val/size/strmap
    parseRuleValStrQT_8KiB_QT,                                                                                    // val/size/strmap
    parseRuleValStrQT_16KiB_QT,                                                                                   // val/size/strmap
    parseRuleValStrQT_32KiB_QT,                                                                                   // val/size/strmap
    parseRuleValStrQT_64KiB_QT,                                                                                   // val/size/strmap
    parseRuleValStrQT_128KiB_QT,                                                                                  // val/size/strmap
    parseRuleValStrQT_256KiB_QT,                                                                                  // val/size/strmap
    parseRuleValStrQT_512KiB_QT,                                                                                  // val/size/strmap
    parseRuleValStrQT_1MiB_QT,                                                                                    // val/size/strmap
    parseRuleValStrQT_2MiB_QT,                                                                                    // val/size/strmap
    parseRuleValStrQT_4MiB_QT,                                                                                    // val/size/strmap
    parseRuleValStrQT_5MiB_QT,                                                                                    // val/size/strmap
    parseRuleValStrQT_8MiB_QT,                                                                                    // val/size/strmap
    parseRuleValStrQT_16MiB_QT,                                                                                   // val/size/strmap
    parseRuleValStrQT_20MiB_QT,                                                                                   // val/size/strmap
    parseRuleValStrQT_128MiB_QT,                                                                                  // val/size/strmap
    parseRuleValStrQT_1GiB_QT,                                                                                    // val/size/strmap
    parseRuleValStrQT_1TiB_QT,                                                                                    // val/size/strmap
    parseRuleValStrQT_1PiB_QT,                                                                                    // val/size/strmap
    parseRuleValStrQT_4PiB_QT,                                                                                    // val/size/strmap
};

typedef enum
{
    parseRuleValSize0B,                                                                                             // val/size/enum
    parseRuleValSize1B,                                                                                             // val/size/enum
    parseRuleValSize8KiB,                                                                                           // val/size/enum
    parseRuleValSize16KiB,                                                                                          // val/size/enum
    parseRuleValSize32KiB,                                                                                          // val/size/enum
    parseRuleValSize64KiB,                                                                                          // val/size/enum
    parseRuleValSize128KiB,                                                                                         // val/size/enum
    parseRuleValSize256KiB,                                                                                         // val/size/enum
    parseRuleValSize512KiB,                                                                                         // val/size/enum
    parseRuleValSize1MiB,                                                                                           // val/size/enum
    parseRuleValSize2MiB,                                                                                           // val/size/enum
    parseRuleValSize4MiB,                                                                                           // val/size/enum
    parseRuleValSize5MiB,                                                                                           // val/size/enum
    parseRuleValSize8MiB,                                                                                           // val/size/enum
    parseRuleValSize16MiB,                                                                                          // val/size/enum
    parseRuleValSize20MiB,                                                                                          // val/size/enum
    parseRuleValSize128MiB,                                                                                         // val/size/enum
    parseRuleValSize1GiB,                                                                                           // val/size/enum
    parseRuleValSize1TiB,                                                                                           // val/size/enum
    parseRuleValSize1PiB,                                                                                           // val/size/enum
    parseRuleValSize4PiB,                                                                                           // val/size/enum
} ParseRuleValueSize;

/***********************************************************************************************************************************
Rule Times
***********************************************************************************************************************************/
#define PARSE_RULE_VAL_TIME(value)                                  PARSE_RULE_U32_1(parseRuleValTime##value)

static const unsigned int parseRuleValueTime[] =
{
    0,                                                                                                                   // val/time
    100,                                                                                                                 // val/time
    15000,                                                                                                               // val/time
    60000,                                                                                                               // val/time
    900000,                                                                                                              // val/time
    1800000,                                                                                                             // val/time
    1860000,                                                                                                             // val/time
    3600000,                                                                                                             // val/time
    86400000,                                                                                                            // val/time
    604800000,                                                                                                           // val/time
};

static const uint8_t parseRuleValueTimeStrMap[] =
{
    parseRuleValStrQT_0s_QT,                                                                                      // val/time/strmap
    parseRuleValStrQT_100ms_QT,                                                                                   // val/time/strmap
    parseRuleValStrQT_15s_QT,                                                                                     // val/time/strmap
    parseRuleValStrQT_1m_QT,                                                                                      // val/time/strmap
    parseRuleValStrQT_15m_QT,                                                                                     // val/time/strmap
    parseRuleValStrQT_30m_QT,                                                                                     // val/time/strmap
    parseRuleValStrQT_31m_QT,                                                                                     // val/time/strmap
    parseRuleValStrQT_1h_QT,                                                                                      // val/time/strmap
    parseRuleValStrQT_1d_QT,                                                                                      // val/time/strmap
    parseRuleValStrQT_7d_QT,                                                                                      // val/time/strmap
};

typedef enum
{
    parseRuleValTime0s,                                                                                             // val/time/enum
    parseRuleValTime100ms,                                                                                          // val/time/enum
    parseRuleValTime15s,                                                                                            // val/time/enum
    parseRuleValTime1m,                                                                                             // val/time/enum
    parseRuleValTime15m,                                                                                            // val/time/enum
    parseRuleValTime30m,                                                                                            // val/time/enum
    parseRuleValTime31m,                                                                                            // val/time/enum
    parseRuleValTime1h,                                                                                             // val/time/enum
    parseRuleValTime1d,                                                                                             // val/time/enum
    parseRuleValTime7d,                                                                                             // val/time/enum
} ParseRuleValueTime;

/***********************************************************************************************************************************
Dynamic default values
***********************************************************************************************************************************/
#define PARSE_RULE_DEFAULT_DYNAMIC(value)                           PARSE_RULE_U32_1(parseRuleDefaultDynamic##value)

typedef enum
{
    parseRuleDefaultDynamicBin,
} ParseRuleDefaultDynamic;

/***********************************************************************************************************************************
Command parse data
***********************************************************************************************************************************/
#define PARSE_RULE_VAL_CMD(value)                                   PARSE_RULE_U32_1(cfgCmd##value)

static const ParseRuleCommand parseRuleCommand[CFG_COMMAND_TOTAL] =
{
    PARSE_RULE_COMMAND                                                                                               // cmd/annotate
    (                                                                                                                // cmd/annotate
        PARSE_RULE_COMMAND_NAME("annotate"),                                                                         // cmd/annotate
        PARSE_RULE_COMMAND_LOCK_REQUIRED(true),                                                                      // cmd/annotate
        PARSE_RULE_COMMAND_LOCK_REMOTE_REQUIRED(true),                                                               // cmd/annotate
        PARSE_RULE_COMMAND_LOCK_TYPE(Backup),                                                                        // cmd/annotate
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                                  // cmd/annotate
                                                                                                                     // cmd/annotate
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                           // cmd/annotate
        (                                                                                                            // cmd/annotate
            PARSE_RULE_COMMAND_ROLE(Main)                                                                            // cmd/annotate
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                          // cmd/annotate
        ),                                                                                                           // cmd/annotate
    ),                                                                                                               // cmd/annotate
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                            // cmd/archive-get
    (                                                                                                             // cmd/archive-get
        PARSE_RULE_COMMAND_NAME("archive-get"),                                                                   // cmd/archive-get
        PARSE_RULE_COMMAND_LOCK_TYPE(Archive),                                                                    // cmd/archive-get
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                               // cmd/archive-get
        PARSE_RULE_COMMAND_PARAMETER_ALLOWED(true),                                                               // cmd/archive-get
                                                                                                                  // cmd/archive-get
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                        // cmd/archive-get
        (                                                                                                         // cmd/archive-get
            PARSE_RULE_COMMAND_ROLE(Async)                                                                        // cmd/archive-get
            PARSE_RULE_COMMAND_ROLE(Local)                                                                        // cmd/archive-get
            PARSE_RULE_COMMAND_ROLE(Main)                                                                         // cmd/archive-get
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                       // cmd/archive-get
        ),                                                                                                        // cmd/archive-get
    ),                                                                                                            // cmd/archive-get
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                           // cmd/archive-push
    (                                                                                                            // cmd/archive-push
        PARSE_RULE_COMMAND_NAME("archive-push"),                                                                 // cmd/archive-push
        PARSE_RULE_COMMAND_LOCK_REMOTE_REQUIRED(true),                                                           // cmd/archive-push
        PARSE_RULE_COMMAND_LOCK_TYPE(Archive),                                                                   // cmd/archive-push
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                              // cmd/archive-push
        PARSE_RULE_COMMAND_PARAMETER_ALLOWED(true),                                                              // cmd/archive-push
                                                                                                                 // cmd/archive-push
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                       // cmd/archive-push
        (                                                                                                        // cmd/archive-push
            PARSE_RULE_COMMAND_ROLE(Async)                                                                       // cmd/archive-push
            PARSE_RULE_COMMAND_ROLE(Local)                                                                       // cmd/archive-push
            PARSE_RULE_COMMAND_ROLE(Main)                                                                        // cmd/archive-push
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                      // cmd/archive-push
        ),                                                                                                       // cmd/archive-push
    ),                                                                                                           // cmd/archive-push
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                 // cmd/backup
    (                                                                                                                  // cmd/backup
        PARSE_RULE_COMMAND_NAME("backup"),                                                                             // cmd/backup
        PARSE_RULE_COMMAND_LOCK_REQUIRED(true),                                                                        // cmd/backup
        PARSE_RULE_COMMAND_LOCK_REMOTE_REQUIRED(true),                                                                 // cmd/backup
        PARSE_RULE_COMMAND_LOCK_TYPE(Backup),                                                                          // cmd/backup
        PARSE_RULE_COMMAND_LOG_FILE(true),                                                                             // cmd/backup
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                                    // cmd/backup
                                                                                                                       // cmd/backup
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                             // cmd/backup
        (                                                                                                              // cmd/backup
            PARSE_RULE_COMMAND_ROLE(Local)                                                                             // cmd/backup
            PARSE_RULE_COMMAND_ROLE(Main)                                                                              // cmd/backup
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                            // cmd/backup
        ),                                                                                                             // cmd/backup
    ),                                                                                                                 // cmd/backup
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                  // cmd/check
    (                                                                                                                   // cmd/check
        PARSE_RULE_COMMAND_NAME("check"),                                                                               // cmd/check
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                             // cmd/check
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                                     // cmd/check
                                                                                                                        // cmd/check
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                              // cmd/check
        (                                                                                                               // cmd/check
            PARSE_RULE_COMMAND_ROLE(Main)                                                                               // cmd/check
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                             // cmd/check
        ),                                                                                                              // cmd/check
    ),                                                                                                                  // cmd/check
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                 // cmd/expire
    (                                                                                                                  // cmd/expire
        PARSE_RULE_COMMAND_NAME("expire"),                                                                             // cmd/expire
        PARSE_RULE_COMMAND_LOCK_REQUIRED(true),                                                                        // cmd/expire
        PARSE_RULE_COMMAND_LOCK_REMOTE_REQUIRED(true),                                                                 // cmd/expire
        PARSE_RULE_COMMAND_LOCK_TYPE(Backup),                                                                          // cmd/expire
        PARSE_RULE_COMMAND_LOG_FILE(true),                                                                             // cmd/expire
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                                    // cmd/expire
                                                                                                                       // cmd/expire
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                             // cmd/expire
        (                                                                                                              // cmd/expire
            PARSE_RULE_COMMAND_ROLE(Main)                                                                              // cmd/expire
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                            // cmd/expire
        ),                                                                                                             // cmd/expire
    ),                                                                                                                 // cmd/expire
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                   // cmd/help
    (                                                                                                                    // cmd/help
        PARSE_RULE_COMMAND_NAME("help"),                                                                                 // cmd/help
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                              // cmd/help
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Debug),                                                                     // cmd/help
        PARSE_RULE_COMMAND_PARAMETER_ALLOWED(true),                                                                      // cmd/help
                                                                                                                         // cmd/help
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                               // cmd/help
        (                                                                                                                // cmd/help
            PARSE_RULE_COMMAND_ROLE(Main)                                                                                // cmd/help
        ),                                                                                                               // cmd/help
    ),                                                                                                                   // cmd/help
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                   // cmd/info
    (                                                                                                                    // cmd/info
        PARSE_RULE_COMMAND_NAME("info"),                                                                                 // cmd/info
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                              // cmd/info
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Debug),                                                                     // cmd/info
                                                                                                                         // cmd/info
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                               // cmd/info
        (                                                                                                                // cmd/info
            PARSE_RULE_COMMAND_ROLE(Main)                                                                                // cmd/info
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                              // cmd/info
        ),                                                                                                               // cmd/info
    ),                                                                                                                   // cmd/info
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                               // cmd/manifest
    (                                                                                                                // cmd/manifest
        PARSE_RULE_COMMAND_NAME("manifest"),                                                                         // cmd/manifest
        PARSE_RULE_COMMAND_INTERNAL(true)                                                                            // cmd/manifest
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                          // cmd/manifest
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Debug),                                                                 // cmd/manifest
                                                                                                                     // cmd/manifest
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                           // cmd/manifest
        (                                                                                                            // cmd/manifest
            PARSE_RULE_COMMAND_ROLE(Main)                                                                            // cmd/manifest
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                          // cmd/manifest
        ),                                                                                                           // cmd/manifest
    ),                                                                                                               // cmd/manifest
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                               // cmd/repo-get
    (                                                                                                                // cmd/repo-get
        PARSE_RULE_COMMAND_NAME("repo-get"),                                                                         // cmd/repo-get
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                          // cmd/repo-get
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Debug),                                                                 // cmd/repo-get
        PARSE_RULE_COMMAND_PARAMETER_ALLOWED(true),                                                                  // cmd/repo-get
                                                                                                                     // cmd/repo-get
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                           // cmd/repo-get
        (                                                                                                            // cmd/repo-get
            PARSE_RULE_COMMAND_ROLE(Main)                                                                            // cmd/repo-get
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                          // cmd/repo-get
        ),                                                                                                           // cmd/repo-get
    ),                                                                                                               // cmd/repo-get
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                // cmd/repo-ls
    (                                                                                                                 // cmd/repo-ls
        PARSE_RULE_COMMAND_NAME("repo-ls"),                                                                           // cmd/repo-ls
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                           // cmd/repo-ls
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Debug),                                                                  // cmd/repo-ls
        PARSE_RULE_COMMAND_PARAMETER_ALLOWED(true),                                                                   // cmd/repo-ls
                                                                                                                      // cmd/repo-ls
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                            // cmd/repo-ls
        (                                                                                                             // cmd/repo-ls
            PARSE_RULE_COMMAND_ROLE(Main)                                                                             // cmd/repo-ls
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                           // cmd/repo-ls
        ),                                                                                                            // cmd/repo-ls
    ),                                                                                                                // cmd/repo-ls
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                               // cmd/repo-put
    (                                                                                                                // cmd/repo-put
        PARSE_RULE_COMMAND_NAME("repo-put"),                                                                         // cmd/repo-put
        PARSE_RULE_COMMAND_INTERNAL(true)                                                                            // cmd/repo-put
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                          // cmd/repo-put
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Debug),                                                                 // cmd/repo-put
        PARSE_RULE_COMMAND_PARAMETER_ALLOWED(true),                                                                  // cmd/repo-put
                                                                                                                     // cmd/repo-put
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                           // cmd/repo-put
        (                                                                                                            // cmd/repo-put
            PARSE_RULE_COMMAND_ROLE(Main)                                                                            // cmd/repo-put
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                          // cmd/repo-put
        ),                                                                                                           // cmd/repo-put
    ),                                                                                                               // cmd/repo-put
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                // cmd/repo-rm
    (                                                                                                                 // cmd/repo-rm
        PARSE_RULE_COMMAND_NAME("repo-rm"),                                                                           // cmd/repo-rm
        PARSE_RULE_COMMAND_INTERNAL(true)                                                                             // cmd/repo-rm
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                           // cmd/repo-rm
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Debug),                                                                  // cmd/repo-rm
        PARSE_RULE_COMMAND_PARAMETER_ALLOWED(true),                                                                   // cmd/repo-rm
                                                                                                                      // cmd/repo-rm
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                            // cmd/repo-rm
        (                                                                                                             // cmd/repo-rm
            PARSE_RULE_COMMAND_ROLE(Main)                                                                             // cmd/repo-rm
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                           // cmd/repo-rm
        ),                                                                                                            // cmd/repo-rm
    ),                                                                                                                // cmd/repo-rm
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                // cmd/restore
    (                                                                                                                 // cmd/restore
        PARSE_RULE_COMMAND_NAME("restore"),                                                                           // cmd/restore
        PARSE_RULE_COMMAND_LOCK_REQUIRED(true),                                                                       // cmd/restore
        PARSE_RULE_COMMAND_LOCK_TYPE(Restore),                                                                        // cmd/restore
        PARSE_RULE_COMMAND_LOG_FILE(true),                                                                            // cmd/restore
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                                   // cmd/restore
                                                                                                                      // cmd/restore
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                            // cmd/restore
        (                                                                                                             // cmd/restore
            PARSE_RULE_COMMAND_ROLE(Local)                                                                            // cmd/restore
            PARSE_RULE_COMMAND_ROLE(Main)                                                                             // cmd/restore
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                           // cmd/restore
        ),                                                                                                            // cmd/restore
    ),                                                                                                                // cmd/restore
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                 // cmd/server
    (                                                                                                                  // cmd/server
        PARSE_RULE_COMMAND_NAME("server"),                                                                             // cmd/server
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                            // cmd/server
        PARSE_RULE_COMMAND_LOG_FILE(true),                                                                             // cmd/server
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                                    // cmd/server
                                                                                                                       // cmd/server
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                             // cmd/server
        (                                                                                                              // cmd/server
            PARSE_RULE_COMMAND_ROLE(Main)                                                                              // cmd/server
        ),                                                                                                             // cmd/server
    ),                                                                                                                 // cmd/server
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                            // cmd/server-ping
    (                                                                                                             // cmd/server-ping
        PARSE_RULE_COMMAND_NAME("server-ping"),                                                                   // cmd/server-ping
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                       // cmd/server-ping
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                               // cmd/server-ping
        PARSE_RULE_COMMAND_PARAMETER_ALLOWED(true),                                                               // cmd/server-ping
                                                                                                                  // cmd/server-ping
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                        // cmd/server-ping
        (                                                                                                         // cmd/server-ping
            PARSE_RULE_COMMAND_ROLE(Main)                                                                         // cmd/server-ping
        ),                                                                                                        // cmd/server-ping
    ),                                                                                                            // cmd/server-ping
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                          // cmd/stanza-create
    (                                                                                                           // cmd/stanza-create
        PARSE_RULE_COMMAND_NAME("stanza-create"),                                                               // cmd/stanza-create
        PARSE_RULE_COMMAND_LOCK_REQUIRED(true),                                                                 // cmd/stanza-create
        PARSE_RULE_COMMAND_LOCK_REMOTE_REQUIRED(true),                                                          // cmd/stanza-create
        PARSE_RULE_COMMAND_LOCK_TYPE(All),                                                                      // cmd/stanza-create
        PARSE_RULE_COMMAND_LOG_FILE(true),                                                                      // cmd/stanza-create
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                             // cmd/stanza-create
                                                                                                                // cmd/stanza-create
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                      // cmd/stanza-create
        (                                                                                                       // cmd/stanza-create
            PARSE_RULE_COMMAND_ROLE(Main)                                                                       // cmd/stanza-create
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                     // cmd/stanza-create
        ),                                                                                                      // cmd/stanza-create
    ),                                                                                                          // cmd/stanza-create
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                          // cmd/stanza-delete
    (                                                                                                           // cmd/stanza-delete
        PARSE_RULE_COMMAND_NAME("stanza-delete"),                                                               // cmd/stanza-delete
        PARSE_RULE_COMMAND_LOCK_REQUIRED(true),                                                                 // cmd/stanza-delete
        PARSE_RULE_COMMAND_LOCK_REMOTE_REQUIRED(true),                                                          // cmd/stanza-delete
        PARSE_RULE_COMMAND_LOCK_TYPE(All),                                                                      // cmd/stanza-delete
        PARSE_RULE_COMMAND_LOG_FILE(true),                                                                      // cmd/stanza-delete
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                             // cmd/stanza-delete
                                                                                                                // cmd/stanza-delete
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                      // cmd/stanza-delete
        (                                                                                                       // cmd/stanza-delete
            PARSE_RULE_COMMAND_ROLE(Main)                                                                       // cmd/stanza-delete
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                     // cmd/stanza-delete
        ),                                                                                                      // cmd/stanza-delete
    ),                                                                                                          // cmd/stanza-delete
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                         // cmd/stanza-upgrade
    (                                                                                                          // cmd/stanza-upgrade
        PARSE_RULE_COMMAND_NAME("stanza-upgrade"),                                                             // cmd/stanza-upgrade
        PARSE_RULE_COMMAND_LOCK_REQUIRED(true),                                                                // cmd/stanza-upgrade
        PARSE_RULE_COMMAND_LOCK_REMOTE_REQUIRED(true),                                                         // cmd/stanza-upgrade
        PARSE_RULE_COMMAND_LOCK_TYPE(All),                                                                     // cmd/stanza-upgrade
        PARSE_RULE_COMMAND_LOG_FILE(true),                                                                     // cmd/stanza-upgrade
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                            // cmd/stanza-upgrade
                                                                                                               // cmd/stanza-upgrade
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                     // cmd/stanza-upgrade
        (                                                                                                      // cmd/stanza-upgrade
            PARSE_RULE_COMMAND_ROLE(Main)                                                                      // cmd/stanza-upgrade
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                    // cmd/stanza-upgrade
        ),                                                                                                     // cmd/stanza-upgrade
    ),                                                                                                         // cmd/stanza-upgrade
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                  // cmd/start
    (                                                                                                                   // cmd/start
        PARSE_RULE_COMMAND_NAME("start"),                                                                               // cmd/start
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                             // cmd/start
        PARSE_RULE_COMMAND_LOG_FILE(true),                                                                              // cmd/start
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                                     // cmd/start
                                                                                                                        // cmd/start
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                              // cmd/start
        (                                                                                                               // cmd/start
            PARSE_RULE_COMMAND_ROLE(Main)                                                                               // cmd/start
        ),                                                                                                              // cmd/start
    ),                                                                                                                  // cmd/start
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                   // cmd/stop
    (                                                                                                                    // cmd/stop
        PARSE_RULE_COMMAND_NAME("stop"),                                                                                 // cmd/stop
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                              // cmd/stop
        PARSE_RULE_COMMAND_LOG_FILE(true),                                                                               // cmd/stop
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                                      // cmd/stop
                                                                                                                         // cmd/stop
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                               // cmd/stop
        (                                                                                                                // cmd/stop
            PARSE_RULE_COMMAND_ROLE(Main)                                                                                // cmd/stop
        ),                                                                                                               // cmd/stop
    ),                                                                                                                   // cmd/stop
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                 // cmd/verify
    (                                                                                                                  // cmd/verify
        PARSE_RULE_COMMAND_NAME("verify"),                                                                             // cmd/verify
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                            // cmd/verify
        PARSE_RULE_COMMAND_LOG_FILE(true),                                                                             // cmd/verify
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Info),                                                                    // cmd/verify
                                                                                                                       // cmd/verify
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                             // cmd/verify
        (                                                                                                              // cmd/verify
            PARSE_RULE_COMMAND_ROLE(Local)                                                                             // cmd/verify
            PARSE_RULE_COMMAND_ROLE(Main)                                                                              // cmd/verify
            PARSE_RULE_COMMAND_ROLE(Remote)                                                                            // cmd/verify
        ),                                                                                                             // cmd/verify
    ),                                                                                                                 // cmd/verify
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_COMMAND                                                                                                // cmd/version
    (                                                                                                                 // cmd/version
        PARSE_RULE_COMMAND_NAME("version"),                                                                           // cmd/version
        PARSE_RULE_COMMAND_LOCK_TYPE(None),                                                                           // cmd/version
        PARSE_RULE_COMMAND_LOG_LEVEL_DEFAULT(Debug),                                                                  // cmd/version
                                                                                                                      // cmd/version
        PARSE_RULE_COMMAND_ROLE_VALID_LIST                                                                            // cmd/version
        (                                                                                                             // cmd/version
            PARSE_RULE_COMMAND_ROLE(Main)                                                                             // cmd/version
        ),                                                                                                            // cmd/version
    ),                                                                                                                // cmd/version
};

/***********************************************************************************************************************************
Option group parse data
***********************************************************************************************************************************/
static const ParseRuleOptionGroup parseRuleOptionGroup[CFG_OPTION_GROUP_TOTAL] =
{
    PARSE_RULE_OPTION_GROUP                                                                                            // opt-grp/pg
    (                                                                                                                  // opt-grp/pg
        PARSE_RULE_OPTION_GROUP_NAME("pg"),                                                                            // opt-grp/pg
    ),                                                                                                                 // opt-grp/pg
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION_GROUP                                                                                          // opt-grp/repo
    (                                                                                                                // opt-grp/repo
        PARSE_RULE_OPTION_GROUP_NAME("repo"),                                                                        // opt-grp/repo
    ),                                                                                                               // opt-grp/repo
};

/***********************************************************************************************************************************
Option parse data
***********************************************************************************************************************************/
#define PARSE_RULE_VAL_OPT(value)                                   PARSE_RULE_U32_2(cfgOpt##value)

static const ParseRuleOption parseRuleOption[CFG_OPTION_TOTAL] =
{
    PARSE_RULE_OPTION                                                                                              // opt/annotation
    (                                                                                                              // opt/annotation
        PARSE_RULE_OPTION_NAME("annotation"),                                                                      // opt/annotation
        PARSE_RULE_OPTION_TYPE(Hash),                                                                              // opt/annotation
        PARSE_RULE_OPTION_RESET(true),                                                                             // opt/annotation
        PARSE_RULE_OPTION_REQUIRED(false),                                                                         // opt/annotation
        PARSE_RULE_OPTION_SECTION(Global),                                                                         // opt/annotation
        PARSE_RULE_OPTION_MULTI(true),                                                                             // opt/annotation
                                                                                                                   // opt/annotation
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                             // opt/annotation
        (                                                                                                          // opt/annotation
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                    // opt/annotation
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                      // opt/annotation
        ),                                                                                                         // opt/annotation
                                                                                                                   // opt/annotation
        PARSE_RULE_OPTIONAL                                                                                        // opt/annotation
        (                                                                                                          // opt/annotation
            PARSE_RULE_OPTIONAL_GROUP                                                                              // opt/annotation
            (                                                                                                      // opt/annotation
                PARSE_RULE_FILTER_CMD                                                                              // opt/annotation
                (                                                                                                  // opt/annotation
                    PARSE_RULE_VAL_CMD(Annotate),                                                                  // opt/annotation
                ),                                                                                                 // opt/annotation
                                                                                                                   // opt/annotation
                PARSE_RULE_OPTIONAL_REQUIRED(),                                                                    // opt/annotation
            ),                                                                                                     // opt/annotation
        ),                                                                                                         // opt/annotation
    ),                                                                                                             // opt/annotation
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                           // opt/archive-async
    (                                                                                                           // opt/archive-async
        PARSE_RULE_OPTION_NAME("archive-async"),                                                                // opt/archive-async
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                        // opt/archive-async
        PARSE_RULE_OPTION_NEGATE(true),                                                                         // opt/archive-async
        PARSE_RULE_OPTION_RESET(true),                                                                          // opt/archive-async
        PARSE_RULE_OPTION_REQUIRED(true),                                                                       // opt/archive-async
        PARSE_RULE_OPTION_SECTION(Global),                                                                      // opt/archive-async
                                                                                                                // opt/archive-async
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                          // opt/archive-async
        (                                                                                                       // opt/archive-async
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/archive-async
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/archive-async
        ),                                                                                                      // opt/archive-async
                                                                                                                // opt/archive-async
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                         // opt/archive-async
        (                                                                                                       // opt/archive-async
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/archive-async
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/archive-async
        ),                                                                                                      // opt/archive-async
                                                                                                                // opt/archive-async
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                         // opt/archive-async
        (                                                                                                       // opt/archive-async
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/archive-async
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/archive-async
        ),                                                                                                      // opt/archive-async
                                                                                                                // opt/archive-async
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                        // opt/archive-async
        (                                                                                                       // opt/archive-async
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/archive-async
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/archive-async
        ),                                                                                                      // opt/archive-async
                                                                                                                // opt/archive-async
        PARSE_RULE_OPTIONAL                                                                                     // opt/archive-async
        (                                                                                                       // opt/archive-async
            PARSE_RULE_OPTIONAL_GROUP                                                                           // opt/archive-async
            (                                                                                                   // opt/archive-async
                PARSE_RULE_OPTIONAL_DEFAULT                                                                     // opt/archive-async
                (                                                                                               // opt/archive-async
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                  // opt/archive-async
                ),                                                                                              // opt/archive-async
            ),                                                                                                  // opt/archive-async
        ),                                                                                                      // opt/archive-async
    ),                                                                                                          // opt/archive-async
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                           // opt/archive-check
    (                                                                                                           // opt/archive-check
        PARSE_RULE_OPTION_NAME("archive-check"),                                                                // opt/archive-check
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                        // opt/archive-check
        PARSE_RULE_OPTION_NEGATE(true),                                                                         // opt/archive-check
        PARSE_RULE_OPTION_RESET(true),                                                                          // opt/archive-check
        PARSE_RULE_OPTION_REQUIRED(true),                                                                       // opt/archive-check
        PARSE_RULE_OPTION_SECTION(Global),                                                                      // opt/archive-check
                                                                                                                // opt/archive-check
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                          // opt/archive-check
        (                                                                                                       // opt/archive-check
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/archive-check
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/archive-check
            PARSE_RULE_OPTION_COMMAND(Check)                                                                    // opt/archive-check
        ),                                                                                                      // opt/archive-check
                                                                                                                // opt/archive-check
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                         // opt/archive-check
        (                                                                                                       // opt/archive-check
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/archive-check
        ),                                                                                                      // opt/archive-check
                                                                                                                // opt/archive-check
        PARSE_RULE_OPTIONAL                                                                                     // opt/archive-check
        (                                                                                                       // opt/archive-check
            PARSE_RULE_OPTIONAL_GROUP                                                                           // opt/archive-check
            (                                                                                                   // opt/archive-check
                PARSE_RULE_FILTER_CMD                                                                           // opt/archive-check
                (                                                                                               // opt/archive-check
                    PARSE_RULE_VAL_CMD(Backup),                                                                 // opt/archive-check
                ),                                                                                              // opt/archive-check
                                                                                                                // opt/archive-check
                PARSE_RULE_OPTIONAL_DEPEND                                                                      // opt/archive-check
                (                                                                                               // opt/archive-check
                    PARSE_RULE_OPTIONAL_DEPEND_DEFAULT(PARSE_RULE_VAL_BOOL_FALSE),                              // opt/archive-check
                    PARSE_RULE_VAL_OPT(Online),                                                                 // opt/archive-check
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                   // opt/archive-check
                ),                                                                                              // opt/archive-check
                                                                                                                // opt/archive-check
                PARSE_RULE_OPTIONAL_DEFAULT                                                                     // opt/archive-check
                (                                                                                               // opt/archive-check
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                   // opt/archive-check
                ),                                                                                              // opt/archive-check
            ),                                                                                                  // opt/archive-check
                                                                                                                // opt/archive-check
            PARSE_RULE_OPTIONAL_GROUP                                                                           // opt/archive-check
            (                                                                                                   // opt/archive-check
                PARSE_RULE_OPTIONAL_DEFAULT                                                                     // opt/archive-check
                (                                                                                               // opt/archive-check
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                   // opt/archive-check
                ),                                                                                              // opt/archive-check
            ),                                                                                                  // opt/archive-check
        ),                                                                                                      // opt/archive-check
    ),                                                                                                          // opt/archive-check
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                            // opt/archive-copy
    (                                                                                                            // opt/archive-copy
        PARSE_RULE_OPTION_NAME("archive-copy"),                                                                  // opt/archive-copy
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                         // opt/archive-copy
        PARSE_RULE_OPTION_NEGATE(true),                                                                          // opt/archive-copy
        PARSE_RULE_OPTION_RESET(true),                                                                           // opt/archive-copy
        PARSE_RULE_OPTION_REQUIRED(true),                                                                        // opt/archive-copy
        PARSE_RULE_OPTION_SECTION(Global),                                                                       // opt/archive-copy
                                                                                                                 // opt/archive-copy
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                           // opt/archive-copy
        (                                                                                                        // opt/archive-copy
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/archive-copy
        ),                                                                                                       // opt/archive-copy
                                                                                                                 // opt/archive-copy
        PARSE_RULE_OPTIONAL                                                                                      // opt/archive-copy
        (                                                                                                        // opt/archive-copy
            PARSE_RULE_OPTIONAL_GROUP                                                                            // opt/archive-copy
            (                                                                                                    // opt/archive-copy
                PARSE_RULE_FILTER_CMD                                                                            // opt/archive-copy
                (                                                                                                // opt/archive-copy
                    PARSE_RULE_VAL_CMD(Backup),                                                                  // opt/archive-copy
                ),                                                                                               // opt/archive-copy
                                                                                                                 // opt/archive-copy
                PARSE_RULE_OPTIONAL_DEPEND                                                                       // opt/archive-copy
                (                                                                                                // opt/archive-copy
                    PARSE_RULE_OPTIONAL_DEPEND_DEFAULT(PARSE_RULE_VAL_BOOL_FALSE),                               // opt/archive-copy
                    PARSE_RULE_VAL_OPT(ArchiveCheck),                                                            // opt/archive-copy
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                    // opt/archive-copy
                ),                                                                                               // opt/archive-copy
                                                                                                                 // opt/archive-copy
                PARSE_RULE_OPTIONAL_DEFAULT                                                                      // opt/archive-copy
                (                                                                                                // opt/archive-copy
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                   // opt/archive-copy
                ),                                                                                               // opt/archive-copy
            ),                                                                                                   // opt/archive-copy
                                                                                                                 // opt/archive-copy
            PARSE_RULE_OPTIONAL_GROUP                                                                            // opt/archive-copy
            (                                                                                                    // opt/archive-copy
                PARSE_RULE_OPTIONAL_DEFAULT                                                                      // opt/archive-copy
                (                                                                                                // opt/archive-copy
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                   // opt/archive-copy
                ),                                                                                               // opt/archive-copy
            ),                                                                                                   // opt/archive-copy
        ),                                                                                                       // opt/archive-copy
    ),                                                                                                           // opt/archive-copy
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                   // opt/archive-get-queue-max
    (                                                                                                   // opt/archive-get-queue-max
        PARSE_RULE_OPTION_NAME("archive-get-queue-max"),                                                // opt/archive-get-queue-max
        PARSE_RULE_OPTION_TYPE(Size),                                                                   // opt/archive-get-queue-max
        PARSE_RULE_OPTION_RESET(true),                                                                  // opt/archive-get-queue-max
        PARSE_RULE_OPTION_REQUIRED(true),                                                               // opt/archive-get-queue-max
        PARSE_RULE_OPTION_SECTION(Global),                                                              // opt/archive-get-queue-max
                                                                                                        // opt/archive-get-queue-max
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                  // opt/archive-get-queue-max
        (                                                                                               // opt/archive-get-queue-max
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                       // opt/archive-get-queue-max
        ),                                                                                              // opt/archive-get-queue-max
                                                                                                        // opt/archive-get-queue-max
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                 // opt/archive-get-queue-max
        (                                                                                               // opt/archive-get-queue-max
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                       // opt/archive-get-queue-max
        ),                                                                                              // opt/archive-get-queue-max
                                                                                                        // opt/archive-get-queue-max
        PARSE_RULE_OPTIONAL                                                                             // opt/archive-get-queue-max
        (                                                                                               // opt/archive-get-queue-max
            PARSE_RULE_OPTIONAL_GROUP                                                                   // opt/archive-get-queue-max
            (                                                                                           // opt/archive-get-queue-max
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                         // opt/archive-get-queue-max
                (                                                                                       // opt/archive-get-queue-max
                    PARSE_RULE_VAL_SIZE(0B),                                                            // opt/archive-get-queue-max
                    PARSE_RULE_VAL_SIZE(4PiB),                                                          // opt/archive-get-queue-max
                ),                                                                                      // opt/archive-get-queue-max
                                                                                                        // opt/archive-get-queue-max
                PARSE_RULE_OPTIONAL_DEFAULT                                                             // opt/archive-get-queue-max
                (                                                                                       // opt/archive-get-queue-max
                    PARSE_RULE_VAL_SIZE(128MiB),                                                        // opt/archive-get-queue-max
                ),                                                                                      // opt/archive-get-queue-max
            ),                                                                                          // opt/archive-get-queue-max
        ),                                                                                              // opt/archive-get-queue-max
    ),                                                                                                  // opt/archive-get-queue-max
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                    // opt/archive-header-check
    (                                                                                                    // opt/archive-header-check
        PARSE_RULE_OPTION_NAME("archive-header-check"),                                                  // opt/archive-header-check
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                 // opt/archive-header-check
        PARSE_RULE_OPTION_NEGATE(true),                                                                  // opt/archive-header-check
        PARSE_RULE_OPTION_RESET(true),                                                                   // opt/archive-header-check
        PARSE_RULE_OPTION_REQUIRED(true),                                                                // opt/archive-header-check
        PARSE_RULE_OPTION_SECTION(Global),                                                               // opt/archive-header-check
                                                                                                         // opt/archive-header-check
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                   // opt/archive-header-check
        (                                                                                                // opt/archive-header-check
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/archive-header-check
        ),                                                                                               // opt/archive-header-check
                                                                                                         // opt/archive-header-check
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                  // opt/archive-header-check
        (                                                                                                // opt/archive-header-check
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/archive-header-check
        ),                                                                                               // opt/archive-header-check
                                                                                                         // opt/archive-header-check
        PARSE_RULE_OPTIONAL                                                                              // opt/archive-header-check
        (                                                                                                // opt/archive-header-check
            PARSE_RULE_OPTIONAL_GROUP                                                                    // opt/archive-header-check
            (                                                                                            // opt/archive-header-check
                PARSE_RULE_OPTIONAL_DEFAULT                                                              // opt/archive-header-check
                (                                                                                        // opt/archive-header-check
                    PARSE_RULE_VAL_BOOL_TRUE,                                                            // opt/archive-header-check
                ),                                                                                       // opt/archive-header-check
            ),                                                                                           // opt/archive-header-check
        ),                                                                                               // opt/archive-header-check
    ),                                                                                                   // opt/archive-header-check
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                   // opt/archive-missing-retry
    (                                                                                                   // opt/archive-missing-retry
        PARSE_RULE_OPTION_NAME("archive-missing-retry"),                                                // opt/archive-missing-retry
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                // opt/archive-missing-retry
        PARSE_RULE_OPTION_NEGATE(true),                                                                 // opt/archive-missing-retry
        PARSE_RULE_OPTION_RESET(true),                                                                  // opt/archive-missing-retry
        PARSE_RULE_OPTION_REQUIRED(true),                                                               // opt/archive-missing-retry
        PARSE_RULE_OPTION_SECTION(Global),                                                              // opt/archive-missing-retry
                                                                                                        // opt/archive-missing-retry
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                  // opt/archive-missing-retry
        (                                                                                               // opt/archive-missing-retry
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                       // opt/archive-missing-retry
        ),                                                                                              // opt/archive-missing-retry
                                                                                                        // opt/archive-missing-retry
        PARSE_RULE_OPTIONAL                                                                             // opt/archive-missing-retry
        (                                                                                               // opt/archive-missing-retry
            PARSE_RULE_OPTIONAL_GROUP                                                                   // opt/archive-missing-retry
            (                                                                                           // opt/archive-missing-retry
                PARSE_RULE_OPTIONAL_DEFAULT                                                             // opt/archive-missing-retry
                (                                                                                       // opt/archive-missing-retry
                    PARSE_RULE_VAL_BOOL_TRUE,                                                           // opt/archive-missing-retry
                ),                                                                                      // opt/archive-missing-retry
            ),                                                                                          // opt/archive-missing-retry
        ),                                                                                              // opt/archive-missing-retry
    ),                                                                                                  // opt/archive-missing-retry
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                            // opt/archive-mode
    (                                                                                                            // opt/archive-mode
        PARSE_RULE_OPTION_NAME("archive-mode"),                                                                  // opt/archive-mode
        PARSE_RULE_OPTION_TYPE(StringId),                                                                        // opt/archive-mode
        PARSE_RULE_OPTION_RESET(true),                                                                           // opt/archive-mode
        PARSE_RULE_OPTION_REQUIRED(true),                                                                        // opt/archive-mode
        PARSE_RULE_OPTION_SECTION(Global),                                                                       // opt/archive-mode
                                                                                                                 // opt/archive-mode
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                           // opt/archive-mode
        (                                                                                                        // opt/archive-mode
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                   // opt/archive-mode
        ),                                                                                                       // opt/archive-mode
                                                                                                                 // opt/archive-mode
        PARSE_RULE_OPTIONAL                                                                                      // opt/archive-mode
        (                                                                                                        // opt/archive-mode
            PARSE_RULE_OPTIONAL_GROUP                                                                            // opt/archive-mode
            (                                                                                                    // opt/archive-mode
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                   // opt/archive-mode
                (                                                                                                // opt/archive-mode
                    PARSE_RULE_VAL_STRID(off),                                                                   // opt/archive-mode
                    PARSE_RULE_VAL_STRID(preserve),                                                              // opt/archive-mode
                ),                                                                                               // opt/archive-mode
                                                                                                                 // opt/archive-mode
                PARSE_RULE_OPTIONAL_DEFAULT                                                                      // opt/archive-mode
                (                                                                                                // opt/archive-mode
                    PARSE_RULE_VAL_STRID(preserve),                                                              // opt/archive-mode
                ),                                                                                               // opt/archive-mode
            ),                                                                                                   // opt/archive-mode
        ),                                                                                                       // opt/archive-mode
    ),                                                                                                           // opt/archive-mode
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                      // opt/archive-mode-check
    (                                                                                                      // opt/archive-mode-check
        PARSE_RULE_OPTION_NAME("archive-mode-check"),                                                      // opt/archive-mode-check
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                   // opt/archive-mode-check
        PARSE_RULE_OPTION_NEGATE(true),                                                                    // opt/archive-mode-check
        PARSE_RULE_OPTION_RESET(true),                                                                     // opt/archive-mode-check
        PARSE_RULE_OPTION_REQUIRED(true),                                                                  // opt/archive-mode-check
        PARSE_RULE_OPTION_SECTION(Global),                                                                 // opt/archive-mode-check
                                                                                                           // opt/archive-mode-check
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                     // opt/archive-mode-check
        (                                                                                                  // opt/archive-mode-check
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/archive-mode-check
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/archive-mode-check
            PARSE_RULE_OPTION_COMMAND(Check)                                                               // opt/archive-mode-check
        ),                                                                                                 // opt/archive-mode-check
                                                                                                           // opt/archive-mode-check
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                    // opt/archive-mode-check
        (                                                                                                  // opt/archive-mode-check
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/archive-mode-check
        ),                                                                                                 // opt/archive-mode-check
                                                                                                           // opt/archive-mode-check
        PARSE_RULE_OPTIONAL                                                                                // opt/archive-mode-check
        (                                                                                                  // opt/archive-mode-check
            PARSE_RULE_OPTIONAL_GROUP                                                                      // opt/archive-mode-check
            (                                                                                              // opt/archive-mode-check
                PARSE_RULE_OPTIONAL_DEPEND                                                                 // opt/archive-mode-check
                (                                                                                          // opt/archive-mode-check
                    PARSE_RULE_OPTIONAL_DEPEND_DEFAULT(PARSE_RULE_VAL_BOOL_FALSE),                         // opt/archive-mode-check
                    PARSE_RULE_VAL_OPT(ArchiveCheck),                                                      // opt/archive-mode-check
                    PARSE_RULE_VAL_BOOL_TRUE,                                                              // opt/archive-mode-check
                ),                                                                                         // opt/archive-mode-check
                                                                                                           // opt/archive-mode-check
                PARSE_RULE_OPTIONAL_DEFAULT                                                                // opt/archive-mode-check
                (                                                                                          // opt/archive-mode-check
                    PARSE_RULE_VAL_BOOL_TRUE,                                                              // opt/archive-mode-check
                ),                                                                                         // opt/archive-mode-check
            ),                                                                                             // opt/archive-mode-check
        ),                                                                                                 // opt/archive-mode-check
    ),                                                                                                     // opt/archive-mode-check
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                  // opt/archive-push-queue-max
    (                                                                                                  // opt/archive-push-queue-max
        PARSE_RULE_OPTION_NAME("archive-push-queue-max"),                                              // opt/archive-push-queue-max
        PARSE_RULE_OPTION_TYPE(Size),                                                                  // opt/archive-push-queue-max
        PARSE_RULE_OPTION_RESET(true),                                                                 // opt/archive-push-queue-max
        PARSE_RULE_OPTION_REQUIRED(false),                                                             // opt/archive-push-queue-max
        PARSE_RULE_OPTION_SECTION(Global),                                                             // opt/archive-push-queue-max
                                                                                                       // opt/archive-push-queue-max
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                 // opt/archive-push-queue-max
        (                                                                                              // opt/archive-push-queue-max
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                     // opt/archive-push-queue-max
        ),                                                                                             // opt/archive-push-queue-max
                                                                                                       // opt/archive-push-queue-max
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                // opt/archive-push-queue-max
        (                                                                                              // opt/archive-push-queue-max
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                     // opt/archive-push-queue-max
        ),                                                                                             // opt/archive-push-queue-max
                                                                                                       // opt/archive-push-queue-max
        PARSE_RULE_OPTIONAL                                                                            // opt/archive-push-queue-max
        (                                                                                              // opt/archive-push-queue-max
            PARSE_RULE_OPTIONAL_GROUP                                                                  // opt/archive-push-queue-max
            (                                                                                          // opt/archive-push-queue-max
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                        // opt/archive-push-queue-max
                (                                                                                      // opt/archive-push-queue-max
                    PARSE_RULE_VAL_SIZE(0B),                                                           // opt/archive-push-queue-max
                    PARSE_RULE_VAL_SIZE(4PiB),                                                         // opt/archive-push-queue-max
                ),                                                                                     // opt/archive-push-queue-max
            ),                                                                                         // opt/archive-push-queue-max
        ),                                                                                             // opt/archive-push-queue-max
    ),                                                                                                 // opt/archive-push-queue-max
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                         // opt/archive-timeout
    (                                                                                                         // opt/archive-timeout
        PARSE_RULE_OPTION_NAME("archive-timeout"),                                                            // opt/archive-timeout
        PARSE_RULE_OPTION_TYPE(Time),                                                                         // opt/archive-timeout
        PARSE_RULE_OPTION_RESET(true),                                                                        // opt/archive-timeout
        PARSE_RULE_OPTION_REQUIRED(true),                                                                     // opt/archive-timeout
        PARSE_RULE_OPTION_SECTION(Global),                                                                    // opt/archive-timeout
                                                                                                              // opt/archive-timeout
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                        // opt/archive-timeout
        (                                                                                                     // opt/archive-timeout
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                             // opt/archive-timeout
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                            // opt/archive-timeout
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                 // opt/archive-timeout
            PARSE_RULE_OPTION_COMMAND(Check)                                                                  // opt/archive-timeout
        ),                                                                                                    // opt/archive-timeout
                                                                                                              // opt/archive-timeout
        PARSE_RULE_OPTIONAL                                                                                   // opt/archive-timeout
        (                                                                                                     // opt/archive-timeout
            PARSE_RULE_OPTIONAL_GROUP                                                                         // opt/archive-timeout
            (                                                                                                 // opt/archive-timeout
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                               // opt/archive-timeout
                (                                                                                             // opt/archive-timeout
                    PARSE_RULE_VAL_TIME(100ms),                                                               // opt/archive-timeout
                    PARSE_RULE_VAL_TIME(1d),                                                                  // opt/archive-timeout
                ),                                                                                            // opt/archive-timeout
                                                                                                              // opt/archive-timeout
                PARSE_RULE_OPTIONAL_DEFAULT                                                                   // opt/archive-timeout
                (                                                                                             // opt/archive-timeout
                    PARSE_RULE_VAL_TIME(1m),                                                                  // opt/archive-timeout
                ),                                                                                            // opt/archive-timeout
            ),                                                                                                // opt/archive-timeout
        ),                                                                                                    // opt/archive-timeout
    ),                                                                                                        // opt/archive-timeout
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/backup-standby
    (                                                                                                          // opt/backup-standby
        PARSE_RULE_OPTION_NAME("backup-standby"),                                                              // opt/backup-standby
        PARSE_RULE_OPTION_TYPE(StringId),                                                                      // opt/backup-standby
        PARSE_RULE_OPTION_BOOL_LIKE(true),                                                                     // opt/backup-standby
        PARSE_RULE_OPTION_NEGATE(true),                                                                        // opt/backup-standby
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/backup-standby
        PARSE_RULE_OPTION_REQUIRED(true),                                                                      // opt/backup-standby
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/backup-standby
                                                                                                               // opt/backup-standby
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/backup-standby
        (                                                                                                      // opt/backup-standby
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/backup-standby
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/backup-standby
        ),                                                                                                     // opt/backup-standby
                                                                                                               // opt/backup-standby
        PARSE_RULE_OPTIONAL                                                                                    // opt/backup-standby
        (                                                                                                      // opt/backup-standby
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/backup-standby
            (                                                                                                  // opt/backup-standby
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                 // opt/backup-standby
                (                                                                                              // opt/backup-standby
                    PARSE_RULE_VAL_STRID(n),                                                                   // opt/backup-standby
                    PARSE_RULE_VAL_STRID(prefer),                                                              // opt/backup-standby
                    PARSE_RULE_VAL_STRID(y),                                                                   // opt/backup-standby
                ),                                                                                             // opt/backup-standby
                                                                                                               // opt/backup-standby
                PARSE_RULE_OPTIONAL_DEFAULT                                                                    // opt/backup-standby
                (                                                                                              // opt/backup-standby
                    PARSE_RULE_VAL_STRID(n),                                                                   // opt/backup-standby
                ),                                                                                             // opt/backup-standby
            ),                                                                                                 // opt/backup-standby
        ),                                                                                                     // opt/backup-standby
    ),                                                                                                         // opt/backup-standby
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                    // opt/beta
    (                                                                                                                    // opt/beta
        PARSE_RULE_OPTION_NAME("beta"),                                                                                  // opt/beta
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                                 // opt/beta
        PARSE_RULE_OPTION_INTERNAL(true)                                                                                 // opt/beta
        PARSE_RULE_OPTION_BETA(true),                                                                                    // opt/beta
        PARSE_RULE_OPTION_NEGATE(true),                                                                                  // opt/beta
        PARSE_RULE_OPTION_RESET(true),                                                                                   // opt/beta
        PARSE_RULE_OPTION_REQUIRED(true),                                                                                // opt/beta
        PARSE_RULE_OPTION_SECTION(Global),                                                                               // opt/beta
                                                                                                                         // opt/beta
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                   // opt/beta
        (                                                                                                                // opt/beta
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                          // opt/beta
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                        // opt/beta
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                       // opt/beta
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                            // opt/beta
            PARSE_RULE_OPTION_COMMAND(Check)                                                                             // opt/beta
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                            // opt/beta
            PARSE_RULE_OPTION_COMMAND(Info)                                                                              // opt/beta
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                          // opt/beta
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                           // opt/beta
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                            // opt/beta
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                           // opt/beta
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                            // opt/beta
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                           // opt/beta
            PARSE_RULE_OPTION_COMMAND(Server)                                                                            // opt/beta
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                                        // opt/beta
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                      // opt/beta
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                      // opt/beta
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                     // opt/beta
            PARSE_RULE_OPTION_COMMAND(Start)                                                                             // opt/beta
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                              // opt/beta
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                            // opt/beta
        ),                                                                                                               // opt/beta
                                                                                                                         // opt/beta
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                                  // opt/beta
        (                                                                                                                // opt/beta
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                        // opt/beta
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                       // opt/beta
        ),                                                                                                               // opt/beta
                                                                                                                         // opt/beta
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                                  // opt/beta
        (                                                                                                                // opt/beta
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                        // opt/beta
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                       // opt/beta
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                            // opt/beta
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                           // opt/beta
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                            // opt/beta
        ),                                                                                                               // opt/beta
                                                                                                                         // opt/beta
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                                 // opt/beta
        (                                                                                                                // opt/beta
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                          // opt/beta
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                        // opt/beta
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                       // opt/beta
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                            // opt/beta
            PARSE_RULE_OPTION_COMMAND(Check)                                                                             // opt/beta
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                            // opt/beta
            PARSE_RULE_OPTION_COMMAND(Info)                                                                              // opt/beta
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                          // opt/beta
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                           // opt/beta
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                            // opt/beta
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                           // opt/beta
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                            // opt/beta
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                           // opt/beta
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                      // opt/beta
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                      // opt/beta
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                     // opt/beta
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                            // opt/beta
        ),                                                                                                               // opt/beta
                                                                                                                         // opt/beta
        PARSE_RULE_OPTIONAL                                                                                              // opt/beta
        (                                                                                                                // opt/beta
            PARSE_RULE_OPTIONAL_GROUP                                                                                    // opt/beta
            (                                                                                                            // opt/beta
                PARSE_RULE_OPTIONAL_DEFAULT                                                                              // opt/beta
                (                                                                                                        // opt/beta
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                           // opt/beta
                ),                                                                                                       // opt/beta
            ),                                                                                                           // opt/beta
        ),                                                                                                               // opt/beta
    ),                                                                                                                   // opt/beta
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                             // opt/buffer-size
    (                                                                                                             // opt/buffer-size
        PARSE_RULE_OPTION_NAME("buffer-size"),                                                                    // opt/buffer-size
        PARSE_RULE_OPTION_TYPE(Size),                                                                             // opt/buffer-size
        PARSE_RULE_OPTION_RESET(true),                                                                            // opt/buffer-size
        PARSE_RULE_OPTION_REQUIRED(true),                                                                         // opt/buffer-size
        PARSE_RULE_OPTION_SECTION(Global),                                                                        // opt/buffer-size
                                                                                                                  // opt/buffer-size
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                            // opt/buffer-size
        (                                                                                                         // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                   // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Check)                                                                      // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                     // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Info)                                                                       // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                   // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                    // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                     // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                    // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                     // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                    // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Server)                                                                     // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                                 // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                               // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                               // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                              // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                     // opt/buffer-size
        ),                                                                                                        // opt/buffer-size
                                                                                                                  // opt/buffer-size
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                           // opt/buffer-size
        (                                                                                                         // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/buffer-size
        ),                                                                                                        // opt/buffer-size
                                                                                                                  // opt/buffer-size
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                           // opt/buffer-size
        (                                                                                                         // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                    // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                     // opt/buffer-size
        ),                                                                                                        // opt/buffer-size
                                                                                                                  // opt/buffer-size
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                          // opt/buffer-size
        (                                                                                                         // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                   // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Check)                                                                      // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                     // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Info)                                                                       // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                   // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                    // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                     // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                    // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                     // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                    // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                               // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                               // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                              // opt/buffer-size
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                     // opt/buffer-size
        ),                                                                                                        // opt/buffer-size
                                                                                                                  // opt/buffer-size
        PARSE_RULE_OPTIONAL                                                                                       // opt/buffer-size
        (                                                                                                         // opt/buffer-size
            PARSE_RULE_OPTIONAL_GROUP                                                                             // opt/buffer-size
            (                                                                                                     // opt/buffer-size
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                    // opt/buffer-size
                (                                                                                                 // opt/buffer-size
                    PARSE_RULE_VAL_SIZE(16KiB),                                                                   // opt/buffer-size
                    PARSE_RULE_VAL_SIZE(32KiB),                                                                   // opt/buffer-size
                    PARSE_RULE_VAL_SIZE(64KiB),                                                                   // opt/buffer-size
                    PARSE_RULE_VAL_SIZE(128KiB),                                                                  // opt/buffer-size
                    PARSE_RULE_VAL_SIZE(256KiB),                                                                  // opt/buffer-size
                    PARSE_RULE_VAL_SIZE(512KiB),                                                                  // opt/buffer-size
                    PARSE_RULE_VAL_SIZE(1MiB),                                                                    // opt/buffer-size
                    PARSE_RULE_VAL_SIZE(2MiB),                                                                    // opt/buffer-size
                    PARSE_RULE_VAL_SIZE(4MiB),                                                                    // opt/buffer-size
                    PARSE_RULE_VAL_SIZE(8MiB),                                                                    // opt/buffer-size
                    PARSE_RULE_VAL_SIZE(16MiB),                                                                   // opt/buffer-size
                ),                                                                                                // opt/buffer-size
                                                                                                                  // opt/buffer-size
                PARSE_RULE_OPTIONAL_DEFAULT                                                                       // opt/buffer-size
                (                                                                                                 // opt/buffer-size
                    PARSE_RULE_VAL_SIZE(1MiB),                                                                    // opt/buffer-size
                ),                                                                                                // opt/buffer-size
            ),                                                                                                    // opt/buffer-size
        ),                                                                                                        // opt/buffer-size
    ),                                                                                                            // opt/buffer-size
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                           // opt/checksum-page
    (                                                                                                           // opt/checksum-page
        PARSE_RULE_OPTION_NAME("checksum-page"),                                                                // opt/checksum-page
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                        // opt/checksum-page
        PARSE_RULE_OPTION_NEGATE(true),                                                                         // opt/checksum-page
        PARSE_RULE_OPTION_RESET(true),                                                                          // opt/checksum-page
        PARSE_RULE_OPTION_REQUIRED(false),                                                                      // opt/checksum-page
        PARSE_RULE_OPTION_SECTION(Global),                                                                      // opt/checksum-page
                                                                                                                // opt/checksum-page
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                          // opt/checksum-page
        (                                                                                                       // opt/checksum-page
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/checksum-page
        ),                                                                                                      // opt/checksum-page
    ),                                                                                                          // opt/checksum-page
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                             // opt/cipher-pass
    (                                                                                                             // opt/cipher-pass
        PARSE_RULE_OPTION_NAME("cipher-pass"),                                                                    // opt/cipher-pass
        PARSE_RULE_OPTION_TYPE(String),                                                                           // opt/cipher-pass
        PARSE_RULE_OPTION_INTERNAL(true)                                                                          // opt/cipher-pass
        PARSE_RULE_OPTION_REQUIRED(false),                                                                        // opt/cipher-pass
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                   // opt/cipher-pass
                                                                                                                  // opt/cipher-pass
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                            // opt/cipher-pass
        (                                                                                                         // opt/cipher-pass
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                    // opt/cipher-pass
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                    // opt/cipher-pass
        ),                                                                                                        // opt/cipher-pass
    ),                                                                                                            // opt/cipher-pass
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                     // opt/cmd
    (                                                                                                                     // opt/cmd
        PARSE_RULE_OPTION_NAME("cmd"),                                                                                    // opt/cmd
        PARSE_RULE_OPTION_TYPE(String),                                                                                   // opt/cmd
        PARSE_RULE_OPTION_DEFAULT_TYPE(Dynamic),                                                                          // opt/cmd
        PARSE_RULE_OPTION_RESET(true),                                                                                    // opt/cmd
        PARSE_RULE_OPTION_REQUIRED(true),                                                                                 // opt/cmd
        PARSE_RULE_OPTION_SECTION(Global),                                                                                // opt/cmd
                                                                                                                          // opt/cmd
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                    // opt/cmd
        (                                                                                                                 // opt/cmd
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                         // opt/cmd
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                        // opt/cmd
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                             // opt/cmd
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                            // opt/cmd
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                             // opt/cmd
        ),                                                                                                                // opt/cmd
                                                                                                                          // opt/cmd
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                                   // opt/cmd
        (                                                                                                                 // opt/cmd
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                         // opt/cmd
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                        // opt/cmd
        ),                                                                                                                // opt/cmd
                                                                                                                          // opt/cmd
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                                   // opt/cmd
        (                                                                                                                 // opt/cmd
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                         // opt/cmd
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                        // opt/cmd
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                             // opt/cmd
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                            // opt/cmd
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                             // opt/cmd
        ),                                                                                                                // opt/cmd
                                                                                                                          // opt/cmd
        PARSE_RULE_OPTIONAL                                                                                               // opt/cmd
        (                                                                                                                 // opt/cmd
            PARSE_RULE_OPTIONAL_GROUP                                                                                     // opt/cmd
            (                                                                                                             // opt/cmd
                PARSE_RULE_OPTIONAL_DEFAULT                                                                               // opt/cmd
                (                                                                                                         // opt/cmd
                    PARSE_RULE_DEFAULT_DYNAMIC(Bin),                                                                      // opt/cmd
                ),                                                                                                        // opt/cmd
            ),                                                                                                            // opt/cmd
        ),                                                                                                                // opt/cmd
    ),                                                                                                                    // opt/cmd
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                 // opt/cmd-ssh
    (                                                                                                                 // opt/cmd-ssh
        PARSE_RULE_OPTION_NAME("cmd-ssh"),                                                                            // opt/cmd-ssh
        PARSE_RULE_OPTION_TYPE(String),                                                                               // opt/cmd-ssh
        PARSE_RULE_OPTION_RESET(true),                                                                                // opt/cmd-ssh
        PARSE_RULE_OPTION_REQUIRED(true),                                                                             // opt/cmd-ssh
        PARSE_RULE_OPTION_SECTION(Global),                                                                            // opt/cmd-ssh
                                                                                                                      // opt/cmd-ssh
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                // opt/cmd-ssh
        (                                                                                                             // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                       // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(Check)                                                                          // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                         // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(Info)                                                                           // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                       // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                        // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                         // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                        // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                         // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                        // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                   // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                   // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                  // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                         // opt/cmd-ssh
        ),                                                                                                            // opt/cmd-ssh
                                                                                                                      // opt/cmd-ssh
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                               // opt/cmd-ssh
        (                                                                                                             // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/cmd-ssh
        ),                                                                                                            // opt/cmd-ssh
                                                                                                                      // opt/cmd-ssh
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                               // opt/cmd-ssh
        (                                                                                                             // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                        // opt/cmd-ssh
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                         // opt/cmd-ssh
        ),                                                                                                            // opt/cmd-ssh
                                                                                                                      // opt/cmd-ssh
        PARSE_RULE_OPTIONAL                                                                                           // opt/cmd-ssh
        (                                                                                                             // opt/cmd-ssh
            PARSE_RULE_OPTIONAL_GROUP                                                                                 // opt/cmd-ssh
            (                                                                                                         // opt/cmd-ssh
                PARSE_RULE_OPTIONAL_DEFAULT                                                                           // opt/cmd-ssh
                (                                                                                                     // opt/cmd-ssh
                    PARSE_RULE_VAL_STR(QT_ssh_QT),                                                                    // opt/cmd-ssh
                ),                                                                                                    // opt/cmd-ssh
            ),                                                                                                        // opt/cmd-ssh
        ),                                                                                                            // opt/cmd-ssh
    ),                                                                                                                // opt/cmd-ssh
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                // opt/compress
    (                                                                                                                // opt/compress
        PARSE_RULE_OPTION_NAME("compress"),                                                                          // opt/compress
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                             // opt/compress
        PARSE_RULE_OPTION_NEGATE(true),                                                                              // opt/compress
        PARSE_RULE_OPTION_RESET(true),                                                                               // opt/compress
        PARSE_RULE_OPTION_REQUIRED(true),                                                                            // opt/compress
        PARSE_RULE_OPTION_SECTION(Global),                                                                           // opt/compress
                                                                                                                     // opt/compress
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                               // opt/compress
        (                                                                                                            // opt/compress
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                   // opt/compress
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                        // opt/compress
        ),                                                                                                           // opt/compress
                                                                                                                     // opt/compress
        PARSE_RULE_OPTIONAL                                                                                          // opt/compress
        (                                                                                                            // opt/compress
            PARSE_RULE_OPTIONAL_GROUP                                                                                // opt/compress
            (                                                                                                        // opt/compress
                PARSE_RULE_OPTIONAL_DEFAULT                                                                          // opt/compress
                (                                                                                                    // opt/compress
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                        // opt/compress
                ),                                                                                                   // opt/compress
            ),                                                                                                       // opt/compress
        ),                                                                                                           // opt/compress
    ),                                                                                                               // opt/compress
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/compress-level
    (                                                                                                          // opt/compress-level
        PARSE_RULE_OPTION_NAME("compress-level"),                                                              // opt/compress-level
        PARSE_RULE_OPTION_TYPE(Integer),                                                                       // opt/compress-level
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/compress-level
        PARSE_RULE_OPTION_REQUIRED(false),                                                                     // opt/compress-level
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/compress-level
                                                                                                               // opt/compress-level
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/compress-level
        (                                                                                                      // opt/compress-level
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/compress-level
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/compress-level
        ),                                                                                                     // opt/compress-level
                                                                                                               // opt/compress-level
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                        // opt/compress-level
        (                                                                                                      // opt/compress-level
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/compress-level
        ),                                                                                                     // opt/compress-level
                                                                                                               // opt/compress-level
        PARSE_RULE_OPTIONAL                                                                                    // opt/compress-level
        (                                                                                                      // opt/compress-level
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/compress-level
            (                                                                                                  // opt/compress-level
                PARSE_RULE_OPTIONAL_DEPEND                                                                     // opt/compress-level
                (                                                                                              // opt/compress-level
                    PARSE_RULE_OPTIONAL_DEPEND_DEFAULT(PARSE_RULE_VAL_INT(0)),                                 // opt/compress-level
                    PARSE_RULE_VAL_OPT(CompressType),                                                          // opt/compress-level
                    PARSE_RULE_VAL_STRID(bz2),                                                                 // opt/compress-level
                    PARSE_RULE_VAL_STRID(gz),                                                                  // opt/compress-level
                    PARSE_RULE_VAL_STRID(lz4),                                                                 // opt/compress-level
                    PARSE_RULE_VAL_STRID(zst),                                                                 // opt/compress-level
                ),                                                                                             // opt/compress-level
                                                                                                               // opt/compress-level
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                // opt/compress-level
                (                                                                                              // opt/compress-level
                    PARSE_RULE_OPTIONAL_ALLOW_RANGE_MAP                                                        // opt/compress-level
                    (                                                                                          // opt/compress-level
                        PARSE_RULE_VAL_STRID(bz2),                                                             // opt/compress-level
                        PARSE_RULE_VAL_INT(1),                                                                 // opt/compress-level
                        PARSE_RULE_VAL_INT(9),                                                                 // opt/compress-level
                        PARSE_RULE_VAL_STRID(gz),                                                              // opt/compress-level
                        PARSE_RULE_VAL_INT(DS_1),                                                              // opt/compress-level
                        PARSE_RULE_VAL_INT(9),                                                                 // opt/compress-level
                        PARSE_RULE_VAL_STRID(lz4),                                                             // opt/compress-level
                        PARSE_RULE_VAL_INT(DS_5),                                                              // opt/compress-level
                        PARSE_RULE_VAL_INT(12),                                                                // opt/compress-level
                        PARSE_RULE_VAL_STRID(zst),                                                             // opt/compress-level
                        PARSE_RULE_VAL_INT(DS_7),                                                              // opt/compress-level
                        PARSE_RULE_VAL_INT(22),                                                                // opt/compress-level
                    ),                                                                                         // opt/compress-level
                ),                                                                                             // opt/compress-level
                                                                                                               // opt/compress-level
                PARSE_RULE_OPTIONAL_DEFAULT                                                                    // opt/compress-level
                (                                                                                              // opt/compress-level
                    PARSE_RULE_OPTIONAL_DEFAULT_MAP                                                            // opt/compress-level
                    (                                                                                          // opt/compress-level
                        PARSE_RULE_VAL_STRID(bz2),                                                             // opt/compress-level
                        PARSE_RULE_VAL_INT(9),                                                                 // opt/compress-level
                        PARSE_RULE_VAL_STRID(gz),                                                              // opt/compress-level
                        PARSE_RULE_VAL_INT(6),                                                                 // opt/compress-level
                        PARSE_RULE_VAL_STRID(lz4),                                                             // opt/compress-level
                        PARSE_RULE_VAL_INT(1),                                                                 // opt/compress-level
                        PARSE_RULE_VAL_STRID(zst),                                                             // opt/compress-level
                        PARSE_RULE_VAL_INT(3),                                                                 // opt/compress-level
                    ),                                                                                         // opt/compress-level
                ),                                                                                             // opt/compress-level
            ),                                                                                                 // opt/compress-level
        ),                                                                                                     // opt/compress-level
    ),                                                                                                         // opt/compress-level
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                  // opt/compress-level-network
    (                                                                                                  // opt/compress-level-network
        PARSE_RULE_OPTION_NAME("compress-level-network"),                                              // opt/compress-level-network
        PARSE_RULE_OPTION_TYPE(Integer),                                                               // opt/compress-level-network
        PARSE_RULE_OPTION_RESET(true),                                                                 // opt/compress-level-network
        PARSE_RULE_OPTION_REQUIRED(true),                                                              // opt/compress-level-network
        PARSE_RULE_OPTION_SECTION(Global),                                                             // opt/compress-level-network
                                                                                                       // opt/compress-level-network
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                 // opt/compress-level-network
        (                                                                                              // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                        // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                      // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                     // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(Backup)                                                          // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(Check)                                                           // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(Expire)                                                          // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(Info)                                                            // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                        // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                         // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                          // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                         // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                          // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(Restore)                                                         // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                    // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                    // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                   // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(Verify)                                                          // opt/compress-level-network
        ),                                                                                             // opt/compress-level-network
                                                                                                       // opt/compress-level-network
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                // opt/compress-level-network
        (                                                                                              // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                      // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                     // opt/compress-level-network
        ),                                                                                             // opt/compress-level-network
                                                                                                       // opt/compress-level-network
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                // opt/compress-level-network
        (                                                                                              // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                      // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                     // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(Backup)                                                          // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(Restore)                                                         // opt/compress-level-network
            PARSE_RULE_OPTION_COMMAND(Verify)                                                          // opt/compress-level-network
        ),                                                                                             // opt/compress-level-network
                                                                                                       // opt/compress-level-network
        PARSE_RULE_OPTIONAL                                                                            // opt/compress-level-network
        (                                                                                              // opt/compress-level-network
            PARSE_RULE_OPTIONAL_GROUP                                                                  // opt/compress-level-network
            (                                                                                          // opt/compress-level-network
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                        // opt/compress-level-network
                (                                                                                      // opt/compress-level-network
                    PARSE_RULE_VAL_INT(DS_5),                                                          // opt/compress-level-network
                    PARSE_RULE_VAL_INT(12),                                                            // opt/compress-level-network
                ),                                                                                     // opt/compress-level-network
                                                                                                       // opt/compress-level-network
                PARSE_RULE_OPTIONAL_DEFAULT                                                            // opt/compress-level-network
                (                                                                                      // opt/compress-level-network
                    PARSE_RULE_VAL_INT(1),                                                             // opt/compress-level-network
                ),                                                                                     // opt/compress-level-network
            ),                                                                                         // opt/compress-level-network
        ),                                                                                             // opt/compress-level-network
    ),                                                                                                 // opt/compress-level-network
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                           // opt/compress-type
    (                                                                                                           // opt/compress-type
        PARSE_RULE_OPTION_NAME("compress-type"),                                                                // opt/compress-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                                       // opt/compress-type
        PARSE_RULE_OPTION_RESET(true),                                                                          // opt/compress-type
        PARSE_RULE_OPTION_REQUIRED(true),                                                                       // opt/compress-type
        PARSE_RULE_OPTION_SECTION(Global),                                                                      // opt/compress-type
                                                                                                                // opt/compress-type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                          // opt/compress-type
        (                                                                                                       // opt/compress-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/compress-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/compress-type
        ),                                                                                                      // opt/compress-type
                                                                                                                // opt/compress-type
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                         // opt/compress-type
        (                                                                                                       // opt/compress-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/compress-type
        ),                                                                                                      // opt/compress-type
                                                                                                                // opt/compress-type
        PARSE_RULE_OPTIONAL                                                                                     // opt/compress-type
        (                                                                                                       // opt/compress-type
            PARSE_RULE_OPTIONAL_GROUP                                                                           // opt/compress-type
            (                                                                                                   // opt/compress-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                  // opt/compress-type
                (                                                                                               // opt/compress-type
                    PARSE_RULE_VAL_STRID(none),                                                                 // opt/compress-type
                    PARSE_RULE_VAL_STRID(bz2),                                                                  // opt/compress-type
                    PARSE_RULE_VAL_STRID(gz),                                                                   // opt/compress-type
                    PARSE_RULE_VAL_STRID(lz4),                                                                  // opt/compress-type
                    PARSE_RULE_VAL_STRID(zst),                                                                  // opt/compress-type
#ifndef HAVE_LIBZST                                                                                             // opt/compress-type
                        PARSE_RULE_BOOL_FALSE,                                                                  // opt/compress-type
#endif                                                                                                          // opt/compress-type
                ),                                                                                              // opt/compress-type
                                                                                                                // opt/compress-type
                PARSE_RULE_OPTIONAL_DEFAULT                                                                     // opt/compress-type
                (                                                                                               // opt/compress-type
                    PARSE_RULE_VAL_STRID(gz),                                                                   // opt/compress-type
                ),                                                                                              // opt/compress-type
            ),                                                                                                  // opt/compress-type
        ),                                                                                                      // opt/compress-type
    ),                                                                                                          // opt/compress-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                  // opt/config
    (                                                                                                                  // opt/config
        PARSE_RULE_OPTION_NAME("config"),                                                                              // opt/config
        PARSE_RULE_OPTION_TYPE(String),                                                                                // opt/config
        PARSE_RULE_OPTION_NEGATE(true),                                                                                // opt/config
        PARSE_RULE_OPTION_REQUIRED(true),                                                                              // opt/config
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                        // opt/config
                                                                                                                       // opt/config
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                 // opt/config
        (                                                                                                              // opt/config
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                        // opt/config
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                      // opt/config
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                     // opt/config
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                          // opt/config
            PARSE_RULE_OPTION_COMMAND(Check)                                                                           // opt/config
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                          // opt/config
            PARSE_RULE_OPTION_COMMAND(Info)                                                                            // opt/config
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                        // opt/config
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                         // opt/config
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                          // opt/config
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                         // opt/config
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                          // opt/config
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                         // opt/config
            PARSE_RULE_OPTION_COMMAND(Server)                                                                          // opt/config
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                                      // opt/config
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                    // opt/config
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                    // opt/config
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                   // opt/config
            PARSE_RULE_OPTION_COMMAND(Start)                                                                           // opt/config
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                            // opt/config
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                          // opt/config
        ),                                                                                                             // opt/config
                                                                                                                       // opt/config
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                                // opt/config
        (                                                                                                              // opt/config
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                      // opt/config
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                     // opt/config
        ),                                                                                                             // opt/config
                                                                                                                       // opt/config
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                                // opt/config
        (                                                                                                              // opt/config
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                      // opt/config
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                     // opt/config
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                          // opt/config
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                         // opt/config
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                          // opt/config
        ),                                                                                                             // opt/config
                                                                                                                       // opt/config
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                               // opt/config
        (                                                                                                              // opt/config
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                        // opt/config
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                      // opt/config
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                     // opt/config
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                          // opt/config
            PARSE_RULE_OPTION_COMMAND(Check)                                                                           // opt/config
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                          // opt/config
            PARSE_RULE_OPTION_COMMAND(Info)                                                                            // opt/config
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                        // opt/config
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                         // opt/config
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                          // opt/config
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                         // opt/config
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                          // opt/config
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                         // opt/config
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                    // opt/config
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                    // opt/config
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                   // opt/config
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                          // opt/config
        ),                                                                                                             // opt/config
                                                                                                                       // opt/config
        PARSE_RULE_OPTIONAL                                                                                            // opt/config
        (                                                                                                              // opt/config
            PARSE_RULE_OPTIONAL_GROUP                                                                                  // opt/config
            (                                                                                                          // opt/config
                PARSE_RULE_OPTIONAL_DEFAULT                                                                            // opt/config
                (                                                                                                      // opt/config
                    PARSE_RULE_VAL_STR(CFGOPTDEF_CONFIG_PATH_SP_QT_FS_QT_SP_PROJECT_CONFIG_FILE),                      // opt/config
                ),                                                                                                     // opt/config
            ),                                                                                                         // opt/config
        ),                                                                                                             // opt/config
    ),                                                                                                                 // opt/config
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                     // opt/config-include-path
    (                                                                                                     // opt/config-include-path
        PARSE_RULE_OPTION_NAME("config-include-path"),                                                    // opt/config-include-path
        PARSE_RULE_OPTION_TYPE(Path),                                                                     // opt/config-include-path
        PARSE_RULE_OPTION_REQUIRED(true),                                                                 // opt/config-include-path
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                           // opt/config-include-path
                                                                                                          // opt/config-include-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                    // opt/config-include-path
        (                                                                                                 // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Server)                                                             // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                         // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Start)                                                              // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Stop)                                                               // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/config-include-path
        ),                                                                                                // opt/config-include-path
                                                                                                          // opt/config-include-path
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                   // opt/config-include-path
        (                                                                                                 // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/config-include-path
        ),                                                                                                // opt/config-include-path
                                                                                                          // opt/config-include-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                   // opt/config-include-path
        (                                                                                                 // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/config-include-path
        ),                                                                                                // opt/config-include-path
                                                                                                          // opt/config-include-path
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                  // opt/config-include-path
        (                                                                                                 // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/config-include-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/config-include-path
        ),                                                                                                // opt/config-include-path
                                                                                                          // opt/config-include-path
        PARSE_RULE_OPTIONAL                                                                               // opt/config-include-path
        (                                                                                                 // opt/config-include-path
            PARSE_RULE_OPTIONAL_GROUP                                                                     // opt/config-include-path
            (                                                                                             // opt/config-include-path
                PARSE_RULE_OPTIONAL_DEFAULT                                                               // opt/config-include-path
                (                                                                                         // opt/config-include-path
                    PARSE_RULE_VAL_STR(CFGOPTDEF_CONFIG_PATH_SP_QT_FS_QT_SP_PROJECT_CONFIG_INCLUDE_PATH), // opt/config-include-path
                ),                                                                                        // opt/config-include-path
            ),                                                                                            // opt/config-include-path
        ),                                                                                                // opt/config-include-path
    ),                                                                                                    // opt/config-include-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                             // opt/config-path
    (                                                                                                             // opt/config-path
        PARSE_RULE_OPTION_NAME("config-path"),                                                                    // opt/config-path
        PARSE_RULE_OPTION_TYPE(Path),                                                                             // opt/config-path
        PARSE_RULE_OPTION_REQUIRED(true),                                                                         // opt/config-path
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                   // opt/config-path
                                                                                                                  // opt/config-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                            // opt/config-path
        (                                                                                                         // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                   // opt/config-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/config-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                                      // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                     // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                                       // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                   // opt/config-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                    // opt/config-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                     // opt/config-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                    // opt/config-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                     // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                    // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Server)                                                                     // opt/config-path
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                                 // opt/config-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                               // opt/config-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                               // opt/config-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                              // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Start)                                                                      // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                       // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                     // opt/config-path
        ),                                                                                                        // opt/config-path
                                                                                                                  // opt/config-path
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                           // opt/config-path
        (                                                                                                         // opt/config-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/config-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/config-path
        ),                                                                                                        // opt/config-path
                                                                                                                  // opt/config-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                           // opt/config-path
        (                                                                                                         // opt/config-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/config-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                    // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                     // opt/config-path
        ),                                                                                                        // opt/config-path
                                                                                                                  // opt/config-path
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                          // opt/config-path
        (                                                                                                         // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                   // opt/config-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/config-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                                      // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                     // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                                       // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                   // opt/config-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                    // opt/config-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                     // opt/config-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                    // opt/config-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                     // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                    // opt/config-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                               // opt/config-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                               // opt/config-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                              // opt/config-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                     // opt/config-path
        ),                                                                                                        // opt/config-path
                                                                                                                  // opt/config-path
        PARSE_RULE_OPTIONAL                                                                                       // opt/config-path
        (                                                                                                         // opt/config-path
            PARSE_RULE_OPTIONAL_GROUP                                                                             // opt/config-path
            (                                                                                                     // opt/config-path
                PARSE_RULE_OPTIONAL_DEFAULT                                                                       // opt/config-path
                (                                                                                                 // opt/config-path
                    PARSE_RULE_VAL_STR(CFGOPTDEF_CONFIG_PATH),                                                    // opt/config-path
                ),                                                                                                // opt/config-path
            ),                                                                                                    // opt/config-path
        ),                                                                                                        // opt/config-path
    ),                                                                                                            // opt/config-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                              // opt/db-exclude
    (                                                                                                              // opt/db-exclude
        PARSE_RULE_OPTION_NAME("db-exclude"),                                                                      // opt/db-exclude
        PARSE_RULE_OPTION_TYPE(List),                                                                              // opt/db-exclude
        PARSE_RULE_OPTION_RESET(true),                                                                             // opt/db-exclude
        PARSE_RULE_OPTION_REQUIRED(false),                                                                         // opt/db-exclude
        PARSE_RULE_OPTION_SECTION(Global),                                                                         // opt/db-exclude
        PARSE_RULE_OPTION_MULTI(true),                                                                             // opt/db-exclude
                                                                                                                   // opt/db-exclude
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                             // opt/db-exclude
        (                                                                                                          // opt/db-exclude
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                     // opt/db-exclude
        ),                                                                                                         // opt/db-exclude
    ),                                                                                                             // opt/db-exclude
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                              // opt/db-include
    (                                                                                                              // opt/db-include
        PARSE_RULE_OPTION_NAME("db-include"),                                                                      // opt/db-include
        PARSE_RULE_OPTION_TYPE(List),                                                                              // opt/db-include
        PARSE_RULE_OPTION_RESET(true),                                                                             // opt/db-include
        PARSE_RULE_OPTION_REQUIRED(false),                                                                         // opt/db-include
        PARSE_RULE_OPTION_SECTION(Global),                                                                         // opt/db-include
        PARSE_RULE_OPTION_MULTI(true),                                                                             // opt/db-include
                                                                                                                   // opt/db-include
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                             // opt/db-include
        (                                                                                                          // opt/db-include
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                     // opt/db-include
        ),                                                                                                         // opt/db-include
    ),                                                                                                             // opt/db-include
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                              // opt/db-timeout
    (                                                                                                              // opt/db-timeout
        PARSE_RULE_OPTION_NAME("db-timeout"),                                                                      // opt/db-timeout
        PARSE_RULE_OPTION_TYPE(Time),                                                                              // opt/db-timeout
        PARSE_RULE_OPTION_RESET(true),                                                                             // opt/db-timeout
        PARSE_RULE_OPTION_REQUIRED(true),                                                                          // opt/db-timeout
        PARSE_RULE_OPTION_SECTION(Global),                                                                         // opt/db-timeout
                                                                                                                   // opt/db-timeout
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                             // opt/db-timeout
        (                                                                                                          // opt/db-timeout
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                      // opt/db-timeout
            PARSE_RULE_OPTION_COMMAND(Check)                                                                       // opt/db-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                // opt/db-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                // opt/db-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                               // opt/db-timeout
        ),                                                                                                         // opt/db-timeout
                                                                                                                   // opt/db-timeout
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                            // opt/db-timeout
        (                                                                                                          // opt/db-timeout
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                      // opt/db-timeout
        ),                                                                                                         // opt/db-timeout
                                                                                                                   // opt/db-timeout
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                           // opt/db-timeout
        (                                                                                                          // opt/db-timeout
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                      // opt/db-timeout
            PARSE_RULE_OPTION_COMMAND(Check)                                                                       // opt/db-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                // opt/db-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                // opt/db-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                               // opt/db-timeout
        ),                                                                                                         // opt/db-timeout
                                                                                                                   // opt/db-timeout
        PARSE_RULE_OPTIONAL                                                                                        // opt/db-timeout
        (                                                                                                          // opt/db-timeout
            PARSE_RULE_OPTIONAL_GROUP                                                                              // opt/db-timeout
            (                                                                                                      // opt/db-timeout
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                    // opt/db-timeout
                (                                                                                                  // opt/db-timeout
                    PARSE_RULE_VAL_TIME(100ms),                                                                    // opt/db-timeout
                    PARSE_RULE_VAL_TIME(7d),                                                                       // opt/db-timeout
                ),                                                                                                 // opt/db-timeout
                                                                                                                   // opt/db-timeout
                PARSE_RULE_OPTIONAL_DEFAULT                                                                        // opt/db-timeout
                (                                                                                                  // opt/db-timeout
                    PARSE_RULE_VAL_TIME(30m),                                                                      // opt/db-timeout
                ),                                                                                                 // opt/db-timeout
            ),                                                                                                     // opt/db-timeout
        ),                                                                                                         // opt/db-timeout
    ),                                                                                                             // opt/db-timeout
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                   // opt/delta
    (                                                                                                                   // opt/delta
        PARSE_RULE_OPTION_NAME("delta"),                                                                                // opt/delta
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                                // opt/delta
        PARSE_RULE_OPTION_NEGATE(true),                                                                                 // opt/delta
        PARSE_RULE_OPTION_RESET(true),                                                                                  // opt/delta
        PARSE_RULE_OPTION_REQUIRED(true),                                                                               // opt/delta
        PARSE_RULE_OPTION_SECTION(Global),                                                                              // opt/delta
                                                                                                                        // opt/delta
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                  // opt/delta
        (                                                                                                               // opt/delta
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                           // opt/delta
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                          // opt/delta
        ),                                                                                                              // opt/delta
                                                                                                                        // opt/delta
        PARSE_RULE_OPTIONAL                                                                                             // opt/delta
        (                                                                                                               // opt/delta
            PARSE_RULE_OPTIONAL_GROUP                                                                                   // opt/delta
            (                                                                                                           // opt/delta
                PARSE_RULE_OPTIONAL_DEFAULT                                                                             // opt/delta
                (                                                                                                       // opt/delta
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                          // opt/delta
                ),                                                                                                      // opt/delta
            ),                                                                                                          // opt/delta
        ),                                                                                                              // opt/delta
    ),                                                                                                                  // opt/delta
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                            // opt/detail-level
    (                                                                                                            // opt/detail-level
        PARSE_RULE_OPTION_NAME("detail-level"),                                                                  // opt/detail-level
        PARSE_RULE_OPTION_TYPE(StringId),                                                                        // opt/detail-level
        PARSE_RULE_OPTION_REQUIRED(true),                                                                        // opt/detail-level
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                  // opt/detail-level
                                                                                                                 // opt/detail-level
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                           // opt/detail-level
        (                                                                                                        // opt/detail-level
            PARSE_RULE_OPTION_COMMAND(Info)                                                                      // opt/detail-level
        ),                                                                                                       // opt/detail-level
                                                                                                                 // opt/detail-level
        PARSE_RULE_OPTIONAL                                                                                      // opt/detail-level
        (                                                                                                        // opt/detail-level
            PARSE_RULE_OPTIONAL_GROUP                                                                            // opt/detail-level
            (                                                                                                    // opt/detail-level
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                   // opt/detail-level
                (                                                                                                // opt/detail-level
                    PARSE_RULE_VAL_STRID(progress),                                                              // opt/detail-level
                    PARSE_RULE_VAL_STRID(full),                                                                  // opt/detail-level
                ),                                                                                               // opt/detail-level
                                                                                                                 // opt/detail-level
                PARSE_RULE_OPTIONAL_DEFAULT                                                                      // opt/detail-level
                (                                                                                                // opt/detail-level
                    PARSE_RULE_VAL_STRID(full),                                                                  // opt/detail-level
                ),                                                                                               // opt/detail-level
            ),                                                                                                   // opt/detail-level
        ),                                                                                                       // opt/detail-level
    ),                                                                                                           // opt/detail-level
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                 // opt/dry-run
    (                                                                                                                 // opt/dry-run
        PARSE_RULE_OPTION_NAME("dry-run"),                                                                            // opt/dry-run
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                              // opt/dry-run
        PARSE_RULE_OPTION_REQUIRED(true),                                                                             // opt/dry-run
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                       // opt/dry-run
                                                                                                                      // opt/dry-run
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                // opt/dry-run
        (                                                                                                             // opt/dry-run
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                         // opt/dry-run
        ),                                                                                                            // opt/dry-run
                                                                                                                      // opt/dry-run
        PARSE_RULE_OPTIONAL                                                                                           // opt/dry-run
        (                                                                                                             // opt/dry-run
            PARSE_RULE_OPTIONAL_GROUP                                                                                 // opt/dry-run
            (                                                                                                         // opt/dry-run
                PARSE_RULE_OPTIONAL_DEFAULT                                                                           // opt/dry-run
                (                                                                                                     // opt/dry-run
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                        // opt/dry-run
                ),                                                                                                    // opt/dry-run
            ),                                                                                                        // opt/dry-run
        ),                                                                                                            // opt/dry-run
    ),                                                                                                                // opt/dry-run
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                 // opt/exclude
    (                                                                                                                 // opt/exclude
        PARSE_RULE_OPTION_NAME("exclude"),                                                                            // opt/exclude
        PARSE_RULE_OPTION_TYPE(List),                                                                                 // opt/exclude
        PARSE_RULE_OPTION_RESET(true),                                                                                // opt/exclude
        PARSE_RULE_OPTION_REQUIRED(false),                                                                            // opt/exclude
        PARSE_RULE_OPTION_SECTION(Global),                                                                            // opt/exclude
        PARSE_RULE_OPTION_MULTI(true),                                                                                // opt/exclude
                                                                                                                      // opt/exclude
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                // opt/exclude
        (                                                                                                             // opt/exclude
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/exclude
        ),                                                                                                            // opt/exclude
    ),                                                                                                                // opt/exclude
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                 // opt/exec-id
    (                                                                                                                 // opt/exec-id
        PARSE_RULE_OPTION_NAME("exec-id"),                                                                            // opt/exec-id
        PARSE_RULE_OPTION_TYPE(String),                                                                               // opt/exec-id
        PARSE_RULE_OPTION_INTERNAL(true)                                                                              // opt/exec-id
        PARSE_RULE_OPTION_REQUIRED(false),                                                                            // opt/exec-id
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                       // opt/exec-id
                                                                                                                      // opt/exec-id
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                // opt/exec-id
        (                                                                                                             // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                       // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Check)                                                                          // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                         // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Info)                                                                           // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                       // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                        // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                         // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                        // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                         // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                        // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Server)                                                                         // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                                     // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                   // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                   // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                  // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Start)                                                                          // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                           // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                         // opt/exec-id
        ),                                                                                                            // opt/exec-id
                                                                                                                      // opt/exec-id
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                               // opt/exec-id
        (                                                                                                             // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/exec-id
        ),                                                                                                            // opt/exec-id
                                                                                                                      // opt/exec-id
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                               // opt/exec-id
        (                                                                                                             // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                        // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                         // opt/exec-id
        ),                                                                                                            // opt/exec-id
                                                                                                                      // opt/exec-id
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                              // opt/exec-id
        (                                                                                                             // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                       // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Check)                                                                          // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                         // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Info)                                                                           // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                       // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                        // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                         // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                        // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                         // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                        // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                   // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                   // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                  // opt/exec-id
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                         // opt/exec-id
        ),                                                                                                            // opt/exec-id
    ),                                                                                                                // opt/exec-id
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                             // opt/expire-auto
    (                                                                                                             // opt/expire-auto
        PARSE_RULE_OPTION_NAME("expire-auto"),                                                                    // opt/expire-auto
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                          // opt/expire-auto
        PARSE_RULE_OPTION_NEGATE(true),                                                                           // opt/expire-auto
        PARSE_RULE_OPTION_RESET(true),                                                                            // opt/expire-auto
        PARSE_RULE_OPTION_REQUIRED(true),                                                                         // opt/expire-auto
        PARSE_RULE_OPTION_SECTION(Global),                                                                        // opt/expire-auto
                                                                                                                  // opt/expire-auto
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                            // opt/expire-auto
        (                                                                                                         // opt/expire-auto
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/expire-auto
        ),                                                                                                        // opt/expire-auto
                                                                                                                  // opt/expire-auto
        PARSE_RULE_OPTIONAL                                                                                       // opt/expire-auto
        (                                                                                                         // opt/expire-auto
            PARSE_RULE_OPTIONAL_GROUP                                                                             // opt/expire-auto
            (                                                                                                     // opt/expire-auto
                PARSE_RULE_OPTIONAL_DEFAULT                                                                       // opt/expire-auto
                (                                                                                                 // opt/expire-auto
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                     // opt/expire-auto
                ),                                                                                                // opt/expire-auto
            ),                                                                                                    // opt/expire-auto
        ),                                                                                                        // opt/expire-auto
    ),                                                                                                            // opt/expire-auto
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                  // opt/filter
    (                                                                                                                  // opt/filter
        PARSE_RULE_OPTION_NAME("filter"),                                                                              // opt/filter
        PARSE_RULE_OPTION_TYPE(String),                                                                                // opt/filter
        PARSE_RULE_OPTION_REQUIRED(false),                                                                             // opt/filter
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                        // opt/filter
                                                                                                                       // opt/filter
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                 // opt/filter
        (                                                                                                              // opt/filter
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                        // opt/filter
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                          // opt/filter
        ),                                                                                                             // opt/filter
    ),                                                                                                                 // opt/filter
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                   // opt/force
    (                                                                                                                   // opt/force
        PARSE_RULE_OPTION_NAME("force"),                                                                                // opt/force
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                                // opt/force
        PARSE_RULE_OPTION_REQUIRED(true),                                                                               // opt/force
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                         // opt/force
                                                                                                                        // opt/force
        PARSE_RULE_OPTION_COMMAND_INTERNAL_LIST                                                                         // opt/force
        (                                                                                                               // opt/force
            PARSE_RULE_OPTION_COMMAND_INTERNAL(StanzaCreate, true),                                                     // opt/force
        )                                                                                                               // opt/force
                                                                                                                        // opt/force
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                  // opt/force
        (                                                                                                               // opt/force
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                           // opt/force
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                          // opt/force
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                     // opt/force
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                     // opt/force
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                             // opt/force
        ),                                                                                                              // opt/force
                                                                                                                        // opt/force
        PARSE_RULE_OPTIONAL                                                                                             // opt/force
        (                                                                                                               // opt/force
            PARSE_RULE_OPTIONAL_GROUP                                                                                   // opt/force
            (                                                                                                           // opt/force
                PARSE_RULE_FILTER_CMD                                                                                   // opt/force
                (                                                                                                       // opt/force
                    PARSE_RULE_VAL_CMD(Backup),                                                                         // opt/force
                ),                                                                                                      // opt/force
                                                                                                                        // opt/force
                PARSE_RULE_OPTIONAL_DEPEND                                                                              // opt/force
                (                                                                                                       // opt/force
                    PARSE_RULE_VAL_OPT(Online),                                                                         // opt/force
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                          // opt/force
                ),                                                                                                      // opt/force
                                                                                                                        // opt/force
                PARSE_RULE_OPTIONAL_DEFAULT                                                                             // opt/force
                (                                                                                                       // opt/force
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                          // opt/force
                ),                                                                                                      // opt/force
            ),                                                                                                          // opt/force
                                                                                                                        // opt/force
            PARSE_RULE_OPTIONAL_GROUP                                                                                   // opt/force
            (                                                                                                           // opt/force
                PARSE_RULE_OPTIONAL_DEFAULT                                                                             // opt/force
                (                                                                                                       // opt/force
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                          // opt/force
                ),                                                                                                      // opt/force
            ),                                                                                                          // opt/force
        ),                                                                                                              // opt/force
    ),                                                                                                                  // opt/force
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                    // opt/help
    (                                                                                                                    // opt/help
        PARSE_RULE_OPTION_NAME("help"),                                                                                  // opt/help
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                                 // opt/help
        PARSE_RULE_OPTION_REQUIRED(true),                                                                                // opt/help
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                          // opt/help
                                                                                                                         // opt/help
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                   // opt/help
        (                                                                                                                // opt/help
            PARSE_RULE_OPTION_COMMAND(Help)                                                                              // opt/help
        ),                                                                                                               // opt/help
                                                                                                                         // opt/help
        PARSE_RULE_OPTIONAL                                                                                              // opt/help
        (                                                                                                                // opt/help
            PARSE_RULE_OPTIONAL_GROUP                                                                                    // opt/help
            (                                                                                                            // opt/help
                PARSE_RULE_OPTIONAL_DEFAULT                                                                              // opt/help
                (                                                                                                        // opt/help
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                           // opt/help
                ),                                                                                                       // opt/help
            ),                                                                                                           // opt/help
        ),                                                                                                               // opt/help
    ),                                                                                                                   // opt/help
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/ignore-missing
    (                                                                                                          // opt/ignore-missing
        PARSE_RULE_OPTION_NAME("ignore-missing"),                                                              // opt/ignore-missing
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                       // opt/ignore-missing
        PARSE_RULE_OPTION_REQUIRED(true),                                                                      // opt/ignore-missing
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                // opt/ignore-missing
                                                                                                               // opt/ignore-missing
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/ignore-missing
        (                                                                                                      // opt/ignore-missing
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/ignore-missing
        ),                                                                                                     // opt/ignore-missing
                                                                                                               // opt/ignore-missing
        PARSE_RULE_OPTIONAL                                                                                    // opt/ignore-missing
        (                                                                                                      // opt/ignore-missing
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/ignore-missing
            (                                                                                                  // opt/ignore-missing
                PARSE_RULE_OPTIONAL_DEFAULT                                                                    // opt/ignore-missing
                (                                                                                              // opt/ignore-missing
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                 // opt/ignore-missing
                ),                                                                                             // opt/ignore-missing
            ),                                                                                                 // opt/ignore-missing
        ),                                                                                                     // opt/ignore-missing
    ),                                                                                                         // opt/ignore-missing
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                              // opt/io-timeout
    (                                                                                                              // opt/io-timeout
        PARSE_RULE_OPTION_NAME("io-timeout"),                                                                      // opt/io-timeout
        PARSE_RULE_OPTION_TYPE(Time),                                                                              // opt/io-timeout
        PARSE_RULE_OPTION_RESET(true),                                                                             // opt/io-timeout
        PARSE_RULE_OPTION_REQUIRED(true),                                                                          // opt/io-timeout
        PARSE_RULE_OPTION_SECTION(Global),                                                                         // opt/io-timeout
                                                                                                                   // opt/io-timeout
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                             // opt/io-timeout
        (                                                                                                          // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                    // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                  // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                 // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                      // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Check)                                                                       // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                      // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Info)                                                                        // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                    // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                     // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                      // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                     // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                      // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                     // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Server)                                                                      // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                                  // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                               // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                      // opt/io-timeout
        ),                                                                                                         // opt/io-timeout
                                                                                                                   // opt/io-timeout
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                            // opt/io-timeout
        (                                                                                                          // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                  // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                 // opt/io-timeout
        ),                                                                                                         // opt/io-timeout
                                                                                                                   // opt/io-timeout
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                            // opt/io-timeout
        (                                                                                                          // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                  // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                 // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                      // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                     // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                      // opt/io-timeout
        ),                                                                                                         // opt/io-timeout
                                                                                                                   // opt/io-timeout
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                           // opt/io-timeout
        (                                                                                                          // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                    // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                  // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                 // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                      // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Check)                                                                       // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                      // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Info)                                                                        // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                    // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                     // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                      // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                     // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                      // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                     // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                               // opt/io-timeout
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                      // opt/io-timeout
        ),                                                                                                         // opt/io-timeout
                                                                                                                   // opt/io-timeout
        PARSE_RULE_OPTIONAL                                                                                        // opt/io-timeout
        (                                                                                                          // opt/io-timeout
            PARSE_RULE_OPTIONAL_GROUP                                                                              // opt/io-timeout
            (                                                                                                      // opt/io-timeout
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                    // opt/io-timeout
                (                                                                                                  // opt/io-timeout
                    PARSE_RULE_VAL_TIME(100ms),                                                                    // opt/io-timeout
                    PARSE_RULE_VAL_TIME(1h),                                                                       // opt/io-timeout
                ),                                                                                                 // opt/io-timeout
                                                                                                                   // opt/io-timeout
                PARSE_RULE_OPTIONAL_DEFAULT                                                                        // opt/io-timeout
                (                                                                                                  // opt/io-timeout
                    PARSE_RULE_VAL_TIME(1m),                                                                       // opt/io-timeout
                ),                                                                                                 // opt/io-timeout
            ),                                                                                                     // opt/io-timeout
        ),                                                                                                         // opt/io-timeout
    ),                                                                                                             // opt/io-timeout
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                               // opt/job-retry
    (                                                                                                               // opt/job-retry
        PARSE_RULE_OPTION_NAME("job-retry"),                                                                        // opt/job-retry
        PARSE_RULE_OPTION_TYPE(Integer),                                                                            // opt/job-retry
        PARSE_RULE_OPTION_INTERNAL(true)                                                                            // opt/job-retry
        PARSE_RULE_OPTION_RESET(true),                                                                              // opt/job-retry
        PARSE_RULE_OPTION_REQUIRED(true),                                                                           // opt/job-retry
        PARSE_RULE_OPTION_SECTION(Global),                                                                          // opt/job-retry
                                                                                                                    // opt/job-retry
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                              // opt/job-retry
        (                                                                                                           // opt/job-retry
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/job-retry
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/job-retry
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/job-retry
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/job-retry
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/job-retry
        ),                                                                                                          // opt/job-retry
                                                                                                                    // opt/job-retry
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                             // opt/job-retry
        (                                                                                                           // opt/job-retry
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/job-retry
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/job-retry
        ),                                                                                                          // opt/job-retry
                                                                                                                    // opt/job-retry
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                             // opt/job-retry
        (                                                                                                           // opt/job-retry
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/job-retry
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/job-retry
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/job-retry
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/job-retry
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/job-retry
        ),                                                                                                          // opt/job-retry
                                                                                                                    // opt/job-retry
        PARSE_RULE_OPTIONAL                                                                                         // opt/job-retry
        (                                                                                                           // opt/job-retry
            PARSE_RULE_OPTIONAL_GROUP                                                                               // opt/job-retry
            (                                                                                                       // opt/job-retry
                PARSE_RULE_FILTER_CMD                                                                               // opt/job-retry
                (                                                                                                   // opt/job-retry
                    PARSE_RULE_VAL_CMD(ArchiveGet),                                                                 // opt/job-retry
                    PARSE_RULE_VAL_CMD(ArchivePush),                                                                // opt/job-retry
                ),                                                                                                  // opt/job-retry
                                                                                                                    // opt/job-retry
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                     // opt/job-retry
                (                                                                                                   // opt/job-retry
                    PARSE_RULE_VAL_INT(0),                                                                          // opt/job-retry
                    PARSE_RULE_VAL_INT(360),                                                                        // opt/job-retry
                ),                                                                                                  // opt/job-retry
                                                                                                                    // opt/job-retry
                PARSE_RULE_OPTIONAL_DEFAULT                                                                         // opt/job-retry
                (                                                                                                   // opt/job-retry
                    PARSE_RULE_VAL_INT(1),                                                                          // opt/job-retry
                ),                                                                                                  // opt/job-retry
            ),                                                                                                      // opt/job-retry
                                                                                                                    // opt/job-retry
            PARSE_RULE_OPTIONAL_GROUP                                                                               // opt/job-retry
            (                                                                                                       // opt/job-retry
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                     // opt/job-retry
                (                                                                                                   // opt/job-retry
                    PARSE_RULE_VAL_INT(0),                                                                          // opt/job-retry
                    PARSE_RULE_VAL_INT(360),                                                                        // opt/job-retry
                ),                                                                                                  // opt/job-retry
                                                                                                                    // opt/job-retry
                PARSE_RULE_OPTIONAL_DEFAULT                                                                         // opt/job-retry
                (                                                                                                   // opt/job-retry
                    PARSE_RULE_VAL_INT(2),                                                                          // opt/job-retry
                ),                                                                                                  // opt/job-retry
            ),                                                                                                      // opt/job-retry
        ),                                                                                                          // opt/job-retry
    ),                                                                                                              // opt/job-retry
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                      // opt/job-retry-interval
    (                                                                                                      // opt/job-retry-interval
        PARSE_RULE_OPTION_NAME("job-retry-interval"),                                                      // opt/job-retry-interval
        PARSE_RULE_OPTION_TYPE(Time),                                                                      // opt/job-retry-interval
        PARSE_RULE_OPTION_INTERNAL(true)                                                                   // opt/job-retry-interval
        PARSE_RULE_OPTION_RESET(true),                                                                     // opt/job-retry-interval
        PARSE_RULE_OPTION_REQUIRED(true),                                                                  // opt/job-retry-interval
        PARSE_RULE_OPTION_SECTION(Global),                                                                 // opt/job-retry-interval
                                                                                                           // opt/job-retry-interval
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                     // opt/job-retry-interval
        (                                                                                                  // opt/job-retry-interval
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/job-retry-interval
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/job-retry-interval
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/job-retry-interval
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/job-retry-interval
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/job-retry-interval
        ),                                                                                                 // opt/job-retry-interval
                                                                                                           // opt/job-retry-interval
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                    // opt/job-retry-interval
        (                                                                                                  // opt/job-retry-interval
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/job-retry-interval
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/job-retry-interval
        ),                                                                                                 // opt/job-retry-interval
                                                                                                           // opt/job-retry-interval
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                    // opt/job-retry-interval
        (                                                                                                  // opt/job-retry-interval
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/job-retry-interval
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/job-retry-interval
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/job-retry-interval
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/job-retry-interval
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/job-retry-interval
        ),                                                                                                 // opt/job-retry-interval
                                                                                                           // opt/job-retry-interval
        PARSE_RULE_OPTIONAL                                                                                // opt/job-retry-interval
        (                                                                                                  // opt/job-retry-interval
            PARSE_RULE_OPTIONAL_GROUP                                                                      // opt/job-retry-interval
            (                                                                                              // opt/job-retry-interval
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                            // opt/job-retry-interval
                (                                                                                          // opt/job-retry-interval
                    PARSE_RULE_VAL_TIME(0s),                                                               // opt/job-retry-interval
                    PARSE_RULE_VAL_TIME(15m),                                                              // opt/job-retry-interval
                ),                                                                                         // opt/job-retry-interval
                                                                                                           // opt/job-retry-interval
                PARSE_RULE_OPTIONAL_DEFAULT                                                                // opt/job-retry-interval
                (                                                                                          // opt/job-retry-interval
                    PARSE_RULE_VAL_TIME(15s),                                                              // opt/job-retry-interval
                ),                                                                                         // opt/job-retry-interval
            ),                                                                                             // opt/job-retry-interval
        ),                                                                                                 // opt/job-retry-interval
    ),                                                                                                     // opt/job-retry-interval
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                // opt/link-all
    (                                                                                                                // opt/link-all
        PARSE_RULE_OPTION_NAME("link-all"),                                                                          // opt/link-all
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                             // opt/link-all
        PARSE_RULE_OPTION_NEGATE(true),                                                                              // opt/link-all
        PARSE_RULE_OPTION_RESET(true),                                                                               // opt/link-all
        PARSE_RULE_OPTION_REQUIRED(true),                                                                            // opt/link-all
        PARSE_RULE_OPTION_SECTION(Global),                                                                           // opt/link-all
                                                                                                                     // opt/link-all
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                               // opt/link-all
        (                                                                                                            // opt/link-all
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                       // opt/link-all
        ),                                                                                                           // opt/link-all
                                                                                                                     // opt/link-all
        PARSE_RULE_OPTIONAL                                                                                          // opt/link-all
        (                                                                                                            // opt/link-all
            PARSE_RULE_OPTIONAL_GROUP                                                                                // opt/link-all
            (                                                                                                        // opt/link-all
                PARSE_RULE_OPTIONAL_DEFAULT                                                                          // opt/link-all
                (                                                                                                    // opt/link-all
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                       // opt/link-all
                ),                                                                                                   // opt/link-all
            ),                                                                                                       // opt/link-all
        ),                                                                                                           // opt/link-all
    ),                                                                                                               // opt/link-all
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                // opt/link-map
    (                                                                                                                // opt/link-map
        PARSE_RULE_OPTION_NAME("link-map"),                                                                          // opt/link-map
        PARSE_RULE_OPTION_TYPE(Hash),                                                                                // opt/link-map
        PARSE_RULE_OPTION_RESET(true),                                                                               // opt/link-map
        PARSE_RULE_OPTION_REQUIRED(false),                                                                           // opt/link-map
        PARSE_RULE_OPTION_SECTION(Global),                                                                           // opt/link-map
        PARSE_RULE_OPTION_MULTI(true),                                                                               // opt/link-map
                                                                                                                     // opt/link-map
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                               // opt/link-map
        (                                                                                                            // opt/link-map
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                       // opt/link-map
        ),                                                                                                           // opt/link-map
    ),                                                                                                               // opt/link-map
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                    // opt/lock
    (                                                                                                                    // opt/lock
        PARSE_RULE_OPTION_NAME("lock"),                                                                                  // opt/lock
        PARSE_RULE_OPTION_TYPE(List),                                                                                    // opt/lock
        PARSE_RULE_OPTION_INTERNAL(true)                                                                                 // opt/lock
        PARSE_RULE_OPTION_REQUIRED(false),                                                                               // opt/lock
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                          // opt/lock
        PARSE_RULE_OPTION_MULTI(true),                                                                                   // opt/lock
                                                                                                                         // opt/lock
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                                 // opt/lock
        (                                                                                                                // opt/lock
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                          // opt/lock
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                       // opt/lock
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                            // opt/lock
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                            // opt/lock
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                      // opt/lock
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                      // opt/lock
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                     // opt/lock
        ),                                                                                                               // opt/lock
    ),                                                                                                                   // opt/lock
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                               // opt/lock-path
    (                                                                                                               // opt/lock-path
        PARSE_RULE_OPTION_NAME("lock-path"),                                                                        // opt/lock-path
        PARSE_RULE_OPTION_TYPE(Path),                                                                               // opt/lock-path
        PARSE_RULE_OPTION_RESET(true),                                                                              // opt/lock-path
        PARSE_RULE_OPTION_REQUIRED(true),                                                                           // opt/lock-path
        PARSE_RULE_OPTION_SECTION(Global),                                                                          // opt/lock-path
                                                                                                                    // opt/lock-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                              // opt/lock-path
        (                                                                                                           // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                     // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                       // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                                         // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                 // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                 // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Start)                                                                        // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                         // opt/lock-path
        ),                                                                                                          // opt/lock-path
                                                                                                                    // opt/lock-path
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                             // opt/lock-path
        (                                                                                                           // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/lock-path
        ),                                                                                                          // opt/lock-path
                                                                                                                    // opt/lock-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                             // opt/lock-path
        (                                                                                                           // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/lock-path
        ),                                                                                                          // opt/lock-path
                                                                                                                    // opt/lock-path
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                            // opt/lock-path
        (                                                                                                           // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                     // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                       // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                                         // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                 // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                 // opt/lock-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                // opt/lock-path
        ),                                                                                                          // opt/lock-path
                                                                                                                    // opt/lock-path
        PARSE_RULE_OPTIONAL                                                                                         // opt/lock-path
        (                                                                                                           // opt/lock-path
            PARSE_RULE_OPTIONAL_GROUP                                                                               // opt/lock-path
            (                                                                                                       // opt/lock-path
                PARSE_RULE_OPTIONAL_DEFAULT                                                                         // opt/lock-path
                (                                                                                                   // opt/lock-path
                    PARSE_RULE_VAL_STR(QT_FS_tmp_FS_pgbackrest_QT),                                                 // opt/lock-path
                ),                                                                                                  // opt/lock-path
            ),                                                                                                      // opt/lock-path
        ),                                                                                                          // opt/lock-path
    ),                                                                                                              // opt/lock-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                       // opt/log-level-console
    (                                                                                                       // opt/log-level-console
        PARSE_RULE_OPTION_NAME("log-level-console"),                                                        // opt/log-level-console
        PARSE_RULE_OPTION_TYPE(StringId),                                                                   // opt/log-level-console
        PARSE_RULE_OPTION_RESET(true),                                                                      // opt/log-level-console
        PARSE_RULE_OPTION_REQUIRED(true),                                                                   // opt/log-level-console
        PARSE_RULE_OPTION_SECTION(Global),                                                                  // opt/log-level-console
                                                                                                            // opt/log-level-console
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                      // opt/log-level-console
        (                                                                                                   // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Server)                                                               // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                           // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Start)                                                                // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                 // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/log-level-console
        ),                                                                                                  // opt/log-level-console
                                                                                                            // opt/log-level-console
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                     // opt/log-level-console
        (                                                                                                   // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/log-level-console
        ),                                                                                                  // opt/log-level-console
                                                                                                            // opt/log-level-console
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                     // opt/log-level-console
        (                                                                                                   // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/log-level-console
        ),                                                                                                  // opt/log-level-console
                                                                                                            // opt/log-level-console
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                    // opt/log-level-console
        (                                                                                                   // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/log-level-console
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/log-level-console
        ),                                                                                                  // opt/log-level-console
                                                                                                            // opt/log-level-console
        PARSE_RULE_OPTIONAL                                                                                 // opt/log-level-console
        (                                                                                                   // opt/log-level-console
            PARSE_RULE_OPTIONAL_GROUP                                                                       // opt/log-level-console
            (                                                                                               // opt/log-level-console
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                              // opt/log-level-console
                (                                                                                           // opt/log-level-console
                    PARSE_RULE_VAL_STRID(off),                                                              // opt/log-level-console
                    PARSE_RULE_VAL_STRID(error),                                                            // opt/log-level-console
                    PARSE_RULE_VAL_STRID(warn),                                                             // opt/log-level-console
                    PARSE_RULE_VAL_STRID(info),                                                             // opt/log-level-console
                    PARSE_RULE_VAL_STRID(detail),                                                           // opt/log-level-console
                    PARSE_RULE_VAL_STRID(debug),                                                            // opt/log-level-console
                    PARSE_RULE_VAL_STRID(trace),                                                            // opt/log-level-console
                ),                                                                                          // opt/log-level-console
                                                                                                            // opt/log-level-console
                PARSE_RULE_OPTIONAL_DEFAULT                                                                 // opt/log-level-console
                (                                                                                           // opt/log-level-console
                    PARSE_RULE_VAL_STRID(warn),                                                             // opt/log-level-console
                ),                                                                                          // opt/log-level-console
            ),                                                                                              // opt/log-level-console
        ),                                                                                                  // opt/log-level-console
    ),                                                                                                      // opt/log-level-console
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/log-level-file
    (                                                                                                          // opt/log-level-file
        PARSE_RULE_OPTION_NAME("log-level-file"),                                                              // opt/log-level-file
        PARSE_RULE_OPTION_TYPE(StringId),                                                                      // opt/log-level-file
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/log-level-file
        PARSE_RULE_OPTION_REQUIRED(true),                                                                      // opt/log-level-file
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/log-level-file
                                                                                                               // opt/log-level-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/log-level-file
        (                                                                                                      // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Server)                                                                  // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                              // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Start)                                                                   // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                    // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/log-level-file
        ),                                                                                                     // opt/log-level-file
                                                                                                               // opt/log-level-file
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                        // opt/log-level-file
        (                                                                                                      // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/log-level-file
        ),                                                                                                     // opt/log-level-file
                                                                                                               // opt/log-level-file
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                        // opt/log-level-file
        (                                                                                                      // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/log-level-file
        ),                                                                                                     // opt/log-level-file
                                                                                                               // opt/log-level-file
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                       // opt/log-level-file
        (                                                                                                      // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/log-level-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/log-level-file
        ),                                                                                                     // opt/log-level-file
                                                                                                               // opt/log-level-file
        PARSE_RULE_OPTIONAL                                                                                    // opt/log-level-file
        (                                                                                                      // opt/log-level-file
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/log-level-file
            (                                                                                                  // opt/log-level-file
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                 // opt/log-level-file
                (                                                                                              // opt/log-level-file
                    PARSE_RULE_VAL_STRID(off),                                                                 // opt/log-level-file
                    PARSE_RULE_VAL_STRID(error),                                                               // opt/log-level-file
                    PARSE_RULE_VAL_STRID(warn),                                                                // opt/log-level-file
                    PARSE_RULE_VAL_STRID(info),                                                                // opt/log-level-file
                    PARSE_RULE_VAL_STRID(detail),                                                              // opt/log-level-file
                    PARSE_RULE_VAL_STRID(debug),                                                               // opt/log-level-file
                    PARSE_RULE_VAL_STRID(trace),                                                               // opt/log-level-file
                ),                                                                                             // opt/log-level-file
                                                                                                               // opt/log-level-file
                PARSE_RULE_OPTIONAL_DEFAULT                                                                    // opt/log-level-file
                (                                                                                              // opt/log-level-file
                    PARSE_RULE_VAL_STRID(info),                                                                // opt/log-level-file
                ),                                                                                             // opt/log-level-file
            ),                                                                                                 // opt/log-level-file
        ),                                                                                                     // opt/log-level-file
    ),                                                                                                         // opt/log-level-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/log-level-stderr
    (                                                                                                        // opt/log-level-stderr
        PARSE_RULE_OPTION_NAME("log-level-stderr"),                                                          // opt/log-level-stderr
        PARSE_RULE_OPTION_TYPE(StringId),                                                                    // opt/log-level-stderr
        PARSE_RULE_OPTION_RESET(true),                                                                       // opt/log-level-stderr
        PARSE_RULE_OPTION_REQUIRED(true),                                                                    // opt/log-level-stderr
        PARSE_RULE_OPTION_SECTION(Global),                                                                   // opt/log-level-stderr
                                                                                                             // opt/log-level-stderr
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/log-level-stderr
        (                                                                                                    // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Server)                                                                // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                            // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Start)                                                                 // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                  // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/log-level-stderr
        ),                                                                                                   // opt/log-level-stderr
                                                                                                             // opt/log-level-stderr
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                      // opt/log-level-stderr
        (                                                                                                    // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/log-level-stderr
        ),                                                                                                   // opt/log-level-stderr
                                                                                                             // opt/log-level-stderr
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                      // opt/log-level-stderr
        (                                                                                                    // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/log-level-stderr
        ),                                                                                                   // opt/log-level-stderr
                                                                                                             // opt/log-level-stderr
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                     // opt/log-level-stderr
        (                                                                                                    // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/log-level-stderr
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/log-level-stderr
        ),                                                                                                   // opt/log-level-stderr
                                                                                                             // opt/log-level-stderr
        PARSE_RULE_OPTIONAL                                                                                  // opt/log-level-stderr
        (                                                                                                    // opt/log-level-stderr
            PARSE_RULE_OPTIONAL_GROUP                                                                        // opt/log-level-stderr
            (                                                                                                // opt/log-level-stderr
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                               // opt/log-level-stderr
                (                                                                                            // opt/log-level-stderr
                    PARSE_RULE_VAL_STRID(off),                                                               // opt/log-level-stderr
                    PARSE_RULE_VAL_STRID(error),                                                             // opt/log-level-stderr
                    PARSE_RULE_VAL_STRID(warn),                                                              // opt/log-level-stderr
                    PARSE_RULE_VAL_STRID(info),                                                              // opt/log-level-stderr
                    PARSE_RULE_VAL_STRID(detail),                                                            // opt/log-level-stderr
                    PARSE_RULE_VAL_STRID(debug),                                                             // opt/log-level-stderr
                    PARSE_RULE_VAL_STRID(trace),                                                             // opt/log-level-stderr
                ),                                                                                           // opt/log-level-stderr
                                                                                                             // opt/log-level-stderr
                PARSE_RULE_OPTIONAL_DEFAULT                                                                  // opt/log-level-stderr
                (                                                                                            // opt/log-level-stderr
                    PARSE_RULE_VAL_STRID(off),                                                               // opt/log-level-stderr
                ),                                                                                           // opt/log-level-stderr
            ),                                                                                               // opt/log-level-stderr
        ),                                                                                                   // opt/log-level-stderr
    ),                                                                                                       // opt/log-level-stderr
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                // opt/log-path
    (                                                                                                                // opt/log-path
        PARSE_RULE_OPTION_NAME("log-path"),                                                                          // opt/log-path
        PARSE_RULE_OPTION_TYPE(Path),                                                                                // opt/log-path
        PARSE_RULE_OPTION_RESET(true),                                                                               // opt/log-path
        PARSE_RULE_OPTION_REQUIRED(true),                                                                            // opt/log-path
        PARSE_RULE_OPTION_SECTION(Global),                                                                           // opt/log-path
                                                                                                                     // opt/log-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                               // opt/log-path
        (                                                                                                            // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                      // opt/log-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                    // opt/log-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                   // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                        // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                                         // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                        // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                                          // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                      // opt/log-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                       // opt/log-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                        // opt/log-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                       // opt/log-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                        // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                       // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Server)                                                                        // opt/log-path
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                                    // opt/log-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                  // opt/log-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                  // opt/log-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                 // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Start)                                                                         // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                          // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                        // opt/log-path
        ),                                                                                                           // opt/log-path
                                                                                                                     // opt/log-path
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                              // opt/log-path
        (                                                                                                            // opt/log-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                    // opt/log-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                   // opt/log-path
        ),                                                                                                           // opt/log-path
                                                                                                                     // opt/log-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                              // opt/log-path
        (                                                                                                            // opt/log-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                    // opt/log-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                   // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                        // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                       // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                        // opt/log-path
        ),                                                                                                           // opt/log-path
                                                                                                                     // opt/log-path
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                             // opt/log-path
        (                                                                                                            // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                      // opt/log-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                    // opt/log-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                   // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                        // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                                         // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                        // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                                          // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                      // opt/log-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                       // opt/log-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                        // opt/log-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                       // opt/log-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                        // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                       // opt/log-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                  // opt/log-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                  // opt/log-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                 // opt/log-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                        // opt/log-path
        ),                                                                                                           // opt/log-path
                                                                                                                     // opt/log-path
        PARSE_RULE_OPTIONAL                                                                                          // opt/log-path
        (                                                                                                            // opt/log-path
            PARSE_RULE_OPTIONAL_GROUP                                                                                // opt/log-path
            (                                                                                                        // opt/log-path
                PARSE_RULE_OPTIONAL_DEFAULT                                                                          // opt/log-path
                (                                                                                                    // opt/log-path
                    PARSE_RULE_VAL_STR(QT_FS_var_FS_log_FS_pgbackrest_QT),                                           // opt/log-path
                ),                                                                                                   // opt/log-path
            ),                                                                                                       // opt/log-path
        ),                                                                                                           // opt/log-path
    ),                                                                                                               // opt/log-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/log-subprocess
    (                                                                                                          // opt/log-subprocess
        PARSE_RULE_OPTION_NAME("log-subprocess"),                                                              // opt/log-subprocess
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                       // opt/log-subprocess
        PARSE_RULE_OPTION_NEGATE(true),                                                                        // opt/log-subprocess
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/log-subprocess
        PARSE_RULE_OPTION_REQUIRED(true),                                                                      // opt/log-subprocess
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/log-subprocess
                                                                                                               // opt/log-subprocess
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/log-subprocess
        (                                                                                                      // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/log-subprocess
        ),                                                                                                     // opt/log-subprocess
                                                                                                               // opt/log-subprocess
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                        // opt/log-subprocess
        (                                                                                                      // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/log-subprocess
        ),                                                                                                     // opt/log-subprocess
                                                                                                               // opt/log-subprocess
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                        // opt/log-subprocess
        (                                                                                                      // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/log-subprocess
        ),                                                                                                     // opt/log-subprocess
                                                                                                               // opt/log-subprocess
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                       // opt/log-subprocess
        (                                                                                                      // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/log-subprocess
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/log-subprocess
        ),                                                                                                     // opt/log-subprocess
                                                                                                               // opt/log-subprocess
        PARSE_RULE_OPTIONAL                                                                                    // opt/log-subprocess
        (                                                                                                      // opt/log-subprocess
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/log-subprocess
            (                                                                                                  // opt/log-subprocess
                PARSE_RULE_OPTIONAL_DEFAULT                                                                    // opt/log-subprocess
                (                                                                                              // opt/log-subprocess
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                 // opt/log-subprocess
                ),                                                                                             // opt/log-subprocess
            ),                                                                                                 // opt/log-subprocess
        ),                                                                                                     // opt/log-subprocess
    ),                                                                                                         // opt/log-subprocess
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                           // opt/log-timestamp
    (                                                                                                           // opt/log-timestamp
        PARSE_RULE_OPTION_NAME("log-timestamp"),                                                                // opt/log-timestamp
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                        // opt/log-timestamp
        PARSE_RULE_OPTION_NEGATE(true),                                                                         // opt/log-timestamp
        PARSE_RULE_OPTION_RESET(true),                                                                          // opt/log-timestamp
        PARSE_RULE_OPTION_REQUIRED(true),                                                                       // opt/log-timestamp
        PARSE_RULE_OPTION_SECTION(Global),                                                                      // opt/log-timestamp
                                                                                                                // opt/log-timestamp
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                          // opt/log-timestamp
        (                                                                                                       // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                 // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Check)                                                                    // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                   // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Info)                                                                     // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                 // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                  // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                   // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                  // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                   // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Server)                                                                   // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                               // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                             // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                             // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                            // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Start)                                                                    // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                     // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/log-timestamp
        ),                                                                                                      // opt/log-timestamp
                                                                                                                // opt/log-timestamp
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                         // opt/log-timestamp
        (                                                                                                       // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/log-timestamp
        ),                                                                                                      // opt/log-timestamp
                                                                                                                // opt/log-timestamp
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                         // opt/log-timestamp
        (                                                                                                       // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/log-timestamp
        ),                                                                                                      // opt/log-timestamp
                                                                                                                // opt/log-timestamp
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                        // opt/log-timestamp
        (                                                                                                       // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                 // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Check)                                                                    // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                   // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Info)                                                                     // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                 // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                  // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                   // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                  // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                   // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                             // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                             // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                            // opt/log-timestamp
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/log-timestamp
        ),                                                                                                      // opt/log-timestamp
                                                                                                                // opt/log-timestamp
        PARSE_RULE_OPTIONAL                                                                                     // opt/log-timestamp
        (                                                                                                       // opt/log-timestamp
            PARSE_RULE_OPTIONAL_GROUP                                                                           // opt/log-timestamp
            (                                                                                                   // opt/log-timestamp
                PARSE_RULE_OPTIONAL_DEFAULT                                                                     // opt/log-timestamp
                (                                                                                               // opt/log-timestamp
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                   // opt/log-timestamp
                ),                                                                                              // opt/log-timestamp
            ),                                                                                                  // opt/log-timestamp
        ),                                                                                                      // opt/log-timestamp
    ),                                                                                                          // opt/log-timestamp
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                 // opt/manifest-save-threshold
    (                                                                                                 // opt/manifest-save-threshold
        PARSE_RULE_OPTION_NAME("manifest-save-threshold"),                                            // opt/manifest-save-threshold
        PARSE_RULE_OPTION_TYPE(Size),                                                                 // opt/manifest-save-threshold
        PARSE_RULE_OPTION_RESET(true),                                                                // opt/manifest-save-threshold
        PARSE_RULE_OPTION_REQUIRED(true),                                                             // opt/manifest-save-threshold
        PARSE_RULE_OPTION_SECTION(Global),                                                            // opt/manifest-save-threshold
                                                                                                      // opt/manifest-save-threshold
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                // opt/manifest-save-threshold
        (                                                                                             // opt/manifest-save-threshold
            PARSE_RULE_OPTION_COMMAND(Backup)                                                         // opt/manifest-save-threshold
        ),                                                                                            // opt/manifest-save-threshold
                                                                                                      // opt/manifest-save-threshold
        PARSE_RULE_OPTIONAL                                                                           // opt/manifest-save-threshold
        (                                                                                             // opt/manifest-save-threshold
            PARSE_RULE_OPTIONAL_GROUP                                                                 // opt/manifest-save-threshold
            (                                                                                         // opt/manifest-save-threshold
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                       // opt/manifest-save-threshold
                (                                                                                     // opt/manifest-save-threshold
                    PARSE_RULE_VAL_SIZE(1B),                                                          // opt/manifest-save-threshold
                    PARSE_RULE_VAL_SIZE(1TiB),                                                        // opt/manifest-save-threshold
                ),                                                                                    // opt/manifest-save-threshold
                                                                                                      // opt/manifest-save-threshold
                PARSE_RULE_OPTIONAL_DEFAULT                                                           // opt/manifest-save-threshold
                (                                                                                     // opt/manifest-save-threshold
                    PARSE_RULE_VAL_SIZE(1GiB),                                                        // opt/manifest-save-threshold
                ),                                                                                    // opt/manifest-save-threshold
            ),                                                                                        // opt/manifest-save-threshold
        ),                                                                                            // opt/manifest-save-threshold
    ),                                                                                                // opt/manifest-save-threshold
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                           // opt/neutral-umask
    (                                                                                                           // opt/neutral-umask
        PARSE_RULE_OPTION_NAME("neutral-umask"),                                                                // opt/neutral-umask
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                        // opt/neutral-umask
        PARSE_RULE_OPTION_NEGATE(true),                                                                         // opt/neutral-umask
        PARSE_RULE_OPTION_RESET(true),                                                                          // opt/neutral-umask
        PARSE_RULE_OPTION_REQUIRED(true),                                                                       // opt/neutral-umask
        PARSE_RULE_OPTION_SECTION(Global),                                                                      // opt/neutral-umask
                                                                                                                // opt/neutral-umask
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                          // opt/neutral-umask
        (                                                                                                       // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                 // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Check)                                                                    // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                   // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                  // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                   // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                  // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                   // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                             // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                             // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                            // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Start)                                                                    // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                     // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/neutral-umask
        ),                                                                                                      // opt/neutral-umask
                                                                                                                // opt/neutral-umask
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                         // opt/neutral-umask
        (                                                                                                       // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/neutral-umask
        ),                                                                                                      // opt/neutral-umask
                                                                                                                // opt/neutral-umask
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                         // opt/neutral-umask
        (                                                                                                       // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/neutral-umask
        ),                                                                                                      // opt/neutral-umask
                                                                                                                // opt/neutral-umask
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                        // opt/neutral-umask
        (                                                                                                       // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                 // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Check)                                                                    // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                   // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                  // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                   // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                  // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                   // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                             // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                             // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                            // opt/neutral-umask
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/neutral-umask
        ),                                                                                                      // opt/neutral-umask
                                                                                                                // opt/neutral-umask
        PARSE_RULE_OPTIONAL                                                                                     // opt/neutral-umask
        (                                                                                                       // opt/neutral-umask
            PARSE_RULE_OPTIONAL_GROUP                                                                           // opt/neutral-umask
            (                                                                                                   // opt/neutral-umask
                PARSE_RULE_OPTIONAL_DEFAULT                                                                     // opt/neutral-umask
                (                                                                                               // opt/neutral-umask
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                   // opt/neutral-umask
                ),                                                                                              // opt/neutral-umask
            ),                                                                                                  // opt/neutral-umask
        ),                                                                                                      // opt/neutral-umask
    ),                                                                                                          // opt/neutral-umask
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                  // opt/oldest
    (                                                                                                                  // opt/oldest
        PARSE_RULE_OPTION_NAME("oldest"),                                                                              // opt/oldest
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                               // opt/oldest
        PARSE_RULE_OPTION_REQUIRED(true),                                                                              // opt/oldest
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                        // opt/oldest
                                                                                                                       // opt/oldest
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                 // opt/oldest
        (                                                                                                              // opt/oldest
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                          // opt/oldest
        ),                                                                                                             // opt/oldest
                                                                                                                       // opt/oldest
        PARSE_RULE_OPTIONAL                                                                                            // opt/oldest
        (                                                                                                              // opt/oldest
            PARSE_RULE_OPTIONAL_GROUP                                                                                  // opt/oldest
            (                                                                                                          // opt/oldest
                PARSE_RULE_FILTER_CMD                                                                                  // opt/oldest
                (                                                                                                      // opt/oldest
                    PARSE_RULE_VAL_CMD(Expire),                                                                        // opt/oldest
                ),                                                                                                     // opt/oldest
                                                                                                                       // opt/oldest
                PARSE_RULE_OPTIONAL_DEPEND                                                                             // opt/oldest
                (                                                                                                      // opt/oldest
                    PARSE_RULE_VAL_OPT(Stanza),                                                                        // opt/oldest
                ),                                                                                                     // opt/oldest
                                                                                                                       // opt/oldest
                PARSE_RULE_OPTIONAL_DEFAULT                                                                            // opt/oldest
                (                                                                                                      // opt/oldest
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                         // opt/oldest
                ),                                                                                                     // opt/oldest
            ),                                                                                                         // opt/oldest
                                                                                                                       // opt/oldest
            PARSE_RULE_OPTIONAL_GROUP                                                                                  // opt/oldest
            (                                                                                                          // opt/oldest
                PARSE_RULE_OPTIONAL_DEFAULT                                                                            // opt/oldest
                (                                                                                                      // opt/oldest
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                         // opt/oldest
                ),                                                                                                     // opt/oldest
            ),                                                                                                         // opt/oldest
        ),                                                                                                             // opt/oldest
    ),                                                                                                                 // opt/oldest
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                  // opt/online
    (                                                                                                                  // opt/online
        PARSE_RULE_OPTION_NAME("online"),                                                                              // opt/online
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                               // opt/online
        PARSE_RULE_OPTION_NEGATE(true),                                                                                // opt/online
        PARSE_RULE_OPTION_REQUIRED(true),                                                                              // opt/online
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                        // opt/online
                                                                                                                       // opt/online
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                 // opt/online
        (                                                                                                              // opt/online
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                          // opt/online
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                    // opt/online
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                   // opt/online
        ),                                                                                                             // opt/online
                                                                                                                       // opt/online
        PARSE_RULE_OPTIONAL                                                                                            // opt/online
        (                                                                                                              // opt/online
            PARSE_RULE_OPTIONAL_GROUP                                                                                  // opt/online
            (                                                                                                          // opt/online
                PARSE_RULE_OPTIONAL_DEFAULT                                                                            // opt/online
                (                                                                                                      // opt/online
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                          // opt/online
                ),                                                                                                     // opt/online
            ),                                                                                                         // opt/online
        ),                                                                                                             // opt/online
    ),                                                                                                                 // opt/online
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                  // opt/output
    (                                                                                                                  // opt/output
        PARSE_RULE_OPTION_NAME("output"),                                                                              // opt/output
        PARSE_RULE_OPTION_TYPE(StringId),                                                                              // opt/output
        PARSE_RULE_OPTION_REQUIRED(true),                                                                              // opt/output
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                        // opt/output
                                                                                                                       // opt/output
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                 // opt/output
        (                                                                                                              // opt/output
            PARSE_RULE_OPTION_COMMAND(Info)                                                                            // opt/output
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                        // opt/output
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                          // opt/output
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                          // opt/output
            PARSE_RULE_OPTION_COMMAND(Version)                                                                         // opt/output
        ),                                                                                                             // opt/output
                                                                                                                       // opt/output
        PARSE_RULE_OPTIONAL                                                                                            // opt/output
        (                                                                                                              // opt/output
            PARSE_RULE_OPTIONAL_GROUP                                                                                  // opt/output
            (                                                                                                          // opt/output
                PARSE_RULE_FILTER_CMD                                                                                  // opt/output
                (                                                                                                      // opt/output
                    PARSE_RULE_VAL_CMD(Verify),                                                                        // opt/output
                ),                                                                                                     // opt/output
                                                                                                                       // opt/output
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                         // opt/output
                (                                                                                                      // opt/output
                    PARSE_RULE_VAL_STRID(none),                                                                        // opt/output
                    PARSE_RULE_VAL_STRID(text),                                                                        // opt/output
                ),                                                                                                     // opt/output
                                                                                                                       // opt/output
                PARSE_RULE_OPTIONAL_DEFAULT                                                                            // opt/output
                (                                                                                                      // opt/output
                    PARSE_RULE_VAL_STRID(none),                                                                        // opt/output
                ),                                                                                                     // opt/output
            ),                                                                                                         // opt/output
                                                                                                                       // opt/output
            PARSE_RULE_OPTIONAL_GROUP                                                                                  // opt/output
            (                                                                                                          // opt/output
                PARSE_RULE_FILTER_CMD                                                                                  // opt/output
                (                                                                                                      // opt/output
                    PARSE_RULE_VAL_CMD(Version),                                                                       // opt/output
                ),                                                                                                     // opt/output
                                                                                                                       // opt/output
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                         // opt/output
                (                                                                                                      // opt/output
                    PARSE_RULE_VAL_STRID(num),                                                                         // opt/output
                    PARSE_RULE_VAL_STRID(text),                                                                        // opt/output
                ),                                                                                                     // opt/output
                                                                                                                       // opt/output
                PARSE_RULE_OPTIONAL_DEFAULT                                                                            // opt/output
                (                                                                                                      // opt/output
                    PARSE_RULE_VAL_STRID(text),                                                                        // opt/output
                ),                                                                                                     // opt/output
            ),                                                                                                         // opt/output
                                                                                                                       // opt/output
            PARSE_RULE_OPTIONAL_GROUP                                                                                  // opt/output
            (                                                                                                          // opt/output
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                         // opt/output
                (                                                                                                      // opt/output
                    PARSE_RULE_VAL_STRID(text),                                                                        // opt/output
                    PARSE_RULE_VAL_STRID(json),                                                                        // opt/output
                ),                                                                                                     // opt/output
                                                                                                                       // opt/output
                PARSE_RULE_OPTIONAL_DEFAULT                                                                            // opt/output
                (                                                                                                      // opt/output
                    PARSE_RULE_VAL_STRID(text),                                                                        // opt/output
                ),                                                                                                     // opt/output
            ),                                                                                                         // opt/output
        ),                                                                                                             // opt/output
    ),                                                                                                                 // opt/output
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                       // opt/page-header-check
    (                                                                                                       // opt/page-header-check
        PARSE_RULE_OPTION_NAME("page-header-check"),                                                        // opt/page-header-check
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                    // opt/page-header-check
        PARSE_RULE_OPTION_NEGATE(true),                                                                     // opt/page-header-check
        PARSE_RULE_OPTION_RESET(true),                                                                      // opt/page-header-check
        PARSE_RULE_OPTION_REQUIRED(true),                                                                   // opt/page-header-check
        PARSE_RULE_OPTION_SECTION(Global),                                                                  // opt/page-header-check
                                                                                                            // opt/page-header-check
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                      // opt/page-header-check
        (                                                                                                   // opt/page-header-check
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/page-header-check
        ),                                                                                                  // opt/page-header-check
                                                                                                            // opt/page-header-check
        PARSE_RULE_OPTIONAL                                                                                 // opt/page-header-check
        (                                                                                                   // opt/page-header-check
            PARSE_RULE_OPTIONAL_GROUP                                                                       // opt/page-header-check
            (                                                                                               // opt/page-header-check
                PARSE_RULE_OPTIONAL_DEFAULT                                                                 // opt/page-header-check
                (                                                                                           // opt/page-header-check
                    PARSE_RULE_VAL_BOOL_TRUE,                                                               // opt/page-header-check
                ),                                                                                          // opt/page-header-check
            ),                                                                                              // opt/page-header-check
        ),                                                                                                  // opt/page-header-check
    ),                                                                                                      // opt/page-header-check
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                      // opt/pg
    (                                                                                                                      // opt/pg
        PARSE_RULE_OPTION_NAME("pg"),                                                                                      // opt/pg
        PARSE_RULE_OPTION_TYPE(Integer),                                                                                   // opt/pg
        PARSE_RULE_OPTION_INTERNAL(true)                                                                                   // opt/pg
        PARSE_RULE_OPTION_REQUIRED(false),                                                                                 // opt/pg
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                            // opt/pg
                                                                                                                           // opt/pg
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                     // opt/pg
        (                                                                                                                  // opt/pg
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                            // opt/pg
        ),                                                                                                                 // opt/pg
                                                                                                                           // opt/pg
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                                    // opt/pg
        (                                                                                                                  // opt/pg
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                              // opt/pg
        ),                                                                                                                 // opt/pg
                                                                                                                           // opt/pg
        PARSE_RULE_OPTIONAL                                                                                                // opt/pg
        (                                                                                                                  // opt/pg
            PARSE_RULE_OPTIONAL_GROUP                                                                                      // opt/pg
            (                                                                                                              // opt/pg
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                            // opt/pg
                (                                                                                                          // opt/pg
                    PARSE_RULE_VAL_INT(1),                                                                                 // opt/pg
                    PARSE_RULE_VAL_INT(256),                                                                               // opt/pg
                ),                                                                                                         // opt/pg
            ),                                                                                                             // opt/pg
        ),                                                                                                                 // opt/pg
    ),                                                                                                                     // opt/pg
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                             // opt/pg-database
    (                                                                                                             // opt/pg-database
        PARSE_RULE_OPTION_NAME("pg-database"),                                                                    // opt/pg-database
        PARSE_RULE_OPTION_TYPE(String),                                                                           // opt/pg-database
        PARSE_RULE_OPTION_RESET(true),                                                                            // opt/pg-database
        PARSE_RULE_OPTION_REQUIRED(true),                                                                         // opt/pg-database
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                        // opt/pg-database
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                           // opt/pg-database
                                                                                                                  // opt/pg-database
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                            // opt/pg-database
        (                                                                                                         // opt/pg-database
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/pg-database
            PARSE_RULE_OPTION_COMMAND(Check)                                                                      // opt/pg-database
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                               // opt/pg-database
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                               // opt/pg-database
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                              // opt/pg-database
        ),                                                                                                        // opt/pg-database
                                                                                                                  // opt/pg-database
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                           // opt/pg-database
        (                                                                                                         // opt/pg-database
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/pg-database
        ),                                                                                                        // opt/pg-database
                                                                                                                  // opt/pg-database
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                          // opt/pg-database
        (                                                                                                         // opt/pg-database
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/pg-database
            PARSE_RULE_OPTION_COMMAND(Check)                                                                      // opt/pg-database
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                               // opt/pg-database
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                               // opt/pg-database
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                              // opt/pg-database
        ),                                                                                                        // opt/pg-database
                                                                                                                  // opt/pg-database
        PARSE_RULE_OPTIONAL                                                                                       // opt/pg-database
        (                                                                                                         // opt/pg-database
            PARSE_RULE_OPTIONAL_GROUP                                                                             // opt/pg-database
            (                                                                                                     // opt/pg-database
                PARSE_RULE_OPTIONAL_DEPEND                                                                        // opt/pg-database
                (                                                                                                 // opt/pg-database
                    PARSE_RULE_VAL_OPT(PgPath),                                                                   // opt/pg-database
                ),                                                                                                // opt/pg-database
                                                                                                                  // opt/pg-database
                PARSE_RULE_OPTIONAL_DEFAULT                                                                       // opt/pg-database
                (                                                                                                 // opt/pg-database
                    PARSE_RULE_VAL_STR(QT_postgres_QT),                                                           // opt/pg-database
                ),                                                                                                // opt/pg-database
            ),                                                                                                    // opt/pg-database
        ),                                                                                                        // opt/pg-database
    ),                                                                                                            // opt/pg-database
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                 // opt/pg-host
    (                                                                                                                 // opt/pg-host
        PARSE_RULE_OPTION_NAME("pg-host"),                                                                            // opt/pg-host
        PARSE_RULE_OPTION_TYPE(String),                                                                               // opt/pg-host
        PARSE_RULE_OPTION_RESET(true),                                                                                // opt/pg-host
        PARSE_RULE_OPTION_REQUIRED(false),                                                                            // opt/pg-host
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                            // opt/pg-host
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                               // opt/pg-host
                                                                                                                      // opt/pg-host
        PARSE_RULE_OPTION_COMMAND_INTERNAL_LIST                                                                       // opt/pg-host
        (                                                                                                             // opt/pg-host
            PARSE_RULE_OPTION_COMMAND_INTERNAL(ArchiveGet, true),                                                     // opt/pg-host
            PARSE_RULE_OPTION_COMMAND_INTERNAL(ArchivePush, true),                                                    // opt/pg-host
            PARSE_RULE_OPTION_COMMAND_INTERNAL(Restore, true),                                                        // opt/pg-host
        )                                                                                                             // opt/pg-host
                                                                                                                      // opt/pg-host
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                // opt/pg-host
        (                                                                                                             // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(Check)                                                                          // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                       // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                        // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                   // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                   // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                  // opt/pg-host
        ),                                                                                                            // opt/pg-host
                                                                                                                      // opt/pg-host
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                               // opt/pg-host
        (                                                                                                             // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/pg-host
        ),                                                                                                            // opt/pg-host
                                                                                                                      // opt/pg-host
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                               // opt/pg-host
        (                                                                                                             // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/pg-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                        // opt/pg-host
        ),                                                                                                            // opt/pg-host
                                                                                                                      // opt/pg-host
        PARSE_RULE_OPTIONAL                                                                                           // opt/pg-host
        (                                                                                                             // opt/pg-host
            PARSE_RULE_OPTIONAL_GROUP                                                                                 // opt/pg-host
            (                                                                                                         // opt/pg-host
                PARSE_RULE_OPTIONAL_DEPEND                                                                            // opt/pg-host
                (                                                                                                     // opt/pg-host
                    PARSE_RULE_VAL_OPT(PgLocal),                                                                      // opt/pg-host
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                        // opt/pg-host
                ),                                                                                                    // opt/pg-host
            ),                                                                                                        // opt/pg-host
        ),                                                                                                            // opt/pg-host
    ),                                                                                                                // opt/pg-host
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                         // opt/pg-host-ca-file
    (                                                                                                         // opt/pg-host-ca-file
        PARSE_RULE_OPTION_NAME("pg-host-ca-file"),                                                            // opt/pg-host-ca-file
        PARSE_RULE_OPTION_TYPE(String),                                                                       // opt/pg-host-ca-file
        PARSE_RULE_OPTION_RESET(true),                                                                        // opt/pg-host-ca-file
        PARSE_RULE_OPTION_REQUIRED(false),                                                                    // opt/pg-host-ca-file
        PARSE_RULE_OPTION_SECTION(Global),                                                                    // opt/pg-host-ca-file
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                       // opt/pg-host-ca-file
                                                                                                              // opt/pg-host-ca-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                        // opt/pg-host-ca-file
        (                                                                                                     // opt/pg-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                 // opt/pg-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                                  // opt/pg-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                               // opt/pg-host-ca-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                           // opt/pg-host-ca-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                           // opt/pg-host-ca-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                          // opt/pg-host-ca-file
        ),                                                                                                    // opt/pg-host-ca-file
                                                                                                              // opt/pg-host-ca-file
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                       // opt/pg-host-ca-file
        (                                                                                                     // opt/pg-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                 // opt/pg-host-ca-file
        ),                                                                                                    // opt/pg-host-ca-file
                                                                                                              // opt/pg-host-ca-file
        PARSE_RULE_OPTIONAL                                                                                   // opt/pg-host-ca-file
        (                                                                                                     // opt/pg-host-ca-file
            PARSE_RULE_OPTIONAL_GROUP                                                                         // opt/pg-host-ca-file
            (                                                                                                 // opt/pg-host-ca-file
                PARSE_RULE_OPTIONAL_DEPEND                                                                    // opt/pg-host-ca-file
                (                                                                                             // opt/pg-host-ca-file
                    PARSE_RULE_VAL_OPT(PgHostType),                                                           // opt/pg-host-ca-file
                    PARSE_RULE_VAL_STRID(tls),                                                                // opt/pg-host-ca-file
                ),                                                                                            // opt/pg-host-ca-file
            ),                                                                                                // opt/pg-host-ca-file
        ),                                                                                                    // opt/pg-host-ca-file
    ),                                                                                                        // opt/pg-host-ca-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                         // opt/pg-host-ca-path
    (                                                                                                         // opt/pg-host-ca-path
        PARSE_RULE_OPTION_NAME("pg-host-ca-path"),                                                            // opt/pg-host-ca-path
        PARSE_RULE_OPTION_TYPE(String),                                                                       // opt/pg-host-ca-path
        PARSE_RULE_OPTION_RESET(true),                                                                        // opt/pg-host-ca-path
        PARSE_RULE_OPTION_REQUIRED(false),                                                                    // opt/pg-host-ca-path
        PARSE_RULE_OPTION_SECTION(Global),                                                                    // opt/pg-host-ca-path
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                       // opt/pg-host-ca-path
                                                                                                              // opt/pg-host-ca-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                        // opt/pg-host-ca-path
        (                                                                                                     // opt/pg-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                 // opt/pg-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                                  // opt/pg-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                               // opt/pg-host-ca-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                           // opt/pg-host-ca-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                           // opt/pg-host-ca-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                          // opt/pg-host-ca-path
        ),                                                                                                    // opt/pg-host-ca-path
                                                                                                              // opt/pg-host-ca-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                       // opt/pg-host-ca-path
        (                                                                                                     // opt/pg-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                 // opt/pg-host-ca-path
        ),                                                                                                    // opt/pg-host-ca-path
                                                                                                              // opt/pg-host-ca-path
        PARSE_RULE_OPTIONAL                                                                                   // opt/pg-host-ca-path
        (                                                                                                     // opt/pg-host-ca-path
            PARSE_RULE_OPTIONAL_GROUP                                                                         // opt/pg-host-ca-path
            (                                                                                                 // opt/pg-host-ca-path
                PARSE_RULE_OPTIONAL_DEPEND                                                                    // opt/pg-host-ca-path
                (                                                                                             // opt/pg-host-ca-path
                    PARSE_RULE_VAL_OPT(PgHostType),                                                           // opt/pg-host-ca-path
                    PARSE_RULE_VAL_STRID(tls),                                                                // opt/pg-host-ca-path
                ),                                                                                            // opt/pg-host-ca-path
            ),                                                                                                // opt/pg-host-ca-path
        ),                                                                                                    // opt/pg-host-ca-path
    ),                                                                                                        // opt/pg-host-ca-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                       // opt/pg-host-cert-file
    (                                                                                                       // opt/pg-host-cert-file
        PARSE_RULE_OPTION_NAME("pg-host-cert-file"),                                                        // opt/pg-host-cert-file
        PARSE_RULE_OPTION_TYPE(String),                                                                     // opt/pg-host-cert-file
        PARSE_RULE_OPTION_RESET(true),                                                                      // opt/pg-host-cert-file
        PARSE_RULE_OPTION_REQUIRED(true),                                                                   // opt/pg-host-cert-file
        PARSE_RULE_OPTION_SECTION(Global),                                                                  // opt/pg-host-cert-file
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                     // opt/pg-host-cert-file
                                                                                                            // opt/pg-host-cert-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                      // opt/pg-host-cert-file
        (                                                                                                   // opt/pg-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/pg-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/pg-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/pg-host-cert-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/pg-host-cert-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/pg-host-cert-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/pg-host-cert-file
        ),                                                                                                  // opt/pg-host-cert-file
                                                                                                            // opt/pg-host-cert-file
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                     // opt/pg-host-cert-file
        (                                                                                                   // opt/pg-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/pg-host-cert-file
        ),                                                                                                  // opt/pg-host-cert-file
                                                                                                            // opt/pg-host-cert-file
        PARSE_RULE_OPTIONAL                                                                                 // opt/pg-host-cert-file
        (                                                                                                   // opt/pg-host-cert-file
            PARSE_RULE_OPTIONAL_GROUP                                                                       // opt/pg-host-cert-file
            (                                                                                               // opt/pg-host-cert-file
                PARSE_RULE_OPTIONAL_DEPEND                                                                  // opt/pg-host-cert-file
                (                                                                                           // opt/pg-host-cert-file
                    PARSE_RULE_VAL_OPT(PgHostType),                                                         // opt/pg-host-cert-file
                    PARSE_RULE_VAL_STRID(tls),                                                              // opt/pg-host-cert-file
                ),                                                                                          // opt/pg-host-cert-file
            ),                                                                                              // opt/pg-host-cert-file
        ),                                                                                                  // opt/pg-host-cert-file
    ),                                                                                                      // opt/pg-host-cert-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                             // opt/pg-host-cmd
    (                                                                                                             // opt/pg-host-cmd
        PARSE_RULE_OPTION_NAME("pg-host-cmd"),                                                                    // opt/pg-host-cmd
        PARSE_RULE_OPTION_TYPE(String),                                                                           // opt/pg-host-cmd
        PARSE_RULE_OPTION_DEFAULT_TYPE(Dynamic),                                                                  // opt/pg-host-cmd
        PARSE_RULE_OPTION_RESET(true),                                                                            // opt/pg-host-cmd
        PARSE_RULE_OPTION_REQUIRED(true),                                                                         // opt/pg-host-cmd
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                        // opt/pg-host-cmd
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                           // opt/pg-host-cmd
                                                                                                                  // opt/pg-host-cmd
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                            // opt/pg-host-cmd
        (                                                                                                         // opt/pg-host-cmd
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/pg-host-cmd
            PARSE_RULE_OPTION_COMMAND(Check)                                                                      // opt/pg-host-cmd
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                   // opt/pg-host-cmd
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                               // opt/pg-host-cmd
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                               // opt/pg-host-cmd
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                              // opt/pg-host-cmd
        ),                                                                                                        // opt/pg-host-cmd
                                                                                                                  // opt/pg-host-cmd
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                           // opt/pg-host-cmd
        (                                                                                                         // opt/pg-host-cmd
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/pg-host-cmd
        ),                                                                                                        // opt/pg-host-cmd
                                                                                                                  // opt/pg-host-cmd
        PARSE_RULE_OPTIONAL                                                                                       // opt/pg-host-cmd
        (                                                                                                         // opt/pg-host-cmd
            PARSE_RULE_OPTIONAL_GROUP                                                                             // opt/pg-host-cmd
            (                                                                                                     // opt/pg-host-cmd
                PARSE_RULE_OPTIONAL_DEPEND                                                                        // opt/pg-host-cmd
                (                                                                                                 // opt/pg-host-cmd
                    PARSE_RULE_VAL_OPT(PgHostType),                                                               // opt/pg-host-cmd
                    PARSE_RULE_VAL_STRID(ssh),                                                                    // opt/pg-host-cmd
                ),                                                                                                // opt/pg-host-cmd
                                                                                                                  // opt/pg-host-cmd
                PARSE_RULE_OPTIONAL_DEFAULT                                                                       // opt/pg-host-cmd
                (                                                                                                 // opt/pg-host-cmd
                    PARSE_RULE_DEFAULT_DYNAMIC(Bin),                                                              // opt/pg-host-cmd
                ),                                                                                                // opt/pg-host-cmd
            ),                                                                                                    // opt/pg-host-cmd
        ),                                                                                                        // opt/pg-host-cmd
    ),                                                                                                            // opt/pg-host-cmd
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/pg-host-config
    (                                                                                                          // opt/pg-host-config
        PARSE_RULE_OPTION_NAME("pg-host-config"),                                                              // opt/pg-host-config
        PARSE_RULE_OPTION_TYPE(Path),                                                                          // opt/pg-host-config
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/pg-host-config
        PARSE_RULE_OPTION_REQUIRED(true),                                                                      // opt/pg-host-config
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                     // opt/pg-host-config
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                        // opt/pg-host-config
                                                                                                               // opt/pg-host-config
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/pg-host-config
        (                                                                                                      // opt/pg-host-config
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/pg-host-config
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/pg-host-config
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/pg-host-config
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/pg-host-config
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/pg-host-config
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/pg-host-config
        ),                                                                                                     // opt/pg-host-config
                                                                                                               // opt/pg-host-config
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                        // opt/pg-host-config
        (                                                                                                      // opt/pg-host-config
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/pg-host-config
        ),                                                                                                     // opt/pg-host-config
                                                                                                               // opt/pg-host-config
        PARSE_RULE_OPTIONAL                                                                                    // opt/pg-host-config
        (                                                                                                      // opt/pg-host-config
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/pg-host-config
            (                                                                                                  // opt/pg-host-config
                PARSE_RULE_OPTIONAL_DEPEND                                                                     // opt/pg-host-config
                (                                                                                              // opt/pg-host-config
                    PARSE_RULE_VAL_OPT(PgHost),                                                                // opt/pg-host-config
                ),                                                                                             // opt/pg-host-config
                                                                                                               // opt/pg-host-config
                PARSE_RULE_OPTIONAL_DEFAULT                                                                    // opt/pg-host-config
                (                                                                                              // opt/pg-host-config
                    PARSE_RULE_VAL_STR(CFGOPTDEF_CONFIG_PATH_SP_QT_FS_QT_SP_PROJECT_CONFIG_FILE),              // opt/pg-host-config
                ),                                                                                             // opt/pg-host-config
            ),                                                                                                 // opt/pg-host-config
        ),                                                                                                     // opt/pg-host-config
    ),                                                                                                         // opt/pg-host-config
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                             // opt/pg-host-config-include-path
    (                                                                                             // opt/pg-host-config-include-path
        PARSE_RULE_OPTION_NAME("pg-host-config-include-path"),                                    // opt/pg-host-config-include-path
        PARSE_RULE_OPTION_TYPE(Path),                                                             // opt/pg-host-config-include-path
        PARSE_RULE_OPTION_RESET(true),                                                            // opt/pg-host-config-include-path
        PARSE_RULE_OPTION_REQUIRED(true),                                                         // opt/pg-host-config-include-path
        PARSE_RULE_OPTION_SECTION(Stanza),                                                        // opt/pg-host-config-include-path
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                           // opt/pg-host-config-include-path
                                                                                                  // opt/pg-host-config-include-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                            // opt/pg-host-config-include-path
        (                                                                                         // opt/pg-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                     // opt/pg-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                      // opt/pg-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                   // opt/pg-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                               // opt/pg-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                               // opt/pg-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                              // opt/pg-host-config-include-path
        ),                                                                                        // opt/pg-host-config-include-path
                                                                                                  // opt/pg-host-config-include-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                           // opt/pg-host-config-include-path
        (                                                                                         // opt/pg-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                     // opt/pg-host-config-include-path
        ),                                                                                        // opt/pg-host-config-include-path
                                                                                                  // opt/pg-host-config-include-path
        PARSE_RULE_OPTIONAL                                                                       // opt/pg-host-config-include-path
        (                                                                                         // opt/pg-host-config-include-path
            PARSE_RULE_OPTIONAL_GROUP                                                             // opt/pg-host-config-include-path
            (                                                                                     // opt/pg-host-config-include-path
                PARSE_RULE_OPTIONAL_DEPEND                                                        // opt/pg-host-config-include-path
                (                                                                                 // opt/pg-host-config-include-path
                    PARSE_RULE_VAL_OPT(PgHost),                                                   // opt/pg-host-config-include-path
                ),                                                                                // opt/pg-host-config-include-path
                                                                                                  // opt/pg-host-config-include-path
                PARSE_RULE_OPTIONAL_DEFAULT                                                       // opt/pg-host-config-include-path
                (                                                                                 // opt/pg-host-config-include-path
                    PARSE_RULE_VAL_STR(CFGOPTDEF_CONFIG_PATH_SP_QT_FS_QT_SP_PROJECT_CONFIG_INCLUDE_PATH),
                ),                                                                                // opt/pg-host-config-include-path
            ),                                                                                    // opt/pg-host-config-include-path
        ),                                                                                        // opt/pg-host-config-include-path
    ),                                                                                            // opt/pg-host-config-include-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                     // opt/pg-host-config-path
    (                                                                                                     // opt/pg-host-config-path
        PARSE_RULE_OPTION_NAME("pg-host-config-path"),                                                    // opt/pg-host-config-path
        PARSE_RULE_OPTION_TYPE(Path),                                                                     // opt/pg-host-config-path
        PARSE_RULE_OPTION_RESET(true),                                                                    // opt/pg-host-config-path
        PARSE_RULE_OPTION_REQUIRED(true),                                                                 // opt/pg-host-config-path
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                // opt/pg-host-config-path
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                   // opt/pg-host-config-path
                                                                                                          // opt/pg-host-config-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                    // opt/pg-host-config-path
        (                                                                                                 // opt/pg-host-config-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/pg-host-config-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/pg-host-config-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/pg-host-config-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/pg-host-config-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/pg-host-config-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/pg-host-config-path
        ),                                                                                                // opt/pg-host-config-path
                                                                                                          // opt/pg-host-config-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                   // opt/pg-host-config-path
        (                                                                                                 // opt/pg-host-config-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/pg-host-config-path
        ),                                                                                                // opt/pg-host-config-path
                                                                                                          // opt/pg-host-config-path
        PARSE_RULE_OPTIONAL                                                                               // opt/pg-host-config-path
        (                                                                                                 // opt/pg-host-config-path
            PARSE_RULE_OPTIONAL_GROUP                                                                     // opt/pg-host-config-path
            (                                                                                             // opt/pg-host-config-path
                PARSE_RULE_OPTIONAL_DEPEND                                                                // opt/pg-host-config-path
                (                                                                                         // opt/pg-host-config-path
                    PARSE_RULE_VAL_OPT(PgHost),                                                           // opt/pg-host-config-path
                ),                                                                                        // opt/pg-host-config-path
                                                                                                          // opt/pg-host-config-path
                PARSE_RULE_OPTIONAL_DEFAULT                                                               // opt/pg-host-config-path
                (                                                                                         // opt/pg-host-config-path
                    PARSE_RULE_VAL_STR(CFGOPTDEF_CONFIG_PATH),                                            // opt/pg-host-config-path
                ),                                                                                        // opt/pg-host-config-path
            ),                                                                                            // opt/pg-host-config-path
        ),                                                                                                // opt/pg-host-config-path
    ),                                                                                                    // opt/pg-host-config-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/pg-host-key-file
    (                                                                                                        // opt/pg-host-key-file
        PARSE_RULE_OPTION_NAME("pg-host-key-file"),                                                          // opt/pg-host-key-file
        PARSE_RULE_OPTION_TYPE(String),                                                                      // opt/pg-host-key-file
        PARSE_RULE_OPTION_RESET(true),                                                                       // opt/pg-host-key-file
        PARSE_RULE_OPTION_REQUIRED(true),                                                                    // opt/pg-host-key-file
        PARSE_RULE_OPTION_SECTION(Global),                                                                   // opt/pg-host-key-file
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                      // opt/pg-host-key-file
                                                                                                             // opt/pg-host-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/pg-host-key-file
        (                                                                                                    // opt/pg-host-key-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/pg-host-key-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/pg-host-key-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/pg-host-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/pg-host-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/pg-host-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/pg-host-key-file
        ),                                                                                                   // opt/pg-host-key-file
                                                                                                             // opt/pg-host-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                      // opt/pg-host-key-file
        (                                                                                                    // opt/pg-host-key-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/pg-host-key-file
        ),                                                                                                   // opt/pg-host-key-file
                                                                                                             // opt/pg-host-key-file
        PARSE_RULE_OPTIONAL                                                                                  // opt/pg-host-key-file
        (                                                                                                    // opt/pg-host-key-file
            PARSE_RULE_OPTIONAL_GROUP                                                                        // opt/pg-host-key-file
            (                                                                                                // opt/pg-host-key-file
                PARSE_RULE_OPTIONAL_DEPEND                                                                   // opt/pg-host-key-file
                (                                                                                            // opt/pg-host-key-file
                    PARSE_RULE_VAL_OPT(PgHostType),                                                          // opt/pg-host-key-file
                    PARSE_RULE_VAL_STRID(tls),                                                               // opt/pg-host-key-file
                ),                                                                                           // opt/pg-host-key-file
            ),                                                                                               // opt/pg-host-key-file
        ),                                                                                                   // opt/pg-host-key-file
    ),                                                                                                       // opt/pg-host-key-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                            // opt/pg-host-port
    (                                                                                                            // opt/pg-host-port
        PARSE_RULE_OPTION_NAME("pg-host-port"),                                                                  // opt/pg-host-port
        PARSE_RULE_OPTION_TYPE(Integer),                                                                         // opt/pg-host-port
        PARSE_RULE_OPTION_RESET(true),                                                                           // opt/pg-host-port
        PARSE_RULE_OPTION_REQUIRED(false),                                                                       // opt/pg-host-port
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                       // opt/pg-host-port
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                          // opt/pg-host-port
                                                                                                                 // opt/pg-host-port
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                           // opt/pg-host-port
        (                                                                                                        // opt/pg-host-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/pg-host-port
            PARSE_RULE_OPTION_COMMAND(Check)                                                                     // opt/pg-host-port
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                  // opt/pg-host-port
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                              // opt/pg-host-port
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                              // opt/pg-host-port
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                             // opt/pg-host-port
        ),                                                                                                       // opt/pg-host-port
                                                                                                                 // opt/pg-host-port
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                          // opt/pg-host-port
        (                                                                                                        // opt/pg-host-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/pg-host-port
        ),                                                                                                       // opt/pg-host-port
                                                                                                                 // opt/pg-host-port
        PARSE_RULE_OPTIONAL                                                                                      // opt/pg-host-port
        (                                                                                                        // opt/pg-host-port
            PARSE_RULE_OPTIONAL_GROUP                                                                            // opt/pg-host-port
            (                                                                                                    // opt/pg-host-port
                PARSE_RULE_OPTIONAL_DEPEND                                                                       // opt/pg-host-port
                (                                                                                                // opt/pg-host-port
                    PARSE_RULE_VAL_OPT(PgHostType),                                                              // opt/pg-host-port
                    PARSE_RULE_VAL_STRID(ssh),                                                                   // opt/pg-host-port
                    PARSE_RULE_VAL_STRID(tls),                                                                   // opt/pg-host-port
                ),                                                                                               // opt/pg-host-port
                                                                                                                 // opt/pg-host-port
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                  // opt/pg-host-port
                (                                                                                                // opt/pg-host-port
                    PARSE_RULE_VAL_INT(0),                                                                       // opt/pg-host-port
                    PARSE_RULE_VAL_INT(65535),                                                                   // opt/pg-host-port
                ),                                                                                               // opt/pg-host-port
                                                                                                                 // opt/pg-host-port
                PARSE_RULE_OPTIONAL_DEFAULT                                                                      // opt/pg-host-port
                (                                                                                                // opt/pg-host-port
                    PARSE_RULE_OPTIONAL_DEFAULT_MAP                                                              // opt/pg-host-port
                    (                                                                                            // opt/pg-host-port
                        PARSE_RULE_VAL_STRID(tls),                                                               // opt/pg-host-port
                        PARSE_RULE_VAL_INT(8432),                                                                // opt/pg-host-port
                    ),                                                                                           // opt/pg-host-port
                ),                                                                                               // opt/pg-host-port
            ),                                                                                                   // opt/pg-host-port
        ),                                                                                                       // opt/pg-host-port
    ),                                                                                                           // opt/pg-host-port
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                            // opt/pg-host-type
    (                                                                                                            // opt/pg-host-type
        PARSE_RULE_OPTION_NAME("pg-host-type"),                                                                  // opt/pg-host-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                                        // opt/pg-host-type
        PARSE_RULE_OPTION_RESET(true),                                                                           // opt/pg-host-type
        PARSE_RULE_OPTION_REQUIRED(true),                                                                        // opt/pg-host-type
        PARSE_RULE_OPTION_SECTION(Global),                                                                       // opt/pg-host-type
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                          // opt/pg-host-type
                                                                                                                 // opt/pg-host-type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                           // opt/pg-host-type
        (                                                                                                        // opt/pg-host-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/pg-host-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                                     // opt/pg-host-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                  // opt/pg-host-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                              // opt/pg-host-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                              // opt/pg-host-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                             // opt/pg-host-type
        ),                                                                                                       // opt/pg-host-type
                                                                                                                 // opt/pg-host-type
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                          // opt/pg-host-type
        (                                                                                                        // opt/pg-host-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/pg-host-type
        ),                                                                                                       // opt/pg-host-type
                                                                                                                 // opt/pg-host-type
        PARSE_RULE_OPTIONAL                                                                                      // opt/pg-host-type
        (                                                                                                        // opt/pg-host-type
            PARSE_RULE_OPTIONAL_GROUP                                                                            // opt/pg-host-type
            (                                                                                                    // opt/pg-host-type
                PARSE_RULE_OPTIONAL_DEPEND                                                                       // opt/pg-host-type
                (                                                                                                // opt/pg-host-type
                    PARSE_RULE_VAL_OPT(PgHost),                                                                  // opt/pg-host-type
                ),                                                                                               // opt/pg-host-type
                                                                                                                 // opt/pg-host-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                   // opt/pg-host-type
                (                                                                                                // opt/pg-host-type
                    PARSE_RULE_VAL_STRID(ssh),                                                                   // opt/pg-host-type
                    PARSE_RULE_VAL_STRID(tls),                                                                   // opt/pg-host-type
                ),                                                                                               // opt/pg-host-type
                                                                                                                 // opt/pg-host-type
                PARSE_RULE_OPTIONAL_DEFAULT                                                                      // opt/pg-host-type
                (                                                                                                // opt/pg-host-type
                    PARSE_RULE_VAL_STRID(ssh),                                                                   // opt/pg-host-type
                ),                                                                                               // opt/pg-host-type
            ),                                                                                                   // opt/pg-host-type
        ),                                                                                                       // opt/pg-host-type
    ),                                                                                                           // opt/pg-host-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                            // opt/pg-host-user
    (                                                                                                            // opt/pg-host-user
        PARSE_RULE_OPTION_NAME("pg-host-user"),                                                                  // opt/pg-host-user
        PARSE_RULE_OPTION_TYPE(String),                                                                          // opt/pg-host-user
        PARSE_RULE_OPTION_RESET(true),                                                                           // opt/pg-host-user
        PARSE_RULE_OPTION_REQUIRED(false),                                                                       // opt/pg-host-user
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                       // opt/pg-host-user
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                          // opt/pg-host-user
                                                                                                                 // opt/pg-host-user
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                           // opt/pg-host-user
        (                                                                                                        // opt/pg-host-user
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/pg-host-user
            PARSE_RULE_OPTION_COMMAND(Check)                                                                     // opt/pg-host-user
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                  // opt/pg-host-user
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                              // opt/pg-host-user
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                              // opt/pg-host-user
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                             // opt/pg-host-user
        ),                                                                                                       // opt/pg-host-user
                                                                                                                 // opt/pg-host-user
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                          // opt/pg-host-user
        (                                                                                                        // opt/pg-host-user
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/pg-host-user
        ),                                                                                                       // opt/pg-host-user
                                                                                                                 // opt/pg-host-user
        PARSE_RULE_OPTIONAL                                                                                      // opt/pg-host-user
        (                                                                                                        // opt/pg-host-user
            PARSE_RULE_OPTIONAL_GROUP                                                                            // opt/pg-host-user
            (                                                                                                    // opt/pg-host-user
                PARSE_RULE_OPTIONAL_DEPEND                                                                       // opt/pg-host-user
                (                                                                                                // opt/pg-host-user
                    PARSE_RULE_VAL_OPT(PgHostType),                                                              // opt/pg-host-user
                    PARSE_RULE_VAL_STRID(ssh),                                                                   // opt/pg-host-user
                ),                                                                                               // opt/pg-host-user
                                                                                                                 // opt/pg-host-user
                PARSE_RULE_OPTIONAL_DEFAULT                                                                      // opt/pg-host-user
                (                                                                                                // opt/pg-host-user
                    PARSE_RULE_VAL_STR(QT_postgres_QT),                                                          // opt/pg-host-user
                ),                                                                                               // opt/pg-host-user
            ),                                                                                                   // opt/pg-host-user
        ),                                                                                                       // opt/pg-host-user
    ),                                                                                                           // opt/pg-host-user
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                // opt/pg-local
    (                                                                                                                // opt/pg-local
        PARSE_RULE_OPTION_NAME("pg-local"),                                                                          // opt/pg-local
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                             // opt/pg-local
        PARSE_RULE_OPTION_INTERNAL(true)                                                                             // opt/pg-local
        PARSE_RULE_OPTION_NEGATE(true),                                                                              // opt/pg-local
        PARSE_RULE_OPTION_RESET(true),                                                                               // opt/pg-local
        PARSE_RULE_OPTION_REQUIRED(true),                                                                            // opt/pg-local
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                           // opt/pg-local
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                              // opt/pg-local
                                                                                                                     // opt/pg-local
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                               // opt/pg-local
        (                                                                                                            // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                    // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                   // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                        // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(Check)                                                                         // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                      // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                       // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                  // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                  // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                 // opt/pg-local
        ),                                                                                                           // opt/pg-local
                                                                                                                     // opt/pg-local
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                              // opt/pg-local
        (                                                                                                            // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                    // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                   // opt/pg-local
        ),                                                                                                           // opt/pg-local
                                                                                                                     // opt/pg-local
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                              // opt/pg-local
        (                                                                                                            // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                    // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                   // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                        // opt/pg-local
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                       // opt/pg-local
        ),                                                                                                           // opt/pg-local
                                                                                                                     // opt/pg-local
        PARSE_RULE_OPTIONAL                                                                                          // opt/pg-local
        (                                                                                                            // opt/pg-local
            PARSE_RULE_OPTIONAL_GROUP                                                                                // opt/pg-local
            (                                                                                                        // opt/pg-local
                PARSE_RULE_OPTIONAL_DEFAULT                                                                          // opt/pg-local
                (                                                                                                    // opt/pg-local
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                       // opt/pg-local
                ),                                                                                                   // opt/pg-local
            ),                                                                                                       // opt/pg-local
        ),                                                                                                           // opt/pg-local
    ),                                                                                                               // opt/pg-local
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                 // opt/pg-path
    (                                                                                                                 // opt/pg-path
        PARSE_RULE_OPTION_NAME("pg-path"),                                                                            // opt/pg-path
        PARSE_RULE_OPTION_TYPE(Path),                                                                                 // opt/pg-path
        PARSE_RULE_OPTION_RESET(true),                                                                                // opt/pg-path
        PARSE_RULE_OPTION_REQUIRED(true),                                                                             // opt/pg-path
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                            // opt/pg-path
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                               // opt/pg-path
                                                                                                                      // opt/pg-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                // opt/pg-path
        (                                                                                                             // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                                          // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                       // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                        // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                   // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                   // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                  // opt/pg-path
        ),                                                                                                            // opt/pg-path
                                                                                                                      // opt/pg-path
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                               // opt/pg-path
        (                                                                                                             // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/pg-path
        ),                                                                                                            // opt/pg-path
                                                                                                                      // opt/pg-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                               // opt/pg-path
        (                                                                                                             // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                        // opt/pg-path
        ),                                                                                                            // opt/pg-path
                                                                                                                      // opt/pg-path
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                              // opt/pg-path
        (                                                                                                             // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                                          // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                       // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                        // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                   // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                   // opt/pg-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                  // opt/pg-path
        ),                                                                                                            // opt/pg-path
                                                                                                                      // opt/pg-path
        PARSE_RULE_OPTIONAL                                                                                           // opt/pg-path
        (                                                                                                             // opt/pg-path
            PARSE_RULE_OPTIONAL_GROUP                                                                                 // opt/pg-path
            (                                                                                                         // opt/pg-path
                PARSE_RULE_FILTER_CMD                                                                                 // opt/pg-path
                (                                                                                                     // opt/pg-path
                    PARSE_RULE_VAL_CMD(ArchivePush),                                                                  // opt/pg-path
                    PARSE_RULE_VAL_CMD(Manifest),                                                                     // opt/pg-path
                ),                                                                                                    // opt/pg-path
                                                                                                                      // opt/pg-path
                PARSE_RULE_OPTIONAL_NOT_REQUIRED(),                                                                   // opt/pg-path
            ),                                                                                                        // opt/pg-path
                                                                                                                      // opt/pg-path
            PARSE_RULE_OPTIONAL_GROUP                                                                                 // opt/pg-path
            (                                                                                                         // opt/pg-path
                PARSE_RULE_FILTER_CMD                                                                                 // opt/pg-path
                (                                                                                                     // opt/pg-path
                    PARSE_RULE_VAL_CMD(Check),                                                                        // opt/pg-path
                ),                                                                                                    // opt/pg-path
                                                                                                                      // opt/pg-path
                PARSE_RULE_OPTIONAL_DEPEND                                                                            // opt/pg-path
                (                                                                                                     // opt/pg-path
                    PARSE_RULE_VAL_OPT(Stanza),                                                                       // opt/pg-path
                ),                                                                                                    // opt/pg-path
            ),                                                                                                        // opt/pg-path
        ),                                                                                                            // opt/pg-path
    ),                                                                                                                // opt/pg-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                 // opt/pg-port
    (                                                                                                                 // opt/pg-port
        PARSE_RULE_OPTION_NAME("pg-port"),                                                                            // opt/pg-port
        PARSE_RULE_OPTION_TYPE(Integer),                                                                              // opt/pg-port
        PARSE_RULE_OPTION_RESET(true),                                                                                // opt/pg-port
        PARSE_RULE_OPTION_REQUIRED(true),                                                                             // opt/pg-port
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                            // opt/pg-port
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                               // opt/pg-port
                                                                                                                      // opt/pg-port
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                // opt/pg-port
        (                                                                                                             // opt/pg-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/pg-port
            PARSE_RULE_OPTION_COMMAND(Check)                                                                          // opt/pg-port
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                   // opt/pg-port
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                   // opt/pg-port
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                  // opt/pg-port
        ),                                                                                                            // opt/pg-port
                                                                                                                      // opt/pg-port
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                               // opt/pg-port
        (                                                                                                             // opt/pg-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/pg-port
        ),                                                                                                            // opt/pg-port
                                                                                                                      // opt/pg-port
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                              // opt/pg-port
        (                                                                                                             // opt/pg-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/pg-port
            PARSE_RULE_OPTION_COMMAND(Check)                                                                          // opt/pg-port
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                   // opt/pg-port
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                   // opt/pg-port
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                  // opt/pg-port
        ),                                                                                                            // opt/pg-port
                                                                                                                      // opt/pg-port
        PARSE_RULE_OPTIONAL                                                                                           // opt/pg-port
        (                                                                                                             // opt/pg-port
            PARSE_RULE_OPTIONAL_GROUP                                                                                 // opt/pg-port
            (                                                                                                         // opt/pg-port
                PARSE_RULE_OPTIONAL_DEPEND                                                                            // opt/pg-port
                (                                                                                                     // opt/pg-port
                    PARSE_RULE_VAL_OPT(PgPath),                                                                       // opt/pg-port
                ),                                                                                                    // opt/pg-port
                                                                                                                      // opt/pg-port
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                       // opt/pg-port
                (                                                                                                     // opt/pg-port
                    PARSE_RULE_VAL_INT(0),                                                                            // opt/pg-port
                    PARSE_RULE_VAL_INT(65535),                                                                        // opt/pg-port
                ),                                                                                                    // opt/pg-port
                                                                                                                      // opt/pg-port
                PARSE_RULE_OPTIONAL_DEFAULT                                                                           // opt/pg-port
                (                                                                                                     // opt/pg-port
                    PARSE_RULE_VAL_INT(5432),                                                                         // opt/pg-port
                ),                                                                                                    // opt/pg-port
            ),                                                                                                        // opt/pg-port
        ),                                                                                                            // opt/pg-port
    ),                                                                                                                // opt/pg-port
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/pg-socket-path
    (                                                                                                          // opt/pg-socket-path
        PARSE_RULE_OPTION_NAME("pg-socket-path"),                                                              // opt/pg-socket-path
        PARSE_RULE_OPTION_TYPE(String),                                                                        // opt/pg-socket-path
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/pg-socket-path
        PARSE_RULE_OPTION_REQUIRED(false),                                                                     // opt/pg-socket-path
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                     // opt/pg-socket-path
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                        // opt/pg-socket-path
                                                                                                               // opt/pg-socket-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/pg-socket-path
        (                                                                                                      // opt/pg-socket-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/pg-socket-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/pg-socket-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/pg-socket-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/pg-socket-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/pg-socket-path
        ),                                                                                                     // opt/pg-socket-path
                                                                                                               // opt/pg-socket-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                        // opt/pg-socket-path
        (                                                                                                      // opt/pg-socket-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/pg-socket-path
        ),                                                                                                     // opt/pg-socket-path
                                                                                                               // opt/pg-socket-path
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                       // opt/pg-socket-path
        (                                                                                                      // opt/pg-socket-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/pg-socket-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/pg-socket-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/pg-socket-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/pg-socket-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/pg-socket-path
        ),                                                                                                     // opt/pg-socket-path
                                                                                                               // opt/pg-socket-path
        PARSE_RULE_OPTIONAL                                                                                    // opt/pg-socket-path
        (                                                                                                      // opt/pg-socket-path
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/pg-socket-path
            (                                                                                                  // opt/pg-socket-path
                PARSE_RULE_OPTIONAL_DEPEND                                                                     // opt/pg-socket-path
                (                                                                                              // opt/pg-socket-path
                    PARSE_RULE_VAL_OPT(PgPath),                                                                // opt/pg-socket-path
                ),                                                                                             // opt/pg-socket-path
            ),                                                                                                 // opt/pg-socket-path
        ),                                                                                                     // opt/pg-socket-path
    ),                                                                                                         // opt/pg-socket-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                 // opt/pg-user
    (                                                                                                                 // opt/pg-user
        PARSE_RULE_OPTION_NAME("pg-user"),                                                                            // opt/pg-user
        PARSE_RULE_OPTION_TYPE(String),                                                                               // opt/pg-user
        PARSE_RULE_OPTION_RESET(true),                                                                                // opt/pg-user
        PARSE_RULE_OPTION_REQUIRED(false),                                                                            // opt/pg-user
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                            // opt/pg-user
        PARSE_RULE_OPTION_GROUP_ID(Pg),                                                                               // opt/pg-user
                                                                                                                      // opt/pg-user
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                // opt/pg-user
        (                                                                                                             // opt/pg-user
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/pg-user
            PARSE_RULE_OPTION_COMMAND(Check)                                                                          // opt/pg-user
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                   // opt/pg-user
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                   // opt/pg-user
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                  // opt/pg-user
        ),                                                                                                            // opt/pg-user
                                                                                                                      // opt/pg-user
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                               // opt/pg-user
        (                                                                                                             // opt/pg-user
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/pg-user
        ),                                                                                                            // opt/pg-user
                                                                                                                      // opt/pg-user
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                              // opt/pg-user
        (                                                                                                             // opt/pg-user
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/pg-user
            PARSE_RULE_OPTION_COMMAND(Check)                                                                          // opt/pg-user
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                   // opt/pg-user
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                   // opt/pg-user
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                  // opt/pg-user
        ),                                                                                                            // opt/pg-user
                                                                                                                      // opt/pg-user
        PARSE_RULE_OPTIONAL                                                                                           // opt/pg-user
        (                                                                                                             // opt/pg-user
            PARSE_RULE_OPTIONAL_GROUP                                                                                 // opt/pg-user
            (                                                                                                         // opt/pg-user
                PARSE_RULE_OPTIONAL_DEPEND                                                                            // opt/pg-user
                (                                                                                                     // opt/pg-user
                    PARSE_RULE_VAL_OPT(PgPath),                                                                       // opt/pg-user
                ),                                                                                                    // opt/pg-user
            ),                                                                                                        // opt/pg-user
        ),                                                                                                            // opt/pg-user
    ),                                                                                                                // opt/pg-user
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/pg-version-force
    (                                                                                                        // opt/pg-version-force
        PARSE_RULE_OPTION_NAME("pg-version-force"),                                                          // opt/pg-version-force
        PARSE_RULE_OPTION_TYPE(String),                                                                      // opt/pg-version-force
        PARSE_RULE_OPTION_RESET(true),                                                                       // opt/pg-version-force
        PARSE_RULE_OPTION_REQUIRED(false),                                                                   // opt/pg-version-force
        PARSE_RULE_OPTION_SECTION(Stanza),                                                                   // opt/pg-version-force
                                                                                                             // opt/pg-version-force
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/pg-version-force
        (                                                                                                    // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/pg-version-force
        ),                                                                                                   // opt/pg-version-force
                                                                                                             // opt/pg-version-force
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                      // opt/pg-version-force
        (                                                                                                    // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/pg-version-force
        ),                                                                                                   // opt/pg-version-force
                                                                                                             // opt/pg-version-force
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                      // opt/pg-version-force
        (                                                                                                    // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/pg-version-force
        ),                                                                                                   // opt/pg-version-force
                                                                                                             // opt/pg-version-force
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                     // opt/pg-version-force
        (                                                                                                    // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/pg-version-force
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/pg-version-force
        ),                                                                                                   // opt/pg-version-force
    ),                                                                                                       // opt/pg-version-force
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                // opt/priority
    (                                                                                                                // opt/priority
        PARSE_RULE_OPTION_NAME("priority"),                                                                          // opt/priority
        PARSE_RULE_OPTION_TYPE(Integer),                                                                             // opt/priority
        PARSE_RULE_OPTION_RESET(true),                                                                               // opt/priority
        PARSE_RULE_OPTION_REQUIRED(false),                                                                           // opt/priority
        PARSE_RULE_OPTION_SECTION(Global),                                                                           // opt/priority
                                                                                                                     // opt/priority
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                               // opt/priority
        (                                                                                                            // opt/priority
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                      // opt/priority
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                    // opt/priority
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                   // opt/priority
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                        // opt/priority
            PARSE_RULE_OPTION_COMMAND(Check)                                                                         // opt/priority
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                        // opt/priority
            PARSE_RULE_OPTION_COMMAND(Info)                                                                          // opt/priority
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                      // opt/priority
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                       // opt/priority
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                        // opt/priority
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                       // opt/priority
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                        // opt/priority
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                       // opt/priority
            PARSE_RULE_OPTION_COMMAND(Server)                                                                        // opt/priority
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                                    // opt/priority
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                  // opt/priority
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                  // opt/priority
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                 // opt/priority
            PARSE_RULE_OPTION_COMMAND(Start)                                                                         // opt/priority
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                          // opt/priority
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                        // opt/priority
        ),                                                                                                           // opt/priority
                                                                                                                     // opt/priority
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                              // opt/priority
        (                                                                                                            // opt/priority
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                    // opt/priority
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                   // opt/priority
        ),                                                                                                           // opt/priority
                                                                                                                     // opt/priority
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                              // opt/priority
        (                                                                                                            // opt/priority
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                    // opt/priority
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                   // opt/priority
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                        // opt/priority
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                       // opt/priority
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                        // opt/priority
        ),                                                                                                           // opt/priority
                                                                                                                     // opt/priority
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                             // opt/priority
        (                                                                                                            // opt/priority
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                      // opt/priority
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                    // opt/priority
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                   // opt/priority
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                        // opt/priority
            PARSE_RULE_OPTION_COMMAND(Check)                                                                         // opt/priority
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                        // opt/priority
            PARSE_RULE_OPTION_COMMAND(Info)                                                                          // opt/priority
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                      // opt/priority
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                       // opt/priority
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                        // opt/priority
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                       // opt/priority
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                        // opt/priority
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                       // opt/priority
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                  // opt/priority
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                  // opt/priority
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                 // opt/priority
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                        // opt/priority
        ),                                                                                                           // opt/priority
                                                                                                                     // opt/priority
        PARSE_RULE_OPTIONAL                                                                                          // opt/priority
        (                                                                                                            // opt/priority
            PARSE_RULE_OPTIONAL_GROUP                                                                                // opt/priority
            (                                                                                                        // opt/priority
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                      // opt/priority
                (                                                                                                    // opt/priority
                    PARSE_RULE_VAL_INT(DS_20),                                                                       // opt/priority
                    PARSE_RULE_VAL_INT(19),                                                                          // opt/priority
                ),                                                                                                   // opt/priority
            ),                                                                                                       // opt/priority
        ),                                                                                                           // opt/priority
    ),                                                                                                               // opt/priority
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                 // opt/process
    (                                                                                                                 // opt/process
        PARSE_RULE_OPTION_NAME("process"),                                                                            // opt/process
        PARSE_RULE_OPTION_TYPE(Integer),                                                                              // opt/process
        PARSE_RULE_OPTION_INTERNAL(true)                                                                              // opt/process
        PARSE_RULE_OPTION_REQUIRED(false),                                                                            // opt/process
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                       // opt/process
                                                                                                                      // opt/process
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                               // opt/process
        (                                                                                                             // opt/process
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/process
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/process
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/process
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                        // opt/process
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                         // opt/process
        ),                                                                                                            // opt/process
                                                                                                                      // opt/process
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                              // opt/process
        (                                                                                                             // opt/process
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                       // opt/process
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                     // opt/process
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                    // opt/process
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                         // opt/process
            PARSE_RULE_OPTION_COMMAND(Check)                                                                          // opt/process
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                         // opt/process
            PARSE_RULE_OPTION_COMMAND(Info)                                                                           // opt/process
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                       // opt/process
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                        // opt/process
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                         // opt/process
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                        // opt/process
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                         // opt/process
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                        // opt/process
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                   // opt/process
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                   // opt/process
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                  // opt/process
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                         // opt/process
        ),                                                                                                            // opt/process
                                                                                                                      // opt/process
        PARSE_RULE_OPTIONAL                                                                                           // opt/process
        (                                                                                                             // opt/process
            PARSE_RULE_OPTIONAL_GROUP                                                                                 // opt/process
            (                                                                                                         // opt/process
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                       // opt/process
                (                                                                                                     // opt/process
                    PARSE_RULE_VAL_INT(0),                                                                            // opt/process
                    PARSE_RULE_VAL_INT(1024),                                                                         // opt/process
                ),                                                                                                    // opt/process
            ),                                                                                                        // opt/process
        ),                                                                                                            // opt/process
    ),                                                                                                                // opt/process
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                             // opt/process-max
    (                                                                                                             // opt/process-max
        PARSE_RULE_OPTION_NAME("process-max"),                                                                    // opt/process-max
        PARSE_RULE_OPTION_TYPE(Integer),                                                                          // opt/process-max
        PARSE_RULE_OPTION_RESET(true),                                                                            // opt/process-max
        PARSE_RULE_OPTION_REQUIRED(true),                                                                         // opt/process-max
        PARSE_RULE_OPTION_SECTION(Global),                                                                        // opt/process-max
                                                                                                                  // opt/process-max
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                            // opt/process-max
        (                                                                                                         // opt/process-max
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/process-max
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/process-max
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/process-max
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                    // opt/process-max
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                     // opt/process-max
        ),                                                                                                        // opt/process-max
                                                                                                                  // opt/process-max
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                           // opt/process-max
        (                                                                                                         // opt/process-max
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/process-max
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/process-max
        ),                                                                                                        // opt/process-max
                                                                                                                  // opt/process-max
        PARSE_RULE_OPTIONAL                                                                                       // opt/process-max
        (                                                                                                         // opt/process-max
            PARSE_RULE_OPTIONAL_GROUP                                                                             // opt/process-max
            (                                                                                                     // opt/process-max
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                   // opt/process-max
                (                                                                                                 // opt/process-max
                    PARSE_RULE_VAL_INT(1),                                                                        // opt/process-max
                    PARSE_RULE_VAL_INT(999),                                                                      // opt/process-max
                ),                                                                                                // opt/process-max
                                                                                                                  // opt/process-max
                PARSE_RULE_OPTIONAL_DEFAULT                                                                       // opt/process-max
                (                                                                                                 // opt/process-max
                    PARSE_RULE_VAL_INT(1),                                                                        // opt/process-max
                ),                                                                                                // opt/process-max
            ),                                                                                                    // opt/process-max
        ),                                                                                                        // opt/process-max
    ),                                                                                                            // opt/process-max
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/protocol-timeout
    (                                                                                                        // opt/protocol-timeout
        PARSE_RULE_OPTION_NAME("protocol-timeout"),                                                          // opt/protocol-timeout
        PARSE_RULE_OPTION_TYPE(Time),                                                                        // opt/protocol-timeout
        PARSE_RULE_OPTION_RESET(true),                                                                       // opt/protocol-timeout
        PARSE_RULE_OPTION_REQUIRED(true),                                                                    // opt/protocol-timeout
        PARSE_RULE_OPTION_SECTION(Global),                                                                   // opt/protocol-timeout
                                                                                                             // opt/protocol-timeout
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/protocol-timeout
        (                                                                                                    // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Server)                                                                // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/protocol-timeout
        ),                                                                                                   // opt/protocol-timeout
                                                                                                             // opt/protocol-timeout
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                      // opt/protocol-timeout
        (                                                                                                    // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/protocol-timeout
        ),                                                                                                   // opt/protocol-timeout
                                                                                                             // opt/protocol-timeout
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                      // opt/protocol-timeout
        (                                                                                                    // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/protocol-timeout
        ),                                                                                                   // opt/protocol-timeout
                                                                                                             // opt/protocol-timeout
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                     // opt/protocol-timeout
        (                                                                                                    // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/protocol-timeout
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/protocol-timeout
        ),                                                                                                   // opt/protocol-timeout
                                                                                                             // opt/protocol-timeout
        PARSE_RULE_OPTIONAL                                                                                  // opt/protocol-timeout
        (                                                                                                    // opt/protocol-timeout
            PARSE_RULE_OPTIONAL_GROUP                                                                        // opt/protocol-timeout
            (                                                                                                // opt/protocol-timeout
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                              // opt/protocol-timeout
                (                                                                                            // opt/protocol-timeout
                    PARSE_RULE_VAL_TIME(100ms),                                                              // opt/protocol-timeout
                    PARSE_RULE_VAL_TIME(7d),                                                                 // opt/protocol-timeout
                ),                                                                                           // opt/protocol-timeout
                                                                                                             // opt/protocol-timeout
                PARSE_RULE_OPTIONAL_DEFAULT                                                                  // opt/protocol-timeout
                (                                                                                            // opt/protocol-timeout
                    PARSE_RULE_VAL_TIME(31m),                                                                // opt/protocol-timeout
                ),                                                                                           // opt/protocol-timeout
            ),                                                                                               // opt/protocol-timeout
        ),                                                                                                   // opt/protocol-timeout
    ),                                                                                                       // opt/protocol-timeout
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                     // opt/raw
    (                                                                                                                     // opt/raw
        PARSE_RULE_OPTION_NAME("raw"),                                                                                    // opt/raw
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                                  // opt/raw
        PARSE_RULE_OPTION_REQUIRED(true),                                                                                 // opt/raw
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                           // opt/raw
                                                                                                                          // opt/raw
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                    // opt/raw
        (                                                                                                                 // opt/raw
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                            // opt/raw
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                            // opt/raw
        ),                                                                                                                // opt/raw
                                                                                                                          // opt/raw
        PARSE_RULE_OPTIONAL                                                                                               // opt/raw
        (                                                                                                                 // opt/raw
            PARSE_RULE_OPTIONAL_GROUP                                                                                     // opt/raw
            (                                                                                                             // opt/raw
                PARSE_RULE_OPTIONAL_DEFAULT                                                                               // opt/raw
                (                                                                                                         // opt/raw
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                            // opt/raw
                ),                                                                                                        // opt/raw
            ),                                                                                                            // opt/raw
        ),                                                                                                                // opt/raw
    ),                                                                                                                    // opt/raw
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                         // opt/recovery-option
    (                                                                                                         // opt/recovery-option
        PARSE_RULE_OPTION_NAME("recovery-option"),                                                            // opt/recovery-option
        PARSE_RULE_OPTION_TYPE(Hash),                                                                         // opt/recovery-option
        PARSE_RULE_OPTION_RESET(true),                                                                        // opt/recovery-option
        PARSE_RULE_OPTION_REQUIRED(false),                                                                    // opt/recovery-option
        PARSE_RULE_OPTION_SECTION(Global),                                                                    // opt/recovery-option
        PARSE_RULE_OPTION_MULTI(true),                                                                        // opt/recovery-option
                                                                                                              // opt/recovery-option
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                        // opt/recovery-option
        (                                                                                                     // opt/recovery-option
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                // opt/recovery-option
        ),                                                                                                    // opt/recovery-option
                                                                                                              // opt/recovery-option
        PARSE_RULE_OPTIONAL                                                                                   // opt/recovery-option
        (                                                                                                     // opt/recovery-option
            PARSE_RULE_OPTIONAL_GROUP                                                                         // opt/recovery-option
            (                                                                                                 // opt/recovery-option
                PARSE_RULE_OPTIONAL_DEPEND                                                                    // opt/recovery-option
                (                                                                                             // opt/recovery-option
                    PARSE_RULE_VAL_OPT(Type),                                                                 // opt/recovery-option
                    PARSE_RULE_VAL_STRID(default),                                                            // opt/recovery-option
                    PARSE_RULE_VAL_STRID(immediate),                                                          // opt/recovery-option
                    PARSE_RULE_VAL_STRID(lsn),                                                                // opt/recovery-option
                    PARSE_RULE_VAL_STRID(name),                                                               // opt/recovery-option
                    PARSE_RULE_VAL_STRID(time),                                                               // opt/recovery-option
                    PARSE_RULE_VAL_STRID(standby),                                                            // opt/recovery-option
                    PARSE_RULE_VAL_STRID(xid),                                                                // opt/recovery-option
                ),                                                                                            // opt/recovery-option
            ),                                                                                                // opt/recovery-option
        ),                                                                                                    // opt/recovery-option
    ),                                                                                                        // opt/recovery-option
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                 // opt/recurse
    (                                                                                                                 // opt/recurse
        PARSE_RULE_OPTION_NAME("recurse"),                                                                            // opt/recurse
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                              // opt/recurse
        PARSE_RULE_OPTION_REQUIRED(true),                                                                             // opt/recurse
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                       // opt/recurse
                                                                                                                      // opt/recurse
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                // opt/recurse
        (                                                                                                             // opt/recurse
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                         // opt/recurse
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                         // opt/recurse
        ),                                                                                                            // opt/recurse
                                                                                                                      // opt/recurse
        PARSE_RULE_OPTIONAL                                                                                           // opt/recurse
        (                                                                                                             // opt/recurse
            PARSE_RULE_OPTIONAL_GROUP                                                                                 // opt/recurse
            (                                                                                                         // opt/recurse
                PARSE_RULE_OPTIONAL_DEFAULT                                                                           // opt/recurse
                (                                                                                                     // opt/recurse
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                        // opt/recurse
                ),                                                                                                    // opt/recurse
            ),                                                                                                        // opt/recurse
        ),                                                                                                            // opt/recurse
    ),                                                                                                                // opt/recurse
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                               // opt/reference
    (                                                                                                               // opt/reference
        PARSE_RULE_OPTION_NAME("reference"),                                                                        // opt/reference
        PARSE_RULE_OPTION_TYPE(String),                                                                             // opt/reference
        PARSE_RULE_OPTION_REQUIRED(false),                                                                          // opt/reference
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                     // opt/reference
                                                                                                                    // opt/reference
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                              // opt/reference
        (                                                                                                           // opt/reference
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                     // opt/reference
        ),                                                                                                          // opt/reference
    ),                                                                                                              // opt/reference
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                             // opt/remote-type
    (                                                                                                             // opt/remote-type
        PARSE_RULE_OPTION_NAME("remote-type"),                                                                    // opt/remote-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                                         // opt/remote-type
        PARSE_RULE_OPTION_INTERNAL(true)                                                                          // opt/remote-type
        PARSE_RULE_OPTION_REQUIRED(false),                                                                        // opt/remote-type
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                   // opt/remote-type
                                                                                                                  // opt/remote-type
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                           // opt/remote-type
        (                                                                                                         // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                    // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                     // opt/remote-type
        ),                                                                                                        // opt/remote-type
                                                                                                                  // opt/remote-type
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                          // opt/remote-type
        (                                                                                                         // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                   // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                                      // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                     // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                                       // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                   // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                    // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                     // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                    // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                     // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                    // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                               // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                               // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                              // opt/remote-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                     // opt/remote-type
        ),                                                                                                        // opt/remote-type
                                                                                                                  // opt/remote-type
        PARSE_RULE_OPTIONAL                                                                                       // opt/remote-type
        (                                                                                                         // opt/remote-type
            PARSE_RULE_OPTIONAL_GROUP                                                                             // opt/remote-type
            (                                                                                                     // opt/remote-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                    // opt/remote-type
                (                                                                                                 // opt/remote-type
                    PARSE_RULE_VAL_STRID(pg),                                                                     // opt/remote-type
                    PARSE_RULE_VAL_STRID(repo),                                                                   // opt/remote-type
                ),                                                                                                // opt/remote-type
            ),                                                                                                    // opt/remote-type
        ),                                                                                                        // opt/remote-type
    ),                                                                                                            // opt/remote-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                    // opt/repo
    (                                                                                                                    // opt/repo
        PARSE_RULE_OPTION_NAME("repo"),                                                                                  // opt/repo
        PARSE_RULE_OPTION_TYPE(Integer),                                                                                 // opt/repo
        PARSE_RULE_OPTION_REQUIRED(false),                                                                               // opt/repo
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                          // opt/repo
                                                                                                                         // opt/repo
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                   // opt/repo
        (                                                                                                                // opt/repo
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                          // opt/repo
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                        // opt/repo
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                            // opt/repo
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                            // opt/repo
            PARSE_RULE_OPTION_COMMAND(Info)                                                                              // opt/repo
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                          // opt/repo
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                           // opt/repo
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                            // opt/repo
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                           // opt/repo
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                            // opt/repo
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                           // opt/repo
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                      // opt/repo
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                            // opt/repo
        ),                                                                                                               // opt/repo
                                                                                                                         // opt/repo
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                                  // opt/repo
        (                                                                                                                // opt/repo
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                        // opt/repo
        ),                                                                                                               // opt/repo
                                                                                                                         // opt/repo
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                                  // opt/repo
        (                                                                                                                // opt/repo
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                        // opt/repo
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                            // opt/repo
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                           // opt/repo
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                            // opt/repo
        ),                                                                                                               // opt/repo
                                                                                                                         // opt/repo
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                                 // opt/repo
        (                                                                                                                // opt/repo
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                          // opt/repo
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                        // opt/repo
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                       // opt/repo
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                            // opt/repo
            PARSE_RULE_OPTION_COMMAND(Check)                                                                             // opt/repo
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                            // opt/repo
            PARSE_RULE_OPTION_COMMAND(Info)                                                                              // opt/repo
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                          // opt/repo
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                           // opt/repo
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                            // opt/repo
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                           // opt/repo
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                            // opt/repo
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                           // opt/repo
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                      // opt/repo
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                      // opt/repo
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                     // opt/repo
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                            // opt/repo
        ),                                                                                                               // opt/repo
                                                                                                                         // opt/repo
        PARSE_RULE_OPTIONAL                                                                                              // opt/repo
        (                                                                                                                // opt/repo
            PARSE_RULE_OPTIONAL_GROUP                                                                                    // opt/repo
            (                                                                                                            // opt/repo
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                          // opt/repo
                (                                                                                                        // opt/repo
                    PARSE_RULE_VAL_INT(1),                                                                               // opt/repo
                    PARSE_RULE_VAL_INT(256),                                                                             // opt/repo
                ),                                                                                                       // opt/repo
            ),                                                                                                           // opt/repo
        ),                                                                                                               // opt/repo
    ),                                                                                                                   // opt/repo
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                      // opt/repo-azure-account
    (                                                                                                      // opt/repo-azure-account
        PARSE_RULE_OPTION_NAME("repo-azure-account"),                                                      // opt/repo-azure-account
        PARSE_RULE_OPTION_TYPE(String),                                                                    // opt/repo-azure-account
        PARSE_RULE_OPTION_RESET(true),                                                                     // opt/repo-azure-account
        PARSE_RULE_OPTION_REQUIRED(true),                                                                  // opt/repo-azure-account
        PARSE_RULE_OPTION_SECTION(Global),                                                                 // opt/repo-azure-account
        PARSE_RULE_OPTION_SECURE(true),                                                                    // opt/repo-azure-account
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                  // opt/repo-azure-account
                                                                                                           // opt/repo-azure-account
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                     // opt/repo-azure-account
        (                                                                                                  // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                            // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Check)                                                               // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Expire)                                                              // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Info)                                                                // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                            // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                             // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                              // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                             // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                              // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                        // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                        // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                       // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/repo-azure-account
        ),                                                                                                 // opt/repo-azure-account
                                                                                                           // opt/repo-azure-account
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                    // opt/repo-azure-account
        (                                                                                                  // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-azure-account
        ),                                                                                                 // opt/repo-azure-account
                                                                                                           // opt/repo-azure-account
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                    // opt/repo-azure-account
        (                                                                                                  // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/repo-azure-account
        ),                                                                                                 // opt/repo-azure-account
                                                                                                           // opt/repo-azure-account
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                   // opt/repo-azure-account
        (                                                                                                  // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                            // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Check)                                                               // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Expire)                                                              // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Info)                                                                // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                            // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                             // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                              // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                             // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                              // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                        // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                        // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                       // opt/repo-azure-account
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/repo-azure-account
        ),                                                                                                 // opt/repo-azure-account
                                                                                                           // opt/repo-azure-account
        PARSE_RULE_OPTIONAL                                                                                // opt/repo-azure-account
        (                                                                                                  // opt/repo-azure-account
            PARSE_RULE_OPTIONAL_GROUP                                                                      // opt/repo-azure-account
            (                                                                                              // opt/repo-azure-account
                PARSE_RULE_OPTIONAL_DEPEND                                                                 // opt/repo-azure-account
                (                                                                                          // opt/repo-azure-account
                    PARSE_RULE_VAL_OPT(RepoType),                                                          // opt/repo-azure-account
                    PARSE_RULE_VAL_STRID(azure),                                                           // opt/repo-azure-account
                ),                                                                                         // opt/repo-azure-account
            ),                                                                                             // opt/repo-azure-account
        ),                                                                                                 // opt/repo-azure-account
    ),                                                                                                     // opt/repo-azure-account
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                    // opt/repo-azure-container
    (                                                                                                    // opt/repo-azure-container
        PARSE_RULE_OPTION_NAME("repo-azure-container"),                                                  // opt/repo-azure-container
        PARSE_RULE_OPTION_TYPE(String),                                                                  // opt/repo-azure-container
        PARSE_RULE_OPTION_RESET(true),                                                                   // opt/repo-azure-container
        PARSE_RULE_OPTION_REQUIRED(true),                                                                // opt/repo-azure-container
        PARSE_RULE_OPTION_SECTION(Global),                                                               // opt/repo-azure-container
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                // opt/repo-azure-container
                                                                                                         // opt/repo-azure-container
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                   // opt/repo-azure-container
        (                                                                                                // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                          // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Check)                                                             // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Expire)                                                            // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Info)                                                              // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                          // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                           // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                            // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                           // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                            // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                      // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                      // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                     // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-azure-container
        ),                                                                                               // opt/repo-azure-container
                                                                                                         // opt/repo-azure-container
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                  // opt/repo-azure-container
        (                                                                                                // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-azure-container
        ),                                                                                               // opt/repo-azure-container
                                                                                                         // opt/repo-azure-container
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                  // opt/repo-azure-container
        (                                                                                                // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-azure-container
        ),                                                                                               // opt/repo-azure-container
                                                                                                         // opt/repo-azure-container
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                 // opt/repo-azure-container
        (                                                                                                // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                          // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Check)                                                             // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Expire)                                                            // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Info)                                                              // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                          // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                           // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                            // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                           // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                            // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                      // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                      // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                     // opt/repo-azure-container
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-azure-container
        ),                                                                                               // opt/repo-azure-container
                                                                                                         // opt/repo-azure-container
        PARSE_RULE_OPTIONAL                                                                              // opt/repo-azure-container
        (                                                                                                // opt/repo-azure-container
            PARSE_RULE_OPTIONAL_GROUP                                                                    // opt/repo-azure-container
            (                                                                                            // opt/repo-azure-container
                PARSE_RULE_OPTIONAL_DEPEND                                                               // opt/repo-azure-container
                (                                                                                        // opt/repo-azure-container
                    PARSE_RULE_VAL_OPT(RepoType),                                                        // opt/repo-azure-container
                    PARSE_RULE_VAL_STRID(azure),                                                         // opt/repo-azure-container
                ),                                                                                       // opt/repo-azure-container
            ),                                                                                           // opt/repo-azure-container
        ),                                                                                               // opt/repo-azure-container
    ),                                                                                                   // opt/repo-azure-container
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                     // opt/repo-azure-endpoint
    (                                                                                                     // opt/repo-azure-endpoint
        PARSE_RULE_OPTION_NAME("repo-azure-endpoint"),                                                    // opt/repo-azure-endpoint
        PARSE_RULE_OPTION_TYPE(String),                                                                   // opt/repo-azure-endpoint
        PARSE_RULE_OPTION_RESET(true),                                                                    // opt/repo-azure-endpoint
        PARSE_RULE_OPTION_REQUIRED(true),                                                                 // opt/repo-azure-endpoint
        PARSE_RULE_OPTION_SECTION(Global),                                                                // opt/repo-azure-endpoint
        PARSE_RULE_OPTION_SECURE(true),                                                                   // opt/repo-azure-endpoint
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                 // opt/repo-azure-endpoint
                                                                                                          // opt/repo-azure-endpoint
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                    // opt/repo-azure-endpoint
        (                                                                                                 // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-azure-endpoint
        ),                                                                                                // opt/repo-azure-endpoint
                                                                                                          // opt/repo-azure-endpoint
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                   // opt/repo-azure-endpoint
        (                                                                                                 // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-azure-endpoint
        ),                                                                                                // opt/repo-azure-endpoint
                                                                                                          // opt/repo-azure-endpoint
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                   // opt/repo-azure-endpoint
        (                                                                                                 // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-azure-endpoint
        ),                                                                                                // opt/repo-azure-endpoint
                                                                                                          // opt/repo-azure-endpoint
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                  // opt/repo-azure-endpoint
        (                                                                                                 // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/repo-azure-endpoint
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-azure-endpoint
        ),                                                                                                // opt/repo-azure-endpoint
                                                                                                          // opt/repo-azure-endpoint
        PARSE_RULE_OPTIONAL                                                                               // opt/repo-azure-endpoint
        (                                                                                                 // opt/repo-azure-endpoint
            PARSE_RULE_OPTIONAL_GROUP                                                                     // opt/repo-azure-endpoint
            (                                                                                             // opt/repo-azure-endpoint
                PARSE_RULE_OPTIONAL_DEPEND                                                                // opt/repo-azure-endpoint
                (                                                                                         // opt/repo-azure-endpoint
                    PARSE_RULE_VAL_OPT(RepoType),                                                         // opt/repo-azure-endpoint
                    PARSE_RULE_VAL_STRID(azure),                                                          // opt/repo-azure-endpoint
                ),                                                                                        // opt/repo-azure-endpoint
                                                                                                          // opt/repo-azure-endpoint
                PARSE_RULE_OPTIONAL_DEFAULT                                                               // opt/repo-azure-endpoint
                (                                                                                         // opt/repo-azure-endpoint
                    PARSE_RULE_VAL_STR(QT_blob_DT_core_DT_windows_DT_net_QT),                             // opt/repo-azure-endpoint
                ),                                                                                        // opt/repo-azure-endpoint
            ),                                                                                            // opt/repo-azure-endpoint
        ),                                                                                                // opt/repo-azure-endpoint
    ),                                                                                                    // opt/repo-azure-endpoint
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/repo-azure-key
    (                                                                                                          // opt/repo-azure-key
        PARSE_RULE_OPTION_NAME("repo-azure-key"),                                                              // opt/repo-azure-key
        PARSE_RULE_OPTION_TYPE(String),                                                                        // opt/repo-azure-key
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/repo-azure-key
        PARSE_RULE_OPTION_REQUIRED(false),                                                                     // opt/repo-azure-key
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/repo-azure-key
        PARSE_RULE_OPTION_SECURE(true),                                                                        // opt/repo-azure-key
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                      // opt/repo-azure-key
                                                                                                               // opt/repo-azure-key
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/repo-azure-key
        (                                                                                                      // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-azure-key
        ),                                                                                                     // opt/repo-azure-key
                                                                                                               // opt/repo-azure-key
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                        // opt/repo-azure-key
        (                                                                                                      // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-azure-key
        ),                                                                                                     // opt/repo-azure-key
                                                                                                               // opt/repo-azure-key
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                        // opt/repo-azure-key
        (                                                                                                      // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-azure-key
        ),                                                                                                     // opt/repo-azure-key
                                                                                                               // opt/repo-azure-key
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                       // opt/repo-azure-key
        (                                                                                                      // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/repo-azure-key
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-azure-key
        ),                                                                                                     // opt/repo-azure-key
                                                                                                               // opt/repo-azure-key
        PARSE_RULE_OPTIONAL                                                                                    // opt/repo-azure-key
        (                                                                                                      // opt/repo-azure-key
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/repo-azure-key
            (                                                                                                  // opt/repo-azure-key
                PARSE_RULE_OPTIONAL_DEPEND                                                                     // opt/repo-azure-key
                (                                                                                              // opt/repo-azure-key
                    PARSE_RULE_VAL_OPT(RepoAzureKeyType),                                                      // opt/repo-azure-key
                    PARSE_RULE_VAL_STRID(shared),                                                              // opt/repo-azure-key
                    PARSE_RULE_VAL_STRID(sas),                                                                 // opt/repo-azure-key
                ),                                                                                             // opt/repo-azure-key
            ),                                                                                                 // opt/repo-azure-key
        ),                                                                                                     // opt/repo-azure-key
    ),                                                                                                         // opt/repo-azure-key
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                     // opt/repo-azure-key-type
    (                                                                                                     // opt/repo-azure-key-type
        PARSE_RULE_OPTION_NAME("repo-azure-key-type"),                                                    // opt/repo-azure-key-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                                 // opt/repo-azure-key-type
        PARSE_RULE_OPTION_RESET(true),                                                                    // opt/repo-azure-key-type
        PARSE_RULE_OPTION_REQUIRED(true),                                                                 // opt/repo-azure-key-type
        PARSE_RULE_OPTION_SECTION(Global),                                                                // opt/repo-azure-key-type
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                 // opt/repo-azure-key-type
                                                                                                          // opt/repo-azure-key-type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                    // opt/repo-azure-key-type
        (                                                                                                 // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-azure-key-type
        ),                                                                                                // opt/repo-azure-key-type
                                                                                                          // opt/repo-azure-key-type
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                   // opt/repo-azure-key-type
        (                                                                                                 // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-azure-key-type
        ),                                                                                                // opt/repo-azure-key-type
                                                                                                          // opt/repo-azure-key-type
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                   // opt/repo-azure-key-type
        (                                                                                                 // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-azure-key-type
        ),                                                                                                // opt/repo-azure-key-type
                                                                                                          // opt/repo-azure-key-type
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                  // opt/repo-azure-key-type
        (                                                                                                 // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/repo-azure-key-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-azure-key-type
        ),                                                                                                // opt/repo-azure-key-type
                                                                                                          // opt/repo-azure-key-type
        PARSE_RULE_OPTIONAL                                                                               // opt/repo-azure-key-type
        (                                                                                                 // opt/repo-azure-key-type
            PARSE_RULE_OPTIONAL_GROUP                                                                     // opt/repo-azure-key-type
            (                                                                                             // opt/repo-azure-key-type
                PARSE_RULE_OPTIONAL_DEPEND                                                                // opt/repo-azure-key-type
                (                                                                                         // opt/repo-azure-key-type
                    PARSE_RULE_VAL_OPT(RepoType),                                                         // opt/repo-azure-key-type
                    PARSE_RULE_VAL_STRID(azure),                                                          // opt/repo-azure-key-type
                ),                                                                                        // opt/repo-azure-key-type
                                                                                                          // opt/repo-azure-key-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                            // opt/repo-azure-key-type
                (                                                                                         // opt/repo-azure-key-type
                    PARSE_RULE_VAL_STRID(shared),                                                         // opt/repo-azure-key-type
                    PARSE_RULE_VAL_STRID(sas),                                                            // opt/repo-azure-key-type
                    PARSE_RULE_VAL_STRID(auto),                                                           // opt/repo-azure-key-type
                ),                                                                                        // opt/repo-azure-key-type
                                                                                                          // opt/repo-azure-key-type
                PARSE_RULE_OPTIONAL_DEFAULT                                                               // opt/repo-azure-key-type
                (                                                                                         // opt/repo-azure-key-type
                    PARSE_RULE_VAL_STRID(shared),                                                         // opt/repo-azure-key-type
                ),                                                                                        // opt/repo-azure-key-type
            ),                                                                                            // opt/repo-azure-key-type
        ),                                                                                                // opt/repo-azure-key-type
    ),                                                                                                    // opt/repo-azure-key-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                    // opt/repo-azure-uri-style
    (                                                                                                    // opt/repo-azure-uri-style
        PARSE_RULE_OPTION_NAME("repo-azure-uri-style"),                                                  // opt/repo-azure-uri-style
        PARSE_RULE_OPTION_TYPE(StringId),                                                                // opt/repo-azure-uri-style
        PARSE_RULE_OPTION_RESET(true),                                                                   // opt/repo-azure-uri-style
        PARSE_RULE_OPTION_REQUIRED(true),                                                                // opt/repo-azure-uri-style
        PARSE_RULE_OPTION_SECTION(Global),                                                               // opt/repo-azure-uri-style
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                // opt/repo-azure-uri-style
                                                                                                         // opt/repo-azure-uri-style
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                   // opt/repo-azure-uri-style
        (                                                                                                // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                          // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Check)                                                             // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Expire)                                                            // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Info)                                                              // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                          // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                           // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                            // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                           // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                            // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                      // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                      // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                     // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-azure-uri-style
        ),                                                                                               // opt/repo-azure-uri-style
                                                                                                         // opt/repo-azure-uri-style
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                  // opt/repo-azure-uri-style
        (                                                                                                // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-azure-uri-style
        ),                                                                                               // opt/repo-azure-uri-style
                                                                                                         // opt/repo-azure-uri-style
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                  // opt/repo-azure-uri-style
        (                                                                                                // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-azure-uri-style
        ),                                                                                               // opt/repo-azure-uri-style
                                                                                                         // opt/repo-azure-uri-style
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                 // opt/repo-azure-uri-style
        (                                                                                                // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                          // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Check)                                                             // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Expire)                                                            // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Info)                                                              // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                          // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                           // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                            // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                           // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                            // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                      // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                      // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                     // opt/repo-azure-uri-style
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-azure-uri-style
        ),                                                                                               // opt/repo-azure-uri-style
                                                                                                         // opt/repo-azure-uri-style
        PARSE_RULE_OPTIONAL                                                                              // opt/repo-azure-uri-style
        (                                                                                                // opt/repo-azure-uri-style
            PARSE_RULE_OPTIONAL_GROUP                                                                    // opt/repo-azure-uri-style
            (                                                                                            // opt/repo-azure-uri-style
                PARSE_RULE_OPTIONAL_DEPEND                                                               // opt/repo-azure-uri-style
                (                                                                                        // opt/repo-azure-uri-style
                    PARSE_RULE_VAL_OPT(RepoType),                                                        // opt/repo-azure-uri-style
                    PARSE_RULE_VAL_STRID(azure),                                                         // opt/repo-azure-uri-style
                ),                                                                                       // opt/repo-azure-uri-style
                                                                                                         // opt/repo-azure-uri-style
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                           // opt/repo-azure-uri-style
                (                                                                                        // opt/repo-azure-uri-style
                    PARSE_RULE_VAL_STRID(host),                                                          // opt/repo-azure-uri-style
                    PARSE_RULE_VAL_STRID(path),                                                          // opt/repo-azure-uri-style
                ),                                                                                       // opt/repo-azure-uri-style
                                                                                                         // opt/repo-azure-uri-style
                PARSE_RULE_OPTIONAL_DEFAULT                                                              // opt/repo-azure-uri-style
                (                                                                                        // opt/repo-azure-uri-style
                    PARSE_RULE_VAL_STRID(host),                                                          // opt/repo-azure-uri-style
                ),                                                                                       // opt/repo-azure-uri-style
            ),                                                                                           // opt/repo-azure-uri-style
        ),                                                                                               // opt/repo-azure-uri-style
    ),                                                                                                   // opt/repo-azure-uri-style
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                              // opt/repo-block
    (                                                                                                              // opt/repo-block
        PARSE_RULE_OPTION_NAME("repo-block"),                                                                      // opt/repo-block
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                           // opt/repo-block
        PARSE_RULE_OPTION_NEGATE(true),                                                                            // opt/repo-block
        PARSE_RULE_OPTION_RESET(true),                                                                             // opt/repo-block
        PARSE_RULE_OPTION_REQUIRED(true),                                                                          // opt/repo-block
        PARSE_RULE_OPTION_SECTION(Global),                                                                         // opt/repo-block
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                          // opt/repo-block
                                                                                                                   // opt/repo-block
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                             // opt/repo-block
        (                                                                                                          // opt/repo-block
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                      // opt/repo-block
        ),                                                                                                         // opt/repo-block
                                                                                                                   // opt/repo-block
        PARSE_RULE_OPTIONAL                                                                                        // opt/repo-block
        (                                                                                                          // opt/repo-block
            PARSE_RULE_OPTIONAL_GROUP                                                                              // opt/repo-block
            (                                                                                                      // opt/repo-block
                PARSE_RULE_OPTIONAL_DEPEND                                                                         // opt/repo-block
                (                                                                                                  // opt/repo-block
                    PARSE_RULE_OPTIONAL_DEPEND_DEFAULT(PARSE_RULE_VAL_BOOL_FALSE),                                 // opt/repo-block
                    PARSE_RULE_VAL_OPT(RepoBundle),                                                                // opt/repo-block
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                      // opt/repo-block
                ),                                                                                                 // opt/repo-block
                                                                                                                   // opt/repo-block
                PARSE_RULE_OPTIONAL_DEFAULT                                                                        // opt/repo-block
                (                                                                                                  // opt/repo-block
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                     // opt/repo-block
                ),                                                                                                 // opt/repo-block
            ),                                                                                                     // opt/repo-block
        ),                                                                                                         // opt/repo-block
    ),                                                                                                             // opt/repo-block
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                      // opt/repo-block-age-map
    (                                                                                                      // opt/repo-block-age-map
        PARSE_RULE_OPTION_NAME("repo-block-age-map"),                                                      // opt/repo-block-age-map
        PARSE_RULE_OPTION_TYPE(Hash),                                                                      // opt/repo-block-age-map
        PARSE_RULE_OPTION_INTERNAL(true)                                                                   // opt/repo-block-age-map
        PARSE_RULE_OPTION_RESET(true),                                                                     // opt/repo-block-age-map
        PARSE_RULE_OPTION_REQUIRED(false),                                                                 // opt/repo-block-age-map
        PARSE_RULE_OPTION_SECTION(Global),                                                                 // opt/repo-block-age-map
        PARSE_RULE_OPTION_MULTI(true),                                                                     // opt/repo-block-age-map
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                  // opt/repo-block-age-map
                                                                                                           // opt/repo-block-age-map
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                     // opt/repo-block-age-map
        (                                                                                                  // opt/repo-block-age-map
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/repo-block-age-map
        ),                                                                                                 // opt/repo-block-age-map
                                                                                                           // opt/repo-block-age-map
        PARSE_RULE_OPTIONAL                                                                                // opt/repo-block-age-map
        (                                                                                                  // opt/repo-block-age-map
            PARSE_RULE_OPTIONAL_GROUP                                                                      // opt/repo-block-age-map
            (                                                                                              // opt/repo-block-age-map
                PARSE_RULE_OPTIONAL_DEPEND                                                                 // opt/repo-block-age-map
                (                                                                                          // opt/repo-block-age-map
                    PARSE_RULE_VAL_OPT(RepoBlock),                                                         // opt/repo-block-age-map
                    PARSE_RULE_VAL_BOOL_TRUE,                                                              // opt/repo-block-age-map
                ),                                                                                         // opt/repo-block-age-map
            ),                                                                                             // opt/repo-block-age-map
        ),                                                                                                 // opt/repo-block-age-map
    ),                                                                                                     // opt/repo-block-age-map
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                            // opt/repo-block-checksum-size-map
    (                                                                                            // opt/repo-block-checksum-size-map
        PARSE_RULE_OPTION_NAME("repo-block-checksum-size-map"),                                  // opt/repo-block-checksum-size-map
        PARSE_RULE_OPTION_TYPE(Hash),                                                            // opt/repo-block-checksum-size-map
        PARSE_RULE_OPTION_INTERNAL(true)                                                         // opt/repo-block-checksum-size-map
        PARSE_RULE_OPTION_RESET(true),                                                           // opt/repo-block-checksum-size-map
        PARSE_RULE_OPTION_REQUIRED(false),                                                       // opt/repo-block-checksum-size-map
        PARSE_RULE_OPTION_SECTION(Global),                                                       // opt/repo-block-checksum-size-map
        PARSE_RULE_OPTION_MULTI(true),                                                           // opt/repo-block-checksum-size-map
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                        // opt/repo-block-checksum-size-map
                                                                                                 // opt/repo-block-checksum-size-map
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                           // opt/repo-block-checksum-size-map
        (                                                                                        // opt/repo-block-checksum-size-map
            PARSE_RULE_OPTION_COMMAND(Backup)                                                    // opt/repo-block-checksum-size-map
        ),                                                                                       // opt/repo-block-checksum-size-map
                                                                                                 // opt/repo-block-checksum-size-map
        PARSE_RULE_OPTIONAL                                                                      // opt/repo-block-checksum-size-map
        (                                                                                        // opt/repo-block-checksum-size-map
            PARSE_RULE_OPTIONAL_GROUP                                                            // opt/repo-block-checksum-size-map
            (                                                                                    // opt/repo-block-checksum-size-map
                PARSE_RULE_OPTIONAL_DEPEND                                                       // opt/repo-block-checksum-size-map
                (                                                                                // opt/repo-block-checksum-size-map
                    PARSE_RULE_VAL_OPT(RepoBlock),                                               // opt/repo-block-checksum-size-map
                    PARSE_RULE_VAL_BOOL_TRUE,                                                    // opt/repo-block-checksum-size-map
                ),                                                                               // opt/repo-block-checksum-size-map
            ),                                                                                   // opt/repo-block-checksum-size-map
        ),                                                                                       // opt/repo-block-checksum-size-map
    ),                                                                                           // opt/repo-block-checksum-size-map
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                     // opt/repo-block-size-map
    (                                                                                                     // opt/repo-block-size-map
        PARSE_RULE_OPTION_NAME("repo-block-size-map"),                                                    // opt/repo-block-size-map
        PARSE_RULE_OPTION_TYPE(Hash),                                                                     // opt/repo-block-size-map
        PARSE_RULE_OPTION_INTERNAL(true)                                                                  // opt/repo-block-size-map
        PARSE_RULE_OPTION_RESET(true),                                                                    // opt/repo-block-size-map
        PARSE_RULE_OPTION_REQUIRED(false),                                                                // opt/repo-block-size-map
        PARSE_RULE_OPTION_SECTION(Global),                                                                // opt/repo-block-size-map
        PARSE_RULE_OPTION_MULTI(true),                                                                    // opt/repo-block-size-map
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                 // opt/repo-block-size-map
                                                                                                          // opt/repo-block-size-map
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                    // opt/repo-block-size-map
        (                                                                                                 // opt/repo-block-size-map
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-block-size-map
        ),                                                                                                // opt/repo-block-size-map
                                                                                                          // opt/repo-block-size-map
        PARSE_RULE_OPTIONAL                                                                               // opt/repo-block-size-map
        (                                                                                                 // opt/repo-block-size-map
            PARSE_RULE_OPTIONAL_GROUP                                                                     // opt/repo-block-size-map
            (                                                                                             // opt/repo-block-size-map
                PARSE_RULE_OPTIONAL_DEPEND                                                                // opt/repo-block-size-map
                (                                                                                         // opt/repo-block-size-map
                    PARSE_RULE_VAL_OPT(RepoBlock),                                                        // opt/repo-block-size-map
                    PARSE_RULE_VAL_BOOL_TRUE,                                                             // opt/repo-block-size-map
                ),                                                                                        // opt/repo-block-size-map
            ),                                                                                            // opt/repo-block-size-map
        ),                                                                                                // opt/repo-block-size-map
    ),                                                                                                    // opt/repo-block-size-map
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                   // opt/repo-block-size-super
    (                                                                                                   // opt/repo-block-size-super
        PARSE_RULE_OPTION_NAME("repo-block-size-super"),                                                // opt/repo-block-size-super
        PARSE_RULE_OPTION_TYPE(Size),                                                                   // opt/repo-block-size-super
        PARSE_RULE_OPTION_INTERNAL(true)                                                                // opt/repo-block-size-super
        PARSE_RULE_OPTION_RESET(true),                                                                  // opt/repo-block-size-super
        PARSE_RULE_OPTION_REQUIRED(true),                                                               // opt/repo-block-size-super
        PARSE_RULE_OPTION_SECTION(Global),                                                              // opt/repo-block-size-super
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                               // opt/repo-block-size-super
                                                                                                        // opt/repo-block-size-super
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                  // opt/repo-block-size-super
        (                                                                                               // opt/repo-block-size-super
            PARSE_RULE_OPTION_COMMAND(Backup)                                                           // opt/repo-block-size-super
        ),                                                                                              // opt/repo-block-size-super
                                                                                                        // opt/repo-block-size-super
        PARSE_RULE_OPTIONAL                                                                             // opt/repo-block-size-super
        (                                                                                               // opt/repo-block-size-super
            PARSE_RULE_OPTIONAL_GROUP                                                                   // opt/repo-block-size-super
            (                                                                                           // opt/repo-block-size-super
                PARSE_RULE_OPTIONAL_DEPEND                                                              // opt/repo-block-size-super
                (                                                                                       // opt/repo-block-size-super
                    PARSE_RULE_VAL_OPT(RepoBlock),                                                      // opt/repo-block-size-super
                    PARSE_RULE_VAL_BOOL_TRUE,                                                           // opt/repo-block-size-super
                ),                                                                                      // opt/repo-block-size-super
                                                                                                        // opt/repo-block-size-super
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                         // opt/repo-block-size-super
                (                                                                                       // opt/repo-block-size-super
                    PARSE_RULE_VAL_SIZE(32KiB),                                                         // opt/repo-block-size-super
                    PARSE_RULE_VAL_SIZE(16MiB),                                                         // opt/repo-block-size-super
                ),                                                                                      // opt/repo-block-size-super
                                                                                                        // opt/repo-block-size-super
                PARSE_RULE_OPTIONAL_DEFAULT                                                             // opt/repo-block-size-super
                (                                                                                       // opt/repo-block-size-super
                    PARSE_RULE_VAL_SIZE(256KiB),                                                        // opt/repo-block-size-super
                ),                                                                                      // opt/repo-block-size-super
            ),                                                                                          // opt/repo-block-size-super
        ),                                                                                              // opt/repo-block-size-super
    ),                                                                                                  // opt/repo-block-size-super
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                              // opt/repo-block-size-super-full
    (                                                                                              // opt/repo-block-size-super-full
        PARSE_RULE_OPTION_NAME("repo-block-size-super-full"),                                      // opt/repo-block-size-super-full
        PARSE_RULE_OPTION_TYPE(Size),                                                              // opt/repo-block-size-super-full
        PARSE_RULE_OPTION_INTERNAL(true)                                                           // opt/repo-block-size-super-full
        PARSE_RULE_OPTION_RESET(true),                                                             // opt/repo-block-size-super-full
        PARSE_RULE_OPTION_REQUIRED(true),                                                          // opt/repo-block-size-super-full
        PARSE_RULE_OPTION_SECTION(Global),                                                         // opt/repo-block-size-super-full
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                          // opt/repo-block-size-super-full
                                                                                                   // opt/repo-block-size-super-full
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                             // opt/repo-block-size-super-full
        (                                                                                          // opt/repo-block-size-super-full
            PARSE_RULE_OPTION_COMMAND(Backup)                                                      // opt/repo-block-size-super-full
        ),                                                                                         // opt/repo-block-size-super-full
                                                                                                   // opt/repo-block-size-super-full
        PARSE_RULE_OPTIONAL                                                                        // opt/repo-block-size-super-full
        (                                                                                          // opt/repo-block-size-super-full
            PARSE_RULE_OPTIONAL_GROUP                                                              // opt/repo-block-size-super-full
            (                                                                                      // opt/repo-block-size-super-full
                PARSE_RULE_OPTIONAL_DEPEND                                                         // opt/repo-block-size-super-full
                (                                                                                  // opt/repo-block-size-super-full
                    PARSE_RULE_VAL_OPT(RepoBlock),                                                 // opt/repo-block-size-super-full
                    PARSE_RULE_VAL_BOOL_TRUE,                                                      // opt/repo-block-size-super-full
                ),                                                                                 // opt/repo-block-size-super-full
                                                                                                   // opt/repo-block-size-super-full
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                    // opt/repo-block-size-super-full
                (                                                                                  // opt/repo-block-size-super-full
                    PARSE_RULE_VAL_SIZE(32KiB),                                                    // opt/repo-block-size-super-full
                    PARSE_RULE_VAL_SIZE(16MiB),                                                    // opt/repo-block-size-super-full
                ),                                                                                 // opt/repo-block-size-super-full
                                                                                                   // opt/repo-block-size-super-full
                PARSE_RULE_OPTIONAL_DEFAULT                                                        // opt/repo-block-size-super-full
                (                                                                                  // opt/repo-block-size-super-full
                    PARSE_RULE_VAL_SIZE(1MiB),                                                     // opt/repo-block-size-super-full
                ),                                                                                 // opt/repo-block-size-super-full
            ),                                                                                     // opt/repo-block-size-super-full
        ),                                                                                         // opt/repo-block-size-super-full
    ),                                                                                             // opt/repo-block-size-super-full
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                             // opt/repo-bundle
    (                                                                                                             // opt/repo-bundle
        PARSE_RULE_OPTION_NAME("repo-bundle"),                                                                    // opt/repo-bundle
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                          // opt/repo-bundle
        PARSE_RULE_OPTION_NEGATE(true),                                                                           // opt/repo-bundle
        PARSE_RULE_OPTION_RESET(true),                                                                            // opt/repo-bundle
        PARSE_RULE_OPTION_REQUIRED(true),                                                                         // opt/repo-bundle
        PARSE_RULE_OPTION_SECTION(Global),                                                                        // opt/repo-bundle
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                         // opt/repo-bundle
                                                                                                                  // opt/repo-bundle
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                            // opt/repo-bundle
        (                                                                                                         // opt/repo-bundle
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/repo-bundle
        ),                                                                                                        // opt/repo-bundle
                                                                                                                  // opt/repo-bundle
        PARSE_RULE_OPTIONAL                                                                                       // opt/repo-bundle
        (                                                                                                         // opt/repo-bundle
            PARSE_RULE_OPTIONAL_GROUP                                                                             // opt/repo-bundle
            (                                                                                                     // opt/repo-bundle
                PARSE_RULE_OPTIONAL_DEFAULT                                                                       // opt/repo-bundle
                (                                                                                                 // opt/repo-bundle
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                    // opt/repo-bundle
                ),                                                                                                // opt/repo-bundle
            ),                                                                                                    // opt/repo-bundle
        ),                                                                                                        // opt/repo-bundle
    ),                                                                                                            // opt/repo-bundle
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                       // opt/repo-bundle-limit
    (                                                                                                       // opt/repo-bundle-limit
        PARSE_RULE_OPTION_NAME("repo-bundle-limit"),                                                        // opt/repo-bundle-limit
        PARSE_RULE_OPTION_TYPE(Size),                                                                       // opt/repo-bundle-limit
        PARSE_RULE_OPTION_RESET(true),                                                                      // opt/repo-bundle-limit
        PARSE_RULE_OPTION_REQUIRED(true),                                                                   // opt/repo-bundle-limit
        PARSE_RULE_OPTION_SECTION(Global),                                                                  // opt/repo-bundle-limit
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                   // opt/repo-bundle-limit
                                                                                                            // opt/repo-bundle-limit
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                      // opt/repo-bundle-limit
        (                                                                                                   // opt/repo-bundle-limit
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-bundle-limit
        ),                                                                                                  // opt/repo-bundle-limit
                                                                                                            // opt/repo-bundle-limit
        PARSE_RULE_OPTIONAL                                                                                 // opt/repo-bundle-limit
        (                                                                                                   // opt/repo-bundle-limit
            PARSE_RULE_OPTIONAL_GROUP                                                                       // opt/repo-bundle-limit
            (                                                                                               // opt/repo-bundle-limit
                PARSE_RULE_OPTIONAL_DEPEND                                                                  // opt/repo-bundle-limit
                (                                                                                           // opt/repo-bundle-limit
                    PARSE_RULE_VAL_OPT(RepoBundle),                                                         // opt/repo-bundle-limit
                    PARSE_RULE_VAL_BOOL_TRUE,                                                               // opt/repo-bundle-limit
                ),                                                                                          // opt/repo-bundle-limit
                                                                                                            // opt/repo-bundle-limit
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                             // opt/repo-bundle-limit
                (                                                                                           // opt/repo-bundle-limit
                    PARSE_RULE_VAL_SIZE(8KiB),                                                              // opt/repo-bundle-limit
                    PARSE_RULE_VAL_SIZE(1PiB),                                                              // opt/repo-bundle-limit
                ),                                                                                          // opt/repo-bundle-limit
                                                                                                            // opt/repo-bundle-limit
                PARSE_RULE_OPTIONAL_DEFAULT                                                                 // opt/repo-bundle-limit
                (                                                                                           // opt/repo-bundle-limit
                    PARSE_RULE_VAL_SIZE(2MiB),                                                              // opt/repo-bundle-limit
                ),                                                                                          // opt/repo-bundle-limit
            ),                                                                                              // opt/repo-bundle-limit
        ),                                                                                                  // opt/repo-bundle-limit
    ),                                                                                                      // opt/repo-bundle-limit
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/repo-bundle-size
    (                                                                                                        // opt/repo-bundle-size
        PARSE_RULE_OPTION_NAME("repo-bundle-size"),                                                          // opt/repo-bundle-size
        PARSE_RULE_OPTION_TYPE(Size),                                                                        // opt/repo-bundle-size
        PARSE_RULE_OPTION_RESET(true),                                                                       // opt/repo-bundle-size
        PARSE_RULE_OPTION_REQUIRED(true),                                                                    // opt/repo-bundle-size
        PARSE_RULE_OPTION_SECTION(Global),                                                                   // opt/repo-bundle-size
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                    // opt/repo-bundle-size
                                                                                                             // opt/repo-bundle-size
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/repo-bundle-size
        (                                                                                                    // opt/repo-bundle-size
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-bundle-size
        ),                                                                                                   // opt/repo-bundle-size
                                                                                                             // opt/repo-bundle-size
        PARSE_RULE_OPTIONAL                                                                                  // opt/repo-bundle-size
        (                                                                                                    // opt/repo-bundle-size
            PARSE_RULE_OPTIONAL_GROUP                                                                        // opt/repo-bundle-size
            (                                                                                                // opt/repo-bundle-size
                PARSE_RULE_OPTIONAL_DEPEND                                                                   // opt/repo-bundle-size
                (                                                                                            // opt/repo-bundle-size
                    PARSE_RULE_VAL_OPT(RepoBundle),                                                          // opt/repo-bundle-size
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                // opt/repo-bundle-size
                ),                                                                                           // opt/repo-bundle-size
                                                                                                             // opt/repo-bundle-size
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                              // opt/repo-bundle-size
                (                                                                                            // opt/repo-bundle-size
                    PARSE_RULE_VAL_SIZE(1MiB),                                                               // opt/repo-bundle-size
                    PARSE_RULE_VAL_SIZE(1PiB),                                                               // opt/repo-bundle-size
                ),                                                                                           // opt/repo-bundle-size
                                                                                                             // opt/repo-bundle-size
                PARSE_RULE_OPTIONAL_DEFAULT                                                                  // opt/repo-bundle-size
                (                                                                                            // opt/repo-bundle-size
                    PARSE_RULE_VAL_SIZE(20MiB),                                                              // opt/repo-bundle-size
                ),                                                                                           // opt/repo-bundle-size
            ),                                                                                               // opt/repo-bundle-size
        ),                                                                                                   // opt/repo-bundle-size
    ),                                                                                                       // opt/repo-bundle-size
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/repo-cipher-pass
    (                                                                                                        // opt/repo-cipher-pass
        PARSE_RULE_OPTION_NAME("repo-cipher-pass"),                                                          // opt/repo-cipher-pass
        PARSE_RULE_OPTION_TYPE(String),                                                                      // opt/repo-cipher-pass
        PARSE_RULE_OPTION_RESET(true),                                                                       // opt/repo-cipher-pass
        PARSE_RULE_OPTION_REQUIRED(true),                                                                    // opt/repo-cipher-pass
        PARSE_RULE_OPTION_SECTION(Global),                                                                   // opt/repo-cipher-pass
        PARSE_RULE_OPTION_SECURE(true),                                                                      // opt/repo-cipher-pass
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                    // opt/repo-cipher-pass
        PARSE_RULE_OPTION_DEPRECATE_MATCH(true),                                                             // opt/repo-cipher-pass
                                                                                                             // opt/repo-cipher-pass
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/repo-cipher-pass
        (                                                                                                    // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-cipher-pass
        ),                                                                                                   // opt/repo-cipher-pass
                                                                                                             // opt/repo-cipher-pass
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                      // opt/repo-cipher-pass
        (                                                                                                    // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-cipher-pass
        ),                                                                                                   // opt/repo-cipher-pass
                                                                                                             // opt/repo-cipher-pass
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                      // opt/repo-cipher-pass
        (                                                                                                    // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-cipher-pass
        ),                                                                                                   // opt/repo-cipher-pass
                                                                                                             // opt/repo-cipher-pass
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                     // opt/repo-cipher-pass
        (                                                                                                    // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/repo-cipher-pass
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-cipher-pass
        ),                                                                                                   // opt/repo-cipher-pass
                                                                                                             // opt/repo-cipher-pass
        PARSE_RULE_OPTIONAL                                                                                  // opt/repo-cipher-pass
        (                                                                                                    // opt/repo-cipher-pass
            PARSE_RULE_OPTIONAL_GROUP                                                                        // opt/repo-cipher-pass
            (                                                                                                // opt/repo-cipher-pass
                PARSE_RULE_OPTIONAL_DEPEND                                                                   // opt/repo-cipher-pass
                (                                                                                            // opt/repo-cipher-pass
                    PARSE_RULE_VAL_OPT(RepoCipherType),                                                      // opt/repo-cipher-pass
                    PARSE_RULE_VAL_STRID(aes_DS_256_DS_cbc),                                                 // opt/repo-cipher-pass
                ),                                                                                           // opt/repo-cipher-pass
            ),                                                                                               // opt/repo-cipher-pass
        ),                                                                                                   // opt/repo-cipher-pass
    ),                                                                                                       // opt/repo-cipher-pass
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/repo-cipher-type
    (                                                                                                        // opt/repo-cipher-type
        PARSE_RULE_OPTION_NAME("repo-cipher-type"),                                                          // opt/repo-cipher-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                                    // opt/repo-cipher-type
        PARSE_RULE_OPTION_RESET(true),                                                                       // opt/repo-cipher-type
        PARSE_RULE_OPTION_REQUIRED(true),                                                                    // opt/repo-cipher-type
        PARSE_RULE_OPTION_SECTION(Global),                                                                   // opt/repo-cipher-type
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                    // opt/repo-cipher-type
        PARSE_RULE_OPTION_DEPRECATE_MATCH(true),                                                             // opt/repo-cipher-type
                                                                                                             // opt/repo-cipher-type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/repo-cipher-type
        (                                                                                                    // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-cipher-type
        ),                                                                                                   // opt/repo-cipher-type
                                                                                                             // opt/repo-cipher-type
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                      // opt/repo-cipher-type
        (                                                                                                    // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-cipher-type
        ),                                                                                                   // opt/repo-cipher-type
                                                                                                             // opt/repo-cipher-type
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                      // opt/repo-cipher-type
        (                                                                                                    // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-cipher-type
        ),                                                                                                   // opt/repo-cipher-type
                                                                                                             // opt/repo-cipher-type
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                     // opt/repo-cipher-type
        (                                                                                                    // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/repo-cipher-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-cipher-type
        ),                                                                                                   // opt/repo-cipher-type
                                                                                                             // opt/repo-cipher-type
        PARSE_RULE_OPTIONAL                                                                                  // opt/repo-cipher-type
        (                                                                                                    // opt/repo-cipher-type
            PARSE_RULE_OPTIONAL_GROUP                                                                        // opt/repo-cipher-type
            (                                                                                                // opt/repo-cipher-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                               // opt/repo-cipher-type
                (                                                                                            // opt/repo-cipher-type
                    PARSE_RULE_VAL_STRID(none),                                                              // opt/repo-cipher-type
                    PARSE_RULE_VAL_STRID(aes_DS_256_DS_cbc),                                                 // opt/repo-cipher-type
                ),                                                                                           // opt/repo-cipher-type
                                                                                                             // opt/repo-cipher-type
                PARSE_RULE_OPTIONAL_DEFAULT                                                                  // opt/repo-cipher-type
                (                                                                                            // opt/repo-cipher-type
                    PARSE_RULE_VAL_STRID(none),                                                              // opt/repo-cipher-type
                ),                                                                                           // opt/repo-cipher-type
            ),                                                                                               // opt/repo-cipher-type
        ),                                                                                                   // opt/repo-cipher-type
    ),                                                                                                       // opt/repo-cipher-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                         // opt/repo-gcs-bucket
    (                                                                                                         // opt/repo-gcs-bucket
        PARSE_RULE_OPTION_NAME("repo-gcs-bucket"),                                                            // opt/repo-gcs-bucket
        PARSE_RULE_OPTION_TYPE(String),                                                                       // opt/repo-gcs-bucket
        PARSE_RULE_OPTION_RESET(true),                                                                        // opt/repo-gcs-bucket
        PARSE_RULE_OPTION_REQUIRED(true),                                                                     // opt/repo-gcs-bucket
        PARSE_RULE_OPTION_SECTION(Global),                                                                    // opt/repo-gcs-bucket
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                     // opt/repo-gcs-bucket
                                                                                                              // opt/repo-gcs-bucket
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                        // opt/repo-gcs-bucket
        (                                                                                                     // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                               // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                             // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                            // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                 // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Check)                                                                  // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                 // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Info)                                                                   // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                               // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                 // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                 // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                           // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                           // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                          // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                 // opt/repo-gcs-bucket
        ),                                                                                                    // opt/repo-gcs-bucket
                                                                                                              // opt/repo-gcs-bucket
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                       // opt/repo-gcs-bucket
        (                                                                                                     // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                             // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                            // opt/repo-gcs-bucket
        ),                                                                                                    // opt/repo-gcs-bucket
                                                                                                              // opt/repo-gcs-bucket
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                       // opt/repo-gcs-bucket
        (                                                                                                     // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                             // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                            // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                 // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                 // opt/repo-gcs-bucket
        ),                                                                                                    // opt/repo-gcs-bucket
                                                                                                              // opt/repo-gcs-bucket
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                      // opt/repo-gcs-bucket
        (                                                                                                     // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                               // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                             // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                            // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                 // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Check)                                                                  // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                 // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Info)                                                                   // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                               // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                 // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                 // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                           // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                           // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                          // opt/repo-gcs-bucket
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                 // opt/repo-gcs-bucket
        ),                                                                                                    // opt/repo-gcs-bucket
                                                                                                              // opt/repo-gcs-bucket
        PARSE_RULE_OPTIONAL                                                                                   // opt/repo-gcs-bucket
        (                                                                                                     // opt/repo-gcs-bucket
            PARSE_RULE_OPTIONAL_GROUP                                                                         // opt/repo-gcs-bucket
            (                                                                                                 // opt/repo-gcs-bucket
                PARSE_RULE_OPTIONAL_DEPEND                                                                    // opt/repo-gcs-bucket
                (                                                                                             // opt/repo-gcs-bucket
                    PARSE_RULE_VAL_OPT(RepoGcsKeyType),                                                       // opt/repo-gcs-bucket
                ),                                                                                            // opt/repo-gcs-bucket
            ),                                                                                                // opt/repo-gcs-bucket
        ),                                                                                                    // opt/repo-gcs-bucket
    ),                                                                                                        // opt/repo-gcs-bucket
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                       // opt/repo-gcs-endpoint
    (                                                                                                       // opt/repo-gcs-endpoint
        PARSE_RULE_OPTION_NAME("repo-gcs-endpoint"),                                                        // opt/repo-gcs-endpoint
        PARSE_RULE_OPTION_TYPE(String),                                                                     // opt/repo-gcs-endpoint
        PARSE_RULE_OPTION_RESET(true),                                                                      // opt/repo-gcs-endpoint
        PARSE_RULE_OPTION_REQUIRED(true),                                                                   // opt/repo-gcs-endpoint
        PARSE_RULE_OPTION_SECTION(Global),                                                                  // opt/repo-gcs-endpoint
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                   // opt/repo-gcs-endpoint
                                                                                                            // opt/repo-gcs-endpoint
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                      // opt/repo-gcs-endpoint
        (                                                                                                   // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-gcs-endpoint
        ),                                                                                                  // opt/repo-gcs-endpoint
                                                                                                            // opt/repo-gcs-endpoint
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                     // opt/repo-gcs-endpoint
        (                                                                                                   // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-gcs-endpoint
        ),                                                                                                  // opt/repo-gcs-endpoint
                                                                                                            // opt/repo-gcs-endpoint
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                     // opt/repo-gcs-endpoint
        (                                                                                                   // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-gcs-endpoint
        ),                                                                                                  // opt/repo-gcs-endpoint
                                                                                                            // opt/repo-gcs-endpoint
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                    // opt/repo-gcs-endpoint
        (                                                                                                   // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/repo-gcs-endpoint
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-gcs-endpoint
        ),                                                                                                  // opt/repo-gcs-endpoint
                                                                                                            // opt/repo-gcs-endpoint
        PARSE_RULE_OPTIONAL                                                                                 // opt/repo-gcs-endpoint
        (                                                                                                   // opt/repo-gcs-endpoint
            PARSE_RULE_OPTIONAL_GROUP                                                                       // opt/repo-gcs-endpoint
            (                                                                                               // opt/repo-gcs-endpoint
                PARSE_RULE_OPTIONAL_DEPEND                                                                  // opt/repo-gcs-endpoint
                (                                                                                           // opt/repo-gcs-endpoint
                    PARSE_RULE_VAL_OPT(RepoGcsKeyType),                                                     // opt/repo-gcs-endpoint
                ),                                                                                          // opt/repo-gcs-endpoint
                                                                                                            // opt/repo-gcs-endpoint
                PARSE_RULE_OPTIONAL_DEFAULT                                                                 // opt/repo-gcs-endpoint
                (                                                                                           // opt/repo-gcs-endpoint
                    PARSE_RULE_VAL_STR(QT_storage_DT_googleapis_DT_com_QT),                                 // opt/repo-gcs-endpoint
                ),                                                                                          // opt/repo-gcs-endpoint
            ),                                                                                              // opt/repo-gcs-endpoint
        ),                                                                                                  // opt/repo-gcs-endpoint
    ),                                                                                                      // opt/repo-gcs-endpoint
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                            // opt/repo-gcs-key
    (                                                                                                            // opt/repo-gcs-key
        PARSE_RULE_OPTION_NAME("repo-gcs-key"),                                                                  // opt/repo-gcs-key
        PARSE_RULE_OPTION_TYPE(String),                                                                          // opt/repo-gcs-key
        PARSE_RULE_OPTION_RESET(true),                                                                           // opt/repo-gcs-key
        PARSE_RULE_OPTION_REQUIRED(true),                                                                        // opt/repo-gcs-key
        PARSE_RULE_OPTION_SECTION(Global),                                                                       // opt/repo-gcs-key
        PARSE_RULE_OPTION_SECURE(true),                                                                          // opt/repo-gcs-key
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                        // opt/repo-gcs-key
                                                                                                                 // opt/repo-gcs-key
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                           // opt/repo-gcs-key
        (                                                                                                        // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                  // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                               // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Check)                                                                     // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                    // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Info)                                                                      // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                  // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                   // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                    // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                   // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                    // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                   // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                              // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                              // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                             // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                    // opt/repo-gcs-key
        ),                                                                                                       // opt/repo-gcs-key
                                                                                                                 // opt/repo-gcs-key
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                          // opt/repo-gcs-key
        (                                                                                                        // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                               // opt/repo-gcs-key
        ),                                                                                                       // opt/repo-gcs-key
                                                                                                                 // opt/repo-gcs-key
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                          // opt/repo-gcs-key
        (                                                                                                        // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                               // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                   // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                    // opt/repo-gcs-key
        ),                                                                                                       // opt/repo-gcs-key
                                                                                                                 // opt/repo-gcs-key
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                         // opt/repo-gcs-key
        (                                                                                                        // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                  // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                               // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Check)                                                                     // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                    // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Info)                                                                      // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                  // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                   // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                    // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                   // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                    // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                   // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                              // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                              // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                             // opt/repo-gcs-key
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                    // opt/repo-gcs-key
        ),                                                                                                       // opt/repo-gcs-key
                                                                                                                 // opt/repo-gcs-key
        PARSE_RULE_OPTIONAL                                                                                      // opt/repo-gcs-key
        (                                                                                                        // opt/repo-gcs-key
            PARSE_RULE_OPTIONAL_GROUP                                                                            // opt/repo-gcs-key
            (                                                                                                    // opt/repo-gcs-key
                PARSE_RULE_OPTIONAL_DEPEND                                                                       // opt/repo-gcs-key
                (                                                                                                // opt/repo-gcs-key
                    PARSE_RULE_VAL_OPT(RepoGcsKeyType),                                                          // opt/repo-gcs-key
                    PARSE_RULE_VAL_STRID(service),                                                               // opt/repo-gcs-key
                    PARSE_RULE_VAL_STRID(token),                                                                 // opt/repo-gcs-key
                ),                                                                                               // opt/repo-gcs-key
            ),                                                                                                   // opt/repo-gcs-key
        ),                                                                                                       // opt/repo-gcs-key
    ),                                                                                                           // opt/repo-gcs-key
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                       // opt/repo-gcs-key-type
    (                                                                                                       // opt/repo-gcs-key-type
        PARSE_RULE_OPTION_NAME("repo-gcs-key-type"),                                                        // opt/repo-gcs-key-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                                   // opt/repo-gcs-key-type
        PARSE_RULE_OPTION_RESET(true),                                                                      // opt/repo-gcs-key-type
        PARSE_RULE_OPTION_REQUIRED(true),                                                                   // opt/repo-gcs-key-type
        PARSE_RULE_OPTION_SECTION(Global),                                                                  // opt/repo-gcs-key-type
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                   // opt/repo-gcs-key-type
                                                                                                            // opt/repo-gcs-key-type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                      // opt/repo-gcs-key-type
        (                                                                                                   // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-gcs-key-type
        ),                                                                                                  // opt/repo-gcs-key-type
                                                                                                            // opt/repo-gcs-key-type
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                     // opt/repo-gcs-key-type
        (                                                                                                   // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-gcs-key-type
        ),                                                                                                  // opt/repo-gcs-key-type
                                                                                                            // opt/repo-gcs-key-type
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                     // opt/repo-gcs-key-type
        (                                                                                                   // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-gcs-key-type
        ),                                                                                                  // opt/repo-gcs-key-type
                                                                                                            // opt/repo-gcs-key-type
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                    // opt/repo-gcs-key-type
        (                                                                                                   // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/repo-gcs-key-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-gcs-key-type
        ),                                                                                                  // opt/repo-gcs-key-type
                                                                                                            // opt/repo-gcs-key-type
        PARSE_RULE_OPTIONAL                                                                                 // opt/repo-gcs-key-type
        (                                                                                                   // opt/repo-gcs-key-type
            PARSE_RULE_OPTIONAL_GROUP                                                                       // opt/repo-gcs-key-type
            (                                                                                               // opt/repo-gcs-key-type
                PARSE_RULE_OPTIONAL_DEPEND                                                                  // opt/repo-gcs-key-type
                (                                                                                           // opt/repo-gcs-key-type
                    PARSE_RULE_VAL_OPT(RepoType),                                                           // opt/repo-gcs-key-type
                    PARSE_RULE_VAL_STRID(gcs),                                                              // opt/repo-gcs-key-type
                ),                                                                                          // opt/repo-gcs-key-type
                                                                                                            // opt/repo-gcs-key-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                              // opt/repo-gcs-key-type
                (                                                                                           // opt/repo-gcs-key-type
                    PARSE_RULE_VAL_STRID(auto),                                                             // opt/repo-gcs-key-type
                    PARSE_RULE_VAL_STRID(service),                                                          // opt/repo-gcs-key-type
                    PARSE_RULE_VAL_STRID(token),                                                            // opt/repo-gcs-key-type
                ),                                                                                          // opt/repo-gcs-key-type
                                                                                                            // opt/repo-gcs-key-type
                PARSE_RULE_OPTIONAL_DEFAULT                                                                 // opt/repo-gcs-key-type
                (                                                                                           // opt/repo-gcs-key-type
                    PARSE_RULE_VAL_STRID(service),                                                          // opt/repo-gcs-key-type
                ),                                                                                          // opt/repo-gcs-key-type
            ),                                                                                              // opt/repo-gcs-key-type
        ),                                                                                                  // opt/repo-gcs-key-type
    ),                                                                                                      // opt/repo-gcs-key-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                   // opt/repo-gcs-user-project
    (                                                                                                   // opt/repo-gcs-user-project
        PARSE_RULE_OPTION_NAME("repo-gcs-user-project"),                                                // opt/repo-gcs-user-project
        PARSE_RULE_OPTION_TYPE(String),                                                                 // opt/repo-gcs-user-project
        PARSE_RULE_OPTION_RESET(true),                                                                  // opt/repo-gcs-user-project
        PARSE_RULE_OPTION_REQUIRED(false),                                                              // opt/repo-gcs-user-project
        PARSE_RULE_OPTION_SECTION(Global),                                                              // opt/repo-gcs-user-project
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                               // opt/repo-gcs-user-project
                                                                                                        // opt/repo-gcs-user-project
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                  // opt/repo-gcs-user-project
        (                                                                                               // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                         // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                       // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                      // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Backup)                                                           // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Check)                                                            // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Expire)                                                           // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Info)                                                             // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                         // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                          // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                           // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                          // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                           // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Restore)                                                          // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                     // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                     // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                    // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Verify)                                                           // opt/repo-gcs-user-project
        ),                                                                                              // opt/repo-gcs-user-project
                                                                                                        // opt/repo-gcs-user-project
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                 // opt/repo-gcs-user-project
        (                                                                                               // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                       // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                      // opt/repo-gcs-user-project
        ),                                                                                              // opt/repo-gcs-user-project
                                                                                                        // opt/repo-gcs-user-project
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                 // opt/repo-gcs-user-project
        (                                                                                               // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                       // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                      // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Backup)                                                           // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Restore)                                                          // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Verify)                                                           // opt/repo-gcs-user-project
        ),                                                                                              // opt/repo-gcs-user-project
                                                                                                        // opt/repo-gcs-user-project
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                // opt/repo-gcs-user-project
        (                                                                                               // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                         // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                       // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                      // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Backup)                                                           // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Check)                                                            // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Expire)                                                           // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Info)                                                             // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                         // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                          // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                           // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                          // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                           // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Restore)                                                          // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                     // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                     // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                    // opt/repo-gcs-user-project
            PARSE_RULE_OPTION_COMMAND(Verify)                                                           // opt/repo-gcs-user-project
        ),                                                                                              // opt/repo-gcs-user-project
    ),                                                                                                  // opt/repo-gcs-user-project
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                           // opt/repo-hardlink
    (                                                                                                           // opt/repo-hardlink
        PARSE_RULE_OPTION_NAME("repo-hardlink"),                                                                // opt/repo-hardlink
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                        // opt/repo-hardlink
        PARSE_RULE_OPTION_NEGATE(true),                                                                         // opt/repo-hardlink
        PARSE_RULE_OPTION_RESET(true),                                                                          // opt/repo-hardlink
        PARSE_RULE_OPTION_REQUIRED(true),                                                                       // opt/repo-hardlink
        PARSE_RULE_OPTION_SECTION(Global),                                                                      // opt/repo-hardlink
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                       // opt/repo-hardlink
                                                                                                                // opt/repo-hardlink
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                          // opt/repo-hardlink
        (                                                                                                       // opt/repo-hardlink
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/repo-hardlink
        ),                                                                                                      // opt/repo-hardlink
                                                                                                                // opt/repo-hardlink
        PARSE_RULE_OPTIONAL                                                                                     // opt/repo-hardlink
        (                                                                                                       // opt/repo-hardlink
            PARSE_RULE_OPTIONAL_GROUP                                                                           // opt/repo-hardlink
            (                                                                                                   // opt/repo-hardlink
                PARSE_RULE_OPTIONAL_DEPEND                                                                      // opt/repo-hardlink
                (                                                                                               // opt/repo-hardlink
                    PARSE_RULE_OPTIONAL_DEPEND_DEFAULT(PARSE_RULE_VAL_BOOL_FALSE),                              // opt/repo-hardlink
                    PARSE_RULE_VAL_OPT(RepoBundle),                                                             // opt/repo-hardlink
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                  // opt/repo-hardlink
                ),                                                                                              // opt/repo-hardlink
                                                                                                                // opt/repo-hardlink
                PARSE_RULE_OPTIONAL_DEFAULT                                                                     // opt/repo-hardlink
                (                                                                                               // opt/repo-hardlink
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                  // opt/repo-hardlink
                ),                                                                                              // opt/repo-hardlink
            ),                                                                                                  // opt/repo-hardlink
        ),                                                                                                      // opt/repo-hardlink
    ),                                                                                                          // opt/repo-hardlink
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                               // opt/repo-host
    (                                                                                                               // opt/repo-host
        PARSE_RULE_OPTION_NAME("repo-host"),                                                                        // opt/repo-host
        PARSE_RULE_OPTION_TYPE(String),                                                                             // opt/repo-host
        PARSE_RULE_OPTION_RESET(true),                                                                              // opt/repo-host
        PARSE_RULE_OPTION_REQUIRED(false),                                                                          // opt/repo-host
        PARSE_RULE_OPTION_SECTION(Global),                                                                          // opt/repo-host
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                           // opt/repo-host
                                                                                                                    // opt/repo-host
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                              // opt/repo-host
        (                                                                                                           // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                     // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(Check)                                                                        // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                       // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(Info)                                                                         // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                     // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                      // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                       // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                      // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                       // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                 // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                 // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/repo-host
        ),                                                                                                          // opt/repo-host
                                                                                                                    // opt/repo-host
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                             // opt/repo-host
        (                                                                                                           // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/repo-host
        ),                                                                                                          // opt/repo-host
                                                                                                                    // opt/repo-host
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                             // opt/repo-host
        (                                                                                                           // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/repo-host
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/repo-host
        ),                                                                                                          // opt/repo-host
                                                                                                                    // opt/repo-host
        PARSE_RULE_OPTIONAL                                                                                         // opt/repo-host
        (                                                                                                           // opt/repo-host
            PARSE_RULE_OPTIONAL_GROUP                                                                               // opt/repo-host
            (                                                                                                       // opt/repo-host
                PARSE_RULE_OPTIONAL_DEPEND                                                                          // opt/repo-host
                (                                                                                                   // opt/repo-host
                    PARSE_RULE_VAL_OPT(RepoLocal),                                                                  // opt/repo-host
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                      // opt/repo-host
                ),                                                                                                  // opt/repo-host
            ),                                                                                                      // opt/repo-host
        ),                                                                                                          // opt/repo-host
    ),                                                                                                              // opt/repo-host
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                       // opt/repo-host-ca-file
    (                                                                                                       // opt/repo-host-ca-file
        PARSE_RULE_OPTION_NAME("repo-host-ca-file"),                                                        // opt/repo-host-ca-file
        PARSE_RULE_OPTION_TYPE(String),                                                                     // opt/repo-host-ca-file
        PARSE_RULE_OPTION_RESET(true),                                                                      // opt/repo-host-ca-file
        PARSE_RULE_OPTION_REQUIRED(false),                                                                  // opt/repo-host-ca-file
        PARSE_RULE_OPTION_SECTION(Global),                                                                  // opt/repo-host-ca-file
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                   // opt/repo-host-ca-file
                                                                                                            // opt/repo-host-ca-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                      // opt/repo-host-ca-file
        (                                                                                                   // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-host-ca-file
        ),                                                                                                  // opt/repo-host-ca-file
                                                                                                            // opt/repo-host-ca-file
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                     // opt/repo-host-ca-file
        (                                                                                                   // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-host-ca-file
        ),                                                                                                  // opt/repo-host-ca-file
                                                                                                            // opt/repo-host-ca-file
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                     // opt/repo-host-ca-file
        (                                                                                                   // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-host-ca-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-host-ca-file
        ),                                                                                                  // opt/repo-host-ca-file
                                                                                                            // opt/repo-host-ca-file
        PARSE_RULE_OPTIONAL                                                                                 // opt/repo-host-ca-file
        (                                                                                                   // opt/repo-host-ca-file
            PARSE_RULE_OPTIONAL_GROUP                                                                       // opt/repo-host-ca-file
            (                                                                                               // opt/repo-host-ca-file
                PARSE_RULE_OPTIONAL_DEPEND                                                                  // opt/repo-host-ca-file
                (                                                                                           // opt/repo-host-ca-file
                    PARSE_RULE_VAL_OPT(RepoHostType),                                                       // opt/repo-host-ca-file
                    PARSE_RULE_VAL_STRID(tls),                                                              // opt/repo-host-ca-file
                ),                                                                                          // opt/repo-host-ca-file
            ),                                                                                              // opt/repo-host-ca-file
        ),                                                                                                  // opt/repo-host-ca-file
    ),                                                                                                      // opt/repo-host-ca-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                       // opt/repo-host-ca-path
    (                                                                                                       // opt/repo-host-ca-path
        PARSE_RULE_OPTION_NAME("repo-host-ca-path"),                                                        // opt/repo-host-ca-path
        PARSE_RULE_OPTION_TYPE(String),                                                                     // opt/repo-host-ca-path
        PARSE_RULE_OPTION_RESET(true),                                                                      // opt/repo-host-ca-path
        PARSE_RULE_OPTION_REQUIRED(false),                                                                  // opt/repo-host-ca-path
        PARSE_RULE_OPTION_SECTION(Global),                                                                  // opt/repo-host-ca-path
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                   // opt/repo-host-ca-path
                                                                                                            // opt/repo-host-ca-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                      // opt/repo-host-ca-path
        (                                                                                                   // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-host-ca-path
        ),                                                                                                  // opt/repo-host-ca-path
                                                                                                            // opt/repo-host-ca-path
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                     // opt/repo-host-ca-path
        (                                                                                                   // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-host-ca-path
        ),                                                                                                  // opt/repo-host-ca-path
                                                                                                            // opt/repo-host-ca-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                     // opt/repo-host-ca-path
        (                                                                                                   // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-host-ca-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-host-ca-path
        ),                                                                                                  // opt/repo-host-ca-path
                                                                                                            // opt/repo-host-ca-path
        PARSE_RULE_OPTIONAL                                                                                 // opt/repo-host-ca-path
        (                                                                                                   // opt/repo-host-ca-path
            PARSE_RULE_OPTIONAL_GROUP                                                                       // opt/repo-host-ca-path
            (                                                                                               // opt/repo-host-ca-path
                PARSE_RULE_OPTIONAL_DEPEND                                                                  // opt/repo-host-ca-path
                (                                                                                           // opt/repo-host-ca-path
                    PARSE_RULE_VAL_OPT(RepoHostType),                                                       // opt/repo-host-ca-path
                    PARSE_RULE_VAL_STRID(tls),                                                              // opt/repo-host-ca-path
                ),                                                                                          // opt/repo-host-ca-path
            ),                                                                                              // opt/repo-host-ca-path
        ),                                                                                                  // opt/repo-host-ca-path
    ),                                                                                                      // opt/repo-host-ca-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                     // opt/repo-host-cert-file
    (                                                                                                     // opt/repo-host-cert-file
        PARSE_RULE_OPTION_NAME("repo-host-cert-file"),                                                    // opt/repo-host-cert-file
        PARSE_RULE_OPTION_TYPE(String),                                                                   // opt/repo-host-cert-file
        PARSE_RULE_OPTION_RESET(true),                                                                    // opt/repo-host-cert-file
        PARSE_RULE_OPTION_REQUIRED(true),                                                                 // opt/repo-host-cert-file
        PARSE_RULE_OPTION_SECTION(Global),                                                                // opt/repo-host-cert-file
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                 // opt/repo-host-cert-file
                                                                                                          // opt/repo-host-cert-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                    // opt/repo-host-cert-file
        (                                                                                                 // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-host-cert-file
        ),                                                                                                // opt/repo-host-cert-file
                                                                                                          // opt/repo-host-cert-file
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                   // opt/repo-host-cert-file
        (                                                                                                 // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-host-cert-file
        ),                                                                                                // opt/repo-host-cert-file
                                                                                                          // opt/repo-host-cert-file
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                   // opt/repo-host-cert-file
        (                                                                                                 // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-host-cert-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-host-cert-file
        ),                                                                                                // opt/repo-host-cert-file
                                                                                                          // opt/repo-host-cert-file
        PARSE_RULE_OPTIONAL                                                                               // opt/repo-host-cert-file
        (                                                                                                 // opt/repo-host-cert-file
            PARSE_RULE_OPTIONAL_GROUP                                                                     // opt/repo-host-cert-file
            (                                                                                             // opt/repo-host-cert-file
                PARSE_RULE_OPTIONAL_DEPEND                                                                // opt/repo-host-cert-file
                (                                                                                         // opt/repo-host-cert-file
                    PARSE_RULE_VAL_OPT(RepoHostType),                                                     // opt/repo-host-cert-file
                    PARSE_RULE_VAL_STRID(tls),                                                            // opt/repo-host-cert-file
                ),                                                                                        // opt/repo-host-cert-file
            ),                                                                                            // opt/repo-host-cert-file
        ),                                                                                                // opt/repo-host-cert-file
    ),                                                                                                    // opt/repo-host-cert-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                           // opt/repo-host-cmd
    (                                                                                                           // opt/repo-host-cmd
        PARSE_RULE_OPTION_NAME("repo-host-cmd"),                                                                // opt/repo-host-cmd
        PARSE_RULE_OPTION_TYPE(String),                                                                         // opt/repo-host-cmd
        PARSE_RULE_OPTION_DEFAULT_TYPE(Dynamic),                                                                // opt/repo-host-cmd
        PARSE_RULE_OPTION_RESET(true),                                                                          // opt/repo-host-cmd
        PARSE_RULE_OPTION_REQUIRED(true),                                                                       // opt/repo-host-cmd
        PARSE_RULE_OPTION_SECTION(Global),                                                                      // opt/repo-host-cmd
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                       // opt/repo-host-cmd
                                                                                                                // opt/repo-host-cmd
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                          // opt/repo-host-cmd
        (                                                                                                       // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                 // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(Check)                                                                    // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                   // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(Info)                                                                     // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                 // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                  // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                   // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                  // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                   // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                             // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                             // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                            // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/repo-host-cmd
        ),                                                                                                      // opt/repo-host-cmd
                                                                                                                // opt/repo-host-cmd
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                         // opt/repo-host-cmd
        (                                                                                                       // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/repo-host-cmd
        ),                                                                                                      // opt/repo-host-cmd
                                                                                                                // opt/repo-host-cmd
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                         // opt/repo-host-cmd
        (                                                                                                       // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/repo-host-cmd
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/repo-host-cmd
        ),                                                                                                      // opt/repo-host-cmd
                                                                                                                // opt/repo-host-cmd
        PARSE_RULE_OPTIONAL                                                                                     // opt/repo-host-cmd
        (                                                                                                       // opt/repo-host-cmd
            PARSE_RULE_OPTIONAL_GROUP                                                                           // opt/repo-host-cmd
            (                                                                                                   // opt/repo-host-cmd
                PARSE_RULE_OPTIONAL_DEPEND                                                                      // opt/repo-host-cmd
                (                                                                                               // opt/repo-host-cmd
                    PARSE_RULE_VAL_OPT(RepoHostType),                                                           // opt/repo-host-cmd
                    PARSE_RULE_VAL_STRID(ssh),                                                                  // opt/repo-host-cmd
                ),                                                                                              // opt/repo-host-cmd
                                                                                                                // opt/repo-host-cmd
                PARSE_RULE_OPTIONAL_DEFAULT                                                                     // opt/repo-host-cmd
                (                                                                                               // opt/repo-host-cmd
                    PARSE_RULE_DEFAULT_DYNAMIC(Bin),                                                            // opt/repo-host-cmd
                ),                                                                                              // opt/repo-host-cmd
            ),                                                                                                  // opt/repo-host-cmd
        ),                                                                                                      // opt/repo-host-cmd
    ),                                                                                                          // opt/repo-host-cmd
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/repo-host-config
    (                                                                                                        // opt/repo-host-config
        PARSE_RULE_OPTION_NAME("repo-host-config"),                                                          // opt/repo-host-config
        PARSE_RULE_OPTION_TYPE(String),                                                                      // opt/repo-host-config
        PARSE_RULE_OPTION_RESET(true),                                                                       // opt/repo-host-config
        PARSE_RULE_OPTION_REQUIRED(true),                                                                    // opt/repo-host-config
        PARSE_RULE_OPTION_SECTION(Global),                                                                   // opt/repo-host-config
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                    // opt/repo-host-config
                                                                                                             // opt/repo-host-config
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/repo-host-config
        (                                                                                                    // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-host-config
        ),                                                                                                   // opt/repo-host-config
                                                                                                             // opt/repo-host-config
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                      // opt/repo-host-config
        (                                                                                                    // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-host-config
        ),                                                                                                   // opt/repo-host-config
                                                                                                             // opt/repo-host-config
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                      // opt/repo-host-config
        (                                                                                                    // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-host-config
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-host-config
        ),                                                                                                   // opt/repo-host-config
                                                                                                             // opt/repo-host-config
        PARSE_RULE_OPTIONAL                                                                                  // opt/repo-host-config
        (                                                                                                    // opt/repo-host-config
            PARSE_RULE_OPTIONAL_GROUP                                                                        // opt/repo-host-config
            (                                                                                                // opt/repo-host-config
                PARSE_RULE_OPTIONAL_DEPEND                                                                   // opt/repo-host-config
                (                                                                                            // opt/repo-host-config
                    PARSE_RULE_VAL_OPT(RepoHost),                                                            // opt/repo-host-config
                ),                                                                                           // opt/repo-host-config
                                                                                                             // opt/repo-host-config
                PARSE_RULE_OPTIONAL_DEFAULT                                                                  // opt/repo-host-config
                (                                                                                            // opt/repo-host-config
                    PARSE_RULE_VAL_STR(CFGOPTDEF_CONFIG_PATH_SP_QT_FS_QT_SP_PROJECT_CONFIG_FILE),            // opt/repo-host-config
                ),                                                                                           // opt/repo-host-config
            ),                                                                                               // opt/repo-host-config
        ),                                                                                                   // opt/repo-host-config
    ),                                                                                                       // opt/repo-host-config
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                           // opt/repo-host-config-include-path
    (                                                                                           // opt/repo-host-config-include-path
        PARSE_RULE_OPTION_NAME("repo-host-config-include-path"),                                // opt/repo-host-config-include-path
        PARSE_RULE_OPTION_TYPE(Path),                                                           // opt/repo-host-config-include-path
        PARSE_RULE_OPTION_RESET(true),                                                          // opt/repo-host-config-include-path
        PARSE_RULE_OPTION_REQUIRED(true),                                                       // opt/repo-host-config-include-path
        PARSE_RULE_OPTION_SECTION(Global),                                                      // opt/repo-host-config-include-path
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                       // opt/repo-host-config-include-path
                                                                                                // opt/repo-host-config-include-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                          // opt/repo-host-config-include-path
        (                                                                                       // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                 // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                               // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                              // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                   // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                    // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                   // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                     // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                 // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                  // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                   // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                  // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                   // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                  // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                             // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                             // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                            // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                   // opt/repo-host-config-include-path
        ),                                                                                      // opt/repo-host-config-include-path
                                                                                                // opt/repo-host-config-include-path
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                         // opt/repo-host-config-include-path
        (                                                                                       // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                               // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                              // opt/repo-host-config-include-path
        ),                                                                                      // opt/repo-host-config-include-path
                                                                                                // opt/repo-host-config-include-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                         // opt/repo-host-config-include-path
        (                                                                                       // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                               // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                              // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                   // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                  // opt/repo-host-config-include-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                   // opt/repo-host-config-include-path
        ),                                                                                      // opt/repo-host-config-include-path
                                                                                                // opt/repo-host-config-include-path
        PARSE_RULE_OPTIONAL                                                                     // opt/repo-host-config-include-path
        (                                                                                       // opt/repo-host-config-include-path
            PARSE_RULE_OPTIONAL_GROUP                                                           // opt/repo-host-config-include-path
            (                                                                                   // opt/repo-host-config-include-path
                PARSE_RULE_OPTIONAL_DEPEND                                                      // opt/repo-host-config-include-path
                (                                                                               // opt/repo-host-config-include-path
                    PARSE_RULE_VAL_OPT(RepoHost),                                               // opt/repo-host-config-include-path
                ),                                                                              // opt/repo-host-config-include-path
                                                                                                // opt/repo-host-config-include-path
                PARSE_RULE_OPTIONAL_DEFAULT                                                     // opt/repo-host-config-include-path
                (                                                                               // opt/repo-host-config-include-path
                    PARSE_RULE_VAL_STR(CFGOPTDEF_CONFIG_PATH_SP_QT_FS_QT_SP_PROJECT_CONFIG_INCLUDE_PATH),
                ),                                                                              // opt/repo-host-config-include-path
            ),                                                                                  // opt/repo-host-config-include-path
        ),                                                                                      // opt/repo-host-config-include-path
    ),                                                                                          // opt/repo-host-config-include-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                   // opt/repo-host-config-path
    (                                                                                                   // opt/repo-host-config-path
        PARSE_RULE_OPTION_NAME("repo-host-config-path"),                                                // opt/repo-host-config-path
        PARSE_RULE_OPTION_TYPE(Path),                                                                   // opt/repo-host-config-path
        PARSE_RULE_OPTION_RESET(true),                                                                  // opt/repo-host-config-path
        PARSE_RULE_OPTION_REQUIRED(true),                                                               // opt/repo-host-config-path
        PARSE_RULE_OPTION_SECTION(Global),                                                              // opt/repo-host-config-path
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                               // opt/repo-host-config-path
                                                                                                        // opt/repo-host-config-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                  // opt/repo-host-config-path
        (                                                                                               // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                         // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                       // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                      // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                           // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                            // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                           // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                             // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                         // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                          // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                           // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                          // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                           // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                          // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                     // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                     // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                    // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                           // opt/repo-host-config-path
        ),                                                                                              // opt/repo-host-config-path
                                                                                                        // opt/repo-host-config-path
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                 // opt/repo-host-config-path
        (                                                                                               // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                       // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                      // opt/repo-host-config-path
        ),                                                                                              // opt/repo-host-config-path
                                                                                                        // opt/repo-host-config-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                 // opt/repo-host-config-path
        (                                                                                               // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                       // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                      // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                           // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                          // opt/repo-host-config-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                           // opt/repo-host-config-path
        ),                                                                                              // opt/repo-host-config-path
                                                                                                        // opt/repo-host-config-path
        PARSE_RULE_OPTIONAL                                                                             // opt/repo-host-config-path
        (                                                                                               // opt/repo-host-config-path
            PARSE_RULE_OPTIONAL_GROUP                                                                   // opt/repo-host-config-path
            (                                                                                           // opt/repo-host-config-path
                PARSE_RULE_OPTIONAL_DEPEND                                                              // opt/repo-host-config-path
                (                                                                                       // opt/repo-host-config-path
                    PARSE_RULE_VAL_OPT(RepoHost),                                                       // opt/repo-host-config-path
                ),                                                                                      // opt/repo-host-config-path
                                                                                                        // opt/repo-host-config-path
                PARSE_RULE_OPTIONAL_DEFAULT                                                             // opt/repo-host-config-path
                (                                                                                       // opt/repo-host-config-path
                    PARSE_RULE_VAL_STR(CFGOPTDEF_CONFIG_PATH),                                          // opt/repo-host-config-path
                ),                                                                                      // opt/repo-host-config-path
            ),                                                                                          // opt/repo-host-config-path
        ),                                                                                              // opt/repo-host-config-path
    ),                                                                                                  // opt/repo-host-config-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                      // opt/repo-host-key-file
    (                                                                                                      // opt/repo-host-key-file
        PARSE_RULE_OPTION_NAME("repo-host-key-file"),                                                      // opt/repo-host-key-file
        PARSE_RULE_OPTION_TYPE(String),                                                                    // opt/repo-host-key-file
        PARSE_RULE_OPTION_RESET(true),                                                                     // opt/repo-host-key-file
        PARSE_RULE_OPTION_REQUIRED(true),                                                                  // opt/repo-host-key-file
        PARSE_RULE_OPTION_SECTION(Global),                                                                 // opt/repo-host-key-file
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                  // opt/repo-host-key-file
                                                                                                           // opt/repo-host-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                     // opt/repo-host-key-file
        (                                                                                                  // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                            // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                               // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(Expire)                                                              // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(Info)                                                                // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                            // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                             // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                              // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                             // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                              // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                        // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                        // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                       // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/repo-host-key-file
        ),                                                                                                 // opt/repo-host-key-file
                                                                                                           // opt/repo-host-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                    // opt/repo-host-key-file
        (                                                                                                  // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-host-key-file
        ),                                                                                                 // opt/repo-host-key-file
                                                                                                           // opt/repo-host-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                    // opt/repo-host-key-file
        (                                                                                                  // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/repo-host-key-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/repo-host-key-file
        ),                                                                                                 // opt/repo-host-key-file
                                                                                                           // opt/repo-host-key-file
        PARSE_RULE_OPTIONAL                                                                                // opt/repo-host-key-file
        (                                                                                                  // opt/repo-host-key-file
            PARSE_RULE_OPTIONAL_GROUP                                                                      // opt/repo-host-key-file
            (                                                                                              // opt/repo-host-key-file
                PARSE_RULE_OPTIONAL_DEPEND                                                                 // opt/repo-host-key-file
                (                                                                                          // opt/repo-host-key-file
                    PARSE_RULE_VAL_OPT(RepoHostType),                                                      // opt/repo-host-key-file
                    PARSE_RULE_VAL_STRID(tls),                                                             // opt/repo-host-key-file
                ),                                                                                         // opt/repo-host-key-file
            ),                                                                                             // opt/repo-host-key-file
        ),                                                                                                 // opt/repo-host-key-file
    ),                                                                                                     // opt/repo-host-key-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/repo-host-port
    (                                                                                                          // opt/repo-host-port
        PARSE_RULE_OPTION_NAME("repo-host-port"),                                                              // opt/repo-host-port
        PARSE_RULE_OPTION_TYPE(Integer),                                                                       // opt/repo-host-port
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/repo-host-port
        PARSE_RULE_OPTION_REQUIRED(false),                                                                     // opt/repo-host-port
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/repo-host-port
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                      // opt/repo-host-port
                                                                                                               // opt/repo-host-port
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/repo-host-port
        (                                                                                                      // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-host-port
        ),                                                                                                     // opt/repo-host-port
                                                                                                               // opt/repo-host-port
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                        // opt/repo-host-port
        (                                                                                                      // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-host-port
        ),                                                                                                     // opt/repo-host-port
                                                                                                               // opt/repo-host-port
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                        // opt/repo-host-port
        (                                                                                                      // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-host-port
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-host-port
        ),                                                                                                     // opt/repo-host-port
                                                                                                               // opt/repo-host-port
        PARSE_RULE_OPTIONAL                                                                                    // opt/repo-host-port
        (                                                                                                      // opt/repo-host-port
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/repo-host-port
            (                                                                                                  // opt/repo-host-port
                PARSE_RULE_OPTIONAL_DEPEND                                                                     // opt/repo-host-port
                (                                                                                              // opt/repo-host-port
                    PARSE_RULE_VAL_OPT(RepoHostType),                                                          // opt/repo-host-port
                    PARSE_RULE_VAL_STRID(ssh),                                                                 // opt/repo-host-port
                    PARSE_RULE_VAL_STRID(tls),                                                                 // opt/repo-host-port
                ),                                                                                             // opt/repo-host-port
                                                                                                               // opt/repo-host-port
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                                // opt/repo-host-port
                (                                                                                              // opt/repo-host-port
                    PARSE_RULE_VAL_INT(0),                                                                     // opt/repo-host-port
                    PARSE_RULE_VAL_INT(65535),                                                                 // opt/repo-host-port
                ),                                                                                             // opt/repo-host-port
                                                                                                               // opt/repo-host-port
                PARSE_RULE_OPTIONAL_DEFAULT                                                                    // opt/repo-host-port
                (                                                                                              // opt/repo-host-port
                    PARSE_RULE_OPTIONAL_DEFAULT_MAP                                                            // opt/repo-host-port
                    (                                                                                          // opt/repo-host-port
                        PARSE_RULE_VAL_STRID(tls),                                                             // opt/repo-host-port
                        PARSE_RULE_VAL_INT(8432),                                                              // opt/repo-host-port
                    ),                                                                                         // opt/repo-host-port
                ),                                                                                             // opt/repo-host-port
            ),                                                                                                 // opt/repo-host-port
        ),                                                                                                     // opt/repo-host-port
    ),                                                                                                         // opt/repo-host-port
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/repo-host-type
    (                                                                                                          // opt/repo-host-type
        PARSE_RULE_OPTION_NAME("repo-host-type"),                                                              // opt/repo-host-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                                      // opt/repo-host-type
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/repo-host-type
        PARSE_RULE_OPTION_REQUIRED(true),                                                                      // opt/repo-host-type
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/repo-host-type
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                      // opt/repo-host-type
                                                                                                               // opt/repo-host-type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/repo-host-type
        (                                                                                                      // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-host-type
        ),                                                                                                     // opt/repo-host-type
                                                                                                               // opt/repo-host-type
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                        // opt/repo-host-type
        (                                                                                                      // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-host-type
        ),                                                                                                     // opt/repo-host-type
                                                                                                               // opt/repo-host-type
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                        // opt/repo-host-type
        (                                                                                                      // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-host-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-host-type
        ),                                                                                                     // opt/repo-host-type
                                                                                                               // opt/repo-host-type
        PARSE_RULE_OPTIONAL                                                                                    // opt/repo-host-type
        (                                                                                                      // opt/repo-host-type
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/repo-host-type
            (                                                                                                  // opt/repo-host-type
                PARSE_RULE_OPTIONAL_DEPEND                                                                     // opt/repo-host-type
                (                                                                                              // opt/repo-host-type
                    PARSE_RULE_VAL_OPT(RepoHost),                                                              // opt/repo-host-type
                ),                                                                                             // opt/repo-host-type
                                                                                                               // opt/repo-host-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                 // opt/repo-host-type
                (                                                                                              // opt/repo-host-type
                    PARSE_RULE_VAL_STRID(ssh),                                                                 // opt/repo-host-type
                    PARSE_RULE_VAL_STRID(tls),                                                                 // opt/repo-host-type
                ),                                                                                             // opt/repo-host-type
                                                                                                               // opt/repo-host-type
                PARSE_RULE_OPTIONAL_DEFAULT                                                                    // opt/repo-host-type
                (                                                                                              // opt/repo-host-type
                    PARSE_RULE_VAL_STRID(ssh),                                                                 // opt/repo-host-type
                ),                                                                                             // opt/repo-host-type
            ),                                                                                                 // opt/repo-host-type
        ),                                                                                                     // opt/repo-host-type
    ),                                                                                                         // opt/repo-host-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/repo-host-user
    (                                                                                                          // opt/repo-host-user
        PARSE_RULE_OPTION_NAME("repo-host-user"),                                                              // opt/repo-host-user
        PARSE_RULE_OPTION_TYPE(String),                                                                        // opt/repo-host-user
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/repo-host-user
        PARSE_RULE_OPTION_REQUIRED(false),                                                                     // opt/repo-host-user
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/repo-host-user
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                      // opt/repo-host-user
                                                                                                               // opt/repo-host-user
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/repo-host-user
        (                                                                                                      // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-host-user
        ),                                                                                                     // opt/repo-host-user
                                                                                                               // opt/repo-host-user
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                        // opt/repo-host-user
        (                                                                                                      // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-host-user
        ),                                                                                                     // opt/repo-host-user
                                                                                                               // opt/repo-host-user
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                        // opt/repo-host-user
        (                                                                                                      // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-host-user
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-host-user
        ),                                                                                                     // opt/repo-host-user
                                                                                                               // opt/repo-host-user
        PARSE_RULE_OPTIONAL                                                                                    // opt/repo-host-user
        (                                                                                                      // opt/repo-host-user
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/repo-host-user
            (                                                                                                  // opt/repo-host-user
                PARSE_RULE_OPTIONAL_DEPEND                                                                     // opt/repo-host-user
                (                                                                                              // opt/repo-host-user
                    PARSE_RULE_VAL_OPT(RepoHostType),                                                          // opt/repo-host-user
                    PARSE_RULE_VAL_STRID(ssh),                                                                 // opt/repo-host-user
                ),                                                                                             // opt/repo-host-user
                                                                                                               // opt/repo-host-user
                PARSE_RULE_OPTIONAL_DEFAULT                                                                    // opt/repo-host-user
                (                                                                                              // opt/repo-host-user
                    PARSE_RULE_VAL_STR(QT_pgbackrest_QT),                                                      // opt/repo-host-user
                ),                                                                                             // opt/repo-host-user
            ),                                                                                                 // opt/repo-host-user
        ),                                                                                                     // opt/repo-host-user
    ),                                                                                                         // opt/repo-host-user
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                              // opt/repo-local
    (                                                                                                              // opt/repo-local
        PARSE_RULE_OPTION_NAME("repo-local"),                                                                      // opt/repo-local
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                           // opt/repo-local
        PARSE_RULE_OPTION_INTERNAL(true)                                                                           // opt/repo-local
        PARSE_RULE_OPTION_NEGATE(true),                                                                            // opt/repo-local
        PARSE_RULE_OPTION_RESET(true),                                                                             // opt/repo-local
        PARSE_RULE_OPTION_REQUIRED(true),                                                                          // opt/repo-local
        PARSE_RULE_OPTION_SECTION(Global),                                                                         // opt/repo-local
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                          // opt/repo-local
                                                                                                                   // opt/repo-local
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                             // opt/repo-local
        (                                                                                                          // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                    // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                  // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                 // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                      // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(Check)                                                                       // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                      // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(Info)                                                                        // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                    // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                     // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                      // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                     // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                      // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                     // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                               // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                      // opt/repo-local
        ),                                                                                                         // opt/repo-local
                                                                                                                   // opt/repo-local
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                            // opt/repo-local
        (                                                                                                          // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                  // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                 // opt/repo-local
        ),                                                                                                         // opt/repo-local
                                                                                                                   // opt/repo-local
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                            // opt/repo-local
        (                                                                                                          // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                  // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                 // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                      // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                     // opt/repo-local
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                      // opt/repo-local
        ),                                                                                                         // opt/repo-local
                                                                                                                   // opt/repo-local
        PARSE_RULE_OPTIONAL                                                                                        // opt/repo-local
        (                                                                                                          // opt/repo-local
            PARSE_RULE_OPTIONAL_GROUP                                                                              // opt/repo-local
            (                                                                                                      // opt/repo-local
                PARSE_RULE_OPTIONAL_DEFAULT                                                                        // opt/repo-local
                (                                                                                                  // opt/repo-local
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                     // opt/repo-local
                ),                                                                                                 // opt/repo-local
            ),                                                                                                     // opt/repo-local
        ),                                                                                                         // opt/repo-local
    ),                                                                                                             // opt/repo-local
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                               // opt/repo-path
    (                                                                                                               // opt/repo-path
        PARSE_RULE_OPTION_NAME("repo-path"),                                                                        // opt/repo-path
        PARSE_RULE_OPTION_TYPE(Path),                                                                               // opt/repo-path
        PARSE_RULE_OPTION_RESET(true),                                                                              // opt/repo-path
        PARSE_RULE_OPTION_REQUIRED(true),                                                                           // opt/repo-path
        PARSE_RULE_OPTION_SECTION(Global),                                                                          // opt/repo-path
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                           // opt/repo-path
        PARSE_RULE_OPTION_DEPRECATE_MATCH(true),                                                                    // opt/repo-path
                                                                                                                    // opt/repo-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                              // opt/repo-path
        (                                                                                                           // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                     // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                                        // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                       // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                                         // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                     // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                      // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                       // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                      // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                       // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                 // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                 // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/repo-path
        ),                                                                                                          // opt/repo-path
                                                                                                                    // opt/repo-path
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                             // opt/repo-path
        (                                                                                                           // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/repo-path
        ),                                                                                                          // opt/repo-path
                                                                                                                    // opt/repo-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                             // opt/repo-path
        (                                                                                                           // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/repo-path
        ),                                                                                                          // opt/repo-path
                                                                                                                    // opt/repo-path
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                            // opt/repo-path
        (                                                                                                           // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                     // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                                        // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                       // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                                         // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                     // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                      // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                       // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                      // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                       // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                 // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                 // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                // opt/repo-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/repo-path
        ),                                                                                                          // opt/repo-path
                                                                                                                    // opt/repo-path
        PARSE_RULE_OPTIONAL                                                                                         // opt/repo-path
        (                                                                                                           // opt/repo-path
            PARSE_RULE_OPTIONAL_GROUP                                                                               // opt/repo-path
            (                                                                                                       // opt/repo-path
                PARSE_RULE_OPTIONAL_DEFAULT                                                                         // opt/repo-path
                (                                                                                                   // opt/repo-path
                    PARSE_RULE_VAL_STR(QT_FS_var_FS_lib_FS_pgbackrest_QT),                                          // opt/repo-path
                ),                                                                                                  // opt/repo-path
            ),                                                                                                      // opt/repo-path
        ),                                                                                                          // opt/repo-path
    ),                                                                                                              // opt/repo-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                  // opt/repo-retention-archive
    (                                                                                                  // opt/repo-retention-archive
        PARSE_RULE_OPTION_NAME("repo-retention-archive"),                                              // opt/repo-retention-archive
        PARSE_RULE_OPTION_TYPE(Integer),                                                               // opt/repo-retention-archive
        PARSE_RULE_OPTION_RESET(true),                                                                 // opt/repo-retention-archive
        PARSE_RULE_OPTION_REQUIRED(false),                                                             // opt/repo-retention-archive
        PARSE_RULE_OPTION_SECTION(Global),                                                             // opt/repo-retention-archive
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                              // opt/repo-retention-archive
                                                                                                       // opt/repo-retention-archive
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                 // opt/repo-retention-archive
        (                                                                                              // opt/repo-retention-archive
            PARSE_RULE_OPTION_COMMAND(Backup)                                                          // opt/repo-retention-archive
            PARSE_RULE_OPTION_COMMAND(Expire)                                                          // opt/repo-retention-archive
        ),                                                                                             // opt/repo-retention-archive
                                                                                                       // opt/repo-retention-archive
        PARSE_RULE_OPTIONAL                                                                            // opt/repo-retention-archive
        (                                                                                              // opt/repo-retention-archive
            PARSE_RULE_OPTIONAL_GROUP                                                                  // opt/repo-retention-archive
            (                                                                                          // opt/repo-retention-archive
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                        // opt/repo-retention-archive
                (                                                                                      // opt/repo-retention-archive
                    PARSE_RULE_VAL_INT(1),                                                             // opt/repo-retention-archive
                    PARSE_RULE_VAL_INT(9999999),                                                       // opt/repo-retention-archive
                ),                                                                                     // opt/repo-retention-archive
            ),                                                                                         // opt/repo-retention-archive
        ),                                                                                             // opt/repo-retention-archive
    ),                                                                                                 // opt/repo-retention-archive
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                             // opt/repo-retention-archive-type
    (                                                                                             // opt/repo-retention-archive-type
        PARSE_RULE_OPTION_NAME("repo-retention-archive-type"),                                    // opt/repo-retention-archive-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                         // opt/repo-retention-archive-type
        PARSE_RULE_OPTION_RESET(true),                                                            // opt/repo-retention-archive-type
        PARSE_RULE_OPTION_REQUIRED(true),                                                         // opt/repo-retention-archive-type
        PARSE_RULE_OPTION_SECTION(Global),                                                        // opt/repo-retention-archive-type
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                         // opt/repo-retention-archive-type
                                                                                                  // opt/repo-retention-archive-type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                            // opt/repo-retention-archive-type
        (                                                                                         // opt/repo-retention-archive-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                     // opt/repo-retention-archive-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                     // opt/repo-retention-archive-type
        ),                                                                                        // opt/repo-retention-archive-type
                                                                                                  // opt/repo-retention-archive-type
        PARSE_RULE_OPTIONAL                                                                       // opt/repo-retention-archive-type
        (                                                                                         // opt/repo-retention-archive-type
            PARSE_RULE_OPTIONAL_GROUP                                                             // opt/repo-retention-archive-type
            (                                                                                     // opt/repo-retention-archive-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                    // opt/repo-retention-archive-type
                (                                                                                 // opt/repo-retention-archive-type
                    PARSE_RULE_VAL_STRID(full),                                                   // opt/repo-retention-archive-type
                    PARSE_RULE_VAL_STRID(diff),                                                   // opt/repo-retention-archive-type
                    PARSE_RULE_VAL_STRID(incr),                                                   // opt/repo-retention-archive-type
                ),                                                                                // opt/repo-retention-archive-type
                                                                                                  // opt/repo-retention-archive-type
                PARSE_RULE_OPTIONAL_DEFAULT                                                       // opt/repo-retention-archive-type
                (                                                                                 // opt/repo-retention-archive-type
                    PARSE_RULE_VAL_STRID(full),                                                   // opt/repo-retention-archive-type
                ),                                                                                // opt/repo-retention-archive-type
            ),                                                                                    // opt/repo-retention-archive-type
        ),                                                                                        // opt/repo-retention-archive-type
    ),                                                                                            // opt/repo-retention-archive-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                     // opt/repo-retention-diff
    (                                                                                                     // opt/repo-retention-diff
        PARSE_RULE_OPTION_NAME("repo-retention-diff"),                                                    // opt/repo-retention-diff
        PARSE_RULE_OPTION_TYPE(Integer),                                                                  // opt/repo-retention-diff
        PARSE_RULE_OPTION_RESET(true),                                                                    // opt/repo-retention-diff
        PARSE_RULE_OPTION_REQUIRED(false),                                                                // opt/repo-retention-diff
        PARSE_RULE_OPTION_SECTION(Global),                                                                // opt/repo-retention-diff
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                 // opt/repo-retention-diff
                                                                                                          // opt/repo-retention-diff
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                    // opt/repo-retention-diff
        (                                                                                                 // opt/repo-retention-diff
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-retention-diff
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/repo-retention-diff
        ),                                                                                                // opt/repo-retention-diff
                                                                                                          // opt/repo-retention-diff
        PARSE_RULE_OPTIONAL                                                                               // opt/repo-retention-diff
        (                                                                                                 // opt/repo-retention-diff
            PARSE_RULE_OPTIONAL_GROUP                                                                     // opt/repo-retention-diff
            (                                                                                             // opt/repo-retention-diff
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                           // opt/repo-retention-diff
                (                                                                                         // opt/repo-retention-diff
                    PARSE_RULE_VAL_INT(1),                                                                // opt/repo-retention-diff
                    PARSE_RULE_VAL_INT(9999999),                                                          // opt/repo-retention-diff
                ),                                                                                        // opt/repo-retention-diff
            ),                                                                                            // opt/repo-retention-diff
        ),                                                                                                // opt/repo-retention-diff
    ),                                                                                                    // opt/repo-retention-diff
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                     // opt/repo-retention-full
    (                                                                                                     // opt/repo-retention-full
        PARSE_RULE_OPTION_NAME("repo-retention-full"),                                                    // opt/repo-retention-full
        PARSE_RULE_OPTION_TYPE(Integer),                                                                  // opt/repo-retention-full
        PARSE_RULE_OPTION_RESET(true),                                                                    // opt/repo-retention-full
        PARSE_RULE_OPTION_REQUIRED(false),                                                                // opt/repo-retention-full
        PARSE_RULE_OPTION_SECTION(Global),                                                                // opt/repo-retention-full
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                 // opt/repo-retention-full
                                                                                                          // opt/repo-retention-full
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                    // opt/repo-retention-full
        (                                                                                                 // opt/repo-retention-full
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-retention-full
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/repo-retention-full
        ),                                                                                                // opt/repo-retention-full
                                                                                                          // opt/repo-retention-full
        PARSE_RULE_OPTIONAL                                                                               // opt/repo-retention-full
        (                                                                                                 // opt/repo-retention-full
            PARSE_RULE_OPTIONAL_GROUP                                                                     // opt/repo-retention-full
            (                                                                                             // opt/repo-retention-full
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                           // opt/repo-retention-full
                (                                                                                         // opt/repo-retention-full
                    PARSE_RULE_VAL_INT(1),                                                                // opt/repo-retention-full
                    PARSE_RULE_VAL_INT(9999999),                                                          // opt/repo-retention-full
                ),                                                                                        // opt/repo-retention-full
            ),                                                                                            // opt/repo-retention-full
        ),                                                                                                // opt/repo-retention-full
    ),                                                                                                    // opt/repo-retention-full
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                // opt/repo-retention-full-type
    (                                                                                                // opt/repo-retention-full-type
        PARSE_RULE_OPTION_NAME("repo-retention-full-type"),                                          // opt/repo-retention-full-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                            // opt/repo-retention-full-type
        PARSE_RULE_OPTION_RESET(true),                                                               // opt/repo-retention-full-type
        PARSE_RULE_OPTION_REQUIRED(true),                                                            // opt/repo-retention-full-type
        PARSE_RULE_OPTION_SECTION(Global),                                                           // opt/repo-retention-full-type
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                            // opt/repo-retention-full-type
                                                                                                     // opt/repo-retention-full-type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                               // opt/repo-retention-full-type
        (                                                                                            // opt/repo-retention-full-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                        // opt/repo-retention-full-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                        // opt/repo-retention-full-type
        ),                                                                                           // opt/repo-retention-full-type
                                                                                                     // opt/repo-retention-full-type
        PARSE_RULE_OPTIONAL                                                                          // opt/repo-retention-full-type
        (                                                                                            // opt/repo-retention-full-type
            PARSE_RULE_OPTIONAL_GROUP                                                                // opt/repo-retention-full-type
            (                                                                                        // opt/repo-retention-full-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                       // opt/repo-retention-full-type
                (                                                                                    // opt/repo-retention-full-type
                    PARSE_RULE_VAL_STRID(count),                                                     // opt/repo-retention-full-type
                    PARSE_RULE_VAL_STRID(time),                                                      // opt/repo-retention-full-type
                ),                                                                                   // opt/repo-retention-full-type
                                                                                                     // opt/repo-retention-full-type
                PARSE_RULE_OPTIONAL_DEFAULT                                                          // opt/repo-retention-full-type
                (                                                                                    // opt/repo-retention-full-type
                    PARSE_RULE_VAL_STRID(count),                                                     // opt/repo-retention-full-type
                ),                                                                                   // opt/repo-retention-full-type
            ),                                                                                       // opt/repo-retention-full-type
        ),                                                                                           // opt/repo-retention-full-type
    ),                                                                                               // opt/repo-retention-full-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                  // opt/repo-retention-history
    (                                                                                                  // opt/repo-retention-history
        PARSE_RULE_OPTION_NAME("repo-retention-history"),                                              // opt/repo-retention-history
        PARSE_RULE_OPTION_TYPE(Integer),                                                               // opt/repo-retention-history
        PARSE_RULE_OPTION_RESET(true),                                                                 // opt/repo-retention-history
        PARSE_RULE_OPTION_REQUIRED(false),                                                             // opt/repo-retention-history
        PARSE_RULE_OPTION_SECTION(Global),                                                             // opt/repo-retention-history
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                              // opt/repo-retention-history
                                                                                                       // opt/repo-retention-history
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                 // opt/repo-retention-history
        (                                                                                              // opt/repo-retention-history
            PARSE_RULE_OPTION_COMMAND(Backup)                                                          // opt/repo-retention-history
            PARSE_RULE_OPTION_COMMAND(Expire)                                                          // opt/repo-retention-history
        ),                                                                                             // opt/repo-retention-history
                                                                                                       // opt/repo-retention-history
        PARSE_RULE_OPTIONAL                                                                            // opt/repo-retention-history
        (                                                                                              // opt/repo-retention-history
            PARSE_RULE_OPTIONAL_GROUP                                                                  // opt/repo-retention-history
            (                                                                                          // opt/repo-retention-history
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                        // opt/repo-retention-history
                (                                                                                      // opt/repo-retention-history
                    PARSE_RULE_VAL_INT(0),                                                             // opt/repo-retention-history
                    PARSE_RULE_VAL_INT(9999999),                                                       // opt/repo-retention-history
                ),                                                                                     // opt/repo-retention-history
            ),                                                                                         // opt/repo-retention-history
        ),                                                                                             // opt/repo-retention-history
    ),                                                                                                 // opt/repo-retention-history
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/repo-s3-bucket
    (                                                                                                          // opt/repo-s3-bucket
        PARSE_RULE_OPTION_NAME("repo-s3-bucket"),                                                              // opt/repo-s3-bucket
        PARSE_RULE_OPTION_TYPE(String),                                                                        // opt/repo-s3-bucket
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/repo-s3-bucket
        PARSE_RULE_OPTION_REQUIRED(true),                                                                      // opt/repo-s3-bucket
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/repo-s3-bucket
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                      // opt/repo-s3-bucket
        PARSE_RULE_OPTION_DEPRECATE_MATCH(true),                                                               // opt/repo-s3-bucket
                                                                                                               // opt/repo-s3-bucket
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/repo-s3-bucket
        (                                                                                                      // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-s3-bucket
        ),                                                                                                     // opt/repo-s3-bucket
                                                                                                               // opt/repo-s3-bucket
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                        // opt/repo-s3-bucket
        (                                                                                                      // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-s3-bucket
        ),                                                                                                     // opt/repo-s3-bucket
                                                                                                               // opt/repo-s3-bucket
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                        // opt/repo-s3-bucket
        (                                                                                                      // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-s3-bucket
        ),                                                                                                     // opt/repo-s3-bucket
                                                                                                               // opt/repo-s3-bucket
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                       // opt/repo-s3-bucket
        (                                                                                                      // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/repo-s3-bucket
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-s3-bucket
        ),                                                                                                     // opt/repo-s3-bucket
                                                                                                               // opt/repo-s3-bucket
        PARSE_RULE_OPTIONAL                                                                                    // opt/repo-s3-bucket
        (                                                                                                      // opt/repo-s3-bucket
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/repo-s3-bucket
            (                                                                                                  // opt/repo-s3-bucket
                PARSE_RULE_OPTIONAL_DEPEND                                                                     // opt/repo-s3-bucket
                (                                                                                              // opt/repo-s3-bucket
                    PARSE_RULE_VAL_OPT(RepoType),                                                              // opt/repo-s3-bucket
                    PARSE_RULE_VAL_STRID(s3),                                                                  // opt/repo-s3-bucket
                ),                                                                                             // opt/repo-s3-bucket
            ),                                                                                                 // opt/repo-s3-bucket
        ),                                                                                                     // opt/repo-s3-bucket
    ),                                                                                                         // opt/repo-s3-bucket
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/repo-s3-endpoint
    (                                                                                                        // opt/repo-s3-endpoint
        PARSE_RULE_OPTION_NAME("repo-s3-endpoint"),                                                          // opt/repo-s3-endpoint
        PARSE_RULE_OPTION_TYPE(String),                                                                      // opt/repo-s3-endpoint
        PARSE_RULE_OPTION_RESET(true),                                                                       // opt/repo-s3-endpoint
        PARSE_RULE_OPTION_REQUIRED(true),                                                                    // opt/repo-s3-endpoint
        PARSE_RULE_OPTION_SECTION(Global),                                                                   // opt/repo-s3-endpoint
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                    // opt/repo-s3-endpoint
        PARSE_RULE_OPTION_DEPRECATE_MATCH(true),                                                             // opt/repo-s3-endpoint
                                                                                                             // opt/repo-s3-endpoint
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/repo-s3-endpoint
        (                                                                                                    // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-s3-endpoint
        ),                                                                                                   // opt/repo-s3-endpoint
                                                                                                             // opt/repo-s3-endpoint
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                      // opt/repo-s3-endpoint
        (                                                                                                    // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-s3-endpoint
        ),                                                                                                   // opt/repo-s3-endpoint
                                                                                                             // opt/repo-s3-endpoint
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                      // opt/repo-s3-endpoint
        (                                                                                                    // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-s3-endpoint
        ),                                                                                                   // opt/repo-s3-endpoint
                                                                                                             // opt/repo-s3-endpoint
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                     // opt/repo-s3-endpoint
        (                                                                                                    // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/repo-s3-endpoint
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-s3-endpoint
        ),                                                                                                   // opt/repo-s3-endpoint
                                                                                                             // opt/repo-s3-endpoint
        PARSE_RULE_OPTIONAL                                                                                  // opt/repo-s3-endpoint
        (                                                                                                    // opt/repo-s3-endpoint
            PARSE_RULE_OPTIONAL_GROUP                                                                        // opt/repo-s3-endpoint
            (                                                                                                // opt/repo-s3-endpoint
                PARSE_RULE_OPTIONAL_DEPEND                                                                   // opt/repo-s3-endpoint
                (                                                                                            // opt/repo-s3-endpoint
                    PARSE_RULE_VAL_OPT(RepoType),                                                            // opt/repo-s3-endpoint
                    PARSE_RULE_VAL_STRID(s3),                                                                // opt/repo-s3-endpoint
                ),                                                                                           // opt/repo-s3-endpoint
            ),                                                                                               // opt/repo-s3-endpoint
        ),                                                                                                   // opt/repo-s3-endpoint
    ),                                                                                                       // opt/repo-s3-endpoint
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                             // opt/repo-s3-key
    (                                                                                                             // opt/repo-s3-key
        PARSE_RULE_OPTION_NAME("repo-s3-key"),                                                                    // opt/repo-s3-key
        PARSE_RULE_OPTION_TYPE(String),                                                                           // opt/repo-s3-key
        PARSE_RULE_OPTION_RESET(true),                                                                            // opt/repo-s3-key
        PARSE_RULE_OPTION_REQUIRED(true),                                                                         // opt/repo-s3-key
        PARSE_RULE_OPTION_SECTION(Global),                                                                        // opt/repo-s3-key
        PARSE_RULE_OPTION_SECURE(true),                                                                           // opt/repo-s3-key
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                         // opt/repo-s3-key
        PARSE_RULE_OPTION_DEPRECATE_MATCH(true),                                                                  // opt/repo-s3-key
                                                                                                                  // opt/repo-s3-key
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                            // opt/repo-s3-key
        (                                                                                                         // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                   // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Check)                                                                      // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                     // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Info)                                                                       // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                   // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                    // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                     // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                    // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                     // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                    // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                               // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                               // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                              // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                     // opt/repo-s3-key
        ),                                                                                                        // opt/repo-s3-key
                                                                                                                  // opt/repo-s3-key
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                           // opt/repo-s3-key
        (                                                                                                         // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/repo-s3-key
        ),                                                                                                        // opt/repo-s3-key
                                                                                                                  // opt/repo-s3-key
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                           // opt/repo-s3-key
        (                                                                                                         // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                    // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                     // opt/repo-s3-key
        ),                                                                                                        // opt/repo-s3-key
                                                                                                                  // opt/repo-s3-key
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                          // opt/repo-s3-key
        (                                                                                                         // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                   // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                 // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                     // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Check)                                                                      // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                     // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Info)                                                                       // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                   // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                    // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                     // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                    // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                     // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                    // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                               // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                               // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                              // opt/repo-s3-key
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                     // opt/repo-s3-key
        ),                                                                                                        // opt/repo-s3-key
                                                                                                                  // opt/repo-s3-key
        PARSE_RULE_OPTIONAL                                                                                       // opt/repo-s3-key
        (                                                                                                         // opt/repo-s3-key
            PARSE_RULE_OPTIONAL_GROUP                                                                             // opt/repo-s3-key
            (                                                                                                     // opt/repo-s3-key
                PARSE_RULE_OPTIONAL_DEPEND                                                                        // opt/repo-s3-key
                (                                                                                                 // opt/repo-s3-key
                    PARSE_RULE_VAL_OPT(RepoS3KeyType),                                                            // opt/repo-s3-key
                    PARSE_RULE_VAL_STRID(shared),                                                                 // opt/repo-s3-key
                ),                                                                                                // opt/repo-s3-key
            ),                                                                                                    // opt/repo-s3-key
        ),                                                                                                        // opt/repo-s3-key
    ),                                                                                                            // opt/repo-s3-key
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                      // opt/repo-s3-key-secret
    (                                                                                                      // opt/repo-s3-key-secret
        PARSE_RULE_OPTION_NAME("repo-s3-key-secret"),                                                      // opt/repo-s3-key-secret
        PARSE_RULE_OPTION_TYPE(String),                                                                    // opt/repo-s3-key-secret
        PARSE_RULE_OPTION_RESET(true),                                                                     // opt/repo-s3-key-secret
        PARSE_RULE_OPTION_REQUIRED(true),                                                                  // opt/repo-s3-key-secret
        PARSE_RULE_OPTION_SECTION(Global),                                                                 // opt/repo-s3-key-secret
        PARSE_RULE_OPTION_SECURE(true),                                                                    // opt/repo-s3-key-secret
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                  // opt/repo-s3-key-secret
        PARSE_RULE_OPTION_DEPRECATE_MATCH(true),                                                           // opt/repo-s3-key-secret
                                                                                                           // opt/repo-s3-key-secret
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                     // opt/repo-s3-key-secret
        (                                                                                                  // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                            // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Check)                                                               // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Expire)                                                              // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Info)                                                                // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                            // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                             // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                              // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                             // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                              // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                        // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                        // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                       // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/repo-s3-key-secret
        ),                                                                                                 // opt/repo-s3-key-secret
                                                                                                           // opt/repo-s3-key-secret
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                    // opt/repo-s3-key-secret
        (                                                                                                  // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-s3-key-secret
        ),                                                                                                 // opt/repo-s3-key-secret
                                                                                                           // opt/repo-s3-key-secret
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                    // opt/repo-s3-key-secret
        (                                                                                                  // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/repo-s3-key-secret
        ),                                                                                                 // opt/repo-s3-key-secret
                                                                                                           // opt/repo-s3-key-secret
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                   // opt/repo-s3-key-secret
        (                                                                                                  // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                            // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Check)                                                               // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Expire)                                                              // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Info)                                                                // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                            // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                             // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                              // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                             // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                              // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                        // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                        // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                       // opt/repo-s3-key-secret
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/repo-s3-key-secret
        ),                                                                                                 // opt/repo-s3-key-secret
                                                                                                           // opt/repo-s3-key-secret
        PARSE_RULE_OPTIONAL                                                                                // opt/repo-s3-key-secret
        (                                                                                                  // opt/repo-s3-key-secret
            PARSE_RULE_OPTIONAL_GROUP                                                                      // opt/repo-s3-key-secret
            (                                                                                              // opt/repo-s3-key-secret
                PARSE_RULE_OPTIONAL_DEPEND                                                                 // opt/repo-s3-key-secret
                (                                                                                          // opt/repo-s3-key-secret
                    PARSE_RULE_VAL_OPT(RepoS3KeyType),                                                     // opt/repo-s3-key-secret
                    PARSE_RULE_VAL_STRID(shared),                                                          // opt/repo-s3-key-secret
                ),                                                                                         // opt/repo-s3-key-secret
            ),                                                                                             // opt/repo-s3-key-secret
        ),                                                                                                 // opt/repo-s3-key-secret
    ),                                                                                                     // opt/repo-s3-key-secret
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/repo-s3-key-type
    (                                                                                                        // opt/repo-s3-key-type
        PARSE_RULE_OPTION_NAME("repo-s3-key-type"),                                                          // opt/repo-s3-key-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                                    // opt/repo-s3-key-type
        PARSE_RULE_OPTION_RESET(true),                                                                       // opt/repo-s3-key-type
        PARSE_RULE_OPTION_REQUIRED(true),                                                                    // opt/repo-s3-key-type
        PARSE_RULE_OPTION_SECTION(Global),                                                                   // opt/repo-s3-key-type
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                    // opt/repo-s3-key-type
                                                                                                             // opt/repo-s3-key-type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/repo-s3-key-type
        (                                                                                                    // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-s3-key-type
        ),                                                                                                   // opt/repo-s3-key-type
                                                                                                             // opt/repo-s3-key-type
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                      // opt/repo-s3-key-type
        (                                                                                                    // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-s3-key-type
        ),                                                                                                   // opt/repo-s3-key-type
                                                                                                             // opt/repo-s3-key-type
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                      // opt/repo-s3-key-type
        (                                                                                                    // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-s3-key-type
        ),                                                                                                   // opt/repo-s3-key-type
                                                                                                             // opt/repo-s3-key-type
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                     // opt/repo-s3-key-type
        (                                                                                                    // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/repo-s3-key-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-s3-key-type
        ),                                                                                                   // opt/repo-s3-key-type
                                                                                                             // opt/repo-s3-key-type
        PARSE_RULE_OPTIONAL                                                                                  // opt/repo-s3-key-type
        (                                                                                                    // opt/repo-s3-key-type
            PARSE_RULE_OPTIONAL_GROUP                                                                        // opt/repo-s3-key-type
            (                                                                                                // opt/repo-s3-key-type
                PARSE_RULE_OPTIONAL_DEPEND                                                                   // opt/repo-s3-key-type
                (                                                                                            // opt/repo-s3-key-type
                    PARSE_RULE_VAL_OPT(RepoType),                                                            // opt/repo-s3-key-type
                    PARSE_RULE_VAL_STRID(s3),                                                                // opt/repo-s3-key-type
                ),                                                                                           // opt/repo-s3-key-type
                                                                                                             // opt/repo-s3-key-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                               // opt/repo-s3-key-type
                (                                                                                            // opt/repo-s3-key-type
                    PARSE_RULE_VAL_STRID(shared),                                                            // opt/repo-s3-key-type
                    PARSE_RULE_VAL_STRID(auto),                                                              // opt/repo-s3-key-type
                    PARSE_RULE_VAL_STRID(web_DS_id),                                                         // opt/repo-s3-key-type
                    PARSE_RULE_VAL_STRID(pod_DS_id),                                                         // opt/repo-s3-key-type
                ),                                                                                           // opt/repo-s3-key-type
                                                                                                             // opt/repo-s3-key-type
                PARSE_RULE_OPTIONAL_DEFAULT                                                                  // opt/repo-s3-key-type
                (                                                                                            // opt/repo-s3-key-type
                    PARSE_RULE_VAL_STRID(shared),                                                            // opt/repo-s3-key-type
                ),                                                                                           // opt/repo-s3-key-type
            ),                                                                                               // opt/repo-s3-key-type
        ),                                                                                                   // opt/repo-s3-key-type
    ),                                                                                                       // opt/repo-s3-key-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                      // opt/repo-s3-kms-key-id
    (                                                                                                      // opt/repo-s3-kms-key-id
        PARSE_RULE_OPTION_NAME("repo-s3-kms-key-id"),                                                      // opt/repo-s3-kms-key-id
        PARSE_RULE_OPTION_TYPE(String),                                                                    // opt/repo-s3-kms-key-id
        PARSE_RULE_OPTION_RESET(true),                                                                     // opt/repo-s3-kms-key-id
        PARSE_RULE_OPTION_REQUIRED(false),                                                                 // opt/repo-s3-kms-key-id
        PARSE_RULE_OPTION_SECTION(Global),                                                                 // opt/repo-s3-kms-key-id
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                  // opt/repo-s3-kms-key-id
                                                                                                           // opt/repo-s3-kms-key-id
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                     // opt/repo-s3-kms-key-id
        (                                                                                                  // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                            // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Check)                                                               // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Expire)                                                              // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Info)                                                                // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                            // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                             // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                              // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                             // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                              // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                        // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                        // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                       // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/repo-s3-kms-key-id
        ),                                                                                                 // opt/repo-s3-kms-key-id
                                                                                                           // opt/repo-s3-kms-key-id
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                    // opt/repo-s3-kms-key-id
        (                                                                                                  // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-s3-kms-key-id
        ),                                                                                                 // opt/repo-s3-kms-key-id
                                                                                                           // opt/repo-s3-kms-key-id
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                    // opt/repo-s3-kms-key-id
        (                                                                                                  // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/repo-s3-kms-key-id
        ),                                                                                                 // opt/repo-s3-kms-key-id
                                                                                                           // opt/repo-s3-kms-key-id
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                   // opt/repo-s3-kms-key-id
        (                                                                                                  // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                            // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                          // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                         // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Backup)                                                              // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Check)                                                               // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Expire)                                                              // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Info)                                                                // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                            // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                             // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                              // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                             // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                              // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                        // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                        // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                       // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTION_COMMAND(Verify)                                                              // opt/repo-s3-kms-key-id
        ),                                                                                                 // opt/repo-s3-kms-key-id
                                                                                                           // opt/repo-s3-kms-key-id
        PARSE_RULE_OPTIONAL                                                                                // opt/repo-s3-kms-key-id
        (                                                                                                  // opt/repo-s3-kms-key-id
            PARSE_RULE_OPTIONAL_GROUP                                                                      // opt/repo-s3-kms-key-id
            (                                                                                              // opt/repo-s3-kms-key-id
                PARSE_RULE_OPTIONAL_DEPEND                                                                 // opt/repo-s3-kms-key-id
                (                                                                                          // opt/repo-s3-kms-key-id
                    PARSE_RULE_VAL_OPT(RepoType),                                                          // opt/repo-s3-kms-key-id
                    PARSE_RULE_VAL_STRID(s3),                                                              // opt/repo-s3-kms-key-id
                ),                                                                                         // opt/repo-s3-kms-key-id
            ),                                                                                             // opt/repo-s3-kms-key-id
        ),                                                                                                 // opt/repo-s3-kms-key-id
    ),                                                                                                     // opt/repo-s3-kms-key-id
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/repo-s3-region
    (                                                                                                          // opt/repo-s3-region
        PARSE_RULE_OPTION_NAME("repo-s3-region"),                                                              // opt/repo-s3-region
        PARSE_RULE_OPTION_TYPE(String),                                                                        // opt/repo-s3-region
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/repo-s3-region
        PARSE_RULE_OPTION_REQUIRED(true),                                                                      // opt/repo-s3-region
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/repo-s3-region
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                      // opt/repo-s3-region
        PARSE_RULE_OPTION_DEPRECATE_MATCH(true),                                                               // opt/repo-s3-region
                                                                                                               // opt/repo-s3-region
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/repo-s3-region
        (                                                                                                      // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-s3-region
        ),                                                                                                     // opt/repo-s3-region
                                                                                                               // opt/repo-s3-region
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                        // opt/repo-s3-region
        (                                                                                                      // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-s3-region
        ),                                                                                                     // opt/repo-s3-region
                                                                                                               // opt/repo-s3-region
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                        // opt/repo-s3-region
        (                                                                                                      // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-s3-region
        ),                                                                                                     // opt/repo-s3-region
                                                                                                               // opt/repo-s3-region
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                       // opt/repo-s3-region
        (                                                                                                      // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/repo-s3-region
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-s3-region
        ),                                                                                                     // opt/repo-s3-region
                                                                                                               // opt/repo-s3-region
        PARSE_RULE_OPTIONAL                                                                                    // opt/repo-s3-region
        (                                                                                                      // opt/repo-s3-region
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/repo-s3-region
            (                                                                                                  // opt/repo-s3-region
                PARSE_RULE_OPTIONAL_DEPEND                                                                     // opt/repo-s3-region
                (                                                                                              // opt/repo-s3-region
                    PARSE_RULE_VAL_OPT(RepoType),                                                              // opt/repo-s3-region
                    PARSE_RULE_VAL_STRID(s3),                                                                  // opt/repo-s3-region
                ),                                                                                             // opt/repo-s3-region
            ),                                                                                                 // opt/repo-s3-region
        ),                                                                                                     // opt/repo-s3-region
    ),                                                                                                         // opt/repo-s3-region
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                  // opt/repo-s3-requester-pays
    (                                                                                                  // opt/repo-s3-requester-pays
        PARSE_RULE_OPTION_NAME("repo-s3-requester-pays"),                                              // opt/repo-s3-requester-pays
        PARSE_RULE_OPTION_TYPE(Boolean),                                                               // opt/repo-s3-requester-pays
        PARSE_RULE_OPTION_NEGATE(true),                                                                // opt/repo-s3-requester-pays
        PARSE_RULE_OPTION_RESET(true),                                                                 // opt/repo-s3-requester-pays
        PARSE_RULE_OPTION_REQUIRED(true),                                                              // opt/repo-s3-requester-pays
        PARSE_RULE_OPTION_SECTION(Global),                                                             // opt/repo-s3-requester-pays
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                              // opt/repo-s3-requester-pays
                                                                                                       // opt/repo-s3-requester-pays
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                 // opt/repo-s3-requester-pays
        (                                                                                              // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                        // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                      // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                     // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Backup)                                                          // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Check)                                                           // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Expire)                                                          // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Info)                                                            // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                        // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                         // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                          // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                         // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                          // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Restore)                                                         // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                    // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                    // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                   // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Verify)                                                          // opt/repo-s3-requester-pays
        ),                                                                                             // opt/repo-s3-requester-pays
                                                                                                       // opt/repo-s3-requester-pays
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                // opt/repo-s3-requester-pays
        (                                                                                              // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                      // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                     // opt/repo-s3-requester-pays
        ),                                                                                             // opt/repo-s3-requester-pays
                                                                                                       // opt/repo-s3-requester-pays
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                // opt/repo-s3-requester-pays
        (                                                                                              // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                      // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                     // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Backup)                                                          // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Restore)                                                         // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Verify)                                                          // opt/repo-s3-requester-pays
        ),                                                                                             // opt/repo-s3-requester-pays
                                                                                                       // opt/repo-s3-requester-pays
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                               // opt/repo-s3-requester-pays
        (                                                                                              // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                        // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                      // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                     // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Backup)                                                          // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Check)                                                           // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Expire)                                                          // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Info)                                                            // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                        // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                         // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                          // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                         // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                          // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Restore)                                                         // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                    // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                    // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                   // opt/repo-s3-requester-pays
            PARSE_RULE_OPTION_COMMAND(Verify)                                                          // opt/repo-s3-requester-pays
        ),                                                                                             // opt/repo-s3-requester-pays
                                                                                                       // opt/repo-s3-requester-pays
        PARSE_RULE_OPTIONAL                                                                            // opt/repo-s3-requester-pays
        (                                                                                              // opt/repo-s3-requester-pays
            PARSE_RULE_OPTIONAL_GROUP                                                                  // opt/repo-s3-requester-pays
            (                                                                                          // opt/repo-s3-requester-pays
                PARSE_RULE_OPTIONAL_DEPEND                                                             // opt/repo-s3-requester-pays
                (                                                                                      // opt/repo-s3-requester-pays
                    PARSE_RULE_VAL_OPT(RepoType),                                                      // opt/repo-s3-requester-pays
                    PARSE_RULE_VAL_STRID(s3),                                                          // opt/repo-s3-requester-pays
                ),                                                                                     // opt/repo-s3-requester-pays
                                                                                                       // opt/repo-s3-requester-pays
                PARSE_RULE_OPTIONAL_DEFAULT                                                            // opt/repo-s3-requester-pays
                (                                                                                      // opt/repo-s3-requester-pays
                    PARSE_RULE_VAL_BOOL_FALSE,                                                         // opt/repo-s3-requester-pays
                ),                                                                                     // opt/repo-s3-requester-pays
            ),                                                                                         // opt/repo-s3-requester-pays
        ),                                                                                             // opt/repo-s3-requester-pays
    ),                                                                                                 // opt/repo-s3-requester-pays
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                            // opt/repo-s3-role
    (                                                                                                            // opt/repo-s3-role
        PARSE_RULE_OPTION_NAME("repo-s3-role"),                                                                  // opt/repo-s3-role
        PARSE_RULE_OPTION_TYPE(String),                                                                          // opt/repo-s3-role
        PARSE_RULE_OPTION_RESET(true),                                                                           // opt/repo-s3-role
        PARSE_RULE_OPTION_REQUIRED(false),                                                                       // opt/repo-s3-role
        PARSE_RULE_OPTION_SECTION(Global),                                                                       // opt/repo-s3-role
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                        // opt/repo-s3-role
                                                                                                                 // opt/repo-s3-role
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                           // opt/repo-s3-role
        (                                                                                                        // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                  // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                               // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Check)                                                                     // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                    // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Info)                                                                      // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                  // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                   // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                    // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                   // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                    // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                   // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                              // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                              // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                             // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                    // opt/repo-s3-role
        ),                                                                                                       // opt/repo-s3-role
                                                                                                                 // opt/repo-s3-role
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                          // opt/repo-s3-role
        (                                                                                                        // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                               // opt/repo-s3-role
        ),                                                                                                       // opt/repo-s3-role
                                                                                                                 // opt/repo-s3-role
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                          // opt/repo-s3-role
        (                                                                                                        // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                               // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                   // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                    // opt/repo-s3-role
        ),                                                                                                       // opt/repo-s3-role
                                                                                                                 // opt/repo-s3-role
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                         // opt/repo-s3-role
        (                                                                                                        // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                  // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                               // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Check)                                                                     // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                    // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Info)                                                                      // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                  // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                   // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                    // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                   // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                    // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                   // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                              // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                              // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                             // opt/repo-s3-role
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                    // opt/repo-s3-role
        ),                                                                                                       // opt/repo-s3-role
                                                                                                                 // opt/repo-s3-role
        PARSE_RULE_OPTIONAL                                                                                      // opt/repo-s3-role
        (                                                                                                        // opt/repo-s3-role
            PARSE_RULE_OPTIONAL_GROUP                                                                            // opt/repo-s3-role
            (                                                                                                    // opt/repo-s3-role
                PARSE_RULE_OPTIONAL_DEPEND                                                                       // opt/repo-s3-role
                (                                                                                                // opt/repo-s3-role
                    PARSE_RULE_VAL_OPT(RepoS3KeyType),                                                           // opt/repo-s3-role
                    PARSE_RULE_VAL_STRID(auto),                                                                  // opt/repo-s3-role
                ),                                                                                               // opt/repo-s3-role
            ),                                                                                                   // opt/repo-s3-role
        ),                                                                                                       // opt/repo-s3-role
    ),                                                                                                           // opt/repo-s3-role
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                // opt/repo-s3-sse-customer-key
    (                                                                                                // opt/repo-s3-sse-customer-key
        PARSE_RULE_OPTION_NAME("repo-s3-sse-customer-key"),                                          // opt/repo-s3-sse-customer-key
        PARSE_RULE_OPTION_TYPE(String),                                                              // opt/repo-s3-sse-customer-key
        PARSE_RULE_OPTION_RESET(true),                                                               // opt/repo-s3-sse-customer-key
        PARSE_RULE_OPTION_REQUIRED(false),                                                           // opt/repo-s3-sse-customer-key
        PARSE_RULE_OPTION_SECTION(Global),                                                           // opt/repo-s3-sse-customer-key
        PARSE_RULE_OPTION_SECURE(true),                                                              // opt/repo-s3-sse-customer-key
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                            // opt/repo-s3-sse-customer-key
                                                                                                     // opt/repo-s3-sse-customer-key
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                               // opt/repo-s3-sse-customer-key
        (                                                                                            // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                      // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                    // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                   // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Backup)                                                        // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Check)                                                         // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Expire)                                                        // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Info)                                                          // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                      // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                       // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                        // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                       // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                        // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Restore)                                                       // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                  // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                  // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                 // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Verify)                                                        // opt/repo-s3-sse-customer-key
        ),                                                                                           // opt/repo-s3-sse-customer-key
                                                                                                     // opt/repo-s3-sse-customer-key
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                              // opt/repo-s3-sse-customer-key
        (                                                                                            // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                    // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                   // opt/repo-s3-sse-customer-key
        ),                                                                                           // opt/repo-s3-sse-customer-key
                                                                                                     // opt/repo-s3-sse-customer-key
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                              // opt/repo-s3-sse-customer-key
        (                                                                                            // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                    // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                   // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Backup)                                                        // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Restore)                                                       // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Verify)                                                        // opt/repo-s3-sse-customer-key
        ),                                                                                           // opt/repo-s3-sse-customer-key
                                                                                                     // opt/repo-s3-sse-customer-key
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                             // opt/repo-s3-sse-customer-key
        (                                                                                            // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                      // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                    // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                   // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Backup)                                                        // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Check)                                                         // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Expire)                                                        // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Info)                                                          // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                      // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                       // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                        // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                       // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                        // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Restore)                                                       // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                  // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                  // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                 // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTION_COMMAND(Verify)                                                        // opt/repo-s3-sse-customer-key
        ),                                                                                           // opt/repo-s3-sse-customer-key
                                                                                                     // opt/repo-s3-sse-customer-key
        PARSE_RULE_OPTIONAL                                                                          // opt/repo-s3-sse-customer-key
        (                                                                                            // opt/repo-s3-sse-customer-key
            PARSE_RULE_OPTIONAL_GROUP                                                                // opt/repo-s3-sse-customer-key
            (                                                                                        // opt/repo-s3-sse-customer-key
                PARSE_RULE_OPTIONAL_DEPEND                                                           // opt/repo-s3-sse-customer-key
                (                                                                                    // opt/repo-s3-sse-customer-key
                    PARSE_RULE_VAL_OPT(RepoType),                                                    // opt/repo-s3-sse-customer-key
                    PARSE_RULE_VAL_STRID(s3),                                                        // opt/repo-s3-sse-customer-key
                ),                                                                                   // opt/repo-s3-sse-customer-key
            ),                                                                                       // opt/repo-s3-sse-customer-key
        ),                                                                                           // opt/repo-s3-sse-customer-key
    ),                                                                                               // opt/repo-s3-sse-customer-key
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                           // opt/repo-s3-token
    (                                                                                                           // opt/repo-s3-token
        PARSE_RULE_OPTION_NAME("repo-s3-token"),                                                                // opt/repo-s3-token
        PARSE_RULE_OPTION_TYPE(String),                                                                         // opt/repo-s3-token
        PARSE_RULE_OPTION_RESET(true),                                                                          // opt/repo-s3-token
        PARSE_RULE_OPTION_REQUIRED(false),                                                                      // opt/repo-s3-token
        PARSE_RULE_OPTION_SECTION(Global),                                                                      // opt/repo-s3-token
        PARSE_RULE_OPTION_SECURE(true),                                                                         // opt/repo-s3-token
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                       // opt/repo-s3-token
                                                                                                                // opt/repo-s3-token
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                          // opt/repo-s3-token
        (                                                                                                       // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                 // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Check)                                                                    // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                   // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Info)                                                                     // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                 // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                  // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                   // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                  // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                   // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                             // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                             // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                            // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/repo-s3-token
        ),                                                                                                      // opt/repo-s3-token
                                                                                                                // opt/repo-s3-token
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                         // opt/repo-s3-token
        (                                                                                                       // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/repo-s3-token
        ),                                                                                                      // opt/repo-s3-token
                                                                                                                // opt/repo-s3-token
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                         // opt/repo-s3-token
        (                                                                                                       // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/repo-s3-token
        ),                                                                                                      // opt/repo-s3-token
                                                                                                                // opt/repo-s3-token
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                        // opt/repo-s3-token
        (                                                                                                       // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                 // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Check)                                                                    // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                   // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Info)                                                                     // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                 // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                  // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                   // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                  // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                   // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                             // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                             // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                            // opt/repo-s3-token
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/repo-s3-token
        ),                                                                                                      // opt/repo-s3-token
                                                                                                                // opt/repo-s3-token
        PARSE_RULE_OPTIONAL                                                                                     // opt/repo-s3-token
        (                                                                                                       // opt/repo-s3-token
            PARSE_RULE_OPTIONAL_GROUP                                                                           // opt/repo-s3-token
            (                                                                                                   // opt/repo-s3-token
                PARSE_RULE_OPTIONAL_DEPEND                                                                      // opt/repo-s3-token
                (                                                                                               // opt/repo-s3-token
                    PARSE_RULE_VAL_OPT(RepoS3KeyType),                                                          // opt/repo-s3-token
                    PARSE_RULE_VAL_STRID(shared),                                                               // opt/repo-s3-token
                ),                                                                                              // opt/repo-s3-token
            ),                                                                                                  // opt/repo-s3-token
        ),                                                                                                      // opt/repo-s3-token
    ),                                                                                                          // opt/repo-s3-token
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                       // opt/repo-s3-uri-style
    (                                                                                                       // opt/repo-s3-uri-style
        PARSE_RULE_OPTION_NAME("repo-s3-uri-style"),                                                        // opt/repo-s3-uri-style
        PARSE_RULE_OPTION_TYPE(StringId),                                                                   // opt/repo-s3-uri-style
        PARSE_RULE_OPTION_RESET(true),                                                                      // opt/repo-s3-uri-style
        PARSE_RULE_OPTION_REQUIRED(true),                                                                   // opt/repo-s3-uri-style
        PARSE_RULE_OPTION_SECTION(Global),                                                                  // opt/repo-s3-uri-style
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                   // opt/repo-s3-uri-style
                                                                                                            // opt/repo-s3-uri-style
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                      // opt/repo-s3-uri-style
        (                                                                                                   // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-s3-uri-style
        ),                                                                                                  // opt/repo-s3-uri-style
                                                                                                            // opt/repo-s3-uri-style
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                     // opt/repo-s3-uri-style
        (                                                                                                   // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-s3-uri-style
        ),                                                                                                  // opt/repo-s3-uri-style
                                                                                                            // opt/repo-s3-uri-style
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                     // opt/repo-s3-uri-style
        (                                                                                                   // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-s3-uri-style
        ),                                                                                                  // opt/repo-s3-uri-style
                                                                                                            // opt/repo-s3-uri-style
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                    // opt/repo-s3-uri-style
        (                                                                                                   // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/repo-s3-uri-style
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-s3-uri-style
        ),                                                                                                  // opt/repo-s3-uri-style
                                                                                                            // opt/repo-s3-uri-style
        PARSE_RULE_OPTIONAL                                                                                 // opt/repo-s3-uri-style
        (                                                                                                   // opt/repo-s3-uri-style
            PARSE_RULE_OPTIONAL_GROUP                                                                       // opt/repo-s3-uri-style
            (                                                                                               // opt/repo-s3-uri-style
                PARSE_RULE_OPTIONAL_DEPEND                                                                  // opt/repo-s3-uri-style
                (                                                                                           // opt/repo-s3-uri-style
                    PARSE_RULE_VAL_OPT(RepoType),                                                           // opt/repo-s3-uri-style
                    PARSE_RULE_VAL_STRID(s3),                                                               // opt/repo-s3-uri-style
                ),                                                                                          // opt/repo-s3-uri-style
                                                                                                            // opt/repo-s3-uri-style
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                              // opt/repo-s3-uri-style
                (                                                                                           // opt/repo-s3-uri-style
                    PARSE_RULE_VAL_STRID(host),                                                             // opt/repo-s3-uri-style
                    PARSE_RULE_VAL_STRID(path),                                                             // opt/repo-s3-uri-style
                ),                                                                                          // opt/repo-s3-uri-style
                                                                                                            // opt/repo-s3-uri-style
                PARSE_RULE_OPTIONAL_DEFAULT                                                                 // opt/repo-s3-uri-style
                (                                                                                           // opt/repo-s3-uri-style
                    PARSE_RULE_VAL_STRID(host),                                                             // opt/repo-s3-uri-style
                ),                                                                                          // opt/repo-s3-uri-style
            ),                                                                                              // opt/repo-s3-uri-style
        ),                                                                                                  // opt/repo-s3-uri-style
    ),                                                                                                      // opt/repo-s3-uri-style
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/repo-sftp-host
    (                                                                                                          // opt/repo-sftp-host
        PARSE_RULE_OPTION_NAME("repo-sftp-host"),                                                              // opt/repo-sftp-host
        PARSE_RULE_OPTION_TYPE(String),                                                                        // opt/repo-sftp-host
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/repo-sftp-host
        PARSE_RULE_OPTION_REQUIRED(true),                                                                      // opt/repo-sftp-host
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/repo-sftp-host
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                      // opt/repo-sftp-host
                                                                                                               // opt/repo-sftp-host
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/repo-sftp-host
        (                                                                                                      // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-sftp-host
        ),                                                                                                     // opt/repo-sftp-host
                                                                                                               // opt/repo-sftp-host
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                        // opt/repo-sftp-host
        (                                                                                                      // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-sftp-host
        ),                                                                                                     // opt/repo-sftp-host
                                                                                                               // opt/repo-sftp-host
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                        // opt/repo-sftp-host
        (                                                                                                      // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-sftp-host
        ),                                                                                                     // opt/repo-sftp-host
                                                                                                               // opt/repo-sftp-host
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                       // opt/repo-sftp-host
        (                                                                                                      // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/repo-sftp-host
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/repo-sftp-host
        ),                                                                                                     // opt/repo-sftp-host
                                                                                                               // opt/repo-sftp-host
        PARSE_RULE_OPTIONAL                                                                                    // opt/repo-sftp-host
        (                                                                                                      // opt/repo-sftp-host
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/repo-sftp-host
            (                                                                                                  // opt/repo-sftp-host
                PARSE_RULE_OPTIONAL_DEPEND                                                                     // opt/repo-sftp-host
                (                                                                                              // opt/repo-sftp-host
                    PARSE_RULE_VAL_OPT(RepoType),                                                              // opt/repo-sftp-host
                    PARSE_RULE_VAL_STRID(sftp),                                                                // opt/repo-sftp-host
                ),                                                                                             // opt/repo-sftp-host
            ),                                                                                                 // opt/repo-sftp-host
        ),                                                                                                     // opt/repo-sftp-host
    ),                                                                                                         // opt/repo-sftp-host
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                              // opt/repo-sftp-host-fingerprint
    (                                                                                              // opt/repo-sftp-host-fingerprint
        PARSE_RULE_OPTION_NAME("repo-sftp-host-fingerprint"),                                      // opt/repo-sftp-host-fingerprint
        PARSE_RULE_OPTION_TYPE(String),                                                            // opt/repo-sftp-host-fingerprint
        PARSE_RULE_OPTION_RESET(true),                                                             // opt/repo-sftp-host-fingerprint
        PARSE_RULE_OPTION_REQUIRED(false),                                                         // opt/repo-sftp-host-fingerprint
        PARSE_RULE_OPTION_SECTION(Global),                                                         // opt/repo-sftp-host-fingerprint
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                          // opt/repo-sftp-host-fingerprint
                                                                                                   // opt/repo-sftp-host-fingerprint
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                             // opt/repo-sftp-host-fingerprint
        (                                                                                          // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                    // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                  // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                 // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Backup)                                                      // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Check)                                                       // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Expire)                                                      // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Info)                                                        // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                    // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                     // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                      // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                     // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                      // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Restore)                                                     // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                               // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Verify)                                                      // opt/repo-sftp-host-fingerprint
        ),                                                                                         // opt/repo-sftp-host-fingerprint
                                                                                                   // opt/repo-sftp-host-fingerprint
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                            // opt/repo-sftp-host-fingerprint
        (                                                                                          // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                  // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                 // opt/repo-sftp-host-fingerprint
        ),                                                                                         // opt/repo-sftp-host-fingerprint
                                                                                                   // opt/repo-sftp-host-fingerprint
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                            // opt/repo-sftp-host-fingerprint
        (                                                                                          // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                  // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                 // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Backup)                                                      // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Restore)                                                     // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Verify)                                                      // opt/repo-sftp-host-fingerprint
        ),                                                                                         // opt/repo-sftp-host-fingerprint
                                                                                                   // opt/repo-sftp-host-fingerprint
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                           // opt/repo-sftp-host-fingerprint
        (                                                                                          // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                    // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                  // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                 // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Backup)                                                      // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Check)                                                       // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Expire)                                                      // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Info)                                                        // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                    // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                     // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                      // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                     // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                      // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Restore)                                                     // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                               // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTION_COMMAND(Verify)                                                      // opt/repo-sftp-host-fingerprint
        ),                                                                                         // opt/repo-sftp-host-fingerprint
                                                                                                   // opt/repo-sftp-host-fingerprint
        PARSE_RULE_OPTIONAL                                                                        // opt/repo-sftp-host-fingerprint
        (                                                                                          // opt/repo-sftp-host-fingerprint
            PARSE_RULE_OPTIONAL_GROUP                                                              // opt/repo-sftp-host-fingerprint
            (                                                                                      // opt/repo-sftp-host-fingerprint
                PARSE_RULE_OPTIONAL_DEPEND                                                         // opt/repo-sftp-host-fingerprint
                (                                                                                  // opt/repo-sftp-host-fingerprint
                    PARSE_RULE_VAL_OPT(RepoSftpHostKeyCheckType),                                  // opt/repo-sftp-host-fingerprint
                    PARSE_RULE_VAL_STRID(fingerprint),                                             // opt/repo-sftp-host-fingerprint
                    PARSE_RULE_VAL_STRID(strict),                                                  // opt/repo-sftp-host-fingerprint
                ),                                                                                 // opt/repo-sftp-host-fingerprint
            ),                                                                                     // opt/repo-sftp-host-fingerprint
        ),                                                                                         // opt/repo-sftp-host-fingerprint
    ),                                                                                             // opt/repo-sftp-host-fingerprint
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                           // opt/repo-sftp-host-key-check-type
    (                                                                                           // opt/repo-sftp-host-key-check-type
        PARSE_RULE_OPTION_NAME("repo-sftp-host-key-check-type"),                                // opt/repo-sftp-host-key-check-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                       // opt/repo-sftp-host-key-check-type
        PARSE_RULE_OPTION_RESET(true),                                                          // opt/repo-sftp-host-key-check-type
        PARSE_RULE_OPTION_REQUIRED(true),                                                       // opt/repo-sftp-host-key-check-type
        PARSE_RULE_OPTION_SECTION(Global),                                                      // opt/repo-sftp-host-key-check-type
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                       // opt/repo-sftp-host-key-check-type
                                                                                                // opt/repo-sftp-host-key-check-type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                          // opt/repo-sftp-host-key-check-type
        (                                                                                       // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                 // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                               // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                              // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                   // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                    // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                   // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                     // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                 // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                  // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                   // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                  // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                   // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                  // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                             // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                             // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                            // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                   // opt/repo-sftp-host-key-check-type
        ),                                                                                      // opt/repo-sftp-host-key-check-type
                                                                                                // opt/repo-sftp-host-key-check-type
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                         // opt/repo-sftp-host-key-check-type
        (                                                                                       // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                               // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                              // opt/repo-sftp-host-key-check-type
        ),                                                                                      // opt/repo-sftp-host-key-check-type
                                                                                                // opt/repo-sftp-host-key-check-type
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                         // opt/repo-sftp-host-key-check-type
        (                                                                                       // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                               // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                              // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                   // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                  // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                   // opt/repo-sftp-host-key-check-type
        ),                                                                                      // opt/repo-sftp-host-key-check-type
                                                                                                // opt/repo-sftp-host-key-check-type
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                        // opt/repo-sftp-host-key-check-type
        (                                                                                       // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                 // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                               // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                              // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                   // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                    // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                   // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                     // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                 // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                  // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                   // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                  // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                   // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                  // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                             // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                             // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                            // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                   // opt/repo-sftp-host-key-check-type
        ),                                                                                      // opt/repo-sftp-host-key-check-type
                                                                                                // opt/repo-sftp-host-key-check-type
        PARSE_RULE_OPTIONAL                                                                     // opt/repo-sftp-host-key-check-type
        (                                                                                       // opt/repo-sftp-host-key-check-type
            PARSE_RULE_OPTIONAL_GROUP                                                           // opt/repo-sftp-host-key-check-type
            (                                                                                   // opt/repo-sftp-host-key-check-type
                PARSE_RULE_OPTIONAL_DEPEND                                                      // opt/repo-sftp-host-key-check-type
                (                                                                               // opt/repo-sftp-host-key-check-type
                    PARSE_RULE_VAL_OPT(RepoType),                                               // opt/repo-sftp-host-key-check-type
                    PARSE_RULE_VAL_STRID(sftp),                                                 // opt/repo-sftp-host-key-check-type
                ),                                                                              // opt/repo-sftp-host-key-check-type
                                                                                                // opt/repo-sftp-host-key-check-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                  // opt/repo-sftp-host-key-check-type
                (                                                                               // opt/repo-sftp-host-key-check-type
                    PARSE_RULE_VAL_STRID(accept_DS_new),                                        // opt/repo-sftp-host-key-check-type
                    PARSE_RULE_VAL_STRID(fingerprint),                                          // opt/repo-sftp-host-key-check-type
                    PARSE_RULE_VAL_STRID(none),                                                 // opt/repo-sftp-host-key-check-type
                    PARSE_RULE_VAL_STRID(strict),                                               // opt/repo-sftp-host-key-check-type
                ),                                                                              // opt/repo-sftp-host-key-check-type
                                                                                                // opt/repo-sftp-host-key-check-type
                PARSE_RULE_OPTIONAL_DEFAULT                                                     // opt/repo-sftp-host-key-check-type
                (                                                                               // opt/repo-sftp-host-key-check-type
                    PARSE_RULE_VAL_STRID(strict),                                               // opt/repo-sftp-host-key-check-type
                ),                                                                              // opt/repo-sftp-host-key-check-type
            ),                                                                                  // opt/repo-sftp-host-key-check-type
        ),                                                                                      // opt/repo-sftp-host-key-check-type
    ),                                                                                          // opt/repo-sftp-host-key-check-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                            // opt/repo-sftp-host-key-hash-type
    (                                                                                            // opt/repo-sftp-host-key-hash-type
        PARSE_RULE_OPTION_NAME("repo-sftp-host-key-hash-type"),                                  // opt/repo-sftp-host-key-hash-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                        // opt/repo-sftp-host-key-hash-type
        PARSE_RULE_OPTION_RESET(true),                                                           // opt/repo-sftp-host-key-hash-type
        PARSE_RULE_OPTION_REQUIRED(true),                                                        // opt/repo-sftp-host-key-hash-type
        PARSE_RULE_OPTION_SECTION(Global),                                                       // opt/repo-sftp-host-key-hash-type
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                        // opt/repo-sftp-host-key-hash-type
                                                                                                 // opt/repo-sftp-host-key-hash-type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                           // opt/repo-sftp-host-key-hash-type
        (                                                                                        // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                  // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                               // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                    // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                     // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                    // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                      // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                  // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                   // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                    // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                   // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                    // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                   // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                              // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                              // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                             // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                    // opt/repo-sftp-host-key-hash-type
        ),                                                                                       // opt/repo-sftp-host-key-hash-type
                                                                                                 // opt/repo-sftp-host-key-hash-type
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                          // opt/repo-sftp-host-key-hash-type
        (                                                                                        // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                               // opt/repo-sftp-host-key-hash-type
        ),                                                                                       // opt/repo-sftp-host-key-hash-type
                                                                                                 // opt/repo-sftp-host-key-hash-type
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                          // opt/repo-sftp-host-key-hash-type
        (                                                                                        // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                               // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                    // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                   // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                    // opt/repo-sftp-host-key-hash-type
        ),                                                                                       // opt/repo-sftp-host-key-hash-type
                                                                                                 // opt/repo-sftp-host-key-hash-type
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                         // opt/repo-sftp-host-key-hash-type
        (                                                                                        // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                  // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                               // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                    // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                     // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                    // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                      // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                  // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                   // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                    // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                   // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                    // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                   // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                              // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                              // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                             // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                    // opt/repo-sftp-host-key-hash-type
        ),                                                                                       // opt/repo-sftp-host-key-hash-type
                                                                                                 // opt/repo-sftp-host-key-hash-type
        PARSE_RULE_OPTIONAL                                                                      // opt/repo-sftp-host-key-hash-type
        (                                                                                        // opt/repo-sftp-host-key-hash-type
            PARSE_RULE_OPTIONAL_GROUP                                                            // opt/repo-sftp-host-key-hash-type
            (                                                                                    // opt/repo-sftp-host-key-hash-type
                PARSE_RULE_OPTIONAL_DEPEND                                                       // opt/repo-sftp-host-key-hash-type
                (                                                                                // opt/repo-sftp-host-key-hash-type
                    PARSE_RULE_VAL_OPT(RepoType),                                                // opt/repo-sftp-host-key-hash-type
                    PARSE_RULE_VAL_STRID(sftp),                                                  // opt/repo-sftp-host-key-hash-type
                ),                                                                               // opt/repo-sftp-host-key-hash-type
                                                                                                 // opt/repo-sftp-host-key-hash-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                   // opt/repo-sftp-host-key-hash-type
                (                                                                                // opt/repo-sftp-host-key-hash-type
                    PARSE_RULE_VAL_STRID(md5),                                                   // opt/repo-sftp-host-key-hash-type
                    PARSE_RULE_VAL_STRID(sha1),                                                  // opt/repo-sftp-host-key-hash-type
                    PARSE_RULE_VAL_STRID(sha256),                                                // opt/repo-sftp-host-key-hash-type
                ),                                                                               // opt/repo-sftp-host-key-hash-type
            ),                                                                                   // opt/repo-sftp-host-key-hash-type
        ),                                                                                       // opt/repo-sftp-host-key-hash-type
    ),                                                                                           // opt/repo-sftp-host-key-hash-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                     // opt/repo-sftp-host-port
    (                                                                                                     // opt/repo-sftp-host-port
        PARSE_RULE_OPTION_NAME("repo-sftp-host-port"),                                                    // opt/repo-sftp-host-port
        PARSE_RULE_OPTION_TYPE(Integer),                                                                  // opt/repo-sftp-host-port
        PARSE_RULE_OPTION_RESET(true),                                                                    // opt/repo-sftp-host-port
        PARSE_RULE_OPTION_REQUIRED(true),                                                                 // opt/repo-sftp-host-port
        PARSE_RULE_OPTION_SECTION(Global),                                                                // opt/repo-sftp-host-port
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                 // opt/repo-sftp-host-port
                                                                                                          // opt/repo-sftp-host-port
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                    // opt/repo-sftp-host-port
        (                                                                                                 // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-sftp-host-port
        ),                                                                                                // opt/repo-sftp-host-port
                                                                                                          // opt/repo-sftp-host-port
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                   // opt/repo-sftp-host-port
        (                                                                                                 // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-sftp-host-port
        ),                                                                                                // opt/repo-sftp-host-port
                                                                                                          // opt/repo-sftp-host-port
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                   // opt/repo-sftp-host-port
        (                                                                                                 // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-sftp-host-port
        ),                                                                                                // opt/repo-sftp-host-port
                                                                                                          // opt/repo-sftp-host-port
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                  // opt/repo-sftp-host-port
        (                                                                                                 // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/repo-sftp-host-port
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-sftp-host-port
        ),                                                                                                // opt/repo-sftp-host-port
                                                                                                          // opt/repo-sftp-host-port
        PARSE_RULE_OPTIONAL                                                                               // opt/repo-sftp-host-port
        (                                                                                                 // opt/repo-sftp-host-port
            PARSE_RULE_OPTIONAL_GROUP                                                                     // opt/repo-sftp-host-port
            (                                                                                             // opt/repo-sftp-host-port
                PARSE_RULE_OPTIONAL_DEPEND                                                                // opt/repo-sftp-host-port
                (                                                                                         // opt/repo-sftp-host-port
                    PARSE_RULE_VAL_OPT(RepoType),                                                         // opt/repo-sftp-host-port
                    PARSE_RULE_VAL_STRID(sftp),                                                           // opt/repo-sftp-host-port
                ),                                                                                        // opt/repo-sftp-host-port
                                                                                                          // opt/repo-sftp-host-port
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                           // opt/repo-sftp-host-port
                (                                                                                         // opt/repo-sftp-host-port
                    PARSE_RULE_VAL_INT(1),                                                                // opt/repo-sftp-host-port
                    PARSE_RULE_VAL_INT(65535),                                                            // opt/repo-sftp-host-port
                ),                                                                                        // opt/repo-sftp-host-port
                                                                                                          // opt/repo-sftp-host-port
                PARSE_RULE_OPTIONAL_DEFAULT                                                               // opt/repo-sftp-host-port
                (                                                                                         // opt/repo-sftp-host-port
                    PARSE_RULE_VAL_INT(22),                                                               // opt/repo-sftp-host-port
                ),                                                                                        // opt/repo-sftp-host-port
            ),                                                                                            // opt/repo-sftp-host-port
        ),                                                                                                // opt/repo-sftp-host-port
    ),                                                                                                    // opt/repo-sftp-host-port
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                     // opt/repo-sftp-host-user
    (                                                                                                     // opt/repo-sftp-host-user
        PARSE_RULE_OPTION_NAME("repo-sftp-host-user"),                                                    // opt/repo-sftp-host-user
        PARSE_RULE_OPTION_TYPE(String),                                                                   // opt/repo-sftp-host-user
        PARSE_RULE_OPTION_RESET(true),                                                                    // opt/repo-sftp-host-user
        PARSE_RULE_OPTION_REQUIRED(true),                                                                 // opt/repo-sftp-host-user
        PARSE_RULE_OPTION_SECTION(Global),                                                                // opt/repo-sftp-host-user
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                 // opt/repo-sftp-host-user
                                                                                                          // opt/repo-sftp-host-user
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                    // opt/repo-sftp-host-user
        (                                                                                                 // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-sftp-host-user
        ),                                                                                                // opt/repo-sftp-host-user
                                                                                                          // opt/repo-sftp-host-user
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                   // opt/repo-sftp-host-user
        (                                                                                                 // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-sftp-host-user
        ),                                                                                                // opt/repo-sftp-host-user
                                                                                                          // opt/repo-sftp-host-user
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                   // opt/repo-sftp-host-user
        (                                                                                                 // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-sftp-host-user
        ),                                                                                                // opt/repo-sftp-host-user
                                                                                                          // opt/repo-sftp-host-user
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                  // opt/repo-sftp-host-user
        (                                                                                                 // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/repo-sftp-host-user
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/repo-sftp-host-user
        ),                                                                                                // opt/repo-sftp-host-user
                                                                                                          // opt/repo-sftp-host-user
        PARSE_RULE_OPTIONAL                                                                               // opt/repo-sftp-host-user
        (                                                                                                 // opt/repo-sftp-host-user
            PARSE_RULE_OPTIONAL_GROUP                                                                     // opt/repo-sftp-host-user
            (                                                                                             // opt/repo-sftp-host-user
                PARSE_RULE_OPTIONAL_DEPEND                                                                // opt/repo-sftp-host-user
                (                                                                                         // opt/repo-sftp-host-user
                    PARSE_RULE_VAL_OPT(RepoType),                                                         // opt/repo-sftp-host-user
                    PARSE_RULE_VAL_STRID(sftp),                                                           // opt/repo-sftp-host-user
                ),                                                                                        // opt/repo-sftp-host-user
            ),                                                                                            // opt/repo-sftp-host-user
        ),                                                                                                // opt/repo-sftp-host-user
    ),                                                                                                    // opt/repo-sftp-host-user
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                    // opt/repo-sftp-known-host
    (                                                                                                    // opt/repo-sftp-known-host
        PARSE_RULE_OPTION_NAME("repo-sftp-known-host"),                                                  // opt/repo-sftp-known-host
        PARSE_RULE_OPTION_TYPE(List),                                                                    // opt/repo-sftp-known-host
        PARSE_RULE_OPTION_RESET(true),                                                                   // opt/repo-sftp-known-host
        PARSE_RULE_OPTION_REQUIRED(false),                                                               // opt/repo-sftp-known-host
        PARSE_RULE_OPTION_SECTION(Global),                                                               // opt/repo-sftp-known-host
        PARSE_RULE_OPTION_MULTI(true),                                                                   // opt/repo-sftp-known-host
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                // opt/repo-sftp-known-host
                                                                                                         // opt/repo-sftp-known-host
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                   // opt/repo-sftp-known-host
        (                                                                                                // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                          // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Check)                                                             // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Expire)                                                            // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Info)                                                              // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                          // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                           // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                            // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                           // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                            // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                      // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                      // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                     // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-sftp-known-host
        ),                                                                                               // opt/repo-sftp-known-host
                                                                                                         // opt/repo-sftp-known-host
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                  // opt/repo-sftp-known-host
        (                                                                                                // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-sftp-known-host
        ),                                                                                               // opt/repo-sftp-known-host
                                                                                                         // opt/repo-sftp-known-host
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                  // opt/repo-sftp-known-host
        (                                                                                                // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-sftp-known-host
        ),                                                                                               // opt/repo-sftp-known-host
                                                                                                         // opt/repo-sftp-known-host
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                 // opt/repo-sftp-known-host
        (                                                                                                // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                          // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Check)                                                             // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Expire)                                                            // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Info)                                                              // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                          // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                           // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                            // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                           // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                            // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                      // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                      // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                     // opt/repo-sftp-known-host
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-sftp-known-host
        ),                                                                                               // opt/repo-sftp-known-host
                                                                                                         // opt/repo-sftp-known-host
        PARSE_RULE_OPTIONAL                                                                              // opt/repo-sftp-known-host
        (                                                                                                // opt/repo-sftp-known-host
            PARSE_RULE_OPTIONAL_GROUP                                                                    // opt/repo-sftp-known-host
            (                                                                                            // opt/repo-sftp-known-host
                PARSE_RULE_OPTIONAL_DEPEND                                                               // opt/repo-sftp-known-host
                (                                                                                        // opt/repo-sftp-known-host
                    PARSE_RULE_VAL_OPT(RepoSftpHostKeyCheckType),                                        // opt/repo-sftp-known-host
                    PARSE_RULE_VAL_STRID(accept_DS_new),                                                 // opt/repo-sftp-known-host
                    PARSE_RULE_VAL_STRID(strict),                                                        // opt/repo-sftp-known-host
                ),                                                                                       // opt/repo-sftp-known-host
            ),                                                                                           // opt/repo-sftp-known-host
        ),                                                                                               // opt/repo-sftp-known-host
    ),                                                                                                   // opt/repo-sftp-known-host
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                              // opt/repo-sftp-private-key-file
    (                                                                                              // opt/repo-sftp-private-key-file
        PARSE_RULE_OPTION_NAME("repo-sftp-private-key-file"),                                      // opt/repo-sftp-private-key-file
        PARSE_RULE_OPTION_TYPE(String),                                                            // opt/repo-sftp-private-key-file
        PARSE_RULE_OPTION_RESET(true),                                                             // opt/repo-sftp-private-key-file
        PARSE_RULE_OPTION_REQUIRED(true),                                                          // opt/repo-sftp-private-key-file
        PARSE_RULE_OPTION_SECTION(Global),                                                         // opt/repo-sftp-private-key-file
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                          // opt/repo-sftp-private-key-file
                                                                                                   // opt/repo-sftp-private-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                             // opt/repo-sftp-private-key-file
        (                                                                                          // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                    // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                  // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                 // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                      // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                       // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Expire)                                                      // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Info)                                                        // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                    // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                     // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                      // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                     // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                      // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                     // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                               // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                      // opt/repo-sftp-private-key-file
        ),                                                                                         // opt/repo-sftp-private-key-file
                                                                                                   // opt/repo-sftp-private-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                            // opt/repo-sftp-private-key-file
        (                                                                                          // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                  // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                 // opt/repo-sftp-private-key-file
        ),                                                                                         // opt/repo-sftp-private-key-file
                                                                                                   // opt/repo-sftp-private-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                            // opt/repo-sftp-private-key-file
        (                                                                                          // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                  // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                 // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                      // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                     // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                      // opt/repo-sftp-private-key-file
        ),                                                                                         // opt/repo-sftp-private-key-file
                                                                                                   // opt/repo-sftp-private-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                           // opt/repo-sftp-private-key-file
        (                                                                                          // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                    // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                  // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                 // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                      // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                       // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Expire)                                                      // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Info)                                                        // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                    // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                     // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                      // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                     // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                      // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                     // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                               // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                      // opt/repo-sftp-private-key-file
        ),                                                                                         // opt/repo-sftp-private-key-file
                                                                                                   // opt/repo-sftp-private-key-file
        PARSE_RULE_OPTIONAL                                                                        // opt/repo-sftp-private-key-file
        (                                                                                          // opt/repo-sftp-private-key-file
            PARSE_RULE_OPTIONAL_GROUP                                                              // opt/repo-sftp-private-key-file
            (                                                                                      // opt/repo-sftp-private-key-file
                PARSE_RULE_OPTIONAL_DEPEND                                                         // opt/repo-sftp-private-key-file
                (                                                                                  // opt/repo-sftp-private-key-file
                    PARSE_RULE_VAL_OPT(RepoType),                                                  // opt/repo-sftp-private-key-file
                    PARSE_RULE_VAL_STRID(sftp),                                                    // opt/repo-sftp-private-key-file
                ),                                                                                 // opt/repo-sftp-private-key-file
            ),                                                                                     // opt/repo-sftp-private-key-file
        ),                                                                                         // opt/repo-sftp-private-key-file
    ),                                                                                             // opt/repo-sftp-private-key-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                        // opt/repo-sftp-private-key-passphrase
    (                                                                                        // opt/repo-sftp-private-key-passphrase
        PARSE_RULE_OPTION_NAME("repo-sftp-private-key-passphrase"),                          // opt/repo-sftp-private-key-passphrase
        PARSE_RULE_OPTION_TYPE(String),                                                      // opt/repo-sftp-private-key-passphrase
        PARSE_RULE_OPTION_RESET(true),                                                       // opt/repo-sftp-private-key-passphrase
        PARSE_RULE_OPTION_REQUIRED(false),                                                   // opt/repo-sftp-private-key-passphrase
        PARSE_RULE_OPTION_SECTION(Global),                                                   // opt/repo-sftp-private-key-passphrase
        PARSE_RULE_OPTION_SECURE(true),                                                      // opt/repo-sftp-private-key-passphrase
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                    // opt/repo-sftp-private-key-passphrase
                                                                                             // opt/repo-sftp-private-key-passphrase
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                       // opt/repo-sftp-private-key-passphrase
        (                                                                                    // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Annotate)                                              // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                            // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                           // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Backup)                                                // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Check)                                                 // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Expire)                                                // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Info)                                                  // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Manifest)                                              // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                               // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                               // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Restore)                                               // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                          // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                          // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                         // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Verify)                                                // opt/repo-sftp-private-key-passphrase
        ),                                                                                   // opt/repo-sftp-private-key-passphrase
                                                                                             // opt/repo-sftp-private-key-passphrase
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                      // opt/repo-sftp-private-key-passphrase
        (                                                                                    // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                            // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                           // opt/repo-sftp-private-key-passphrase
        ),                                                                                   // opt/repo-sftp-private-key-passphrase
                                                                                             // opt/repo-sftp-private-key-passphrase
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                      // opt/repo-sftp-private-key-passphrase
        (                                                                                    // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                            // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                           // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Backup)                                                // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Restore)                                               // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Verify)                                                // opt/repo-sftp-private-key-passphrase
        ),                                                                                   // opt/repo-sftp-private-key-passphrase
                                                                                             // opt/repo-sftp-private-key-passphrase
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                     // opt/repo-sftp-private-key-passphrase
        (                                                                                    // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Annotate)                                              // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                            // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                           // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Backup)                                                // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Check)                                                 // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Expire)                                                // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Info)                                                  // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Manifest)                                              // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                               // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                               // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Restore)                                               // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                          // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                          // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                         // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTION_COMMAND(Verify)                                                // opt/repo-sftp-private-key-passphrase
        ),                                                                                   // opt/repo-sftp-private-key-passphrase
                                                                                             // opt/repo-sftp-private-key-passphrase
        PARSE_RULE_OPTIONAL                                                                  // opt/repo-sftp-private-key-passphrase
        (                                                                                    // opt/repo-sftp-private-key-passphrase
            PARSE_RULE_OPTIONAL_GROUP                                                        // opt/repo-sftp-private-key-passphrase
            (                                                                                // opt/repo-sftp-private-key-passphrase
                PARSE_RULE_OPTIONAL_DEPEND                                                   // opt/repo-sftp-private-key-passphrase
                (                                                                            // opt/repo-sftp-private-key-passphrase
                    PARSE_RULE_VAL_OPT(RepoType),                                            // opt/repo-sftp-private-key-passphrase
                    PARSE_RULE_VAL_STRID(sftp),                                              // opt/repo-sftp-private-key-passphrase
                ),                                                                           // opt/repo-sftp-private-key-passphrase
            ),                                                                               // opt/repo-sftp-private-key-passphrase
        ),                                                                                   // opt/repo-sftp-private-key-passphrase
    ),                                                                                       // opt/repo-sftp-private-key-passphrase
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                               // opt/repo-sftp-public-key-file
    (                                                                                               // opt/repo-sftp-public-key-file
        PARSE_RULE_OPTION_NAME("repo-sftp-public-key-file"),                                        // opt/repo-sftp-public-key-file
        PARSE_RULE_OPTION_TYPE(String),                                                             // opt/repo-sftp-public-key-file
        PARSE_RULE_OPTION_RESET(true),                                                              // opt/repo-sftp-public-key-file
        PARSE_RULE_OPTION_REQUIRED(false),                                                          // opt/repo-sftp-public-key-file
        PARSE_RULE_OPTION_SECTION(Global),                                                          // opt/repo-sftp-public-key-file
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                           // opt/repo-sftp-public-key-file
                                                                                                    // opt/repo-sftp-public-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                              // opt/repo-sftp-public-key-file
        (                                                                                           // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                     // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                   // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                  // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                       // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                        // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Expire)                                                       // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Info)                                                         // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                     // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                      // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                       // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                      // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                       // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                      // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                 // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                 // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                       // opt/repo-sftp-public-key-file
        ),                                                                                          // opt/repo-sftp-public-key-file
                                                                                                    // opt/repo-sftp-public-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                             // opt/repo-sftp-public-key-file
        (                                                                                           // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                   // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                  // opt/repo-sftp-public-key-file
        ),                                                                                          // opt/repo-sftp-public-key-file
                                                                                                    // opt/repo-sftp-public-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                             // opt/repo-sftp-public-key-file
        (                                                                                           // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                   // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                  // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                       // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                      // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                       // opt/repo-sftp-public-key-file
        ),                                                                                          // opt/repo-sftp-public-key-file
                                                                                                    // opt/repo-sftp-public-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                            // opt/repo-sftp-public-key-file
        (                                                                                           // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                     // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                   // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                  // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                       // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                        // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Expire)                                                       // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Info)                                                         // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                     // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                      // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                       // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                      // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                       // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                      // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                 // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                 // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                       // opt/repo-sftp-public-key-file
        ),                                                                                          // opt/repo-sftp-public-key-file
                                                                                                    // opt/repo-sftp-public-key-file
        PARSE_RULE_OPTIONAL                                                                         // opt/repo-sftp-public-key-file
        (                                                                                           // opt/repo-sftp-public-key-file
            PARSE_RULE_OPTIONAL_GROUP                                                               // opt/repo-sftp-public-key-file
            (                                                                                       // opt/repo-sftp-public-key-file
                PARSE_RULE_OPTIONAL_DEPEND                                                          // opt/repo-sftp-public-key-file
                (                                                                                   // opt/repo-sftp-public-key-file
                    PARSE_RULE_VAL_OPT(RepoType),                                                   // opt/repo-sftp-public-key-file
                    PARSE_RULE_VAL_STRID(sftp),                                                     // opt/repo-sftp-public-key-file
                ),                                                                                  // opt/repo-sftp-public-key-file
            ),                                                                                      // opt/repo-sftp-public-key-file
        ),                                                                                          // opt/repo-sftp-public-key-file
    ),                                                                                              // opt/repo-sftp-public-key-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                    // opt/repo-storage-ca-file
    (                                                                                                    // opt/repo-storage-ca-file
        PARSE_RULE_OPTION_NAME("repo-storage-ca-file"),                                                  // opt/repo-storage-ca-file
        PARSE_RULE_OPTION_TYPE(String),                                                                  // opt/repo-storage-ca-file
        PARSE_RULE_OPTION_RESET(true),                                                                   // opt/repo-storage-ca-file
        PARSE_RULE_OPTION_REQUIRED(false),                                                               // opt/repo-storage-ca-file
        PARSE_RULE_OPTION_SECTION(Global),                                                               // opt/repo-storage-ca-file
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                // opt/repo-storage-ca-file
                                                                                                         // opt/repo-storage-ca-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                   // opt/repo-storage-ca-file
        (                                                                                                // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                          // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                             // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Expire)                                                            // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Info)                                                              // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                          // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                           // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                            // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                           // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                            // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                      // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                      // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                     // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-storage-ca-file
        ),                                                                                               // opt/repo-storage-ca-file
                                                                                                         // opt/repo-storage-ca-file
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                  // opt/repo-storage-ca-file
        (                                                                                                // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-storage-ca-file
        ),                                                                                               // opt/repo-storage-ca-file
                                                                                                         // opt/repo-storage-ca-file
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                  // opt/repo-storage-ca-file
        (                                                                                                // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-storage-ca-file
        ),                                                                                               // opt/repo-storage-ca-file
                                                                                                         // opt/repo-storage-ca-file
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                 // opt/repo-storage-ca-file
        (                                                                                                // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                          // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Check)                                                             // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Expire)                                                            // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Info)                                                              // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                          // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                           // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                            // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                           // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                            // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                      // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                      // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                     // opt/repo-storage-ca-file
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-storage-ca-file
        ),                                                                                               // opt/repo-storage-ca-file
                                                                                                         // opt/repo-storage-ca-file
        PARSE_RULE_OPTIONAL                                                                              // opt/repo-storage-ca-file
        (                                                                                                // opt/repo-storage-ca-file
            PARSE_RULE_OPTIONAL_GROUP                                                                    // opt/repo-storage-ca-file
            (                                                                                            // opt/repo-storage-ca-file
                PARSE_RULE_OPTIONAL_DEPEND                                                               // opt/repo-storage-ca-file
                (                                                                                        // opt/repo-storage-ca-file
                    PARSE_RULE_VAL_OPT(RepoType),                                                        // opt/repo-storage-ca-file
                    PARSE_RULE_VAL_STRID(azure),                                                         // opt/repo-storage-ca-file
                    PARSE_RULE_VAL_STRID(gcs),                                                           // opt/repo-storage-ca-file
                    PARSE_RULE_VAL_STRID(s3),                                                            // opt/repo-storage-ca-file
                ),                                                                                       // opt/repo-storage-ca-file
            ),                                                                                           // opt/repo-storage-ca-file
        ),                                                                                               // opt/repo-storage-ca-file
    ),                                                                                                   // opt/repo-storage-ca-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                    // opt/repo-storage-ca-path
    (                                                                                                    // opt/repo-storage-ca-path
        PARSE_RULE_OPTION_NAME("repo-storage-ca-path"),                                                  // opt/repo-storage-ca-path
        PARSE_RULE_OPTION_TYPE(Path),                                                                    // opt/repo-storage-ca-path
        PARSE_RULE_OPTION_RESET(true),                                                                   // opt/repo-storage-ca-path
        PARSE_RULE_OPTION_REQUIRED(false),                                                               // opt/repo-storage-ca-path
        PARSE_RULE_OPTION_SECTION(Global),                                                               // opt/repo-storage-ca-path
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                // opt/repo-storage-ca-path
                                                                                                         // opt/repo-storage-ca-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                   // opt/repo-storage-ca-path
        (                                                                                                // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                          // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                             // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                            // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                              // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                          // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                           // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                            // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                           // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                            // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                      // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                      // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                     // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-storage-ca-path
        ),                                                                                               // opt/repo-storage-ca-path
                                                                                                         // opt/repo-storage-ca-path
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                  // opt/repo-storage-ca-path
        (                                                                                                // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-storage-ca-path
        ),                                                                                               // opt/repo-storage-ca-path
                                                                                                         // opt/repo-storage-ca-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                  // opt/repo-storage-ca-path
        (                                                                                                // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-storage-ca-path
        ),                                                                                               // opt/repo-storage-ca-path
                                                                                                         // opt/repo-storage-ca-path
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                 // opt/repo-storage-ca-path
        (                                                                                                // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                          // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Check)                                                             // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Expire)                                                            // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Info)                                                              // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                          // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                           // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                            // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                           // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                            // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                      // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                      // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                     // opt/repo-storage-ca-path
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/repo-storage-ca-path
        ),                                                                                               // opt/repo-storage-ca-path
                                                                                                         // opt/repo-storage-ca-path
        PARSE_RULE_OPTIONAL                                                                              // opt/repo-storage-ca-path
        (                                                                                                // opt/repo-storage-ca-path
            PARSE_RULE_OPTIONAL_GROUP                                                                    // opt/repo-storage-ca-path
            (                                                                                            // opt/repo-storage-ca-path
                PARSE_RULE_OPTIONAL_DEPEND                                                               // opt/repo-storage-ca-path
                (                                                                                        // opt/repo-storage-ca-path
                    PARSE_RULE_VAL_OPT(RepoType),                                                        // opt/repo-storage-ca-path
                    PARSE_RULE_VAL_STRID(azure),                                                         // opt/repo-storage-ca-path
                    PARSE_RULE_VAL_STRID(gcs),                                                           // opt/repo-storage-ca-path
                    PARSE_RULE_VAL_STRID(s3),                                                            // opt/repo-storage-ca-path
                ),                                                                                       // opt/repo-storage-ca-path
            ),                                                                                           // opt/repo-storage-ca-path
        ),                                                                                               // opt/repo-storage-ca-path
    ),                                                                                                   // opt/repo-storage-ca-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                       // opt/repo-storage-host
    (                                                                                                       // opt/repo-storage-host
        PARSE_RULE_OPTION_NAME("repo-storage-host"),                                                        // opt/repo-storage-host
        PARSE_RULE_OPTION_TYPE(String),                                                                     // opt/repo-storage-host
        PARSE_RULE_OPTION_RESET(true),                                                                      // opt/repo-storage-host
        PARSE_RULE_OPTION_REQUIRED(false),                                                                  // opt/repo-storage-host
        PARSE_RULE_OPTION_SECTION(Global),                                                                  // opt/repo-storage-host
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                   // opt/repo-storage-host
                                                                                                            // opt/repo-storage-host
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                      // opt/repo-storage-host
        (                                                                                                   // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-storage-host
        ),                                                                                                  // opt/repo-storage-host
                                                                                                            // opt/repo-storage-host
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                     // opt/repo-storage-host
        (                                                                                                   // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-storage-host
        ),                                                                                                  // opt/repo-storage-host
                                                                                                            // opt/repo-storage-host
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                     // opt/repo-storage-host
        (                                                                                                   // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-storage-host
        ),                                                                                                  // opt/repo-storage-host
                                                                                                            // opt/repo-storage-host
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                    // opt/repo-storage-host
        (                                                                                                   // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/repo-storage-host
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-storage-host
        ),                                                                                                  // opt/repo-storage-host
                                                                                                            // opt/repo-storage-host
        PARSE_RULE_OPTIONAL                                                                                 // opt/repo-storage-host
        (                                                                                                   // opt/repo-storage-host
            PARSE_RULE_OPTIONAL_GROUP                                                                       // opt/repo-storage-host
            (                                                                                               // opt/repo-storage-host
                PARSE_RULE_OPTIONAL_DEPEND                                                                  // opt/repo-storage-host
                (                                                                                           // opt/repo-storage-host
                    PARSE_RULE_VAL_OPT(RepoType),                                                           // opt/repo-storage-host
                    PARSE_RULE_VAL_STRID(azure),                                                            // opt/repo-storage-host
                    PARSE_RULE_VAL_STRID(s3),                                                               // opt/repo-storage-host
                ),                                                                                          // opt/repo-storage-host
            ),                                                                                              // opt/repo-storage-host
        ),                                                                                                  // opt/repo-storage-host
    ),                                                                                                      // opt/repo-storage-host
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                       // opt/repo-storage-port
    (                                                                                                       // opt/repo-storage-port
        PARSE_RULE_OPTION_NAME("repo-storage-port"),                                                        // opt/repo-storage-port
        PARSE_RULE_OPTION_TYPE(Integer),                                                                    // opt/repo-storage-port
        PARSE_RULE_OPTION_RESET(true),                                                                      // opt/repo-storage-port
        PARSE_RULE_OPTION_REQUIRED(true),                                                                   // opt/repo-storage-port
        PARSE_RULE_OPTION_SECTION(Global),                                                                  // opt/repo-storage-port
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                   // opt/repo-storage-port
                                                                                                            // opt/repo-storage-port
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                      // opt/repo-storage-port
        (                                                                                                   // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-storage-port
        ),                                                                                                  // opt/repo-storage-port
                                                                                                            // opt/repo-storage-port
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                     // opt/repo-storage-port
        (                                                                                                   // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-storage-port
        ),                                                                                                  // opt/repo-storage-port
                                                                                                            // opt/repo-storage-port
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                     // opt/repo-storage-port
        (                                                                                                   // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-storage-port
        ),                                                                                                  // opt/repo-storage-port
                                                                                                            // opt/repo-storage-port
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                    // opt/repo-storage-port
        (                                                                                                   // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                             // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                           // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                          // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Backup)                                                               // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Check)                                                                // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Expire)                                                               // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Info)                                                                 // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                             // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                              // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                               // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                              // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                               // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Restore)                                                              // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                         // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                         // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                        // opt/repo-storage-port
            PARSE_RULE_OPTION_COMMAND(Verify)                                                               // opt/repo-storage-port
        ),                                                                                                  // opt/repo-storage-port
                                                                                                            // opt/repo-storage-port
        PARSE_RULE_OPTIONAL                                                                                 // opt/repo-storage-port
        (                                                                                                   // opt/repo-storage-port
            PARSE_RULE_OPTIONAL_GROUP                                                                       // opt/repo-storage-port
            (                                                                                               // opt/repo-storage-port
                PARSE_RULE_OPTIONAL_DEPEND                                                                  // opt/repo-storage-port
                (                                                                                           // opt/repo-storage-port
                    PARSE_RULE_VAL_OPT(RepoType),                                                           // opt/repo-storage-port
                    PARSE_RULE_VAL_STRID(azure),                                                            // opt/repo-storage-port
                    PARSE_RULE_VAL_STRID(s3),                                                               // opt/repo-storage-port
                ),                                                                                          // opt/repo-storage-port
                                                                                                            // opt/repo-storage-port
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                             // opt/repo-storage-port
                (                                                                                           // opt/repo-storage-port
                    PARSE_RULE_VAL_INT(1),                                                                  // opt/repo-storage-port
                    PARSE_RULE_VAL_INT(65535),                                                              // opt/repo-storage-port
                ),                                                                                          // opt/repo-storage-port
                                                                                                            // opt/repo-storage-port
                PARSE_RULE_OPTIONAL_DEFAULT                                                                 // opt/repo-storage-port
                (                                                                                           // opt/repo-storage-port
                    PARSE_RULE_VAL_INT(443),                                                                // opt/repo-storage-port
                ),                                                                                          // opt/repo-storage-port
            ),                                                                                              // opt/repo-storage-port
        ),                                                                                                  // opt/repo-storage-port
    ),                                                                                                      // opt/repo-storage-port
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/repo-storage-tag
    (                                                                                                        // opt/repo-storage-tag
        PARSE_RULE_OPTION_NAME("repo-storage-tag"),                                                          // opt/repo-storage-tag
        PARSE_RULE_OPTION_TYPE(Hash),                                                                        // opt/repo-storage-tag
        PARSE_RULE_OPTION_RESET(true),                                                                       // opt/repo-storage-tag
        PARSE_RULE_OPTION_REQUIRED(false),                                                                   // opt/repo-storage-tag
        PARSE_RULE_OPTION_SECTION(Global),                                                                   // opt/repo-storage-tag
        PARSE_RULE_OPTION_MULTI(true),                                                                       // opt/repo-storage-tag
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                    // opt/repo-storage-tag
                                                                                                             // opt/repo-storage-tag
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/repo-storage-tag
        (                                                                                                    // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-storage-tag
        ),                                                                                                   // opt/repo-storage-tag
                                                                                                             // opt/repo-storage-tag
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                      // opt/repo-storage-tag
        (                                                                                                    // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-storage-tag
        ),                                                                                                   // opt/repo-storage-tag
                                                                                                             // opt/repo-storage-tag
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                      // opt/repo-storage-tag
        (                                                                                                    // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-storage-tag
        ),                                                                                                   // opt/repo-storage-tag
                                                                                                             // opt/repo-storage-tag
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                     // opt/repo-storage-tag
        (                                                                                                    // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                              // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                           // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Check)                                                                 // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                               // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                          // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                          // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                         // opt/repo-storage-tag
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-storage-tag
        ),                                                                                                   // opt/repo-storage-tag
                                                                                                             // opt/repo-storage-tag
        PARSE_RULE_OPTIONAL                                                                                  // opt/repo-storage-tag
        (                                                                                                    // opt/repo-storage-tag
            PARSE_RULE_OPTIONAL_GROUP                                                                        // opt/repo-storage-tag
            (                                                                                                // opt/repo-storage-tag
                PARSE_RULE_OPTIONAL_DEPEND                                                                   // opt/repo-storage-tag
                (                                                                                            // opt/repo-storage-tag
                    PARSE_RULE_VAL_OPT(RepoType),                                                            // opt/repo-storage-tag
                    PARSE_RULE_VAL_STRID(azure),                                                             // opt/repo-storage-tag
                    PARSE_RULE_VAL_STRID(gcs),                                                               // opt/repo-storage-tag
                    PARSE_RULE_VAL_STRID(s3),                                                                // opt/repo-storage-tag
                ),                                                                                           // opt/repo-storage-tag
            ),                                                                                               // opt/repo-storage-tag
        ),                                                                                                   // opt/repo-storage-tag
    ),                                                                                                       // opt/repo-storage-tag
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                          // opt/repo-storage-upload-chunk-size
    (                                                                                          // opt/repo-storage-upload-chunk-size
        PARSE_RULE_OPTION_NAME("repo-storage-upload-chunk-size"),                              // opt/repo-storage-upload-chunk-size
        PARSE_RULE_OPTION_TYPE(Size),                                                          // opt/repo-storage-upload-chunk-size
        PARSE_RULE_OPTION_RESET(true),                                                         // opt/repo-storage-upload-chunk-size
        PARSE_RULE_OPTION_REQUIRED(true),                                                      // opt/repo-storage-upload-chunk-size
        PARSE_RULE_OPTION_SECTION(Global),                                                     // opt/repo-storage-upload-chunk-size
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                      // opt/repo-storage-upload-chunk-size
                                                                                               // opt/repo-storage-upload-chunk-size
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                         // opt/repo-storage-upload-chunk-size
        (                                                                                      // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                              // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                             // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Backup)                                                  // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Check)                                                   // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Expire)                                                  // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Info)                                                    // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                 // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                  // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                 // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                  // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Restore)                                                 // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                            // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                            // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                           // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Verify)                                                  // opt/repo-storage-upload-chunk-size
        ),                                                                                     // opt/repo-storage-upload-chunk-size
                                                                                               // opt/repo-storage-upload-chunk-size
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                        // opt/repo-storage-upload-chunk-size
        (                                                                                      // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                              // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                             // opt/repo-storage-upload-chunk-size
        ),                                                                                     // opt/repo-storage-upload-chunk-size
                                                                                               // opt/repo-storage-upload-chunk-size
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                        // opt/repo-storage-upload-chunk-size
        (                                                                                      // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                              // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                             // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Backup)                                                  // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Restore)                                                 // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Verify)                                                  // opt/repo-storage-upload-chunk-size
        ),                                                                                     // opt/repo-storage-upload-chunk-size
                                                                                               // opt/repo-storage-upload-chunk-size
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                       // opt/repo-storage-upload-chunk-size
        (                                                                                      // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                              // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                             // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Backup)                                                  // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Check)                                                   // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Expire)                                                  // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Info)                                                    // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                 // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                  // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                 // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                  // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Restore)                                                 // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                            // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                            // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                           // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTION_COMMAND(Verify)                                                  // opt/repo-storage-upload-chunk-size
        ),                                                                                     // opt/repo-storage-upload-chunk-size
                                                                                               // opt/repo-storage-upload-chunk-size
        PARSE_RULE_OPTIONAL                                                                    // opt/repo-storage-upload-chunk-size
        (                                                                                      // opt/repo-storage-upload-chunk-size
            PARSE_RULE_OPTIONAL_GROUP                                                          // opt/repo-storage-upload-chunk-size
            (                                                                                  // opt/repo-storage-upload-chunk-size
                PARSE_RULE_OPTIONAL_DEPEND                                                     // opt/repo-storage-upload-chunk-size
                (                                                                              // opt/repo-storage-upload-chunk-size
                    PARSE_RULE_VAL_OPT(RepoType),                                              // opt/repo-storage-upload-chunk-size
                    PARSE_RULE_VAL_STRID(azure),                                               // opt/repo-storage-upload-chunk-size
                    PARSE_RULE_VAL_STRID(gcs),                                                 // opt/repo-storage-upload-chunk-size
                    PARSE_RULE_VAL_STRID(s3),                                                  // opt/repo-storage-upload-chunk-size
                ),                                                                             // opt/repo-storage-upload-chunk-size
                                                                                               // opt/repo-storage-upload-chunk-size
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                // opt/repo-storage-upload-chunk-size
                (                                                                              // opt/repo-storage-upload-chunk-size
                    PARSE_RULE_OPTIONAL_ALLOW_RANGE_MAP                                        // opt/repo-storage-upload-chunk-size
                    (                                                                          // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_STRID(azure),                                           // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_SIZE(4MiB),                                             // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_SIZE(1GiB),                                             // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_STRID(gcs),                                             // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_SIZE(4MiB),                                             // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_SIZE(1GiB),                                             // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_STRID(s3),                                              // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_SIZE(5MiB),                                             // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_SIZE(1GiB),                                             // opt/repo-storage-upload-chunk-size
                    ),                                                                         // opt/repo-storage-upload-chunk-size
                ),                                                                             // opt/repo-storage-upload-chunk-size
                                                                                               // opt/repo-storage-upload-chunk-size
                PARSE_RULE_OPTIONAL_DEFAULT                                                    // opt/repo-storage-upload-chunk-size
                (                                                                              // opt/repo-storage-upload-chunk-size
                    PARSE_RULE_OPTIONAL_DEFAULT_MAP                                            // opt/repo-storage-upload-chunk-size
                    (                                                                          // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_STRID(azure),                                           // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_SIZE(4MiB),                                             // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_STRID(gcs),                                             // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_SIZE(4MiB),                                             // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_STRID(s3),                                              // opt/repo-storage-upload-chunk-size
                        PARSE_RULE_VAL_SIZE(5MiB),                                             // opt/repo-storage-upload-chunk-size
                    ),                                                                         // opt/repo-storage-upload-chunk-size
                ),                                                                             // opt/repo-storage-upload-chunk-size
            ),                                                                                 // opt/repo-storage-upload-chunk-size
        ),                                                                                     // opt/repo-storage-upload-chunk-size
    ),                                                                                         // opt/repo-storage-upload-chunk-size
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                 // opt/repo-storage-verify-tls
    (                                                                                                 // opt/repo-storage-verify-tls
        PARSE_RULE_OPTION_NAME("repo-storage-verify-tls"),                                            // opt/repo-storage-verify-tls
        PARSE_RULE_OPTION_TYPE(Boolean),                                                              // opt/repo-storage-verify-tls
        PARSE_RULE_OPTION_NEGATE(true),                                                               // opt/repo-storage-verify-tls
        PARSE_RULE_OPTION_RESET(true),                                                                // opt/repo-storage-verify-tls
        PARSE_RULE_OPTION_REQUIRED(true),                                                             // opt/repo-storage-verify-tls
        PARSE_RULE_OPTION_SECTION(Global),                                                            // opt/repo-storage-verify-tls
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                             // opt/repo-storage-verify-tls
                                                                                                      // opt/repo-storage-verify-tls
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                // opt/repo-storage-verify-tls
        (                                                                                             // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                       // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                     // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                    // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Backup)                                                         // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Check)                                                          // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Expire)                                                         // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Info)                                                           // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                       // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                        // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                         // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                        // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                         // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Restore)                                                        // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                   // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                   // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                  // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Verify)                                                         // opt/repo-storage-verify-tls
        ),                                                                                            // opt/repo-storage-verify-tls
                                                                                                      // opt/repo-storage-verify-tls
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                               // opt/repo-storage-verify-tls
        (                                                                                             // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                     // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                    // opt/repo-storage-verify-tls
        ),                                                                                            // opt/repo-storage-verify-tls
                                                                                                      // opt/repo-storage-verify-tls
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                               // opt/repo-storage-verify-tls
        (                                                                                             // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                     // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                    // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Backup)                                                         // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Restore)                                                        // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Verify)                                                         // opt/repo-storage-verify-tls
        ),                                                                                            // opt/repo-storage-verify-tls
                                                                                                      // opt/repo-storage-verify-tls
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                              // opt/repo-storage-verify-tls
        (                                                                                             // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                       // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                     // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                    // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Backup)                                                         // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Check)                                                          // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Expire)                                                         // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Info)                                                           // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                       // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                        // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                         // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                        // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                         // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Restore)                                                        // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                   // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                   // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                  // opt/repo-storage-verify-tls
            PARSE_RULE_OPTION_COMMAND(Verify)                                                         // opt/repo-storage-verify-tls
        ),                                                                                            // opt/repo-storage-verify-tls
                                                                                                      // opt/repo-storage-verify-tls
        PARSE_RULE_OPTIONAL                                                                           // opt/repo-storage-verify-tls
        (                                                                                             // opt/repo-storage-verify-tls
            PARSE_RULE_OPTIONAL_GROUP                                                                 // opt/repo-storage-verify-tls
            (                                                                                         // opt/repo-storage-verify-tls
                PARSE_RULE_OPTIONAL_DEPEND                                                            // opt/repo-storage-verify-tls
                (                                                                                     // opt/repo-storage-verify-tls
                    PARSE_RULE_VAL_OPT(RepoType),                                                     // opt/repo-storage-verify-tls
                    PARSE_RULE_VAL_STRID(azure),                                                      // opt/repo-storage-verify-tls
                    PARSE_RULE_VAL_STRID(gcs),                                                        // opt/repo-storage-verify-tls
                    PARSE_RULE_VAL_STRID(s3),                                                         // opt/repo-storage-verify-tls
                ),                                                                                    // opt/repo-storage-verify-tls
                                                                                                      // opt/repo-storage-verify-tls
                PARSE_RULE_OPTIONAL_DEFAULT                                                           // opt/repo-storage-verify-tls
                (                                                                                     // opt/repo-storage-verify-tls
                    PARSE_RULE_VAL_BOOL_TRUE,                                                         // opt/repo-storage-verify-tls
                ),                                                                                    // opt/repo-storage-verify-tls
            ),                                                                                        // opt/repo-storage-verify-tls
        ),                                                                                            // opt/repo-storage-verify-tls
    ),                                                                                                // opt/repo-storage-verify-tls
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                            // opt/repo-symlink
    (                                                                                                            // opt/repo-symlink
        PARSE_RULE_OPTION_NAME("repo-symlink"),                                                                  // opt/repo-symlink
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                         // opt/repo-symlink
        PARSE_RULE_OPTION_NEGATE(true),                                                                          // opt/repo-symlink
        PARSE_RULE_OPTION_RESET(true),                                                                           // opt/repo-symlink
        PARSE_RULE_OPTION_REQUIRED(true),                                                                        // opt/repo-symlink
        PARSE_RULE_OPTION_SECTION(Global),                                                                       // opt/repo-symlink
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                        // opt/repo-symlink
                                                                                                                 // opt/repo-symlink
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                           // opt/repo-symlink
        (                                                                                                        // opt/repo-symlink
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                    // opt/repo-symlink
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                    // opt/repo-symlink
        ),                                                                                                       // opt/repo-symlink
                                                                                                                 // opt/repo-symlink
        PARSE_RULE_OPTIONAL                                                                                      // opt/repo-symlink
        (                                                                                                        // opt/repo-symlink
            PARSE_RULE_OPTIONAL_GROUP                                                                            // opt/repo-symlink
            (                                                                                                    // opt/repo-symlink
                PARSE_RULE_OPTIONAL_DEPEND                                                                       // opt/repo-symlink
                (                                                                                                // opt/repo-symlink
                    PARSE_RULE_OPTIONAL_DEPEND_DEFAULT(PARSE_RULE_VAL_BOOL_FALSE),                               // opt/repo-symlink
                    PARSE_RULE_VAL_OPT(RepoType),                                                                // opt/repo-symlink
                    PARSE_RULE_VAL_STRID(posix),                                                                 // opt/repo-symlink
                ),                                                                                               // opt/repo-symlink
                                                                                                                 // opt/repo-symlink
                PARSE_RULE_OPTIONAL_DEFAULT                                                                      // opt/repo-symlink
                (                                                                                                // opt/repo-symlink
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                    // opt/repo-symlink
                ),                                                                                               // opt/repo-symlink
            ),                                                                                                   // opt/repo-symlink
        ),                                                                                                       // opt/repo-symlink
    ),                                                                                                           // opt/repo-symlink
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/repo-target-time
    (                                                                                                        // opt/repo-target-time
        PARSE_RULE_OPTION_NAME("repo-target-time"),                                                          // opt/repo-target-time
        PARSE_RULE_OPTION_TYPE(String),                                                                      // opt/repo-target-time
        PARSE_RULE_OPTION_RESET(true),                                                                       // opt/repo-target-time
        PARSE_RULE_OPTION_REQUIRED(false),                                                                   // opt/repo-target-time
        PARSE_RULE_OPTION_SECTION(Global),                                                                   // opt/repo-target-time
                                                                                                             // opt/repo-target-time
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/repo-target-time
        (                                                                                                    // opt/repo-target-time
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-target-time
            PARSE_RULE_OPTION_COMMAND(Info)                                                                  // opt/repo-target-time
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                              // opt/repo-target-time
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                               // opt/repo-target-time
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                // opt/repo-target-time
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-target-time
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-target-time
        ),                                                                                                   // opt/repo-target-time
                                                                                                             // opt/repo-target-time
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                      // opt/repo-target-time
        (                                                                                                    // opt/repo-target-time
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-target-time
        ),                                                                                                   // opt/repo-target-time
                                                                                                             // opt/repo-target-time
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                      // opt/repo-target-time
        (                                                                                                    // opt/repo-target-time
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                            // opt/repo-target-time
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/repo-target-time
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                // opt/repo-target-time
        ),                                                                                                   // opt/repo-target-time
    ),                                                                                                       // opt/repo-target-time
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                               // opt/repo-type
    (                                                                                                               // opt/repo-type
        PARSE_RULE_OPTION_NAME("repo-type"),                                                                        // opt/repo-type
        PARSE_RULE_OPTION_TYPE(StringId),                                                                           // opt/repo-type
        PARSE_RULE_OPTION_RESET(true),                                                                              // opt/repo-type
        PARSE_RULE_OPTION_REQUIRED(true),                                                                           // opt/repo-type
        PARSE_RULE_OPTION_SECTION(Global),                                                                          // opt/repo-type
        PARSE_RULE_OPTION_GROUP_ID(Repo),                                                                           // opt/repo-type
        PARSE_RULE_OPTION_DEPRECATE_MATCH(true),                                                                    // opt/repo-type
                                                                                                                    // opt/repo-type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                              // opt/repo-type
        (                                                                                                           // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                     // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                                        // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                       // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                                         // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                     // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                      // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                       // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                      // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                       // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                 // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                 // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/repo-type
        ),                                                                                                          // opt/repo-type
                                                                                                                    // opt/repo-type
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                             // opt/repo-type
        (                                                                                                           // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/repo-type
        ),                                                                                                          // opt/repo-type
                                                                                                                    // opt/repo-type
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                             // opt/repo-type
        (                                                                                                           // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/repo-type
        ),                                                                                                          // opt/repo-type
                                                                                                                    // opt/repo-type
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                            // opt/repo-type
        (                                                                                                           // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                     // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Check)                                                                        // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                       // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Info)                                                                         // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                     // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                      // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                       // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                      // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                       // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                 // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                 // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                // opt/repo-type
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/repo-type
        ),                                                                                                          // opt/repo-type
                                                                                                                    // opt/repo-type
        PARSE_RULE_OPTIONAL                                                                                         // opt/repo-type
        (                                                                                                           // opt/repo-type
            PARSE_RULE_OPTIONAL_GROUP                                                                               // opt/repo-type
            (                                                                                                       // opt/repo-type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                      // opt/repo-type
                (                                                                                                   // opt/repo-type
                    PARSE_RULE_VAL_STRID(azure),                                                                    // opt/repo-type
                    PARSE_RULE_VAL_STRID(cifs),                                                                     // opt/repo-type
                    PARSE_RULE_VAL_STRID(gcs),                                                                      // opt/repo-type
                    PARSE_RULE_VAL_STRID(posix),                                                                    // opt/repo-type
                    PARSE_RULE_VAL_STRID(s3),                                                                       // opt/repo-type
                    PARSE_RULE_VAL_STRID(sftp),                                                                     // opt/repo-type
#ifndef HAVE_LIBSSH2                                                                                                // opt/repo-type
                        PARSE_RULE_BOOL_FALSE,                                                                      // opt/repo-type
#endif                                                                                                              // opt/repo-type
                ),                                                                                                  // opt/repo-type
                                                                                                                    // opt/repo-type
                PARSE_RULE_OPTIONAL_DEFAULT                                                                         // opt/repo-type
                (                                                                                                   // opt/repo-type
                    PARSE_RULE_VAL_STRID(posix),                                                                    // opt/repo-type
                ),                                                                                                  // opt/repo-type
            ),                                                                                                      // opt/repo-type
        ),                                                                                                          // opt/repo-type
    ),                                                                                                              // opt/repo-type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                  // opt/report
    (                                                                                                                  // opt/report
        PARSE_RULE_OPTION_NAME("report"),                                                                              // opt/report
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                               // opt/report
        PARSE_RULE_OPTION_INTERNAL(true)                                                                               // opt/report
        PARSE_RULE_OPTION_REQUIRED(true),                                                                              // opt/report
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                        // opt/report
                                                                                                                       // opt/report
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                 // opt/report
        (                                                                                                              // opt/report
            PARSE_RULE_OPTION_COMMAND(Check)                                                                           // opt/report
        ),                                                                                                             // opt/report
                                                                                                                       // opt/report
        PARSE_RULE_OPTIONAL                                                                                            // opt/report
        (                                                                                                              // opt/report
            PARSE_RULE_OPTIONAL_GROUP                                                                                  // opt/report
            (                                                                                                          // opt/report
                PARSE_RULE_OPTIONAL_DEFAULT                                                                            // opt/report
                (                                                                                                      // opt/report
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                         // opt/report
                ),                                                                                                     // opt/report
            ),                                                                                                         // opt/report
        ),                                                                                                             // opt/report
    ),                                                                                                                 // opt/report
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                  // opt/resume
    (                                                                                                                  // opt/resume
        PARSE_RULE_OPTION_NAME("resume"),                                                                              // opt/resume
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                               // opt/resume
        PARSE_RULE_OPTION_NEGATE(true),                                                                                // opt/resume
        PARSE_RULE_OPTION_RESET(true),                                                                                 // opt/resume
        PARSE_RULE_OPTION_REQUIRED(true),                                                                              // opt/resume
        PARSE_RULE_OPTION_SECTION(Global),                                                                             // opt/resume
                                                                                                                       // opt/resume
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                 // opt/resume
        (                                                                                                              // opt/resume
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                          // opt/resume
        ),                                                                                                             // opt/resume
                                                                                                                       // opt/resume
        PARSE_RULE_OPTIONAL                                                                                            // opt/resume
        (                                                                                                              // opt/resume
            PARSE_RULE_OPTIONAL_GROUP                                                                                  // opt/resume
            (                                                                                                          // opt/resume
                PARSE_RULE_OPTIONAL_DEFAULT                                                                            // opt/resume
                (                                                                                                      // opt/resume
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                          // opt/resume
                ),                                                                                                     // opt/resume
            ),                                                                                                         // opt/resume
        ),                                                                                                             // opt/resume
    ),                                                                                                                 // opt/resume
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                               // opt/sck-block
    (                                                                                                               // opt/sck-block
        PARSE_RULE_OPTION_NAME("sck-block"),                                                                        // opt/sck-block
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                            // opt/sck-block
        PARSE_RULE_OPTION_INTERNAL(true)                                                                            // opt/sck-block
        PARSE_RULE_OPTION_NEGATE(true),                                                                             // opt/sck-block
        PARSE_RULE_OPTION_RESET(true),                                                                              // opt/sck-block
        PARSE_RULE_OPTION_REQUIRED(true),                                                                           // opt/sck-block
        PARSE_RULE_OPTION_SECTION(Global),                                                                          // opt/sck-block
                                                                                                                    // opt/sck-block
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                              // opt/sck-block
        (                                                                                                           // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                     // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Check)                                                                        // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                       // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Info)                                                                         // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                     // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                      // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                       // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                      // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                       // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Server)                                                                       // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                                   // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                 // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                 // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/sck-block
        ),                                                                                                          // opt/sck-block
                                                                                                                    // opt/sck-block
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                             // opt/sck-block
        (                                                                                                           // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/sck-block
        ),                                                                                                          // opt/sck-block
                                                                                                                    // opt/sck-block
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                             // opt/sck-block
        (                                                                                                           // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/sck-block
        ),                                                                                                          // opt/sck-block
                                                                                                                    // opt/sck-block
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                            // opt/sck-block
        (                                                                                                           // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                     // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                   // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                  // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Check)                                                                        // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                       // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Info)                                                                         // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                     // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                      // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                       // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                      // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                       // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                      // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                 // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                 // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                // opt/sck-block
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                       // opt/sck-block
        ),                                                                                                          // opt/sck-block
                                                                                                                    // opt/sck-block
        PARSE_RULE_OPTIONAL                                                                                         // opt/sck-block
        (                                                                                                           // opt/sck-block
            PARSE_RULE_OPTIONAL_GROUP                                                                               // opt/sck-block
            (                                                                                                       // opt/sck-block
                PARSE_RULE_OPTIONAL_DEFAULT                                                                         // opt/sck-block
                (                                                                                                   // opt/sck-block
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                      // opt/sck-block
                ),                                                                                                  // opt/sck-block
            ),                                                                                                      // opt/sck-block
        ),                                                                                                          // opt/sck-block
    ),                                                                                                              // opt/sck-block
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/sck-keep-alive
    (                                                                                                          // opt/sck-keep-alive
        PARSE_RULE_OPTION_NAME("sck-keep-alive"),                                                              // opt/sck-keep-alive
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                       // opt/sck-keep-alive
        PARSE_RULE_OPTION_NEGATE(true),                                                                        // opt/sck-keep-alive
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/sck-keep-alive
        PARSE_RULE_OPTION_REQUIRED(true),                                                                      // opt/sck-keep-alive
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/sck-keep-alive
                                                                                                               // opt/sck-keep-alive
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/sck-keep-alive
        (                                                                                                      // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Server)                                                                  // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                              // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/sck-keep-alive
        ),                                                                                                     // opt/sck-keep-alive
                                                                                                               // opt/sck-keep-alive
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                        // opt/sck-keep-alive
        (                                                                                                      // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/sck-keep-alive
        ),                                                                                                     // opt/sck-keep-alive
                                                                                                               // opt/sck-keep-alive
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                        // opt/sck-keep-alive
        (                                                                                                      // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/sck-keep-alive
        ),                                                                                                     // opt/sck-keep-alive
                                                                                                               // opt/sck-keep-alive
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                       // opt/sck-keep-alive
        (                                                                                                      // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                              // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                             // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                  // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Check)                                                                   // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                  // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Info)                                                                    // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                 // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                  // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                 // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                  // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                            // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                            // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                           // opt/sck-keep-alive
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                  // opt/sck-keep-alive
        ),                                                                                                     // opt/sck-keep-alive
                                                                                                               // opt/sck-keep-alive
        PARSE_RULE_OPTIONAL                                                                                    // opt/sck-keep-alive
        (                                                                                                      // opt/sck-keep-alive
            PARSE_RULE_OPTIONAL_GROUP                                                                          // opt/sck-keep-alive
            (                                                                                                  // opt/sck-keep-alive
                PARSE_RULE_OPTIONAL_DEFAULT                                                                    // opt/sck-keep-alive
                (                                                                                              // opt/sck-keep-alive
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                  // opt/sck-keep-alive
                ),                                                                                             // opt/sck-keep-alive
            ),                                                                                                 // opt/sck-keep-alive
        ),                                                                                                     // opt/sck-keep-alive
    ),                                                                                                         // opt/sck-keep-alive
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                     // opt/set
    (                                                                                                                     // opt/set
        PARSE_RULE_OPTION_NAME("set"),                                                                                    // opt/set
        PARSE_RULE_OPTION_TYPE(String),                                                                                   // opt/set
        PARSE_RULE_OPTION_REQUIRED(false),                                                                                // opt/set
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                           // opt/set
                                                                                                                          // opt/set
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                    // opt/set
        (                                                                                                                 // opt/set
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                           // opt/set
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                             // opt/set
            PARSE_RULE_OPTION_COMMAND(Info)                                                                               // opt/set
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                           // opt/set
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                            // opt/set
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                             // opt/set
        ),                                                                                                                // opt/set
                                                                                                                          // opt/set
        PARSE_RULE_OPTIONAL                                                                                               // opt/set
        (                                                                                                                 // opt/set
            PARSE_RULE_OPTIONAL_GROUP                                                                                     // opt/set
            (                                                                                                             // opt/set
                PARSE_RULE_FILTER_CMD                                                                                     // opt/set
                (                                                                                                         // opt/set
                    PARSE_RULE_VAL_CMD(Annotate),                                                                         // opt/set
                ),                                                                                                        // opt/set
                                                                                                                          // opt/set
                PARSE_RULE_OPTIONAL_REQUIRED(),                                                                           // opt/set
            ),                                                                                                            // opt/set
                                                                                                                          // opt/set
            PARSE_RULE_OPTIONAL_GROUP                                                                                     // opt/set
            (                                                                                                             // opt/set
                PARSE_RULE_FILTER_CMD                                                                                     // opt/set
                (                                                                                                         // opt/set
                    PARSE_RULE_VAL_CMD(Info),                                                                             // opt/set
                ),                                                                                                        // opt/set
                                                                                                                          // opt/set
                PARSE_RULE_OPTIONAL_DEPEND                                                                                // opt/set
                (                                                                                                         // opt/set
                    PARSE_RULE_VAL_OPT(Stanza),                                                                           // opt/set
                ),                                                                                                        // opt/set
            ),                                                                                                            // opt/set
                                                                                                                          // opt/set
            PARSE_RULE_OPTIONAL_GROUP                                                                                     // opt/set
            (                                                                                                             // opt/set
                PARSE_RULE_FILTER_CMD                                                                                     // opt/set
                (                                                                                                         // opt/set
                    PARSE_RULE_VAL_CMD(Manifest),                                                                         // opt/set
                    PARSE_RULE_VAL_CMD(Restore),                                                                          // opt/set
                ),                                                                                                        // opt/set
                                                                                                                          // opt/set
                PARSE_RULE_OPTIONAL_DEFAULT                                                                               // opt/set
                (                                                                                                         // opt/set
                    PARSE_RULE_VAL_STR(QT_latest_QT),                                                                     // opt/set
                ),                                                                                                        // opt/set
                                                                                                                          // opt/set
                PARSE_RULE_OPTIONAL_REQUIRED(),                                                                           // opt/set
            ),                                                                                                            // opt/set
        ),                                                                                                                // opt/set
    ),                                                                                                                    // opt/set
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                    // opt/sort
    (                                                                                                                    // opt/sort
        PARSE_RULE_OPTION_NAME("sort"),                                                                                  // opt/sort
        PARSE_RULE_OPTION_TYPE(StringId),                                                                                // opt/sort
        PARSE_RULE_OPTION_REQUIRED(true),                                                                                // opt/sort
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                          // opt/sort
                                                                                                                         // opt/sort
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                   // opt/sort
        (                                                                                                                // opt/sort
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                            // opt/sort
        ),                                                                                                               // opt/sort
                                                                                                                         // opt/sort
        PARSE_RULE_OPTIONAL                                                                                              // opt/sort
        (                                                                                                                // opt/sort
            PARSE_RULE_OPTIONAL_GROUP                                                                                    // opt/sort
            (                                                                                                            // opt/sort
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                           // opt/sort
                (                                                                                                        // opt/sort
                    PARSE_RULE_VAL_STRID(none),                                                                          // opt/sort
                    PARSE_RULE_VAL_STRID(asc),                                                                           // opt/sort
                    PARSE_RULE_VAL_STRID(desc),                                                                          // opt/sort
                ),                                                                                                       // opt/sort
                                                                                                                         // opt/sort
                PARSE_RULE_OPTIONAL_DEFAULT                                                                              // opt/sort
                (                                                                                                        // opt/sort
                    PARSE_RULE_VAL_STRID(asc),                                                                           // opt/sort
                ),                                                                                                       // opt/sort
            ),                                                                                                           // opt/sort
        ),                                                                                                               // opt/sort
    ),                                                                                                                   // opt/sort
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                              // opt/spool-path
    (                                                                                                              // opt/spool-path
        PARSE_RULE_OPTION_NAME("spool-path"),                                                                      // opt/spool-path
        PARSE_RULE_OPTION_TYPE(Path),                                                                              // opt/spool-path
        PARSE_RULE_OPTION_RESET(true),                                                                             // opt/spool-path
        PARSE_RULE_OPTION_REQUIRED(true),                                                                          // opt/spool-path
        PARSE_RULE_OPTION_SECTION(Global),                                                                         // opt/spool-path
                                                                                                                   // opt/spool-path
        PARSE_RULE_OPTION_COMMAND_INTERNAL_LIST                                                                    // opt/spool-path
        (                                                                                                          // opt/spool-path
            PARSE_RULE_OPTION_COMMAND_INTERNAL(Restore, true),                                                     // opt/spool-path
        )                                                                                                          // opt/spool-path
                                                                                                                   // opt/spool-path
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                             // opt/spool-path
        (                                                                                                          // opt/spool-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                  // opt/spool-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                 // opt/spool-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                     // opt/spool-path
        ),                                                                                                         // opt/spool-path
                                                                                                                   // opt/spool-path
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                            // opt/spool-path
        (                                                                                                          // opt/spool-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                  // opt/spool-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                 // opt/spool-path
        ),                                                                                                         // opt/spool-path
                                                                                                                   // opt/spool-path
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                            // opt/spool-path
        (                                                                                                          // opt/spool-path
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                  // opt/spool-path
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                 // opt/spool-path
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                     // opt/spool-path
        ),                                                                                                         // opt/spool-path
                                                                                                                   // opt/spool-path
        PARSE_RULE_OPTIONAL                                                                                        // opt/spool-path
        (                                                                                                          // opt/spool-path
            PARSE_RULE_OPTIONAL_GROUP                                                                              // opt/spool-path
            (                                                                                                      // opt/spool-path
                PARSE_RULE_FILTER_CMD                                                                              // opt/spool-path
                (                                                                                                  // opt/spool-path
                    PARSE_RULE_VAL_CMD(ArchiveGet),                                                                // opt/spool-path
                    PARSE_RULE_VAL_CMD(ArchivePush),                                                               // opt/spool-path
                ),                                                                                                 // opt/spool-path
                                                                                                                   // opt/spool-path
                PARSE_RULE_OPTIONAL_DEPEND                                                                         // opt/spool-path
                (                                                                                                  // opt/spool-path
                    PARSE_RULE_VAL_OPT(ArchiveAsync),                                                              // opt/spool-path
                    PARSE_RULE_VAL_BOOL_TRUE,                                                                      // opt/spool-path
                ),                                                                                                 // opt/spool-path
                                                                                                                   // opt/spool-path
                PARSE_RULE_OPTIONAL_DEFAULT                                                                        // opt/spool-path
                (                                                                                                  // opt/spool-path
                    PARSE_RULE_VAL_STR(QT_FS_var_FS_spool_FS_pgbackrest_QT),                                       // opt/spool-path
                ),                                                                                                 // opt/spool-path
            ),                                                                                                     // opt/spool-path
                                                                                                                   // opt/spool-path
            PARSE_RULE_OPTIONAL_GROUP                                                                              // opt/spool-path
            (                                                                                                      // opt/spool-path
                PARSE_RULE_OPTIONAL_DEFAULT                                                                        // opt/spool-path
                (                                                                                                  // opt/spool-path
                    PARSE_RULE_VAL_STR(QT_FS_var_FS_spool_FS_pgbackrest_QT),                                       // opt/spool-path
                ),                                                                                                 // opt/spool-path
            ),                                                                                                     // opt/spool-path
        ),                                                                                                         // opt/spool-path
    ),                                                                                                             // opt/spool-path
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                  // opt/stanza
    (                                                                                                                  // opt/stanza
        PARSE_RULE_OPTION_NAME("stanza"),                                                                              // opt/stanza
        PARSE_RULE_OPTION_TYPE(String),                                                                                // opt/stanza
        PARSE_RULE_OPTION_REQUIRED(true),                                                                              // opt/stanza
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                        // opt/stanza
                                                                                                                       // opt/stanza
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                 // opt/stanza
        (                                                                                                              // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                        // opt/stanza
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                      // opt/stanza
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                     // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                          // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Check)                                                                           // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                          // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Info)                                                                            // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                        // opt/stanza
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                         // opt/stanza
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                          // opt/stanza
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                         // opt/stanza
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                          // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                         // opt/stanza
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                    // opt/stanza
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                    // opt/stanza
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                   // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Start)                                                                           // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Stop)                                                                            // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                          // opt/stanza
        ),                                                                                                             // opt/stanza
                                                                                                                       // opt/stanza
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                                // opt/stanza
        (                                                                                                              // opt/stanza
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                      // opt/stanza
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                     // opt/stanza
        ),                                                                                                             // opt/stanza
                                                                                                                       // opt/stanza
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                                // opt/stanza
        (                                                                                                              // opt/stanza
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                      // opt/stanza
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                     // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                          // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                         // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                          // opt/stanza
        ),                                                                                                             // opt/stanza
                                                                                                                       // opt/stanza
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                               // opt/stanza
        (                                                                                                              // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                        // opt/stanza
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                                      // opt/stanza
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                                     // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                          // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Check)                                                                           // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                          // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Info)                                                                            // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                        // opt/stanza
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                         // opt/stanza
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                          // opt/stanza
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                         // opt/stanza
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                          // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                         // opt/stanza
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                                    // opt/stanza
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                                    // opt/stanza
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                                   // opt/stanza
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                          // opt/stanza
        ),                                                                                                             // opt/stanza
                                                                                                                       // opt/stanza
        PARSE_RULE_OPTIONAL                                                                                            // opt/stanza
        (                                                                                                              // opt/stanza
            PARSE_RULE_OPTIONAL_GROUP                                                                                  // opt/stanza
            (                                                                                                          // opt/stanza
                PARSE_RULE_FILTER_CMD                                                                                  // opt/stanza
                (                                                                                                      // opt/stanza
                    PARSE_RULE_VAL_CMD(Check),                                                                         // opt/stanza
                    PARSE_RULE_VAL_CMD(Info),                                                                          // opt/stanza
                    PARSE_RULE_VAL_CMD(RepoGet),                                                                       // opt/stanza
                    PARSE_RULE_VAL_CMD(RepoLs),                                                                        // opt/stanza
                    PARSE_RULE_VAL_CMD(RepoPut),                                                                       // opt/stanza
                    PARSE_RULE_VAL_CMD(RepoRm),                                                                        // opt/stanza
                    PARSE_RULE_VAL_CMD(Start),                                                                         // opt/stanza
                    PARSE_RULE_VAL_CMD(Stop),                                                                          // opt/stanza
                ),                                                                                                     // opt/stanza
                                                                                                                       // opt/stanza
                PARSE_RULE_OPTIONAL_NOT_REQUIRED(),                                                                    // opt/stanza
            ),                                                                                                         // opt/stanza
        ),                                                                                                             // opt/stanza
    ),                                                                                                                 // opt/stanza
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                              // opt/start-fast
    (                                                                                                              // opt/start-fast
        PARSE_RULE_OPTION_NAME("start-fast"),                                                                      // opt/start-fast
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                           // opt/start-fast
        PARSE_RULE_OPTION_NEGATE(true),                                                                            // opt/start-fast
        PARSE_RULE_OPTION_RESET(true),                                                                             // opt/start-fast
        PARSE_RULE_OPTION_REQUIRED(true),                                                                          // opt/start-fast
        PARSE_RULE_OPTION_SECTION(Global),                                                                         // opt/start-fast
                                                                                                                   // opt/start-fast
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                             // opt/start-fast
        (                                                                                                          // opt/start-fast
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                      // opt/start-fast
        ),                                                                                                         // opt/start-fast
                                                                                                                   // opt/start-fast
        PARSE_RULE_OPTIONAL                                                                                        // opt/start-fast
        (                                                                                                          // opt/start-fast
            PARSE_RULE_OPTIONAL_GROUP                                                                              // opt/start-fast
            (                                                                                                      // opt/start-fast
                PARSE_RULE_OPTIONAL_DEFAULT                                                                        // opt/start-fast
                (                                                                                                  // opt/start-fast
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                     // opt/start-fast
                ),                                                                                                 // opt/start-fast
            ),                                                                                                     // opt/start-fast
        ),                                                                                                         // opt/start-fast
    ),                                                                                                             // opt/start-fast
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                               // opt/stop-auto
    (                                                                                                               // opt/stop-auto
        PARSE_RULE_OPTION_NAME("stop-auto"),                                                                        // opt/stop-auto
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                            // opt/stop-auto
        PARSE_RULE_OPTION_INTERNAL(true)                                                                            // opt/stop-auto
        PARSE_RULE_OPTION_NEGATE(true),                                                                             // opt/stop-auto
        PARSE_RULE_OPTION_RESET(true),                                                                              // opt/stop-auto
        PARSE_RULE_OPTION_REQUIRED(true),                                                                           // opt/stop-auto
        PARSE_RULE_OPTION_SECTION(Global),                                                                          // opt/stop-auto
                                                                                                                    // opt/stop-auto
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                              // opt/stop-auto
        (                                                                                                           // opt/stop-auto
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                       // opt/stop-auto
        ),                                                                                                          // opt/stop-auto
                                                                                                                    // opt/stop-auto
        PARSE_RULE_OPTIONAL                                                                                         // opt/stop-auto
        (                                                                                                           // opt/stop-auto
            PARSE_RULE_OPTIONAL_GROUP                                                                               // opt/stop-auto
            (                                                                                                       // opt/stop-auto
                PARSE_RULE_OPTIONAL_DEFAULT                                                                         // opt/stop-auto
                (                                                                                                   // opt/stop-auto
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                      // opt/stop-auto
                ),                                                                                                  // opt/stop-auto
            ),                                                                                                      // opt/stop-auto
        ),                                                                                                          // opt/stop-auto
    ),                                                                                                              // opt/stop-auto
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                          // opt/tablespace-map
    (                                                                                                          // opt/tablespace-map
        PARSE_RULE_OPTION_NAME("tablespace-map"),                                                              // opt/tablespace-map
        PARSE_RULE_OPTION_TYPE(Hash),                                                                          // opt/tablespace-map
        PARSE_RULE_OPTION_RESET(true),                                                                         // opt/tablespace-map
        PARSE_RULE_OPTION_REQUIRED(false),                                                                     // opt/tablespace-map
        PARSE_RULE_OPTION_SECTION(Global),                                                                     // opt/tablespace-map
        PARSE_RULE_OPTION_MULTI(true),                                                                         // opt/tablespace-map
                                                                                                               // opt/tablespace-map
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                         // opt/tablespace-map
        (                                                                                                      // opt/tablespace-map
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                 // opt/tablespace-map
        ),                                                                                                     // opt/tablespace-map
    ),                                                                                                         // opt/tablespace-map
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                      // opt/tablespace-map-all
    (                                                                                                      // opt/tablespace-map-all
        PARSE_RULE_OPTION_NAME("tablespace-map-all"),                                                      // opt/tablespace-map-all
        PARSE_RULE_OPTION_TYPE(String),                                                                    // opt/tablespace-map-all
        PARSE_RULE_OPTION_RESET(true),                                                                     // opt/tablespace-map-all
        PARSE_RULE_OPTION_REQUIRED(false),                                                                 // opt/tablespace-map-all
        PARSE_RULE_OPTION_SECTION(Global),                                                                 // opt/tablespace-map-all
                                                                                                           // opt/tablespace-map-all
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                     // opt/tablespace-map-all
        (                                                                                                  // opt/tablespace-map-all
            PARSE_RULE_OPTION_COMMAND(Restore)                                                             // opt/tablespace-map-all
        ),                                                                                                 // opt/tablespace-map-all
    ),                                                                                                     // opt/tablespace-map-all
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                  // opt/target
    (                                                                                                                  // opt/target
        PARSE_RULE_OPTION_NAME("target"),                                                                              // opt/target
        PARSE_RULE_OPTION_TYPE(String),                                                                                // opt/target
        PARSE_RULE_OPTION_REQUIRED(true),                                                                              // opt/target
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                        // opt/target
                                                                                                                       // opt/target
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                 // opt/target
        (                                                                                                              // opt/target
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                         // opt/target
        ),                                                                                                             // opt/target
                                                                                                                       // opt/target
        PARSE_RULE_OPTIONAL                                                                                            // opt/target
        (                                                                                                              // opt/target
            PARSE_RULE_OPTIONAL_GROUP                                                                                  // opt/target
            (                                                                                                          // opt/target
                PARSE_RULE_OPTIONAL_DEPEND                                                                             // opt/target
                (                                                                                                      // opt/target
                    PARSE_RULE_VAL_OPT(Type),                                                                          // opt/target
                    PARSE_RULE_VAL_STRID(lsn),                                                                         // opt/target
                    PARSE_RULE_VAL_STRID(name),                                                                        // opt/target
                    PARSE_RULE_VAL_STRID(time),                                                                        // opt/target
                    PARSE_RULE_VAL_STRID(xid),                                                                         // opt/target
                ),                                                                                                     // opt/target
            ),                                                                                                         // opt/target
        ),                                                                                                             // opt/target
    ),                                                                                                                 // opt/target
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                           // opt/target-action
    (                                                                                                           // opt/target-action
        PARSE_RULE_OPTION_NAME("target-action"),                                                                // opt/target-action
        PARSE_RULE_OPTION_TYPE(StringId),                                                                       // opt/target-action
        PARSE_RULE_OPTION_REQUIRED(true),                                                                       // opt/target-action
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                 // opt/target-action
                                                                                                                // opt/target-action
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                          // opt/target-action
        (                                                                                                       // opt/target-action
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/target-action
        ),                                                                                                      // opt/target-action
                                                                                                                // opt/target-action
        PARSE_RULE_OPTIONAL                                                                                     // opt/target-action
        (                                                                                                       // opt/target-action
            PARSE_RULE_OPTIONAL_GROUP                                                                           // opt/target-action
            (                                                                                                   // opt/target-action
                PARSE_RULE_OPTIONAL_DEPEND                                                                      // opt/target-action
                (                                                                                               // opt/target-action
                    PARSE_RULE_VAL_OPT(Type),                                                                   // opt/target-action
                    PARSE_RULE_VAL_STRID(immediate),                                                            // opt/target-action
                    PARSE_RULE_VAL_STRID(lsn),                                                                  // opt/target-action
                    PARSE_RULE_VAL_STRID(name),                                                                 // opt/target-action
                    PARSE_RULE_VAL_STRID(time),                                                                 // opt/target-action
                    PARSE_RULE_VAL_STRID(xid),                                                                  // opt/target-action
                ),                                                                                              // opt/target-action
                                                                                                                // opt/target-action
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                  // opt/target-action
                (                                                                                               // opt/target-action
                    PARSE_RULE_VAL_STRID(pause),                                                                // opt/target-action
                    PARSE_RULE_VAL_STRID(promote),                                                              // opt/target-action
                    PARSE_RULE_VAL_STRID(shutdown),                                                             // opt/target-action
                ),                                                                                              // opt/target-action
                                                                                                                // opt/target-action
                PARSE_RULE_OPTIONAL_DEFAULT                                                                     // opt/target-action
                (                                                                                               // opt/target-action
                    PARSE_RULE_VAL_STRID(pause),                                                                // opt/target-action
                ),                                                                                              // opt/target-action
            ),                                                                                                  // opt/target-action
        ),                                                                                                      // opt/target-action
    ),                                                                                                          // opt/target-action
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                        // opt/target-exclusive
    (                                                                                                        // opt/target-exclusive
        PARSE_RULE_OPTION_NAME("target-exclusive"),                                                          // opt/target-exclusive
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                     // opt/target-exclusive
        PARSE_RULE_OPTION_REQUIRED(true),                                                                    // opt/target-exclusive
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                              // opt/target-exclusive
                                                                                                             // opt/target-exclusive
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                       // opt/target-exclusive
        (                                                                                                    // opt/target-exclusive
            PARSE_RULE_OPTION_COMMAND(Restore)                                                               // opt/target-exclusive
        ),                                                                                                   // opt/target-exclusive
                                                                                                             // opt/target-exclusive
        PARSE_RULE_OPTIONAL                                                                                  // opt/target-exclusive
        (                                                                                                    // opt/target-exclusive
            PARSE_RULE_OPTIONAL_GROUP                                                                        // opt/target-exclusive
            (                                                                                                // opt/target-exclusive
                PARSE_RULE_OPTIONAL_DEPEND                                                                   // opt/target-exclusive
                (                                                                                            // opt/target-exclusive
                    PARSE_RULE_VAL_OPT(Type),                                                                // opt/target-exclusive
                    PARSE_RULE_VAL_STRID(lsn),                                                               // opt/target-exclusive
                    PARSE_RULE_VAL_STRID(time),                                                              // opt/target-exclusive
                    PARSE_RULE_VAL_STRID(xid),                                                               // opt/target-exclusive
                ),                                                                                           // opt/target-exclusive
                                                                                                             // opt/target-exclusive
                PARSE_RULE_OPTIONAL_DEFAULT                                                                  // opt/target-exclusive
                (                                                                                            // opt/target-exclusive
                    PARSE_RULE_VAL_BOOL_FALSE,                                                               // opt/target-exclusive
                ),                                                                                           // opt/target-exclusive
            ),                                                                                               // opt/target-exclusive
        ),                                                                                                   // opt/target-exclusive
    ),                                                                                                       // opt/target-exclusive
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                         // opt/target-timeline
    (                                                                                                         // opt/target-timeline
        PARSE_RULE_OPTION_NAME("target-timeline"),                                                            // opt/target-timeline
        PARSE_RULE_OPTION_TYPE(String),                                                                       // opt/target-timeline
        PARSE_RULE_OPTION_REQUIRED(false),                                                                    // opt/target-timeline
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                               // opt/target-timeline
                                                                                                              // opt/target-timeline
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                        // opt/target-timeline
        (                                                                                                     // opt/target-timeline
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                // opt/target-timeline
        ),                                                                                                    // opt/target-timeline
                                                                                                              // opt/target-timeline
        PARSE_RULE_OPTIONAL                                                                                   // opt/target-timeline
        (                                                                                                     // opt/target-timeline
            PARSE_RULE_OPTIONAL_GROUP                                                                         // opt/target-timeline
            (                                                                                                 // opt/target-timeline
                PARSE_RULE_OPTIONAL_DEPEND                                                                    // opt/target-timeline
                (                                                                                             // opt/target-timeline
                    PARSE_RULE_VAL_OPT(Type),                                                                 // opt/target-timeline
                    PARSE_RULE_VAL_STRID(default),                                                            // opt/target-timeline
                    PARSE_RULE_VAL_STRID(lsn),                                                                // opt/target-timeline
                    PARSE_RULE_VAL_STRID(name),                                                               // opt/target-timeline
                    PARSE_RULE_VAL_STRID(standby),                                                            // opt/target-timeline
                    PARSE_RULE_VAL_STRID(time),                                                               // opt/target-timeline
                    PARSE_RULE_VAL_STRID(xid),                                                                // opt/target-timeline
                ),                                                                                            // opt/target-timeline
            ),                                                                                                // opt/target-timeline
        ),                                                                                                    // opt/target-timeline
    ),                                                                                                        // opt/target-timeline
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                    // opt/tcp-keep-alive-count
    (                                                                                                    // opt/tcp-keep-alive-count
        PARSE_RULE_OPTION_NAME("tcp-keep-alive-count"),                                                  // opt/tcp-keep-alive-count
        PARSE_RULE_OPTION_TYPE(Integer),                                                                 // opt/tcp-keep-alive-count
        PARSE_RULE_OPTION_RESET(true),                                                                   // opt/tcp-keep-alive-count
        PARSE_RULE_OPTION_REQUIRED(false),                                                               // opt/tcp-keep-alive-count
        PARSE_RULE_OPTION_SECTION(Global),                                                               // opt/tcp-keep-alive-count
                                                                                                         // opt/tcp-keep-alive-count
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                   // opt/tcp-keep-alive-count
        (                                                                                                // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                          // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Check)                                                             // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Expire)                                                            // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Info)                                                              // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                          // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                           // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                            // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                           // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                            // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Server)                                                            // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                        // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                      // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                      // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                     // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/tcp-keep-alive-count
        ),                                                                                               // opt/tcp-keep-alive-count
                                                                                                         // opt/tcp-keep-alive-count
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                  // opt/tcp-keep-alive-count
        (                                                                                                // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/tcp-keep-alive-count
        ),                                                                                               // opt/tcp-keep-alive-count
                                                                                                         // opt/tcp-keep-alive-count
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                  // opt/tcp-keep-alive-count
        (                                                                                                // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/tcp-keep-alive-count
        ),                                                                                               // opt/tcp-keep-alive-count
                                                                                                         // opt/tcp-keep-alive-count
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                 // opt/tcp-keep-alive-count
        (                                                                                                // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                          // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                        // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                       // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Backup)                                                            // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Check)                                                             // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Expire)                                                            // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Info)                                                              // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                          // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                           // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                            // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                           // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                            // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Restore)                                                           // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                      // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                      // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                     // opt/tcp-keep-alive-count
            PARSE_RULE_OPTION_COMMAND(Verify)                                                            // opt/tcp-keep-alive-count
        ),                                                                                               // opt/tcp-keep-alive-count
                                                                                                         // opt/tcp-keep-alive-count
        PARSE_RULE_OPTIONAL                                                                              // opt/tcp-keep-alive-count
        (                                                                                                // opt/tcp-keep-alive-count
            PARSE_RULE_OPTIONAL_GROUP                                                                    // opt/tcp-keep-alive-count
            (                                                                                            // opt/tcp-keep-alive-count
                PARSE_RULE_OPTIONAL_DEPEND                                                               // opt/tcp-keep-alive-count
                (                                                                                        // opt/tcp-keep-alive-count
                    PARSE_RULE_VAL_OPT(SckKeepAlive),                                                    // opt/tcp-keep-alive-count
                    PARSE_RULE_VAL_BOOL_TRUE,                                                            // opt/tcp-keep-alive-count
                ),                                                                                       // opt/tcp-keep-alive-count
                                                                                                         // opt/tcp-keep-alive-count
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                          // opt/tcp-keep-alive-count
                (                                                                                        // opt/tcp-keep-alive-count
                    PARSE_RULE_VAL_INT(1),                                                               // opt/tcp-keep-alive-count
                    PARSE_RULE_VAL_INT(32),                                                              // opt/tcp-keep-alive-count
                ),                                                                                       // opt/tcp-keep-alive-count
            ),                                                                                           // opt/tcp-keep-alive-count
        ),                                                                                               // opt/tcp-keep-alive-count
    ),                                                                                                   // opt/tcp-keep-alive-count
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                     // opt/tcp-keep-alive-idle
    (                                                                                                     // opt/tcp-keep-alive-idle
        PARSE_RULE_OPTION_NAME("tcp-keep-alive-idle"),                                                    // opt/tcp-keep-alive-idle
        PARSE_RULE_OPTION_TYPE(Integer),                                                                  // opt/tcp-keep-alive-idle
        PARSE_RULE_OPTION_RESET(true),                                                                    // opt/tcp-keep-alive-idle
        PARSE_RULE_OPTION_REQUIRED(false),                                                                // opt/tcp-keep-alive-idle
        PARSE_RULE_OPTION_SECTION(Global),                                                                // opt/tcp-keep-alive-idle
                                                                                                          // opt/tcp-keep-alive-idle
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                    // opt/tcp-keep-alive-idle
        (                                                                                                 // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Server)                                                             // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                         // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/tcp-keep-alive-idle
        ),                                                                                                // opt/tcp-keep-alive-idle
                                                                                                          // opt/tcp-keep-alive-idle
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                   // opt/tcp-keep-alive-idle
        (                                                                                                 // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/tcp-keep-alive-idle
        ),                                                                                                // opt/tcp-keep-alive-idle
                                                                                                          // opt/tcp-keep-alive-idle
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                   // opt/tcp-keep-alive-idle
        (                                                                                                 // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/tcp-keep-alive-idle
        ),                                                                                                // opt/tcp-keep-alive-idle
                                                                                                          // opt/tcp-keep-alive-idle
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                  // opt/tcp-keep-alive-idle
        (                                                                                                 // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                           // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                         // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                        // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Backup)                                                             // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Check)                                                              // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Expire)                                                             // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Info)                                                               // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                           // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                            // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                             // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                            // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                             // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Restore)                                                            // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                       // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                       // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                      // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTION_COMMAND(Verify)                                                             // opt/tcp-keep-alive-idle
        ),                                                                                                // opt/tcp-keep-alive-idle
                                                                                                          // opt/tcp-keep-alive-idle
        PARSE_RULE_OPTIONAL                                                                               // opt/tcp-keep-alive-idle
        (                                                                                                 // opt/tcp-keep-alive-idle
            PARSE_RULE_OPTIONAL_GROUP                                                                     // opt/tcp-keep-alive-idle
            (                                                                                             // opt/tcp-keep-alive-idle
                PARSE_RULE_OPTIONAL_DEPEND                                                                // opt/tcp-keep-alive-idle
                (                                                                                         // opt/tcp-keep-alive-idle
                    PARSE_RULE_VAL_OPT(SckKeepAlive),                                                     // opt/tcp-keep-alive-idle
                    PARSE_RULE_VAL_BOOL_TRUE,                                                             // opt/tcp-keep-alive-idle
                ),                                                                                        // opt/tcp-keep-alive-idle
                                                                                                          // opt/tcp-keep-alive-idle
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                           // opt/tcp-keep-alive-idle
                (                                                                                         // opt/tcp-keep-alive-idle
                    PARSE_RULE_VAL_INT(1),                                                                // opt/tcp-keep-alive-idle
                    PARSE_RULE_VAL_INT(3600),                                                             // opt/tcp-keep-alive-idle
                ),                                                                                        // opt/tcp-keep-alive-idle
            ),                                                                                            // opt/tcp-keep-alive-idle
        ),                                                                                                // opt/tcp-keep-alive-idle
    ),                                                                                                    // opt/tcp-keep-alive-idle
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                 // opt/tcp-keep-alive-interval
    (                                                                                                 // opt/tcp-keep-alive-interval
        PARSE_RULE_OPTION_NAME("tcp-keep-alive-interval"),                                            // opt/tcp-keep-alive-interval
        PARSE_RULE_OPTION_TYPE(Integer),                                                              // opt/tcp-keep-alive-interval
        PARSE_RULE_OPTION_RESET(true),                                                                // opt/tcp-keep-alive-interval
        PARSE_RULE_OPTION_REQUIRED(false),                                                            // opt/tcp-keep-alive-interval
        PARSE_RULE_OPTION_SECTION(Global),                                                            // opt/tcp-keep-alive-interval
                                                                                                      // opt/tcp-keep-alive-interval
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                // opt/tcp-keep-alive-interval
        (                                                                                             // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                       // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                     // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                    // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Backup)                                                         // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Check)                                                          // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Expire)                                                         // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Info)                                                           // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                       // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                        // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                         // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                        // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                         // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Restore)                                                        // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Server)                                                         // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                     // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                   // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                   // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                  // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Verify)                                                         // opt/tcp-keep-alive-interval
        ),                                                                                            // opt/tcp-keep-alive-interval
                                                                                                      // opt/tcp-keep-alive-interval
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                               // opt/tcp-keep-alive-interval
        (                                                                                             // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                     // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                    // opt/tcp-keep-alive-interval
        ),                                                                                            // opt/tcp-keep-alive-interval
                                                                                                      // opt/tcp-keep-alive-interval
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                               // opt/tcp-keep-alive-interval
        (                                                                                             // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                     // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                    // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Backup)                                                         // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Restore)                                                        // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Verify)                                                         // opt/tcp-keep-alive-interval
        ),                                                                                            // opt/tcp-keep-alive-interval
                                                                                                      // opt/tcp-keep-alive-interval
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                              // opt/tcp-keep-alive-interval
        (                                                                                             // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                       // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                     // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                    // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Backup)                                                         // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Check)                                                          // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Expire)                                                         // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Info)                                                           // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                       // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                        // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                         // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                        // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                         // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Restore)                                                        // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                   // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                   // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                  // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTION_COMMAND(Verify)                                                         // opt/tcp-keep-alive-interval
        ),                                                                                            // opt/tcp-keep-alive-interval
                                                                                                      // opt/tcp-keep-alive-interval
        PARSE_RULE_OPTIONAL                                                                           // opt/tcp-keep-alive-interval
        (                                                                                             // opt/tcp-keep-alive-interval
            PARSE_RULE_OPTIONAL_GROUP                                                                 // opt/tcp-keep-alive-interval
            (                                                                                         // opt/tcp-keep-alive-interval
                PARSE_RULE_OPTIONAL_DEPEND                                                            // opt/tcp-keep-alive-interval
                (                                                                                     // opt/tcp-keep-alive-interval
                    PARSE_RULE_VAL_OPT(SckKeepAlive),                                                 // opt/tcp-keep-alive-interval
                    PARSE_RULE_VAL_BOOL_TRUE,                                                         // opt/tcp-keep-alive-interval
                ),                                                                                    // opt/tcp-keep-alive-interval
                                                                                                      // opt/tcp-keep-alive-interval
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                       // opt/tcp-keep-alive-interval
                (                                                                                     // opt/tcp-keep-alive-interval
                    PARSE_RULE_VAL_INT(1),                                                            // opt/tcp-keep-alive-interval
                    PARSE_RULE_VAL_INT(900),                                                          // opt/tcp-keep-alive-interval
                ),                                                                                    // opt/tcp-keep-alive-interval
            ),                                                                                        // opt/tcp-keep-alive-interval
        ),                                                                                            // opt/tcp-keep-alive-interval
    ),                                                                                                // opt/tcp-keep-alive-interval
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                           // opt/tls-cipher-12
    (                                                                                                           // opt/tls-cipher-12
        PARSE_RULE_OPTION_NAME("tls-cipher-12"),                                                                // opt/tls-cipher-12
        PARSE_RULE_OPTION_TYPE(String),                                                                         // opt/tls-cipher-12
        PARSE_RULE_OPTION_RESET(true),                                                                          // opt/tls-cipher-12
        PARSE_RULE_OPTION_REQUIRED(false),                                                                      // opt/tls-cipher-12
        PARSE_RULE_OPTION_SECTION(Global),                                                                      // opt/tls-cipher-12
                                                                                                                // opt/tls-cipher-12
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                          // opt/tls-cipher-12
        (                                                                                                       // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                 // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Check)                                                                    // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                   // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Info)                                                                     // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                 // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                  // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                   // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                  // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                   // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Server)                                                                   // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                               // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                             // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                             // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                            // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/tls-cipher-12
        ),                                                                                                      // opt/tls-cipher-12
                                                                                                                // opt/tls-cipher-12
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                         // opt/tls-cipher-12
        (                                                                                                       // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/tls-cipher-12
        ),                                                                                                      // opt/tls-cipher-12
                                                                                                                // opt/tls-cipher-12
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                         // opt/tls-cipher-12
        (                                                                                                       // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/tls-cipher-12
        ),                                                                                                      // opt/tls-cipher-12
                                                                                                                // opt/tls-cipher-12
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                        // opt/tls-cipher-12
        (                                                                                                       // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                 // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Check)                                                                    // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                   // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Info)                                                                     // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                 // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                  // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                   // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                  // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                   // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                             // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                             // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                            // opt/tls-cipher-12
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/tls-cipher-12
        ),                                                                                                      // opt/tls-cipher-12
    ),                                                                                                          // opt/tls-cipher-12
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                           // opt/tls-cipher-13
    (                                                                                                           // opt/tls-cipher-13
        PARSE_RULE_OPTION_NAME("tls-cipher-13"),                                                                // opt/tls-cipher-13
        PARSE_RULE_OPTION_TYPE(String),                                                                         // opt/tls-cipher-13
        PARSE_RULE_OPTION_RESET(true),                                                                          // opt/tls-cipher-13
        PARSE_RULE_OPTION_REQUIRED(false),                                                                      // opt/tls-cipher-13
        PARSE_RULE_OPTION_SECTION(Global),                                                                      // opt/tls-cipher-13
                                                                                                                // opt/tls-cipher-13
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                          // opt/tls-cipher-13
        (                                                                                                       // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                 // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Check)                                                                    // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                   // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Info)                                                                     // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                 // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                  // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                   // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                  // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                   // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Server)                                                                   // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                               // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                             // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                             // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                            // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/tls-cipher-13
        ),                                                                                                      // opt/tls-cipher-13
                                                                                                                // opt/tls-cipher-13
        PARSE_RULE_OPTION_COMMAND_ROLE_ASYNC_VALID_LIST                                                         // opt/tls-cipher-13
        (                                                                                                       // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/tls-cipher-13
        ),                                                                                                      // opt/tls-cipher-13
                                                                                                                // opt/tls-cipher-13
        PARSE_RULE_OPTION_COMMAND_ROLE_LOCAL_VALID_LIST                                                         // opt/tls-cipher-13
        (                                                                                                       // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/tls-cipher-13
        ),                                                                                                      // opt/tls-cipher-13
                                                                                                                // opt/tls-cipher-13
        PARSE_RULE_OPTION_COMMAND_ROLE_REMOTE_VALID_LIST                                                        // opt/tls-cipher-13
        (                                                                                                       // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Annotate)                                                                 // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(ArchiveGet)                                                               // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(ArchivePush)                                                              // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                   // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Check)                                                                    // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Expire)                                                                   // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Info)                                                                     // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Manifest)                                                                 // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(RepoGet)                                                                  // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(RepoLs)                                                                   // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(RepoPut)                                                                  // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(RepoRm)                                                                   // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                  // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(StanzaCreate)                                                             // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(StanzaDelete)                                                             // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(StanzaUpgrade)                                                            // opt/tls-cipher-13
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                   // opt/tls-cipher-13
        ),                                                                                                      // opt/tls-cipher-13
    ),                                                                                                          // opt/tls-cipher-13
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                      // opt/tls-server-address
    (                                                                                                      // opt/tls-server-address
        PARSE_RULE_OPTION_NAME("tls-server-address"),                                                      // opt/tls-server-address
        PARSE_RULE_OPTION_TYPE(String),                                                                    // opt/tls-server-address
        PARSE_RULE_OPTION_RESET(true),                                                                     // opt/tls-server-address
        PARSE_RULE_OPTION_REQUIRED(true),                                                                  // opt/tls-server-address
        PARSE_RULE_OPTION_SECTION(Global),                                                                 // opt/tls-server-address
                                                                                                           // opt/tls-server-address
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                     // opt/tls-server-address
        (                                                                                                  // opt/tls-server-address
            PARSE_RULE_OPTION_COMMAND(Server)                                                              // opt/tls-server-address
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                          // opt/tls-server-address
        ),                                                                                                 // opt/tls-server-address
                                                                                                           // opt/tls-server-address
        PARSE_RULE_OPTIONAL                                                                                // opt/tls-server-address
        (                                                                                                  // opt/tls-server-address
            PARSE_RULE_OPTIONAL_GROUP                                                                      // opt/tls-server-address
            (                                                                                              // opt/tls-server-address
                PARSE_RULE_OPTIONAL_DEFAULT                                                                // opt/tls-server-address
                (                                                                                          // opt/tls-server-address
                    PARSE_RULE_VAL_STR(QT_localhost_QT),                                                   // opt/tls-server-address
                ),                                                                                         // opt/tls-server-address
            ),                                                                                             // opt/tls-server-address
        ),                                                                                                 // opt/tls-server-address
    ),                                                                                                     // opt/tls-server-address
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                         // opt/tls-server-auth
    (                                                                                                         // opt/tls-server-auth
        PARSE_RULE_OPTION_NAME("tls-server-auth"),                                                            // opt/tls-server-auth
        PARSE_RULE_OPTION_TYPE(Hash),                                                                         // opt/tls-server-auth
        PARSE_RULE_OPTION_RESET(true),                                                                        // opt/tls-server-auth
        PARSE_RULE_OPTION_REQUIRED(true),                                                                     // opt/tls-server-auth
        PARSE_RULE_OPTION_SECTION(Global),                                                                    // opt/tls-server-auth
        PARSE_RULE_OPTION_MULTI(true),                                                                        // opt/tls-server-auth
                                                                                                              // opt/tls-server-auth
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                        // opt/tls-server-auth
        (                                                                                                     // opt/tls-server-auth
            PARSE_RULE_OPTION_COMMAND(Server)                                                                 // opt/tls-server-auth
        ),                                                                                                    // opt/tls-server-auth
    ),                                                                                                        // opt/tls-server-auth
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                      // opt/tls-server-ca-file
    (                                                                                                      // opt/tls-server-ca-file
        PARSE_RULE_OPTION_NAME("tls-server-ca-file"),                                                      // opt/tls-server-ca-file
        PARSE_RULE_OPTION_TYPE(Path),                                                                      // opt/tls-server-ca-file
        PARSE_RULE_OPTION_RESET(true),                                                                     // opt/tls-server-ca-file
        PARSE_RULE_OPTION_REQUIRED(true),                                                                  // opt/tls-server-ca-file
        PARSE_RULE_OPTION_SECTION(Global),                                                                 // opt/tls-server-ca-file
                                                                                                           // opt/tls-server-ca-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                     // opt/tls-server-ca-file
        (                                                                                                  // opt/tls-server-ca-file
            PARSE_RULE_OPTION_COMMAND(Server)                                                              // opt/tls-server-ca-file
        ),                                                                                                 // opt/tls-server-ca-file
    ),                                                                                                     // opt/tls-server-ca-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                    // opt/tls-server-cert-file
    (                                                                                                    // opt/tls-server-cert-file
        PARSE_RULE_OPTION_NAME("tls-server-cert-file"),                                                  // opt/tls-server-cert-file
        PARSE_RULE_OPTION_TYPE(Path),                                                                    // opt/tls-server-cert-file
        PARSE_RULE_OPTION_RESET(true),                                                                   // opt/tls-server-cert-file
        PARSE_RULE_OPTION_REQUIRED(true),                                                                // opt/tls-server-cert-file
        PARSE_RULE_OPTION_SECTION(Global),                                                               // opt/tls-server-cert-file
                                                                                                         // opt/tls-server-cert-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                   // opt/tls-server-cert-file
        (                                                                                                // opt/tls-server-cert-file
            PARSE_RULE_OPTION_COMMAND(Server)                                                            // opt/tls-server-cert-file
        ),                                                                                               // opt/tls-server-cert-file
    ),                                                                                                   // opt/tls-server-cert-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                     // opt/tls-server-key-file
    (                                                                                                     // opt/tls-server-key-file
        PARSE_RULE_OPTION_NAME("tls-server-key-file"),                                                    // opt/tls-server-key-file
        PARSE_RULE_OPTION_TYPE(Path),                                                                     // opt/tls-server-key-file
        PARSE_RULE_OPTION_RESET(true),                                                                    // opt/tls-server-key-file
        PARSE_RULE_OPTION_REQUIRED(true),                                                                 // opt/tls-server-key-file
        PARSE_RULE_OPTION_SECTION(Global),                                                                // opt/tls-server-key-file
                                                                                                          // opt/tls-server-key-file
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                    // opt/tls-server-key-file
        (                                                                                                 // opt/tls-server-key-file
            PARSE_RULE_OPTION_COMMAND(Server)                                                             // opt/tls-server-key-file
        ),                                                                                                // opt/tls-server-key-file
    ),                                                                                                    // opt/tls-server-key-file
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                         // opt/tls-server-port
    (                                                                                                         // opt/tls-server-port
        PARSE_RULE_OPTION_NAME("tls-server-port"),                                                            // opt/tls-server-port
        PARSE_RULE_OPTION_TYPE(Integer),                                                                      // opt/tls-server-port
        PARSE_RULE_OPTION_RESET(true),                                                                        // opt/tls-server-port
        PARSE_RULE_OPTION_REQUIRED(true),                                                                     // opt/tls-server-port
        PARSE_RULE_OPTION_SECTION(Global),                                                                    // opt/tls-server-port
                                                                                                              // opt/tls-server-port
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                        // opt/tls-server-port
        (                                                                                                     // opt/tls-server-port
            PARSE_RULE_OPTION_COMMAND(Server)                                                                 // opt/tls-server-port
            PARSE_RULE_OPTION_COMMAND(ServerPing)                                                             // opt/tls-server-port
        ),                                                                                                    // opt/tls-server-port
                                                                                                              // opt/tls-server-port
        PARSE_RULE_OPTIONAL                                                                                   // opt/tls-server-port
        (                                                                                                     // opt/tls-server-port
            PARSE_RULE_OPTIONAL_GROUP                                                                         // opt/tls-server-port
            (                                                                                                 // opt/tls-server-port
                PARSE_RULE_OPTIONAL_ALLOW_RANGE                                                               // opt/tls-server-port
                (                                                                                             // opt/tls-server-port
                    PARSE_RULE_VAL_INT(1),                                                                    // opt/tls-server-port
                    PARSE_RULE_VAL_INT(65535),                                                                // opt/tls-server-port
                ),                                                                                            // opt/tls-server-port
                                                                                                              // opt/tls-server-port
                PARSE_RULE_OPTIONAL_DEFAULT                                                                   // opt/tls-server-port
                (                                                                                             // opt/tls-server-port
                    PARSE_RULE_VAL_INT(8432),                                                                 // opt/tls-server-port
                ),                                                                                            // opt/tls-server-port
            ),                                                                                                // opt/tls-server-port
        ),                                                                                                    // opt/tls-server-port
    ),                                                                                                        // opt/tls-server-port
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                    // opt/type
    (                                                                                                                    // opt/type
        PARSE_RULE_OPTION_NAME("type"),                                                                                  // opt/type
        PARSE_RULE_OPTION_TYPE(StringId),                                                                                // opt/type
        PARSE_RULE_OPTION_REQUIRED(true),                                                                                // opt/type
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                          // opt/type
                                                                                                                         // opt/type
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                   // opt/type
        (                                                                                                                // opt/type
            PARSE_RULE_OPTION_COMMAND(Backup)                                                                            // opt/type
            PARSE_RULE_OPTION_COMMAND(Info)                                                                              // opt/type
            PARSE_RULE_OPTION_COMMAND(Restore)                                                                           // opt/type
        ),                                                                                                               // opt/type
                                                                                                                         // opt/type
        PARSE_RULE_OPTIONAL                                                                                              // opt/type
        (                                                                                                                // opt/type
            PARSE_RULE_OPTIONAL_GROUP                                                                                    // opt/type
            (                                                                                                            // opt/type
                PARSE_RULE_FILTER_CMD                                                                                    // opt/type
                (                                                                                                        // opt/type
                    PARSE_RULE_VAL_CMD(Backup),                                                                          // opt/type
                ),                                                                                                       // opt/type
                                                                                                                         // opt/type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                           // opt/type
                (                                                                                                        // opt/type
                    PARSE_RULE_VAL_STRID(full),                                                                          // opt/type
                    PARSE_RULE_VAL_STRID(diff),                                                                          // opt/type
                    PARSE_RULE_VAL_STRID(incr),                                                                          // opt/type
                ),                                                                                                       // opt/type
                                                                                                                         // opt/type
                PARSE_RULE_OPTIONAL_DEFAULT                                                                              // opt/type
                (                                                                                                        // opt/type
                    PARSE_RULE_VAL_STRID(incr),                                                                          // opt/type
                ),                                                                                                       // opt/type
            ),                                                                                                           // opt/type
                                                                                                                         // opt/type
            PARSE_RULE_OPTIONAL_GROUP                                                                                    // opt/type
            (                                                                                                            // opt/type
                PARSE_RULE_FILTER_CMD                                                                                    // opt/type
                (                                                                                                        // opt/type
                    PARSE_RULE_VAL_CMD(Info),                                                                            // opt/type
                ),                                                                                                       // opt/type
                                                                                                                         // opt/type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                           // opt/type
                (                                                                                                        // opt/type
                    PARSE_RULE_VAL_STRID(full),                                                                          // opt/type
                    PARSE_RULE_VAL_STRID(diff),                                                                          // opt/type
                    PARSE_RULE_VAL_STRID(incr),                                                                          // opt/type
                ),                                                                                                       // opt/type
                                                                                                                         // opt/type
                PARSE_RULE_OPTIONAL_NOT_REQUIRED(),                                                                      // opt/type
            ),                                                                                                           // opt/type
                                                                                                                         // opt/type
            PARSE_RULE_OPTIONAL_GROUP                                                                                    // opt/type
            (                                                                                                            // opt/type
                PARSE_RULE_FILTER_CMD                                                                                    // opt/type
                (                                                                                                        // opt/type
                    PARSE_RULE_VAL_CMD(Restore),                                                                         // opt/type
                ),                                                                                                       // opt/type
                                                                                                                         // opt/type
                PARSE_RULE_OPTIONAL_ALLOW_LIST                                                                           // opt/type
                (                                                                                                        // opt/type
                    PARSE_RULE_VAL_STRID(lsn),                                                                           // opt/type
                    PARSE_RULE_VAL_STRID(name),                                                                          // opt/type
                    PARSE_RULE_VAL_STRID(time),                                                                          // opt/type
                    PARSE_RULE_VAL_STRID(xid),                                                                           // opt/type
                    PARSE_RULE_VAL_STRID(preserve),                                                                      // opt/type
                    PARSE_RULE_VAL_STRID(none),                                                                          // opt/type
                    PARSE_RULE_VAL_STRID(immediate),                                                                     // opt/type
                    PARSE_RULE_VAL_STRID(default),                                                                       // opt/type
                    PARSE_RULE_VAL_STRID(standby),                                                                       // opt/type
                ),                                                                                                       // opt/type
                                                                                                                         // opt/type
                PARSE_RULE_OPTIONAL_DEFAULT                                                                              // opt/type
                (                                                                                                        // opt/type
                    PARSE_RULE_VAL_STRID(default),                                                                       // opt/type
                ),                                                                                                       // opt/type
            ),                                                                                                           // opt/type
        ),                                                                                                               // opt/type
    ),                                                                                                                   // opt/type
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                 // opt/verbose
    (                                                                                                                 // opt/verbose
        PARSE_RULE_OPTION_NAME("verbose"),                                                                            // opt/verbose
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                              // opt/verbose
        PARSE_RULE_OPTION_REQUIRED(true),                                                                             // opt/verbose
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                       // opt/verbose
                                                                                                                      // opt/verbose
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                // opt/verbose
        (                                                                                                             // opt/verbose
            PARSE_RULE_OPTION_COMMAND(Verify)                                                                         // opt/verbose
        ),                                                                                                            // opt/verbose
                                                                                                                      // opt/verbose
        PARSE_RULE_OPTIONAL                                                                                           // opt/verbose
        (                                                                                                             // opt/verbose
            PARSE_RULE_OPTIONAL_GROUP                                                                                 // opt/verbose
            (                                                                                                         // opt/verbose
                PARSE_RULE_OPTIONAL_DEFAULT                                                                           // opt/verbose
                (                                                                                                     // opt/verbose
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                        // opt/verbose
                ),                                                                                                    // opt/verbose
            ),                                                                                                        // opt/verbose
        ),                                                                                                            // opt/verbose
    ),                                                                                                                // opt/verbose
    // -----------------------------------------------------------------------------------------------------------------------------
    PARSE_RULE_OPTION                                                                                                 // opt/version
    (                                                                                                                 // opt/version
        PARSE_RULE_OPTION_NAME("version"),                                                                            // opt/version
        PARSE_RULE_OPTION_TYPE(Boolean),                                                                              // opt/version
        PARSE_RULE_OPTION_REQUIRED(true),                                                                             // opt/version
        PARSE_RULE_OPTION_SECTION(CommandLine),                                                                       // opt/version
                                                                                                                      // opt/version
        PARSE_RULE_OPTION_COMMAND_ROLE_MAIN_VALID_LIST                                                                // opt/version
        (                                                                                                             // opt/version
            PARSE_RULE_OPTION_COMMAND(Help)                                                                           // opt/version
        ),                                                                                                            // opt/version
                                                                                                                      // opt/version
        PARSE_RULE_OPTIONAL                                                                                           // opt/version
        (                                                                                                             // opt/version
            PARSE_RULE_OPTIONAL_GROUP                                                                                 // opt/version
            (                                                                                                         // opt/version
                PARSE_RULE_OPTIONAL_DEFAULT                                                                           // opt/version
                (                                                                                                     // opt/version
                    PARSE_RULE_VAL_BOOL_FALSE,                                                                        // opt/version
                ),                                                                                                    // opt/version
            ),                                                                                                        // opt/version
        ),                                                                                                            // opt/version
    ),                                                                                                                // opt/version
};

/***********************************************************************************************************************************
Option deprecations
***********************************************************************************************************************************/
#define CFG_OPTION_DEPRECATE_TOTAL                                  39

static const ParseRuleOptionDeprecate parseRuleOptionDeprecate[CFG_OPTION_DEPRECATE_TOTAL] =
{
    {                                                                                        // opt-deprecate/archive-push-queue-max
        .name = "archive-queue-max",                                                         // opt-deprecate/archive-push-queue-max
        .id = cfgOptArchivePushQueueMax,                                                     // opt-deprecate/archive-push-queue-max
        .unindexed = true,                                                                   // opt-deprecate/archive-push-queue-max
    },                                                                                       // opt-deprecate/archive-push-queue-max
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                 // opt-deprecate/repo-host-cmd
        .name = "backup-cmd",                                                                         // opt-deprecate/repo-host-cmd
        .id = cfgOptRepoHostCmd,                                                                      // opt-deprecate/repo-host-cmd
        .unindexed = true,                                                                            // opt-deprecate/repo-host-cmd
    },                                                                                                // opt-deprecate/repo-host-cmd
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                              // opt-deprecate/repo-host-config
        .name = "backup-config",                                                                   // opt-deprecate/repo-host-config
        .id = cfgOptRepoHostConfig,                                                                // opt-deprecate/repo-host-config
        .unindexed = true,                                                                         // opt-deprecate/repo-host-config
    },                                                                                             // opt-deprecate/repo-host-config
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                     // opt-deprecate/repo-host
        .name = "backup-host",                                                                            // opt-deprecate/repo-host
        .id = cfgOptRepoHost,                                                                             // opt-deprecate/repo-host
        .unindexed = true,                                                                                // opt-deprecate/repo-host
    },                                                                                                    // opt-deprecate/repo-host
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                // opt-deprecate/repo-host-port
        .name = "backup-ssh-port",                                                                   // opt-deprecate/repo-host-port
        .id = cfgOptRepoHostPort,                                                                    // opt-deprecate/repo-host-port
        .unindexed = true,                                                                           // opt-deprecate/repo-host-port
    },                                                                                               // opt-deprecate/repo-host-port
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                // opt-deprecate/repo-host-user
        .name = "backup-user",                                                                       // opt-deprecate/repo-host-user
        .id = cfgOptRepoHostUser,                                                                    // opt-deprecate/repo-host-user
        .unindexed = true,                                                                           // opt-deprecate/repo-host-user
    },                                                                                               // opt-deprecate/repo-host-user
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                   // opt-deprecate/pg-host-cmd
        .name = "db-cmd",                                                                               // opt-deprecate/pg-host-cmd
        .id = cfgOptPgHostCmd,                                                                          // opt-deprecate/pg-host-cmd
        .indexed = true,                                                                                // opt-deprecate/pg-host-cmd
        .unindexed = true,                                                                              // opt-deprecate/pg-host-cmd
    },                                                                                                  // opt-deprecate/pg-host-cmd
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                // opt-deprecate/pg-host-config
        .name = "db-config",                                                                         // opt-deprecate/pg-host-config
        .id = cfgOptPgHostConfig,                                                                    // opt-deprecate/pg-host-config
        .indexed = true,                                                                             // opt-deprecate/pg-host-config
        .unindexed = true,                                                                           // opt-deprecate/pg-host-config
    },                                                                                               // opt-deprecate/pg-host-config
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                       // opt-deprecate/pg-host
        .name = "db-host",                                                                                  // opt-deprecate/pg-host
        .id = cfgOptPgHost,                                                                                 // opt-deprecate/pg-host
        .indexed = true,                                                                                    // opt-deprecate/pg-host
        .unindexed = true,                                                                                  // opt-deprecate/pg-host
    },                                                                                                      // opt-deprecate/pg-host
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                       // opt-deprecate/pg-path
        .name = "db-path",                                                                                  // opt-deprecate/pg-path
        .id = cfgOptPgPath,                                                                                 // opt-deprecate/pg-path
        .indexed = true,                                                                                    // opt-deprecate/pg-path
        .unindexed = true,                                                                                  // opt-deprecate/pg-path
    },                                                                                                      // opt-deprecate/pg-path
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                       // opt-deprecate/pg-port
        .name = "db-port",                                                                                  // opt-deprecate/pg-port
        .id = cfgOptPgPort,                                                                                 // opt-deprecate/pg-port
        .indexed = true,                                                                                    // opt-deprecate/pg-port
        .unindexed = true,                                                                                  // opt-deprecate/pg-port
    },                                                                                                      // opt-deprecate/pg-port
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                // opt-deprecate/pg-socket-path
        .name = "db-socket-path",                                                                    // opt-deprecate/pg-socket-path
        .id = cfgOptPgSocketPath,                                                                    // opt-deprecate/pg-socket-path
        .indexed = true,                                                                             // opt-deprecate/pg-socket-path
        .unindexed = true,                                                                           // opt-deprecate/pg-socket-path
    },                                                                                               // opt-deprecate/pg-socket-path
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                  // opt-deprecate/pg-host-port
        .name = "db-ssh-port",                                                                         // opt-deprecate/pg-host-port
        .id = cfgOptPgHostPort,                                                                        // opt-deprecate/pg-host-port
        .indexed = true,                                                                               // opt-deprecate/pg-host-port
        .unindexed = true,                                                                             // opt-deprecate/pg-host-port
    },                                                                                                 // opt-deprecate/pg-host-port
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                  // opt-deprecate/pg-host-user
        .name = "db-user",                                                                             // opt-deprecate/pg-host-user
        .id = cfgOptPgHostUser,                                                                        // opt-deprecate/pg-host-user
        .indexed = true,                                                                               // opt-deprecate/pg-host-user
        .unindexed = true,                                                                             // opt-deprecate/pg-host-user
    },                                                                                                 // opt-deprecate/pg-host-user
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                 // opt-deprecate/repo-hardlink
        .name = "hardlink",                                                                           // opt-deprecate/repo-hardlink
        .id = cfgOptRepoHardlink,                                                                     // opt-deprecate/repo-hardlink
        .unindexed = true,                                                                            // opt-deprecate/repo-hardlink
    },                                                                                                // opt-deprecate/repo-hardlink
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                          // opt-deprecate/repo-storage-ca-file
        .name = "repo-azure-ca-file",                                                          // opt-deprecate/repo-storage-ca-file
        .id = cfgOptRepoStorageCaFile,                                                         // opt-deprecate/repo-storage-ca-file
        .indexed = true,                                                                       // opt-deprecate/repo-storage-ca-file
    },                                                                                         // opt-deprecate/repo-storage-ca-file
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                          // opt-deprecate/repo-storage-ca-path
        .name = "repo-azure-ca-path",                                                          // opt-deprecate/repo-storage-ca-path
        .id = cfgOptRepoStorageCaPath,                                                         // opt-deprecate/repo-storage-ca-path
        .indexed = true,                                                                       // opt-deprecate/repo-storage-ca-path
    },                                                                                         // opt-deprecate/repo-storage-ca-path
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                             // opt-deprecate/repo-storage-host
        .name = "repo-azure-host",                                                                // opt-deprecate/repo-storage-host
        .id = cfgOptRepoStorageHost,                                                              // opt-deprecate/repo-storage-host
        .indexed = true,                                                                          // opt-deprecate/repo-storage-host
    },                                                                                            // opt-deprecate/repo-storage-host
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                             // opt-deprecate/repo-storage-port
        .name = "repo-azure-port",                                                                // opt-deprecate/repo-storage-port
        .id = cfgOptRepoStoragePort,                                                              // opt-deprecate/repo-storage-port
        .indexed = true,                                                                          // opt-deprecate/repo-storage-port
    },                                                                                            // opt-deprecate/repo-storage-port
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                       // opt-deprecate/repo-storage-verify-tls
        .name = "repo-azure-verify-tls",                                                    // opt-deprecate/repo-storage-verify-tls
        .id = cfgOptRepoStorageVerifyTls,                                                   // opt-deprecate/repo-storage-verify-tls
        .indexed = true,                                                                    // opt-deprecate/repo-storage-verify-tls
    },                                                                                      // opt-deprecate/repo-storage-verify-tls
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                              // opt-deprecate/repo-cipher-pass
        .name = "repo-cipher-pass",                                                                // opt-deprecate/repo-cipher-pass
        .id = cfgOptRepoCipherPass,                                                                // opt-deprecate/repo-cipher-pass
        .unindexed = true,                                                                         // opt-deprecate/repo-cipher-pass
    },                                                                                             // opt-deprecate/repo-cipher-pass
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                              // opt-deprecate/repo-cipher-type
        .name = "repo-cipher-type",                                                                // opt-deprecate/repo-cipher-type
        .id = cfgOptRepoCipherType,                                                                // opt-deprecate/repo-cipher-type
        .unindexed = true,                                                                         // opt-deprecate/repo-cipher-type
    },                                                                                             // opt-deprecate/repo-cipher-type
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                     // opt-deprecate/repo-path
        .name = "repo-path",                                                                              // opt-deprecate/repo-path
        .id = cfgOptRepoPath,                                                                             // opt-deprecate/repo-path
        .unindexed = true,                                                                                // opt-deprecate/repo-path
    },                                                                                                    // opt-deprecate/repo-path
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                // opt-deprecate/repo-s3-bucket
        .name = "repo-s3-bucket",                                                                    // opt-deprecate/repo-s3-bucket
        .id = cfgOptRepoS3Bucket,                                                                    // opt-deprecate/repo-s3-bucket
        .unindexed = true,                                                                           // opt-deprecate/repo-s3-bucket
    },                                                                                               // opt-deprecate/repo-s3-bucket
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                          // opt-deprecate/repo-storage-ca-file
        .name = "repo-s3-ca-file",                                                             // opt-deprecate/repo-storage-ca-file
        .id = cfgOptRepoStorageCaFile,                                                         // opt-deprecate/repo-storage-ca-file
        .indexed = true,                                                                       // opt-deprecate/repo-storage-ca-file
        .unindexed = true,                                                                     // opt-deprecate/repo-storage-ca-file
    },                                                                                         // opt-deprecate/repo-storage-ca-file
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                          // opt-deprecate/repo-storage-ca-path
        .name = "repo-s3-ca-path",                                                             // opt-deprecate/repo-storage-ca-path
        .id = cfgOptRepoStorageCaPath,                                                         // opt-deprecate/repo-storage-ca-path
        .indexed = true,                                                                       // opt-deprecate/repo-storage-ca-path
        .unindexed = true,                                                                     // opt-deprecate/repo-storage-ca-path
    },                                                                                         // opt-deprecate/repo-storage-ca-path
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                              // opt-deprecate/repo-s3-endpoint
        .name = "repo-s3-endpoint",                                                                // opt-deprecate/repo-s3-endpoint
        .id = cfgOptRepoS3Endpoint,                                                                // opt-deprecate/repo-s3-endpoint
        .unindexed = true,                                                                         // opt-deprecate/repo-s3-endpoint
    },                                                                                             // opt-deprecate/repo-s3-endpoint
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                             // opt-deprecate/repo-storage-host
        .name = "repo-s3-host",                                                                   // opt-deprecate/repo-storage-host
        .id = cfgOptRepoStorageHost,                                                              // opt-deprecate/repo-storage-host
        .indexed = true,                                                                          // opt-deprecate/repo-storage-host
        .unindexed = true,                                                                        // opt-deprecate/repo-storage-host
    },                                                                                            // opt-deprecate/repo-storage-host
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                   // opt-deprecate/repo-s3-key
        .name = "repo-s3-key",                                                                          // opt-deprecate/repo-s3-key
        .id = cfgOptRepoS3Key,                                                                          // opt-deprecate/repo-s3-key
        .unindexed = true,                                                                              // opt-deprecate/repo-s3-key
    },                                                                                                  // opt-deprecate/repo-s3-key
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                            // opt-deprecate/repo-s3-key-secret
        .name = "repo-s3-key-secret",                                                            // opt-deprecate/repo-s3-key-secret
        .id = cfgOptRepoS3KeySecret,                                                             // opt-deprecate/repo-s3-key-secret
        .unindexed = true,                                                                       // opt-deprecate/repo-s3-key-secret
    },                                                                                           // opt-deprecate/repo-s3-key-secret
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                             // opt-deprecate/repo-storage-port
        .name = "repo-s3-port",                                                                   // opt-deprecate/repo-storage-port
        .id = cfgOptRepoStoragePort,                                                              // opt-deprecate/repo-storage-port
        .indexed = true,                                                                          // opt-deprecate/repo-storage-port
    },                                                                                            // opt-deprecate/repo-storage-port
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                // opt-deprecate/repo-s3-region
        .name = "repo-s3-region",                                                                    // opt-deprecate/repo-s3-region
        .id = cfgOptRepoS3Region,                                                                    // opt-deprecate/repo-s3-region
        .unindexed = true,                                                                           // opt-deprecate/repo-s3-region
    },                                                                                               // opt-deprecate/repo-s3-region
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                       // opt-deprecate/repo-storage-verify-tls
        .name = "repo-s3-verify-ssl",                                                       // opt-deprecate/repo-storage-verify-tls
        .id = cfgOptRepoStorageVerifyTls,                                                   // opt-deprecate/repo-storage-verify-tls
        .indexed = true,                                                                    // opt-deprecate/repo-storage-verify-tls
        .unindexed = true,                                                                  // opt-deprecate/repo-storage-verify-tls
    },                                                                                      // opt-deprecate/repo-storage-verify-tls
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                       // opt-deprecate/repo-storage-verify-tls
        .name = "repo-s3-verify-tls",                                                       // opt-deprecate/repo-storage-verify-tls
        .id = cfgOptRepoStorageVerifyTls,                                                   // opt-deprecate/repo-storage-verify-tls
        .indexed = true,                                                                    // opt-deprecate/repo-storage-verify-tls
    },                                                                                      // opt-deprecate/repo-storage-verify-tls
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                                     // opt-deprecate/repo-type
        .name = "repo-type",                                                                              // opt-deprecate/repo-type
        .id = cfgOptRepoType,                                                                             // opt-deprecate/repo-type
        .unindexed = true,                                                                                // opt-deprecate/repo-type
    },                                                                                                    // opt-deprecate/repo-type
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                        // opt-deprecate/repo-retention-archive
        .name = "retention-archive",                                                         // opt-deprecate/repo-retention-archive
        .id = cfgOptRepoRetentionArchive,                                                    // opt-deprecate/repo-retention-archive
        .unindexed = true,                                                                   // opt-deprecate/repo-retention-archive
    },                                                                                       // opt-deprecate/repo-retention-archive
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                   // opt-deprecate/repo-retention-archive-type
        .name = "retention-archive-type",                                               // opt-deprecate/repo-retention-archive-type
        .id = cfgOptRepoRetentionArchiveType,                                           // opt-deprecate/repo-retention-archive-type
        .unindexed = true,                                                              // opt-deprecate/repo-retention-archive-type
    },                                                                                  // opt-deprecate/repo-retention-archive-type
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                           // opt-deprecate/repo-retention-diff
        .name = "retention-diff",                                                               // opt-deprecate/repo-retention-diff
        .id = cfgOptRepoRetentionDiff,                                                          // opt-deprecate/repo-retention-diff
        .unindexed = true,                                                                      // opt-deprecate/repo-retention-diff
    },                                                                                          // opt-deprecate/repo-retention-diff
    // -----------------------------------------------------------------------------------------------------------------------------
    {                                                                                           // opt-deprecate/repo-retention-full
        .name = "retention-full",                                                               // opt-deprecate/repo-retention-full
        .id = cfgOptRepoRetentionFull,                                                          // opt-deprecate/repo-retention-full
        .unindexed = true,                                                                      // opt-deprecate/repo-retention-full
    },                                                                                          // opt-deprecate/repo-retention-full
};

/***********************************************************************************************************************************
Order for option parse resolution
***********************************************************************************************************************************/
static const uint8_t optionResolveOrder[] =
{
    cfgOptStanza,                                                                                               // opt-resolve-order
    cfgOptAnnotation,                                                                                           // opt-resolve-order
    cfgOptArchiveAsync,                                                                                         // opt-resolve-order
    cfgOptArchiveGetQueueMax,                                                                                   // opt-resolve-order
    cfgOptArchiveHeaderCheck,                                                                                   // opt-resolve-order
    cfgOptArchiveMissingRetry,                                                                                  // opt-resolve-order
    cfgOptArchiveMode,                                                                                          // opt-resolve-order
    cfgOptArchivePushQueueMax,                                                                                  // opt-resolve-order
    cfgOptArchiveTimeout,                                                                                       // opt-resolve-order
    cfgOptBackupStandby,                                                                                        // opt-resolve-order
    cfgOptBeta,                                                                                                 // opt-resolve-order
    cfgOptBufferSize,                                                                                           // opt-resolve-order
    cfgOptChecksumPage,                                                                                         // opt-resolve-order
    cfgOptCipherPass,                                                                                           // opt-resolve-order
    cfgOptCmd,                                                                                                  // opt-resolve-order
    cfgOptCmdSsh,                                                                                               // opt-resolve-order
    cfgOptCompress,                                                                                             // opt-resolve-order
    cfgOptCompressLevelNetwork,                                                                                 // opt-resolve-order
    cfgOptCompressType,                                                                                         // opt-resolve-order
    cfgOptConfig,                                                                                               // opt-resolve-order
    cfgOptConfigIncludePath,                                                                                    // opt-resolve-order
    cfgOptConfigPath,                                                                                           // opt-resolve-order
    cfgOptDbExclude,                                                                                            // opt-resolve-order
    cfgOptDbInclude,                                                                                            // opt-resolve-order
    cfgOptDbTimeout,                                                                                            // opt-resolve-order
    cfgOptDelta,                                                                                                // opt-resolve-order
    cfgOptDetailLevel,                                                                                          // opt-resolve-order
    cfgOptDryRun,                                                                                               // opt-resolve-order
    cfgOptExclude,                                                                                              // opt-resolve-order
    cfgOptExecId,                                                                                               // opt-resolve-order
    cfgOptExpireAuto,                                                                                           // opt-resolve-order
    cfgOptFilter,                                                                                               // opt-resolve-order
    cfgOptHelp,                                                                                                 // opt-resolve-order
    cfgOptIgnoreMissing,                                                                                        // opt-resolve-order
    cfgOptIoTimeout,                                                                                            // opt-resolve-order
    cfgOptJobRetry,                                                                                             // opt-resolve-order
    cfgOptJobRetryInterval,                                                                                     // opt-resolve-order
    cfgOptLinkAll,                                                                                              // opt-resolve-order
    cfgOptLinkMap,                                                                                              // opt-resolve-order
    cfgOptLock,                                                                                                 // opt-resolve-order
    cfgOptLockPath,                                                                                             // opt-resolve-order
    cfgOptLogLevelConsole,                                                                                      // opt-resolve-order
    cfgOptLogLevelFile,                                                                                         // opt-resolve-order
    cfgOptLogLevelStderr,                                                                                       // opt-resolve-order
    cfgOptLogPath,                                                                                              // opt-resolve-order
    cfgOptLogSubprocess,                                                                                        // opt-resolve-order
    cfgOptLogTimestamp,                                                                                         // opt-resolve-order
    cfgOptManifestSaveThreshold,                                                                                // opt-resolve-order
    cfgOptNeutralUmask,                                                                                         // opt-resolve-order
    cfgOptOldest,                                                                                               // opt-resolve-order
    cfgOptOnline,                                                                                               // opt-resolve-order
    cfgOptOutput,                                                                                               // opt-resolve-order
    cfgOptPageHeaderCheck,                                                                                      // opt-resolve-order
    cfgOptPg,                                                                                                   // opt-resolve-order
    cfgOptPgLocal,                                                                                              // opt-resolve-order
    cfgOptPgPath,                                                                                               // opt-resolve-order
    cfgOptPgPort,                                                                                               // opt-resolve-order
    cfgOptPgSocketPath,                                                                                         // opt-resolve-order
    cfgOptPgUser,                                                                                               // opt-resolve-order
    cfgOptPgVersionForce,                                                                                       // opt-resolve-order
    cfgOptPriority,                                                                                             // opt-resolve-order
    cfgOptProcess,                                                                                              // opt-resolve-order
    cfgOptProcessMax,                                                                                           // opt-resolve-order
    cfgOptProtocolTimeout,                                                                                      // opt-resolve-order
    cfgOptRaw,                                                                                                  // opt-resolve-order
    cfgOptRecurse,                                                                                              // opt-resolve-order
    cfgOptReference,                                                                                            // opt-resolve-order
    cfgOptRemoteType,                                                                                           // opt-resolve-order
    cfgOptRepo,                                                                                                 // opt-resolve-order
    cfgOptRepoBundle,                                                                                           // opt-resolve-order
    cfgOptRepoBundleLimit,                                                                                      // opt-resolve-order
    cfgOptRepoBundleSize,                                                                                       // opt-resolve-order
    cfgOptRepoCipherType,                                                                                       // opt-resolve-order
    cfgOptRepoGcsUserProject,                                                                                   // opt-resolve-order
    cfgOptRepoHardlink,                                                                                         // opt-resolve-order
    cfgOptRepoLocal,                                                                                            // opt-resolve-order
    cfgOptRepoPath,                                                                                             // opt-resolve-order
    cfgOptRepoRetentionArchive,                                                                                 // opt-resolve-order
    cfgOptRepoRetentionArchiveType,                                                                             // opt-resolve-order
    cfgOptRepoRetentionDiff,                                                                                    // opt-resolve-order
    cfgOptRepoRetentionFull,                                                                                    // opt-resolve-order
    cfgOptRepoRetentionFullType,                                                                                // opt-resolve-order
    cfgOptRepoRetentionHistory,                                                                                 // opt-resolve-order
    cfgOptRepoTargetTime,                                                                                       // opt-resolve-order
    cfgOptRepoType,                                                                                             // opt-resolve-order
    cfgOptReport,                                                                                               // opt-resolve-order
    cfgOptResume,                                                                                               // opt-resolve-order
    cfgOptSckBlock,                                                                                             // opt-resolve-order
    cfgOptSckKeepAlive,                                                                                         // opt-resolve-order
    cfgOptSet,                                                                                                  // opt-resolve-order
    cfgOptSort,                                                                                                 // opt-resolve-order
    cfgOptSpoolPath,                                                                                            // opt-resolve-order
    cfgOptStartFast,                                                                                            // opt-resolve-order
    cfgOptStopAuto,                                                                                             // opt-resolve-order
    cfgOptTablespaceMap,                                                                                        // opt-resolve-order
    cfgOptTablespaceMapAll,                                                                                     // opt-resolve-order
    cfgOptTcpKeepAliveCount,                                                                                    // opt-resolve-order
    cfgOptTcpKeepAliveIdle,                                                                                     // opt-resolve-order
    cfgOptTcpKeepAliveInterval,                                                                                 // opt-resolve-order
    cfgOptTlsCipher12,                                                                                          // opt-resolve-order
    cfgOptTlsCipher13,                                                                                          // opt-resolve-order
    cfgOptTlsServerAddress,                                                                                     // opt-resolve-order
    cfgOptTlsServerAuth,                                                                                        // opt-resolve-order
    cfgOptTlsServerCaFile,                                                                                      // opt-resolve-order
    cfgOptTlsServerCertFile,                                                                                    // opt-resolve-order
    cfgOptTlsServerKeyFile,                                                                                     // opt-resolve-order
    cfgOptTlsServerPort,                                                                                        // opt-resolve-order
    cfgOptType,                                                                                                 // opt-resolve-order
    cfgOptVerbose,                                                                                              // opt-resolve-order
    cfgOptVersion,                                                                                              // opt-resolve-order
    cfgOptArchiveCheck,                                                                                         // opt-resolve-order
    cfgOptArchiveCopy,                                                                                          // opt-resolve-order
    cfgOptArchiveModeCheck,                                                                                     // opt-resolve-order
    cfgOptCompressLevel,                                                                                        // opt-resolve-order
    cfgOptForce,                                                                                                // opt-resolve-order
    cfgOptPgDatabase,                                                                                           // opt-resolve-order
    cfgOptPgHost,                                                                                               // opt-resolve-order
    cfgOptPgHostConfig,                                                                                         // opt-resolve-order
    cfgOptPgHostConfigIncludePath,                                                                              // opt-resolve-order
    cfgOptPgHostConfigPath,                                                                                     // opt-resolve-order
    cfgOptPgHostType,                                                                                           // opt-resolve-order
    cfgOptPgHostUser,                                                                                           // opt-resolve-order
    cfgOptRecoveryOption,                                                                                       // opt-resolve-order
    cfgOptRepoAzureAccount,                                                                                     // opt-resolve-order
    cfgOptRepoAzureContainer,                                                                                   // opt-resolve-order
    cfgOptRepoAzureEndpoint,                                                                                    // opt-resolve-order
    cfgOptRepoAzureKeyType,                                                                                     // opt-resolve-order
    cfgOptRepoAzureUriStyle,                                                                                    // opt-resolve-order
    cfgOptRepoBlock,                                                                                            // opt-resolve-order
    cfgOptRepoBlockAgeMap,                                                                                      // opt-resolve-order
    cfgOptRepoBlockChecksumSizeMap,                                                                             // opt-resolve-order
    cfgOptRepoBlockSizeMap,                                                                                     // opt-resolve-order
    cfgOptRepoBlockSizeSuper,                                                                                   // opt-resolve-order
    cfgOptRepoBlockSizeSuperFull,                                                                               // opt-resolve-order
    cfgOptRepoCipherPass,                                                                                       // opt-resolve-order
    cfgOptRepoGcsKeyType,                                                                                       // opt-resolve-order
    cfgOptRepoHost,                                                                                             // opt-resolve-order
    cfgOptRepoHostConfig,                                                                                       // opt-resolve-order
    cfgOptRepoHostConfigIncludePath,                                                                            // opt-resolve-order
    cfgOptRepoHostConfigPath,                                                                                   // opt-resolve-order
    cfgOptRepoHostType,                                                                                         // opt-resolve-order
    cfgOptRepoHostUser,                                                                                         // opt-resolve-order
    cfgOptRepoS3Bucket,                                                                                         // opt-resolve-order
    cfgOptRepoS3Endpoint,                                                                                       // opt-resolve-order
    cfgOptRepoS3KeyType,                                                                                        // opt-resolve-order
    cfgOptRepoS3KmsKeyId,                                                                                       // opt-resolve-order
    cfgOptRepoS3Region,                                                                                         // opt-resolve-order
    cfgOptRepoS3RequesterPays,                                                                                  // opt-resolve-order
    cfgOptRepoS3Role,                                                                                           // opt-resolve-order
    cfgOptRepoS3SseCustomerKey,                                                                                 // opt-resolve-order
    cfgOptRepoS3Token,                                                                                          // opt-resolve-order
    cfgOptRepoS3UriStyle,                                                                                       // opt-resolve-order
    cfgOptRepoSftpHost,                                                                                         // opt-resolve-order
    cfgOptRepoSftpHostKeyCheckType,                                                                             // opt-resolve-order
    cfgOptRepoSftpHostKeyHashType,                                                                              // opt-resolve-order
    cfgOptRepoSftpHostPort,                                                                                     // opt-resolve-order
    cfgOptRepoSftpHostUser,                                                                                     // opt-resolve-order
    cfgOptRepoSftpKnownHost,                                                                                    // opt-resolve-order
    cfgOptRepoSftpPrivateKeyFile,                                                                               // opt-resolve-order
    cfgOptRepoSftpPrivateKeyPassphrase,                                                                         // opt-resolve-order
    cfgOptRepoSftpPublicKeyFile,                                                                                // opt-resolve-order
    cfgOptRepoStorageCaFile,                                                                                    // opt-resolve-order
    cfgOptRepoStorageCaPath,                                                                                    // opt-resolve-order
    cfgOptRepoStorageHost,                                                                                      // opt-resolve-order
    cfgOptRepoStoragePort,                                                                                      // opt-resolve-order
    cfgOptRepoStorageTag,                                                                                       // opt-resolve-order
    cfgOptRepoStorageUploadChunkSize,                                                                           // opt-resolve-order
    cfgOptRepoStorageVerifyTls,                                                                                 // opt-resolve-order
    cfgOptRepoSymlink,                                                                                          // opt-resolve-order
    cfgOptTarget,                                                                                               // opt-resolve-order
    cfgOptTargetAction,                                                                                         // opt-resolve-order
    cfgOptTargetExclusive,                                                                                      // opt-resolve-order
    cfgOptTargetTimeline,                                                                                       // opt-resolve-order
    cfgOptPgHostCaFile,                                                                                         // opt-resolve-order
    cfgOptPgHostCaPath,                                                                                         // opt-resolve-order
    cfgOptPgHostCertFile,                                                                                       // opt-resolve-order
    cfgOptPgHostCmd,                                                                                            // opt-resolve-order
    cfgOptPgHostKeyFile,                                                                                        // opt-resolve-order
    cfgOptPgHostPort,                                                                                           // opt-resolve-order
    cfgOptRepoAzureKey,                                                                                         // opt-resolve-order
    cfgOptRepoGcsBucket,                                                                                        // opt-resolve-order
    cfgOptRepoGcsEndpoint,                                                                                      // opt-resolve-order
    cfgOptRepoGcsKey,                                                                                           // opt-resolve-order
    cfgOptRepoHostCaFile,                                                                                       // opt-resolve-order
    cfgOptRepoHostCaPath,                                                                                       // opt-resolve-order
    cfgOptRepoHostCertFile,                                                                                     // opt-resolve-order
    cfgOptRepoHostCmd,                                                                                          // opt-resolve-order
    cfgOptRepoHostKeyFile,                                                                                      // opt-resolve-order
    cfgOptRepoHostPort,                                                                                         // opt-resolve-order
    cfgOptRepoS3Key,                                                                                            // opt-resolve-order
    cfgOptRepoS3KeySecret,                                                                                      // opt-resolve-order
    cfgOptRepoSftpHostFingerprint,                                                                              // opt-resolve-order
};
